/* (C) 2008 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate client tunnel configuration settings in a handy blob.
 *
 * $Id: ClientConfigBlob.cc,v 1.3 2013/03/03 10:05:51 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <sstream>
#include "ClientConfigBlob.hh"
#include "Debug.hh"


namespace HTTPTunnel {


// Construction/destruction
ClientConfigBlob::ClientConfigBlob(Enums::Protocol prot, bool ssl, std::string lhost,
	unsigned int lport, AddrInfo::Family lfam, std::string shost, unsigned int sport)
: ConfigBlob(), m_socks(true), m_prot(prot), m_ssl(ssl), m_listenIP(lhost, lport),
  m_serverIP(shost, sport), m_destIP(), m_cliAuth(false), m_cliUser(""), m_cliPasswd(""),
  m_viaProxy(false), m_proxyIP(), m_proxyAuth(false), m_proxyUser(""), m_proxyPasswd("")
{
}

ClientConfigBlob::ClientConfigBlob(Enums::Protocol prot, bool ssl, std::string lhost,
	unsigned int lport, AddrInfo::Family lfam, std::string shost, unsigned int sport,
	std::string dhost, unsigned int dport)
: ConfigBlob(), m_socks(false), m_prot(prot), m_ssl(ssl), m_listenIP(lhost, lport),
  m_serverIP(shost, sport), m_destIP(dhost, dport), m_cliAuth(false), m_cliUser(""),
  m_cliPasswd(""), m_viaProxy(false), m_proxyIP(), m_proxyAuth(false), m_proxyUser(""),
  m_proxyPasswd("")
{
}

ClientConfigBlob::ClientConfigBlob(const ClientConfigBlob& that)
: ConfigBlob(that), m_socks(that.m_socks), m_prot(that.m_prot), m_ssl(that.m_ssl),
  m_listenIP(that.m_listenIP), m_serverIP(that.m_serverIP), m_destIP(that.m_destIP),
  m_cliAuth(that.m_cliAuth), m_cliUser(that.m_cliUser), m_cliPasswd(that.m_cliPasswd),
  m_viaProxy(that.m_viaProxy), m_proxyIP(that.m_proxyIP), m_proxyAuth(that.m_proxyAuth),
  m_proxyUser(that.m_proxyUser), m_proxyPasswd(that.m_proxyPasswd)
{
}

ClientConfigBlob::~ClientConfigBlob()
{
}


// Assignment
ClientConfigBlob& ClientConfigBlob::operator=(const ClientConfigBlob& that)
{
	if(this != &that) {
		// Base class
		ConfigBlob::operator=(that);

		// Our members
		m_socks = that.m_socks;
		m_prot = that.m_prot;
		m_ssl = that.m_ssl;
		m_listenIP = that.m_listenIP;
		m_serverIP = that.m_serverIP;
		m_destIP = that.m_destIP;
		m_cliAuth = that.m_cliAuth;
		m_cliUser = that.m_cliUser;
		m_cliPasswd = that.m_cliPasswd;
		m_viaProxy = that.m_viaProxy;
		m_proxyIP = that.m_proxyIP;
		m_proxyAuth = that.m_proxyAuth;
		m_proxyUser = that.m_proxyUser;
		m_proxyPasswd = that.m_proxyPasswd;
	}

	return *this;
}


/*************/
/* Accessors */
/*************/


// Get blob type
std::string ClientConfigBlob::getType(void) const
{
	return "CLIENT";
}

// Get blob config as a parsable stream
std::string ClientConfigBlob::getConfig(void) const
{
	std::stringstream cfg;

	/* Format ourselves */

	// Tunnel type
	if(m_socks) {
		// SOCKS
		cfg << "socks" << std::endl;
	}
	else {
		// Fixed
		cfg << "fixed " << m_destIP.getHost() << " " << m_destIP.getPort();
		if(m_prot == Enums::UDP)
			cfg << " udp";
		cfg << std::endl;
	}

	// Use SSL?
	if(m_ssl)
		cfg << "ssl" << std::endl;

	// Local server listen address
	cfg << "listen ";
	std::string lhost = m_listenIP.getHost();
	if(!lhost.empty())
		cfg << lhost << " " << m_listenIP.getPort() << std::endl;
	else
		cfg << m_listenIP.getPort() << " " << m_listenIP.getFamilyStr() << std::endl;

	// Tunnel server address
	cfg << "tunsvr " << m_serverIP.getHost() << " " << m_serverIP.getPort() << std::endl;

	// Client authentication
	if(m_cliAuth)
		cfg << "cliauth " << m_cliUser << " " << m_cliPasswd << std::endl;

	// Proxy settings
	if(m_viaProxy) {
		cfg << "proxy " << m_proxyIP.getHost() << " " << m_proxyIP.getPort() << std::endl;
		if(m_proxyAuth)
			cfg << "proxyauth " << m_proxyUser << " " << m_proxyPasswd << std::endl;
	}

	return cfg.str();
}


/************/
/* Mutators */
/************/


// Enable client authentication to the server and specify username & passwd to use
void ClientConfigBlob::setClientAuth(std::string user, std::string passwd)
{
	m_cliAuth = true;
	m_cliUser = user;
	m_cliPasswd = passwd;
}

// Go via a proxy
void ClientConfigBlob::useProxy(std::string host, unsigned int port)
{
	m_viaProxy = true;
	m_proxyIP = IPAddr(host, port);
	m_proxyAuth = false;
}

// Go via a proxy that requires authentication
void ClientConfigBlob::useProxy(std::string host, unsigned int port, std::string user,
	std::string passwd)
{
	m_viaProxy = true;
	m_proxyIP = IPAddr(host, port);
	m_proxyAuth = true;
	m_proxyUser = user;
	m_proxyPasswd = passwd;
}


}	// namespace HTTPTunnel

