/* (C) 2007 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate a TCP socket.
 *
 * $Id: TCPSock.cc,v 1.8 2012/12/27 13:26:07 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <cerrno>
#include <cstring>
#include <arpa/inet.h>
#include <netdb.h>
#include <poll.h>
#include "TCPSock.hh"
#include "Logger.hh"
#include "Debug.hh"
#include "Utils.hh"
#include "Tunneler.hh"


namespace HTTPTunnel {


// Global references
extern Logger* g_pLogger;


// Construction/destruction
TCPSock::TCPSock(bool blocking) throw(SockError)
: Sock(blocking)
{
}

TCPSock::TCPSock(int fd, std::string cliAddr) throw(SockError)
: Sock(fd, cliAddr)
{
}

TCPSock::~TCPSock()
{
	DPRINTV("TCPSock::~TCPSock()");

	MutexVar mv(m_mtx);

	try {
		// Close socket descriptor
		close();
	}
	catch(SockError& e) {
		g_pLogger->info(e.what());
	}
}


/********************/
/* Public Interface */
/********************/


// Wait for a new incoming TCP connection
void TCPSock::newConnection(Tunneler* tunneler) throw(SockError)
{
	MutexVar mv(m_mtx);

	// Socket ready to accept
	Sock* connectedSock = accept();

	try {
		// Tell tunneler
		if(tunneler)
			tunneler->newConnectionArrived(connectedSock);
		else
			throw SockError("No tunneler to handle new TCP connection", 0);
	}
	catch(TunnelError& e) {
		g_pLogger->info(e.what());
		delete connectedSock;
	}
}


// Set socket listening
void TCPSock::listen(void) throw(SockError)
{
	MutexVar mv(m_mtx);

	// Set listening
	if(::listen(m_fd, 5) < 0)
		throw SockError("Failed to activate listening TCP socket", errno);
}


// Wait for a new incoming connection
Sock* TCPSock::accept(void) throw(SockError)
{
	MutexVar mv(m_mtx);

	// Accept connection
	struct sockaddr_storage cliaddr;
	socklen_t clilen = m_listenAddrLen;
	int connfd = -1;
	if((connfd = ::accept(m_fd, (struct sockaddr*)&cliaddr, &clilen)) < 0) {
		throw SockError("Failed to accept TCP connection", errno);
	}

	// Find client address
	std::string clientAddr = Utils::getIPInfo((struct sockaddr*)&cliaddr);
	DPRINTI("TCPSock::accept() - Got new connection from " + clientAddr);

	// Create and return new TCPSock to handle this connected socket
	return new TCPSock(connfd, clientAddr);
}


// Establish connection out
void TCPSock::connect(IPAddr& ipaddr) throw(SockError, CommsError)
{
	// Create underlying socket
	createSockFD(ipaddr, SOCK_STREAM);

	// Get address details
	std::string addr = ipaddr.getHost();
	unsigned int port = ipaddr.getPort();

	MutexVar mv(m_mtx);

	// Lookup destination address
	struct sockaddr_storage destaddr;
	size_t destaddrLen = 0;
	try {
		AddrInfo addrInfo = Utils::getAddrInfo(addr);
		destaddrLen = addrInfo.getAddrLen();

		// Set up server address structure
		memcpy(&destaddr, addrInfo.getAddr(), destaddrLen);
		if(addrInfo.getFamily() == AddrInfo::IPV6)
			((struct sockaddr_in6*)&destaddr)->sin6_port = htons(port);
		else
			((struct sockaddr_in*)&destaddr)->sin_port = htons(port);
	}
	catch(AddrLookupError& e) {
		std::stringstream err;
		err << "Failed to resolve TCP server address " << addr << ": " << e.what();
		throw ConnectError(err.str());
	}

	// Try the connect
	if(::connect(m_fd, (struct sockaddr*)&destaddr, destaddrLen) < 0) {
		std::stringstream err;
		err << "Failed to connect to TCP server " << ipaddr;
		throw SockError(err.str(), errno);
	}
}


// Read from socket
int TCPSock::read(unsigned char* buff, size_t buffSize)
	throw(SockError, CommsError)
{
	MutexVar mv(m_mtx);

	// Read data from socket
again:
	int noread = ::read(m_fd, buff, buffSize);
	if(noread < 0) {
		if(errno == EINTR)
			goto again;
		else
			throw SockError("Read error from TCP socket", errno);
	}
	else if(noread == 0) {
		// EOF
		throw ConnectionClosed("Other end closed TCP socket");
	}

	return noread;
}


// Write to socket
int TCPSock::write(const unsigned char* buff, size_t buffSize)
	throw(SockError, CommsError)
{
	MutexVar mv(m_mtx);

	// Write data to socket
again:
	int nowritten = ::write(m_fd, buff, buffSize);
	if(nowritten < 0) {
		if(errno == EINTR)
			goto again;
		else
			throw SockError("Write error to TCP socket", errno);
	}

	return nowritten;
}


// Get the result of a non-blocking connect attempt
void TCPSock::getConnectResult(bool& resultAvailable, bool& connected)
	throw(SockError)
{
	MutexVar mv(m_mtx);

	// Poll for connect result
	bool pollError = false;
	resultAvailable = pollWrite(0, pollError);
	if(resultAvailable) {
		// Get connect result
		int cres = -1;
		getSockOpt(SOL_SOCKET, SO_ERROR, cres);
		if(cres == 0) {
			// Success
			connected = true;
		}
		else {
			// Failed
			connected = false;
		}
	}
}


// Bind socket to an address
void TCPSock::bind(IPAddr& ipaddr) throw(SockError)
{
	// Create underlying socket
	createSockFD(ipaddr, SOCK_STREAM);

	// Do the bind
	Sock::bind(ipaddr);
}


}	// namespace

