/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Provide authentication settings and methods.
 *
 * $Id: AuthSettings.cs,v 1.1 2008/07/25 15:22:37 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Collections;
	using System.Text;

	abstract public class AuthSettings {

		/********************/
		/* Public Interface */
		/********************/

		/*********/
		/* Types */
		/*********/

		// Authorisation schemes
		public enum AuthScheme {
			AUTH_NONE, AUTH_UNKNOWN, AUTH_BASIC, AUTH_DIGEST, AUTH_NTLM, AUTH_HOLD
		};


		/***********/
		/* Methods */
		/***********/

		// Construction
		public AuthSettings(Globals glob)
		{
			m_glob = glob;
			m_authScheme = AuthScheme.AUTH_NONE;
			m_triedSchemes = new ArrayList();
			m_user = "";
			m_domain = "";
			m_passwd = "";
			m_realm = "";
			m_digestNonce = "";
			m_digestQop = "auth";
			m_digestCNonce = "";
			m_digestCNonceCount = 1;
			m_ntlmHandler = new NTLM(m_glob);
			m_reqHeader = "";
		}

		public AuthSettings(Globals glob, string user, string domain, string passwd)
		{
			m_glob = glob;
			m_authScheme = AuthScheme.AUTH_UNKNOWN;
			m_triedSchemes = new ArrayList();
			m_user = user;
			m_domain = domain;
			m_passwd = passwd;
			m_realm = "";
			m_digestNonce = "";
			m_digestQop = "";
			m_digestCNonce = "";
			m_digestCNonceCount = 1;
			m_ntlmHandler = new NTLM(m_glob);
			m_reqHeader = "";
		}

		protected AuthSettings(AuthSettings that)
		{
			m_glob = that.m_glob;
			m_authScheme = that.m_authScheme;
			m_triedSchemes = new ArrayList();
			m_user = that.m_user;
			m_domain = that.m_domain;
			m_passwd = that.m_passwd;
			m_realm = that.m_realm;
			m_digestNonce = that.m_digestNonce;
			m_digestQop = that.m_digestQop;
			m_digestCNonce = that.m_digestCNonce;
			m_digestCNonceCount = that.m_digestCNonceCount;
			m_ntlmHandler = new NTLM(m_glob);
			m_reqHeader = that.m_reqHeader;
		}


		/*************/
		/* Accessors */
		/*************/

		// Get whether we have already tried (and failed) some authentication scheme
		public bool triedAuthScheme(AuthScheme scheme)
		{
			// Have we tried this scheme before?
			return (m_triedSchemes.Contains(scheme));
		}

		// Do we have a user/passwd specified?
		public bool gotUserPasswd()
		{
			return (m_user.Length > 0 && m_passwd.Length > 0);
		}

		// Do we have a user domain specified?
		public bool gotDomain()
		{
			return (m_domain.Length > 0);
		}

		// Get proxy authentication user
		public string getUser()
		{
			return m_user;
		}

		// Get proxy authentication domain
		public string getDomain()
		{
			return m_domain;
		}

		// Get proxy authentication passwd
		public string getPasswd()
		{
			return m_passwd;
		}


		/************/
		/* Mutators */
		/************/

		// Set user name & password
		public void setUserPasswd(string user, string passwd)
		{
			m_user = user;
			m_passwd = passwd;
		}


		// Generate a new random nonce
		public void newNonce()
		{
			StringBuilder nonce = new StringBuilder();
			Random rand = new Random();
			nonce.AppendFormat("{0:d8}", rand.Next());
			m_digestNonce = nonce.ToString();
		}


		/*************/
		/* Utilities */
		/*************/

		// Construct tunneling request headers
		public virtual string getRequestHeader(string uri)
		{
			StringBuilder res = new StringBuilder();

			// Using authorization?
			if(m_authScheme == AuthScheme.AUTH_HOLD) {
				// Use previously created request header
				res.Append(m_reqHeader);
			}

			else if(m_authScheme == AuthScheme.AUTH_BASIC) {
				// Create basic authentication fields
				res.AppendFormat("{0}: Basic {1}\r\n", getAuthRequestHeaderID(),
					Utils.Base64Encode(m_user + ":" + m_passwd));
				m_reqHeader = res.ToString();

				// Remember trying basic authentication
				finishedAuthScheme(AuthScheme.AUTH_BASIC);
			}

			else if(m_authScheme == AuthScheme.AUTH_DIGEST) {
				// Get client nonce count
				bool finishedDigest = (m_digestCNonceCount == 1);
				StringBuilder cnoncec = new StringBuilder();
				cnoncec.AppendFormat("{0:d8}", m_digestCNonceCount++);

				// Calculate digest value
				string response = AuthDigest.MD5DigestResponse("MD5", m_user, m_passwd, m_realm,
					m_digestNonce, m_digestCNonce, cnoncec.ToString(), m_digestQop, uri);

				// Create digest authentication fields
				res.AppendFormat("{0}: Digest ", getAuthRequestHeaderID());
				res.AppendFormat("username=\"{0}\"", m_user);
				res.AppendFormat(", realm=\"{0}\"", m_realm);
				res.AppendFormat(", nonce=\"{0}\"", m_digestNonce);
				res.AppendFormat(", uri=\"{0}\"", uri);
				res.Append(", algorithm=\"MD5\"");
				res.AppendFormat(", qop=\"{0}\"", m_digestQop);
				res.AppendFormat(", cnonce=\"{0}\"", m_digestCNonce);
				res.AppendFormat(", nc={0}", cnoncec.ToString());
				res.AppendFormat(", response=\"{0}\"", response);
				res.Append("\r\n");
				m_reqHeader = res.ToString();

				// Remember trying digest authentication
				if(finishedDigest) {
					finishedAuthScheme(AuthScheme.AUTH_DIGEST);
				}
			}

			else if(m_authScheme == AuthScheme.AUTH_NTLM) {
				bool finishedNTLM = false;
				try {
					// Get next set of authentication fields for NTLM
					RawBuff ntlmMsg = new RawBuff();
					finishedNTLM = m_ntlmHandler.createNextMsg(ntlmMsg, m_user, m_domain, m_passwd);
					string encNtlmMsg = "";
					uint encLen = ntlmMsg.getLength();
					ntlmMsg.base64encode(ref encLen, ref encNtlmMsg);

					// Create NTLM authentication fields
					res.AppendFormat("{0}: NTLM ", getAuthRequestHeaderID());
					res.Append(encNtlmMsg);
					res.Append("\r\n");
					m_reqHeader = res.ToString();
				}
				catch(AuthProcessingError e) {
					m_glob.log().info("Error creating NTLM request header: " + e.Message);
					finishedNTLM = true;
				}

				// Remember trying NTLM authentication
				if(finishedNTLM) {
					finishedAuthScheme(AuthScheme.AUTH_NTLM);
				}
			}

			return res.ToString();
		}

		// Construct tunneling response headers
		public virtual string getResponseHeader(string uri)
		{
			return "";
		}


		// Update current authentication scheme state
		public void updateAuthScheme(AuthScheme scheme, ArrayList parms)
		{
			#if DEBUG_TUN
			m_glob.DPRINTV("AuthSettings::updateAuthScheme()");
			StringBuilder dbg = new StringBuilder();
			dbg.AppendFormat("Scheme: {0}, Parms:", scheme);
			for(int dpit = 0; dpit < parms.Count; ++dpit)
				dbg.AppendFormat(" {0}", (string)parms[dpit]);
			m_glob.DPRINTV(dbg.ToString());
			#endif

			// What authentication scheme?
			if(scheme == AuthScheme.AUTH_NONE) {
				// Initialise no authentication
				m_authScheme = AuthScheme.AUTH_NONE;
			}

			else if(scheme == AuthScheme.AUTH_BASIC) {
				// Get parameters for basic authentication
				string realm = "";
				if(Utils.getParmValue(parms, "realm", ref realm)) {
					// Initialise basic authentication
					m_authScheme = AuthScheme.AUTH_BASIC;
					m_realm = realm;
				}
				else {
					throw new AuthProcessingError("Missing basic authentication parameters");
				}
			}

			else if(scheme == AuthScheme.AUTH_DIGEST) {
				// Get parameters for digest authentication
				string realm = "", nonce = "", qop = "";
				if(Utils.getParmValue(parms, "realm", ref realm) && Utils.getParmValue(parms,
				"nonce", ref nonce) && Utils.getParmValue(parms, "qop", ref qop)) {
					// Initialise digest authentication
					m_authScheme = AuthScheme.AUTH_DIGEST;
					m_realm = realm;
					m_digestQop = qop;
					m_digestNonce = nonce;
					m_digestCNonceCount = 1;
					StringBuilder cnonce = new StringBuilder();
					Random rand = new Random();
					cnonce.AppendFormat("{0:d8}", rand.Next());
					m_digestCNonce = cnonce.ToString();
				}
				else {
					throw new AuthProcessingError("Missing digest authentication parameters");
				}
			}

			else if(scheme == AuthScheme.AUTH_NTLM) {
				// Are we starting NTLM?
				if(m_authScheme != AuthScheme.AUTH_NTLM) {
					// Initialise NTLM authentication
					m_authScheme = AuthScheme.AUTH_NTLM;
				}
				else {
					// Update NTLM authentication
					if(parms.Count == 1) {
						m_ntlmHandler.parseRecvdMsg((string)parms[0]);
					}
					else {
						throw new AuthProcessingError("Wrong number of NTLM authentication parameters");
					}
				}
			}
		}


		// Clear and reset any exisiting authentication scheme details
		public abstract void reset();


		/***********************/
		/* Protected Interface */
		/***********************/


		/***********/
		/* Methods */
		/***********/

		// Get authorisation request header field identifier
		protected abstract string getAuthRequestHeaderID();


		// Remember trying an authentication scheme
		protected void finishedAuthScheme(AuthScheme scheme)
		{
			// Check we're not remembering a duplicate
			if(!m_triedSchemes.Contains(scheme)) {
				// Remember trying this scheme
				m_triedSchemes.Add(scheme);
			}
		}


		/********/
		/* Data */
		/********/

		// Globals
		protected Globals					m_glob;

		// Current authentication scheme
		protected AuthScheme				m_authScheme;

		// Tried authentication schemes
		protected ArrayList				m_triedSchemes;

		// Proxy auth parameters
		protected string					m_user;
		protected string					m_domain;
		protected string					m_passwd;
		protected string					m_realm;

		// Digest auth parameters
		protected string					m_digestNonce;
		protected string					m_digestQop;
		protected string					m_digestCNonce;
		protected uint						m_digestCNonceCount;

		// NTLM auth parameters
		protected NTLM						m_ntlmHandler;

		// Last request header
		protected string					m_reqHeader;

	}

}	// namespace

