/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Interface for authenticating a user, and tracking/controling
 * resource usage.
 *
 * $Id: Authenticator.cs,v 1.1 2008/07/25 15:22:37 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Collections;
	using System.Text;

	abstract public class Authenticator {

		// Construction
		public Authenticator(Globals glob)
		{
			m_glob = glob;
			m_users = new Hashtable();
			m_defaultUserName = "DefaultUser";

			// Create server authentication settings
			m_defaultAuthSettings = new ServerAuthSettings(m_glob);

			// Set default username
			m_defaultAuthSettings.setUserPasswd(m_defaultUserName, "");
		}


		/********************/
		/* Public Interface */
		/********************/


		/*************/
		/* Accessors */
		/*************/

		// Check whether user can receive data
		public virtual bool getUserCanRecvBytes(string name, uint bytes)
		{
			bool res = true;

			lock(this) {
				// Lookup user in cached details
				User user = getUser(name);
				if(user != null) {
					res = user.canRecvBytes(bytes);
				}
			}

			return res;
		}

		// Get amount of data user can currently send
		public virtual uint getUserBytesCanSend(string name)
		{
			uint res = 0;

			lock(this) {
				// Lookup user in cached details
				User user = getUser(name);
				if(user != null) {
					res = user.getBytesCanSend();
				}
			}

			return res;
		}

		// Get whether user can create a new tunnel
		public virtual bool getUserCanCreateTunnel(string name)
		{
			bool res = true;

			lock(this) {
				// Lookup user in cached details
				User user = getUser(name);
				if(user != null) {
					res = user.canCreateNewTunnel();
				}
			}

			return res;
		}

		// Get copy of current user details
		public abstract User getUserDetails(string name);

		// Get whether user is authenticated for some tunnel
		public abstract bool getUserAuthenticated(string name, Tunnel tun);

		// Get default authentication settings
		public ServerAuthSettings getDefaultAuthSettings()
		{
			return m_defaultAuthSettings;
		}

		// Get default username
		public string getDefaultUserName()
		{
			lock(this) {
				return m_defaultAuthSettings.getUser();
			}
		}


		/************/
		/* Mutators */
		/************/

		// Set whether a user is authenticated
		public void setUserAuthenticated(string name, bool auth, Tunnel tun)
		{
			lock(this) {
				// Lookup user in cached details
				User user = getUser(name);
				if(user != null) {
					// Set whether they are authenticated
					user.setAuthenticated(auth, tun);
				}
			}
		}

		// Update a users quotas
		public void setUserQuotas(string name, Quotas quotas)
		{
			lock(this) {
				// Lookup user in cached details
				User user = getUser(name);
				if(user != null) {
					// Set their quotas
					user.setQuotas(quotas);
				}
			}
		}

		// Update data user has sent
		public void setUserSentBytes(string name, uint bytes)
		{
			lock(this) {
				// Lookup user in cached details
				User user = getUser(name);
				if(user != null) {
					// Track bytes sent
					user.sentBytes(bytes);
				}
			}
		}


		/*************/
		/* Utilities */
		/*************/


		// A users tunnel has closed
		public virtual void userTunnelClosed(string name, Tunnel tun)
		{
			m_glob.DPRINTV("Authenticator::userTunnelClosed() - " + name + ", " +
				tun.getTidStr());

			lock(this) {
				// Does user exist in cache?
				User user = getUser(name);
				if(user != null) {
					// Let user know it has one less tunnel in use
					user.removeTunnel(tun);

					// If they're no longer required, uncache user details
					if(!user.keepCached()) {
						freeUser(name);
						m_glob.DPRINTV("User " + name + " uncached");
					}
				}
			}
		}

		// Authenticate a new user
		public virtual void authenticate(string userName, Tunnel tun,
		ArrayList authArgs)
		{
			lock(this) {
				// Does user exist in cache?
				User user = getUser(userName);
				if(user == null) {
					// Create user and place in cache
					m_users[userName] = createUser(userName, false);
				}
			}
		}


		/***********************/
		/* Protected Interface */
		/***********************/


		/***********/
		/* Methods */
		/***********/

		// Get cached user details - Caller holds mutex
		protected User getUser(string name)
		{
			User user = null;
			if(m_users.ContainsKey(name)) {
				user = (User)m_users[name];
			}
			#if DEBUG_TUN
			else {
				m_glob.log().info("Authenticator failed to lookup user " + name);
			}
			#endif

			return user;
		}

		// Free cached user details - Caller holds mutex
		protected void freeUser(string name)
		{
			if(m_users.ContainsKey(name)) {
				m_users.Remove(name);
			}
			else {
				m_glob.log().info("Authenticator failed to lookup user " + name);
			}
		}

		// Create a new tunnel user
		protected abstract User createUser(string name, bool keepCached);


		/********/
		/* Data */
		/********/

		// Reference to the globals
		protected Globals						m_glob;

		// Map of cached users
		protected Hashtable					m_users;

		// Default authentication settings to use
		protected string						m_defaultUserName;
		protected ServerAuthSettings		m_defaultAuthSettings;

	}

}	// namespace

