/* (C) 2008 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate client tunnel configuration settings in a handy blob.
 *
 * $Id: ClientConfigBlob.cs,v 1.3 2013/03/10 10:37:26 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Text;

	public class ClientConfigBlob : ConfigBlob {

		// Construction/destruction
		public ClientConfigBlob(Enums.Protocol prot, bool ssl, string lhost, uint lport,
		AddrInfo.Family lfam, string shost, uint sport) : base()
		{
			m_socks = true;
			m_prot = prot;
			m_ssl = ssl;
			m_listenIP = new IPAddr(lhost, lport);
			m_serverIP = new IPAddr(shost, sport);
			m_destIP = new IPAddr();
			m_cliAuth = false;
			m_cliUser = "";
			m_cliPasswd = "";
			m_viaProxy = false;
			m_proxyIP = new IPAddr();
			m_proxyAuth = false;
			m_proxyUser = "";
			m_proxyPasswd = "";
		}

		public ClientConfigBlob(Enums.Protocol prot, bool ssl, string lhost, uint lport,
		AddrInfo.Family lfam, string shost, uint sport, string dhost, uint dport) : base()
		{
			m_socks = false;
			m_prot = prot;
			m_ssl = ssl;
			m_listenIP = new IPAddr(lhost, lport);
			m_serverIP = new IPAddr(shost, sport);
			m_destIP = new IPAddr(dhost, dport);
			m_cliAuth = false;
			m_cliUser = "";
			m_cliPasswd = "";
			m_viaProxy = false;
			m_proxyIP = new IPAddr();
			m_proxyAuth = false;
			m_proxyUser = "";
			m_proxyPasswd = "";
		}


		/********************/
		/* Public Interface */
		/********************/


		/*************/
		/* Accessors */
		/*************/

		// Get blob type
		public override string getType()
		{
			return "CLIENT";
		}

		// Get blob config as a parsable stream
		public override string getConfig()
		{
			StringBuilder cfg = new StringBuilder();

			/* Format ourselves */

			// Tunnel type
			if(m_socks) {
				// SOCKS
				cfg.AppendLine("socks");
			}
			else {
				// Fixed
				cfg.Append("fixed " + m_destIP.getHost() + " " + m_destIP.getPort());
				if(m_prot == Enums.Protocol.UDP)
					cfg.Append(" udp");
				cfg.AppendLine();
			}

			// Use SSL?
			if(m_ssl)
				cfg.AppendLine("ssl");

			// Local server listen address
			cfg.Append("listen ");
			string lhost = m_listenIP.getHost();
			if(lhost.Length > 0)
				cfg.AppendLine(lhost + " " + m_listenIP.getPort().ToString());
			else
				cfg.AppendLine(m_listenIP.getPort().ToString() + " " + m_listenIP.getFamilyStr());

			// Tunnel server address
			cfg.AppendLine("tunsvr " + m_serverIP.getHost() + " " + m_serverIP.getPort());

			// Client authentication
			if(m_cliAuth)
				cfg.AppendLine("cliauth " + m_cliUser + " " + m_cliPasswd);

			// Proxy settings
			if(m_viaProxy) {
				cfg.AppendLine("proxy " + m_proxyIP.getHost() + " " + m_proxyIP.getPort());
				if(m_proxyAuth)
					cfg.AppendLine("proxyauth " + m_proxyUser + " " + m_proxyPasswd);
			}

			return cfg.ToString();
		}


		/************/
		/* Mutators */
		/************/

		// Enable client authentication to the server and specify username & passwd to use
		public void setClientAuth(string user, string passwd)
		{
			m_cliAuth = true;
			m_cliUser = user;
			m_cliPasswd = passwd;
		}

		// Go via a proxy
		public void useProxy(string host, uint port)
		{
			m_viaProxy = true;
			m_proxyIP = new IPAddr(host, port);
			m_proxyAuth = false;
		}

		// Go via a proxy that requires authentication
		public void useProxy(string host, uint port, string user, string passwd)
		{
			m_viaProxy = true;
			m_proxyIP = new IPAddr(host, port);
			m_proxyAuth = true;
			m_proxyUser = user;
			m_proxyPasswd = passwd;
		}


		/***********************/
		/* Protected Interface */
		/***********************/


		/********/
		/* Data */
		/********/

		// SOCKS or direct?
		protected bool					m_socks;

		// TCP/UDP
		protected Enums.Protocol	m_prot;

		// SSL to server?
		protected bool					m_ssl;

		// Local listening address
		protected IPAddr				m_listenIP;

		// Server address
		protected IPAddr				m_serverIP;

		// Direct destination address
		protected IPAddr				m_destIP;

		// Client authentication settings
		protected bool					m_cliAuth;
		protected string				m_cliUser;
		protected string				m_cliPasswd;

		// Proxy settings
		protected bool					m_viaProxy;
		protected IPAddr				m_proxyIP;
		protected bool					m_proxyAuth;
		protected string				m_proxyUser;
		protected string				m_proxyPasswd;

	}

}	// namespace

