/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Client side socket server.
 *
 * $Id: ClientServer.cs,v 1.4 2013/02/17 10:52:29 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Net.Sockets;
	using System.Text;
	using System.Threading;

	public class ClientServer : Server {

		// Construction
		public ClientServer(Globals glob, ClientSideConfig clientConfig)
		: base(glob, clientConfig) 
		{
			m_glob.DPRINTV("ClientServer::ClientServer()");

			try {
				// Take mutexes
				Monitor.Enter(m_serverThreadStartMtx);

				// Setup server socket
				configureSocket(false);

				// Create client tunneler
				string err = "";
				try {
					// Create appropriate tunneler
					if(clientConfig.socksTun()) {
						// Create SOCKS client tunneler
						m_tunneler = new SOCKSClientTunneler(m_glob, this);
					}
					else {
						// Create plain client tunneler
						m_tunneler = new ClientTunneler(m_glob, this);
					}
				}
				catch(TunnelError e) {
					// Error creating tunneler
					err = "Error creating client tunneler: " + e.Message;
				}

				// Wait for server thread to start
				Monitor.Wait(m_serverThreadStartMtx);
				Monitor.Exit(m_serverThreadStartMtx);

				// Let server thread proceed with accepting connections
				Monitor.Enter(m_serverThreadContMtx);
				Monitor.Pulse(m_serverThreadContMtx);
				Monitor.Exit(m_serverThreadContMtx);

				// If we had an error throw it now
				if(err.Length > 0) {
					throw new TunnelError(err);
				}
			}
			catch(TunnelError e) {
				// Clear up cleanly before passing up error
				destroyed();
				throw e;
			}
		}


		// Destruction
		public override void destroyed()
		{
			lock(this) {
			}

			// Base class
			base.destroyed();
		}


		/*************/
		/* Accessors */
		/*************/

		// Get status string
		public override string getStatusString()
		{
			StringBuilder status = new StringBuilder();

			// General base class description
			status.AppendFormat("Client: {0} ", base.getStatusString());

			// Tunnel server
			ClientSideConfig ccfg = (ClientSideConfig)m_serverConfig;
			status.AppendFormat("[Server {0}:{1}] ", ccfg.getServerIP().getHost(),
				ccfg.getServerIP().getPort());

			// Tunnel type
			string tunType = "";
			if(ccfg.socksTun()) {
				tunType = "SOCKS";
			}
			else {
				StringBuilder fxed = new StringBuilder();
				fxed.AppendFormat("FIXED {0}:{1} ", ccfg.getFixedDestIP().getHost(),
					ccfg.getFixedDestIP().getPort());
				tunType = fxed.ToString();
			}
			status.AppendFormat("[{0}]", tunType);

			// Proxy
			if(ccfg.viaProxy()) {
				ProxySettings pcfg = ccfg.getProxySettings();
				status.AppendFormat(" [Proxy {0}:{1}]", pcfg.getIP().getHost(),
					pcfg.getIP().getPort());
			}

			return status.ToString();
		}


		// Get socket descriptor for socket tunnel
		public Sock getSockTunSock()
		{
			lock(this) {
				// Get socket descriptor from tunneler
				ClientTunneler cliTunneler = (ClientTunneler)m_tunneler;
				return cliTunneler.getSockTunSock();
			}
		}

	}

}	// namespace

