/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * A convenient place to store all client side global configuration.
 *
 * $Id: ClientSideConfig.cs,v 1.2 2013/02/17 10:52:29 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.IO;
	using System.Text;

	public class ClientSideConfig : SideConfig {

		// Construction/destruction
		public ClientSideConfig(Globals glob, Enums.Protocol prot, bool ssl,
		bool sockTun, IPAddr listenIP, IPAddr serverIP, string suser, string spasswd,
		string dhost, uint dport, string tune)
		: base(glob, prot, ssl, sockTun, listenIP)
		{
			// Direct to server, fixed tunnel
			m_serverIP = serverIP;
			m_resolvedServerIP = Utils.resolveIP(m_serverIP.getHost());
			m_fixedDestIP = new IPAddr(dhost, dport);
			m_proxySettings = null;
			m_tune = tune;

			// Create authenticator
			m_authenticator = new ClientAuthenticator(m_glob);
			if(suser != "" || spasswd != "")
				setClientAuth(suser, spasswd);
		}

		public ClientSideConfig(Globals glob, Enums.Protocol prot, bool ssl,
		bool sockTun, IPAddr listenIP, IPAddr serverIP, string suser, string spasswd,
		string dhost, uint dport, string tune, IPAddr proxyIP, string puser,
		string ppasswd)
		: base(glob, prot, ssl, sockTun, listenIP)
		{
			// Via explict (possibly authenticating) proxy, fixed tunnel
			m_serverIP = serverIP;
			m_resolvedServerIP = m_serverIP.getHost();
			m_fixedDestIP = new IPAddr(dhost, dport);
			m_tune = tune;

			// Create proxy authentication settings
			string pdomain = lookupUserDomain(ref puser);
			m_proxySettings = new ProxySettings(m_glob, puser, pdomain, ppasswd, proxyIP);

			// Create authenticator
			m_authenticator = new ClientAuthenticator(m_glob);
			if(suser != "" || spasswd != "")
				setClientAuth(suser, spasswd);
		}

		public ClientSideConfig(Globals glob, Enums.Protocol prot, bool ssl,
		IPAddr listenIP, IPAddr serverIP, string suser, string spasswd, string tune)
		: base(glob, prot, ssl, false, listenIP)
		{
			// Direct to server, SOCKS tunnel
			m_serverIP = serverIP;
			m_resolvedServerIP = Utils.resolveIP(m_serverIP.getHost());
			m_fixedDestIP = new IPAddr();
			m_proxySettings = null;
			m_tune = tune;

			// Create authenticator
			m_authenticator = new ClientAuthenticator(m_glob);
			if(suser != "" || spasswd != "")
				setClientAuth(suser, spasswd);
		}

		public ClientSideConfig(Globals glob, Enums.Protocol prot, bool ssl,
		IPAddr listenIP, IPAddr serverIP, string suser, string spasswd, string tune,
		IPAddr proxyIP, string puser, string ppasswd)
		: base(glob, prot, ssl, false, listenIP)
		{
			// Via explict (possibly authenticating) proxy, SOCKS tunnel
			m_serverIP = serverIP;
			m_resolvedServerIP = m_serverIP.getHost();
			m_fixedDestIP = new IPAddr();
			m_tune = tune;

			// Create proxy authentication settings
			string pdomain = lookupUserDomain(ref puser);
			m_proxySettings = new ProxySettings(m_glob, puser, pdomain, ppasswd, proxyIP);

			// Create authenticator
			m_authenticator = new ClientAuthenticator(m_glob);
			if(suser != "" || spasswd != "")
				setClientAuth(suser, spasswd);
		}


		/*************/
		/* Accessors */
		/*************/

		// Client config?
		public override bool isClient()
		{
			return true;
		}

		// Are we going via a proxy?
		public bool viaProxy()
		{
			return (m_proxySettings != null);
		}

		// Are we using a fixed tunnel?
		public bool fixedTun()
		{
			return (m_fixedDestIP.getHost().Length > 0);
		}

		// Are we running a SOCKS proxy?
		public bool socksTun()
		{
			return !fixedTun();
		}

		// Get address for server end of tunnel
		public IPAddr getServerIP()
		{
			return m_serverIP;
		}

		// Get resolved address for server end of tunnel
		public string getResolvedServerIP()
		{
			return m_resolvedServerIP;
		}

		// Get address for fixed tunnel destination
		public IPAddr getFixedDestIP()
		{
			return m_fixedDestIP;
		}

		// Get address for next hop in tunnel
		public IPAddr getNextHopIP()
		{
			if(viaProxy())
				return getProxySettings().getIP();
			else
				return getServerIP();
		}

		// Proxy settings
		public ProxySettings getProxySettings()
		{
			return m_proxySettings;
		}

		// Tunnel tuning
		public string getTunTune()
		{
			return m_tune;
		}


		/************/
		/* Mutators */
		/************/

		// Set client authorisation
		public void setClientAuth(string user, string passwd)
		{
			// Save username & passwd
			ClientAuthenticator auth = (ClientAuthenticator)m_authenticator;
			auth.setUserPasswd(user, passwd);
		}


		/*************/
		/* Utilities */
		/*************/

		// Serialize to a file
		public override void serialize(StreamWriter file)
		{
			// Client header
			file.WriteLine("#######################");
			file.WriteLine("## Client configuration");
			file.WriteLine("#######################");
			file.WriteLine();
			file.WriteLine("[CLIENT]");

			// What sort of client?
			if(fixedTun()) {
				// Fixed tunnel
				file.Write("fixed " + m_fixedDestIP.getHost() + " " + m_fixedDestIP.getPort());
				if(getProtocol() == Enums.Protocol.UDP)
					file.Write(" udp");
				file.WriteLine();
			}
			else {
				// SOCKS tunnel
				file.WriteLine("socks");
			}

			// Local listening address
			file.Write("listen ");
			string host = m_listenIP.getHost();
			if(host.Length > 0) {
				file.Write(host + " ");
			}
			file.Write(m_listenIP.getPort());
			if(host.Length == 0) {
				file.Write(" " + m_listenIP.getFamilyStr());
			}
			file.WriteLine();

			// Tunnel server location
			file.WriteLine("tunsvr " + m_serverIP.getHost() + " " + m_serverIP.getPort());

			// Proxy settings
			if(viaProxy()) {
				ProxySettings proxyCfg = getProxySettings();
				IPAddr proxyip = proxyCfg.getIP();
				file.WriteLine("proxy " + proxyip.getHost() + " " + proxyip.getPort());
				if(proxyCfg.gotUserPasswd()) {
					StringBuilder proxyAuth = new StringBuilder();
					proxyAuth.Append("proxyauth ");
					string user = proxyCfg.getUser();
					if(proxyCfg.gotDomain()) {
						string dfltDomain = lookupUserDomain(ref user);
						if(proxyCfg.getDomain() != dfltDomain)
							proxyAuth.AppendFormat("{0}\\", proxyCfg.getDomain());
					}
					proxyAuth.AppendFormat("{0} {1}", user, proxyCfg.getPasswd());
					file.WriteLine(proxyAuth.ToString());
				}
			}

			// Client authentication
			ServerAuthSettings svrAuth = m_authenticator.getDefaultAuthSettings();
			if(svrAuth.gotUserPasswd()) {
				file.WriteLine("cliauth " + svrAuth.getUser() + " " + svrAuth.getPasswd());
			}

			// SSL certificate
			if(m_ssl)
				file.WriteLine("ssl");

			// Tuning
			if(m_tune != "")
				file.WriteLine("tune " + m_tune);

			file.WriteLine();
		}


		/***********************/
		/* Protected Interface */
		/***********************/


		/***********/
		/* Methods */
		/***********/

		// Get domain for user
		protected string lookupUserDomain(ref string user)
		{
			// By default use domain from environment
			string domain = Environment.UserDomainName;

			// If username contains a '\', it is a combined domain & username
			string[] split = user.Split('\\');
			if(split.Length == 2) {
				domain = split[0];
				user = split[1];
			}

			return domain;
		}


		/********/
		/* Data */
		/********/

		// IP address of server end
		protected IPAddr						m_serverIP;
		protected string						m_resolvedServerIP;

		// IP address for fixed tunnel destination
		protected IPAddr						m_fixedDestIP;

		// Proxy settings
		protected ProxySettings				m_proxySettings;

		// Tunnel tuning
		protected string						m_tune;

	}

}	// namespace

