/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Client side Tunnel.
 *
 * $Id: ClientTunnel.cs,v 1.6 2013/03/31 12:53:55 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Collections;
	using System.Text;

	public class ClientTunnel : Tunnel {

		// Construction
		public ClientTunnel(Globals glob, TunnelConfig tcfg, ClientTunneler tunneler)
		: base(glob, tcfg, tunneler, Direction.RIGHT, TunnelState.CONNECT)
		{
			m_glob.DPRINTI("ClientTunnel::ClientTunnel()");

			m_minPollTimeout = (int)tcfg.getTuning().getMediumPollTimeout();
			m_maxPollTimeout = (int)tcfg.getTuning().getMaxPollTimeout();
			m_pollTimeout = m_minPollTimeout;
			m_httpConnStable = false;
			m_nextPollTime = Time.now();
			m_polling = false;
			m_emptyServerPayloads = 0;
			m_persistent = false;
		}

		// Destruction
		public override void destroyed()
		{
			#if DEBUG_TUN
			StringBuilder str = new StringBuilder();
			str.AppendFormat("ClientTunnel::~ClientTunnel() - {0}", m_tid);
			m_glob.DPRINTI(str.ToString());
			#endif

			// Base class
			base.destroyed();
		}


		/********************/
		/* Public Interface */
		/********************/


		/*************/
		/* Accessors */
		/*************/

		// Are we persistent?
		public override bool isPersistent()
		{
			return m_persistent;
		}


		/*************/
		/* Utilities */
		/*************/

		// Process some newly arrived tunnel msg from one of our connections
		public override void processTunnelMsg(HTTPMsg msg, HTTPConnection httpConn)
		{
			m_glob.DPRINTV("ClientTunnel::processTunnelMsg() - " + getTidStr());

			lock(this) {
				// Check whether HTTP connection is clear
				HTTPMsgParms parms = msg.getParms();
				m_httpConnStable = (parms.getRespCode() == 200);

				// Update our quotas?
				Authenticator authenticator = getAuthenticator();
				if(parms.getState() == TunnelState.AUTHREQ) {
					authenticator.setUserQuotas(m_user, parms.getQuotas());
				}

				// Do we have an indication of whether this connection is persistent?
				if(!httpConn.getPersistent()) {
					if(parms.getConnKeepAlive() || parms.getProxyConnKeepAlive()) {
						#if DEBUG_TUN
						StringBuilder dbg = new StringBuilder();
						dbg.AppendFormat("Tunnel {0} got keep-alive for HTTP connection {1}",
							getTidStr(), httpConn.getID());
						dbg.Append(": Switching to persistent tunneling");
						m_glob.DPRINTV(dbg.ToString());
						#endif

						// Looks like we can use persistent mode
						httpConn.setPersistent();
					}
				}

				// Update count of empty payloads
				uint payloadSize = parms.getPayload().getPayloadSize();
				if(payloadSize == 0)
					m_emptyServerPayloads++;
				else
					m_emptyServerPayloads = 0;

				// Base class
				base.processTunnelMsg(msg, httpConn);

				// Check persistence & see if poll timeout needs to change
				checkPersistent();
				scalePollTimeout();

				// Consider sending an early reply
				considerEarlySend(httpConn, payloadSize);
			}
		}


		// Poll the far end of the tunnel
		public void pollTunnel()
		{
			m_glob.DPRINTV("ClientTunnel::pollTunnel() - " + getTidStr());

			lock(this) {
				// Only send poll msg if we've been quiet
				TimeSpan quietTime = Time.now() - m_lastTunnelAnySendTime;
				int millis = quietTime.Milliseconds;
				if(millis >= m_pollTimeout) {
					#if DEBUG_TUN
					StringBuilder dbg = new StringBuilder();
					dbg.AppendFormat("Last msg sent {0} ago, timeout = {1} - polling",
						millis, m_pollTimeout);
					m_glob.DPRINTV(dbg.ToString());
					#endif

					// Force the sending of a msg
					m_polling = true;
					sendTunnelMsg(true, null);
					m_polling = false;
				}

				// Schedule next poll timeout
				m_nextPollTime = new Time(0, m_pollTimeout);
				m_tunneler.scheduleEvent(new PollTimeoutEvent(m_glob, m_nextPollTime,
					m_tunneler, m_tid));
			}
		}


		// Set whether the tunnel user is authenticated
		public void setUserAuthenticated(bool auth)
		{
			lock(this) {
				Authenticator authenticator = getAuthenticator();
				authenticator.setUserAuthenticated(m_user, auth, this);
			}
		}


		/***********************/
		/* Protected Interface */
		/***********************/


		/*************/
		/* Constants */
		/*************/

		// Time between queued data tunnel msg sends (milli-seconds)
		protected const int		EARLY_MSG_SEND_TIME			= 10;


		/***********/
		/* Methods */
		/***********/

		// Create a new payload msg for sending down the tunnel - Caller holds mutex
		protected override HTTPMsgParms makeTunnelMsg(HTTPMsgParms recvd, bool retry)
		{
			// Get basic parameters from base class
			HTTPMsgParms parms = base.makeTunnelMsg(recvd, retry);

			// If we're connecting, include tuning settings
			if(m_tunnelState == TunnelState.CONNECT || m_tunnelState == TunnelState.RECONNECT) {
				// Set tuning
				parms.setTunTune(m_tunCfg.getTuning().getTuneDesc());
			}

			return parms;
		}


		// Successfully sent a new message down the tunnel - Caller hold mutex
		protected override void sentNewTunnelMsg(HTTPConnection conn, HTTPMsgParms parms)
		{
			// We always want to remember in the client
			base.sentNewTunnelMsg(conn, parms, true);
		}


		// Determine whether a mlti-ack implies we should resend a msg - Caller holds mutex
		protected override bool checkForMultiAckResendRequest(SentMsgParms sentMsgParms)
		{
			/* Since the server only responds to msgs we have sent, any duplicate
			 * acks imply a resend request.
			 */

			m_glob.DPRINTV("ClientTunnel::checkForMultiAckResendRequest() - Multi-Ack resend request");
			return true;
		}


		// Have we got a stable HTTP connection to other end?
		protected override bool httpStable()
		{
			return m_httpConnStable;
		}


		// Can we send a new tunnel msg? - Caller holds mutex
		protected override bool clearToSendNewMsg(HTTPConnection conn)
		{
			bool sendNew = false;

			try {
				// Check for unacked management msgs and consider resending them early
				checkUnackedMgmtMsgs();

				// New message or resend of old message?
				sendNew = !resendUnackedMsgs(conn);
				if(sendNew) {
					// If we're in a special connection state don't check any more, just send
					if(m_tunnelState == TunnelState.CONNECTED || m_tunnelState == TunnelState.CLOSING) {
						// How many outstanding unacked msgs do we already have?
						if(m_unackedMsgQueue.Count >= m_maxUnackedMsgs) {
							m_glob.DPRINTI("Too many unacked msgs on tunnel; not sending new msg");
							sendNew = false;
						}
						else if(!conn.getPersistent() && noUnackedMsgsOnConn(conn.getID()) >= 1) {
							// Non-persistent connection can't send more than 1 msg at a time
							m_glob.DPRINTI("Tunnel connection is non-persistent; not sending new msg");
							sendNew = false;
						}
					}
				}
			}
			catch(ResendError e) {
				// Resending error
				m_glob.log().warn("Error resending msg: " + e.Message);
				recordConnectionError(TunnelErrorType.MSG_RESEND_ERROR);
			}

			return sendNew;
		}


		// Check whether any connections need to be blocked/unblocked - Caller holds mutex
		protected override void checkBlockedConnections()
		{
			// Check raw connections
			ArrayList filt = new ArrayList();
			filt.Add(new SelLeftConn());
			ArrayList conns = new ArrayList();
			getConnsFromSel(filt, conns);

			// Do we want to block or unblock?
			bool unblock = false;
			bool block = (m_rawDataQueueSize >= MAX_RAW_DATA_FILL);
			if(!block) {
				// Only unblock if we're well below full
				uint buffFill = (uint)(((float)m_rawDataQueueSize / (float)MAX_RAW_DATA_FILL)
					* 100);
				unblock = (buffFill <= RAW_DATA_UNBLOCK_LEVEL);
			}

			// Check each connection
			for(int cit = 0; cit != conns.Count; ++cit) {
				// Get connection
				Connection conn = (Connection)conns[cit];

				// Need to change blocked state?
				if(block)
					conn.setBlocked(true);
				else if(unblock)
					conn.setBlocked(false);
			}
		}


		// Advance our internal state machine - Caller holds mutex
		protected override void advanceState(HTTPMsgParms parms)
		{
			// Respond to the reported state from this msg
			switch(parms.getState()) {
				case(TunnelState.AUTHREQ): {
					// Become connected
					setState(TunnelState.CONNECTED);
					break;
				}
				case(TunnelState.CLOSING): {
					// Closing
					setState(TunnelState.CLOSING);
					break;
				}
				case(TunnelState.CLOSED): {
					// Closed
					setState(TunnelState.CLOSED);
					break;
				}
				case(TunnelState.CONNECTED): {
					// Stay connected
					setState(TunnelState.CONNECTED);
					break;
				}
				case(TunnelState.CONNECT):
				case(TunnelState.RECONNECT):
				case(TunnelState.UNKNOWNSTATE):
				default: {
					StringBuilder err = new StringBuilder();
					err.AppendFormat("Unexpected ClientTunnel reported msg state: {0}",
						parms.getState());
					m_glob.log().error(err.ToString());
					break;
				}
			}
		}


		// Get maximum allowable payload size - Caller holds mutex
		protected override uint getMaxPayloadSize(bool useHeader)
		{
			// Get payload size from client tunnel parameters
			ClientTunneler cliTunneler = (ClientTunneler)m_tunneler;
			ClientTunnelParms tunParms = cliTunneler.getTunParms();
			return tunParms.getMaxPayloadSize(useHeader);
		}


		// Something has been sent down the tunnel - Caller holds mutex
		protected override void tunnelAnyMsgSent()
		{
			// Base class
			base.tunnelAnyMsgSent();

			// Reschedule any pending poll timeout
			if(!m_polling) {
				m_nextPollTime = new Time(0, m_pollTimeout);
				m_tunneler.rescheduleEvent<PollTimeoutEvent, TunnelID>(m_tid, m_nextPollTime);
			}
		}


		// Scale poll timeout - Caller hold mutex
		protected void scalePollTimeout()
		{
			// Remember old poll timeout
			int oldPollTimeout = m_pollTimeout;

			// Get tuned poll timeouts
			TuneSettings tuning = m_tunCfg.getTuning();
			int fastPollTimeout = (int)tuning.getFastPollTimeout();
			int mediumPollTimeout = (int)tuning.getMediumPollTimeout();
			int slowPollTimeout = (int)tuning.getSlowPollTimeout();
			int maxPollTimeout = (int)tuning.getMaxPollTimeout();

			// Check and update min/max poll timeouts
			if(m_emptyServerPayloads > THROTTLE_POLL_COUNT_HI) {
				// Poll slowest
				m_minPollTimeout = slowPollTimeout;
				m_maxPollTimeout = maxPollTimeout;
			}
			else if(m_emptyServerPayloads > THROTTLE_POLL_COUNT_LO) {
				// Poll medium
				m_minPollTimeout = mediumPollTimeout;
				m_maxPollTimeout = maxPollTimeout;
			}
			else {
				// Poll fastest
				m_minPollTimeout = fastPollTimeout;
				m_maxPollTimeout = maxPollTimeout;
			}

			// Set poll timeout at 10% below average ack rtt
			m_pollTimeout = m_biasAvgAckRTT - (m_biasAvgAckRTT / 10);

			// Check we're still within bounds
			if(m_pollTimeout > m_maxPollTimeout)
				m_pollTimeout = m_maxPollTimeout;
			else if(m_pollTimeout < m_minPollTimeout)
				m_pollTimeout = m_minPollTimeout;

			if(oldPollTimeout != m_pollTimeout) {
				#if DEBUG_TUN
				StringBuilder nfo = new StringBuilder();
				nfo.AppendFormat("ClientTunnel::scalePollTimeout() - Now {0}", m_pollTimeout);
				nfo.AppendFormat(" for tunnel {0}", getTidStr());
				m_glob.DPRINTV(nfo.ToString());
				#endif

				// Work out new time for next poll timeout
				if(m_pollTimeout > oldPollTimeout) {
					// Reschedule for later
					m_nextPollTime.addInterval(0, (m_pollTimeout - oldPollTimeout));
				}
				else {
					// Reschedule for earlier
					m_nextPollTime.subInterval(0, (oldPollTimeout - m_pollTimeout));
				}

				// Reschedule poll timeout for new time
				m_tunneler.rescheduleEvent<PollTimeoutEvent, TunnelID>(m_tid, m_nextPollTime);
			}
		}


		/********/
		/* Data */
		/********/

		// Min/max poll timeouts to use
		protected int					m_minPollTimeout;
		protected int					m_maxPollTimeout;

		// Time between poll timeout events
		protected int					m_pollTimeout;


		/*********************/
		/* Private Interface */
		/*********************/


		/*************/
		/* Constants */
		/*************/

		// No. of empty payloads to receive before throttling poll timeout
		private const uint		THROTTLE_POLL_COUNT_LO		= 10;
		private const uint		THROTTLE_POLL_COUNT_HI		= 50;


		/***********/
		/* Methods */
		/***********/

		// Work out our persistent state - Caller holds mutex
		private void checkPersistent()
		{
			// Assume non-persistent to start
			m_persistent = false;

			// Get all connections out to server end
			ArrayList filt = new ArrayList();
			filt.Add(new SelRightConn());
			ArrayList conns = new ArrayList();
			getConnsFromSel(filt, conns);

			// Check each connection for persitence
			for(int cit = 0; cit != conns.Count; ++cit) {
				// We are persistent if any connection is
				if(conns[cit] is HTTPConnection) {
					HTTPConnection conn = (HTTPConnection)conns[cit];
					m_persistent |= conn.getPersistent();
				}
			}
		}


		// Consider whether we should schedule an early send before the next poll timeout
		private void considerEarlySend(HTTPConnection httpConn, uint payloadSize)
		{
			// Consider early send if we have data to send or server has sent us data
			if(m_rawDataQueueSize > 0 || payloadSize >= TestServerTunnel.MIN_MAX_PAYLOAD_SIZE) {
				// Check there isn't already a send msg pending
				uint connID = httpConn.getID();
				if(!m_tunneler.isScheduled<SendMsgEvent, uint>(connID)) {
					// Don't start sending new msgs if we're in the process of resending old ones
					if(m_resendQueue.Count == 0) {
						// Check our quotas will allow us to send more just now
						Authenticator authenticator = getAuthenticator();
						if(authenticator.getUserBytesCanSend(m_user) > 0) {
							#if DEBUG_TUN
							StringBuilder dbg = new StringBuilder();
							if(m_rawDataQueueSize > 0) {
								dbg.AppendFormat("ClientTunnel has {0} bytes queued, scheduling early send",
									m_rawDataQueueSize);
							}
							else {
								dbg.AppendFormat("ClientTunnel got {0} bytes from server, scheduling early send",
									payloadSize);
							}
							m_glob.DPRINTV(dbg.ToString());
							#endif

							// Schedule msg send to server
							m_tunneler.scheduleEvent(new SendMsgEvent(m_glob, new Time(0,EARLY_MSG_SEND_TIME),
								connID, false, getTid(), m_tunneler));
						}
						else {
							m_glob.DPRINTV("ClientTunnel prevented from early send of more data by quotas");
						}
					}
					else {
						#if DEBUG_TUN
						StringBuilder dbg = new StringBuilder();
						if(m_rawDataQueueSize > 0)
							dbg.AppendFormat("ClientTunnel has {0} bytes queued ", m_rawDataQueueSize);
						else
							dbg.AppendFormat("ClientTunnel got {0} bytes from server ", payloadSize);
						dbg.AppendFormat("but {0} msgs queued to resend, not sending more early", m_resendQueue.Count);
						m_glob.DPRINTV(dbg.ToString());
						#endif
					}
				}
			}
		}


		/********/
		/* Data */
		/********/

		// Flag to indicate if we have a clear & stable HTTP connection
		private bool					m_httpConnStable;

		// Time next poll is scheduled for
		private Time					m_nextPollTime;
		private bool					m_polling;

		// Number of empty server payloads received in a row
		private uint					m_emptyServerPayloads;

		// Are we persistent?
		private bool					m_persistent;

	}

}	// namespace

