/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Tunnel controller for client end of tunnel.
 *
 * $Id: ClientTunneler.cs,v 1.7 2013/02/17 10:52:29 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Collections;
	using System.Net.Sockets;
	using System.Text;

	public class ClientTunneler : Tunneler {

		// Construction
		public ClientTunneler(Globals glob, Server server) : base(glob, server)
		{
			m_glob.DPRINTV("ClientTunneler::ClientTunneler()");

			m_tunParms = TestClientTunnel.getDefaultTunParms();
			m_sockTunClient = null;
			m_oldTunParmsAge = OLD_TUN_PARMS_AGE;

			try {
				// Slightly randomise old tunnel parms age
				Random rand = new Random();
				int otpaRand = rand.Next(OLD_TUN_PARMS_AGE / 10);
				m_oldTunParmsAge += otpaRand;

				lock(m_tunnels) {
					// Schedule test of connection to server
					checkScheduleServerTest(null);

					// See if we need a socket tunnel client
					ClientSideConfig clientCfg = getClientSideConfig();
					if(clientCfg.isSockTun()) {
						// Get address socket tunnel is listening on
						IPAddr addr = clientCfg.getListenIP();

						// Create appropriate client for socket tunnel
						if(clientCfg.getProtocol() == Enums.Protocol.TCP) {
							// TCP socket tunnel
							m_sockTunClient = new TCPSockTunClient(m_glob, addr);
						}
						else {
							// UDP socket tunnel
							m_sockTunClient = new UDPSockTunClient(m_glob, addr);
						}
					}
				}
			}
			catch(TunnelError e) {
				// Clear up and rethrow
				destroyed();
				throw e;
			}
		}

		// Destruction
		public override void destroyed()
		{
			m_glob.DPRINTV("ClientTunneler::~ClientTunneler()");

			// Shutdown event scheduling
			if(!shuttingDown())
				shutdownSchedulers();

			// Shutdown any socket tunnel
			if(m_sockTunClient != null) {
				m_sockTunClient.destroyed();
				m_sockTunClient = null;
			}

			// Base class
			base.destroyed();
		}


		/*************/
		/* Accessors */
		/*************/

		// Get reference to current tunneling parameters
		public ClientTunnelParms getTunParms()
		{
			lock(m_tunnels) {
				return m_tunParms;
			}
		}


		// Get socket descriptor for client socket tunnel
		public Sock getSockTunSock()
		{
			lock(m_tunnels) {
				// Check we have a socket tunnel client
				if(m_sockTunClient == null)
					throw new TunnelError("No socket tunnel client in use");

				// Get socket descriptor from socket tunnel client
				return m_sockTunClient.getSock();
			}
		}


		/************/
		/* Mutators */
		/************/

		// Enable/disable testing based on current config setting
		public override void setServerTesting()
		{
			m_glob.DPRINTV("ClientTunneler::setServerTesting()");

			lock(m_tunnels) {
				// Do we have a test tunnel running?
				Tunnel testTun = getTunFromPred(new PredTestTun());
				bool testRunning = testTun != null;

				// Do we have a test scheduled?
				bool testScheduled = isScheduled<LaunchServerTestEvent>();

				// Are we starting or stopping testing?
				if(m_glob.cfg().getTesting()) {
					m_glob.DPRINTV("Enabling testing...");

					// Start testing if it's stopped
					if(!testRunning && !testScheduled) {
						// Schedule test of connection to server
						m_glob.DPRINTV("Starting testing...");
						checkScheduleServerTest(null);
					}
				}
				else {
					m_glob.DPRINTV("Disabling testing...");

					// Stop testing if it's started
					if(testScheduled) {
						// Cancel scheduled test
						m_glob.DPRINTV("Cancelling scheduled test...");
						cancelAllEvents<LaunchServerTestEvent>();
					}
					if(testRunning) {
						// Shutdown runnign test tunnel
						m_glob.DPRINTV("Aborting running test...");
						scheduleEvent(new ShutdownTunnelEvent(m_glob, new Time(0,0), this, testTun.getTid(), false));
					}
				}
			}
		}


		/**********/
		/* Events */
		/**********/

		// Notification of a new socket connection
		public override void newConnectionArrived(Sock insock)
		{
			m_glob.DPRINTI("ClientTunneler::newConnectionArrived()");

			///////////////////////////////////////////////////////////////
			// In ClientTunneler a new connection always means a new tunnel
			///////////////////////////////////////////////////////////////

			// Do our users quotas allow another tunnel?
			Authenticator authenticator = m_server.getConfig().getAuthenticator();
			string userName = authenticator.getDefaultUserName();
			bool userQuotaOk = authenticator.getUserCanCreateTunnel(userName);
			if(userQuotaOk) {
				Tunnel tun = null;

				// Create new tunnel config
				TunnelConfig tcfg = createTunnelConfig();

				// Lock connection details
				lock(m_connections) {
					// Create raw server connection to handle incoming traffic
					Connection rawconn = new RawConnection(m_glob, insock, this,
						tcfg.getTunnelID());

					// Create HTTP client connection to handle tunneled traffic
					HTTPConnection httpconn = new HTTPConnection(m_glob, this);

					// Save connection details
					addConnection(rawconn);
					addConnection(httpconn);

					// Create tunnel
					tun = new ClientTunnel(m_glob, tcfg, this);
					tun.addLeftConnection(rawconn);
					tun.addRightConnection(httpconn);

					// Schedule connect out on HTTP side
					ClientSideConfig clientCfg = getClientSideConfig();
					m_connectionMgr.addProxyConnect(httpconn, new Time(0,0),
						clientCfg.getNextHopIP(), Enums.Protocol.TCP, clientCfg.getSSL());
				}

				// Store Tunnel object in our list of active tunnels
				lock(m_tunnels) {
					m_tunnels.Add(tun);
				}
			}
			else {
				// Tunnel creation quota exceeded
				StringBuilder err = new StringBuilder();
				err.AppendFormat("Tunnel creation quota exceeded for user {0}", userName);
				throw new CommsError(err.ToString());
			}
		}


		// Notification that a tunnel has insufficient left/right connections to be viable
		public override void unviableTunnel(TunnelID tid)
		{
			m_glob.DPRINTI("ClientTunneler::unviableTunnel() - " + tid.ToString());

			// Lookup tunnel
			lock(m_connections) {
				lock(m_tunnels) {
					Tunnel tun = getTunFromPred(new PredTunID(tid));
					if(tun != null) {
						// If tunnel is already closed don't bother checking any further
						if(!tun.isClosed()) {
							// Get details about state of tunnel
							int leftConns = tun.noLeftConnections();
							int rightConns = tun.noRightConnections();

							// Check whether our left side warrants preserving
							bool leftPreserve = true;
							if(leftConns == 0) {
								/* In the client, no remaining left connections means we must
								 * abandon this tunnel and close it down.
				 				 */

								// Is this tunnel in the process of closing?
								if(tun.isClosing()) {
									// Give us a chance to finish closing
									m_glob.DPRINTI(tid + " - No lconns, closing: Left preserve yes");
									leftPreserve = true;
								}
								else {
									// No left connections and not closing, so don't preserve
									m_glob.DPRINTI(tid + " - No lconns, not closing: Left preserve no");
									leftPreserve = false;
								}
							}

							// Check whether our right side warrants preserving
							bool rightPreserve = true;
							if(rightConns == 0) {
								/* In the client, no remaining right connections means that if
								 * we want to try and preserve this tunnel we should try and
								 * re-establish a connection to the server.
								 */

								// Do we want to try and preserve this tunnel?
								if(!tun.errorLevelExceeded(Tunnel.Direction.RIGHT)) {
									// Consider preserving right side
									m_glob.DPRINTI(tid + " - No rconns, not erred: Right preserve yes");
									rightPreserve = true;
								}
								else {
									// Don't consider preserving right side
									m_glob.DPRINTI(tid + " - No rconns, erred: Right preserve no");
									rightPreserve = false;
								}
							}

							// Both sides agree to preserve?
							if(leftPreserve && rightPreserve) {
								m_glob.DPRINTI(tid + " - Trying to preserve");

								// Should we delay preservation of tunnel?
								Tunnel.TunnelState oldTunState = tun.getState();
								int preserveDelay = 0;
								if(oldTunState != Tunnel.TunnelState.CONNECTED) {
									// Tunnel wasn't connected
									m_glob.DPRINTV("Tunnel wasn't CONNECTED - delaying preserve connect");
									preserveDelay = NC_PRESERVE_RECONNECT_DELAY;
								}
								else if(!tun.isPersistent()) {
									// Tunnel wasn't persistent
									m_glob.DPRINTV("Tunnel is non-persistent - delaying preserve connect");
									preserveDelay = NP_PRESERVE_RECONNECT_DELAY;
								}

								// Let tunnel know
								tun.reconnecting();

								// Create replacement HTTP client connection to server
								HTTPConnection httpconn = new HTTPConnection(m_glob, this);

								// Save connection details
								addConnection(httpconn);
								tun.addRightConnection(httpconn);

								// Schedule initial connect out HTTP side
								ClientSideConfig clientCfg = getClientSideConfig();
								m_connectionMgr.addProxyConnect(httpconn, new Time(0,preserveDelay),
									clientCfg.getNextHopIP(), Enums.Protocol.TCP, clientCfg.getSSL());
							}
							else {
								m_glob.DPRINTI(tid + " - Not preserving");
								scheduleEvent(new ShutdownTunnelEvent(m_glob, new Time(0,0), this, tid, false));
							}
						}
						else {
							m_glob.DPRINTI(tid + " - Already closed, not preserving further");
						}
					}
					else {
						StringBuilder err = new StringBuilder();
						err.AppendFormat("ClientTunneler::unviableTunnel() - Unknown tunnel {0}", tid);
						m_glob.log().warn(err.ToString());
					}
				}
			}
		}


		// Notification of new tunneled/raw data arrival
		public override void rawDataArrived(uint connID, TunnelID tid, RawBuff raw)
		{
			#if DEBUG_TUN
			StringBuilder nfo = new StringBuilder();
			nfo.AppendFormat("ClientTunneler::rawDataArrived() - {0} bytes ", raw.getLength());
			nfo.AppendFormat("on connection {0}", connID);
			m_glob.DPRINTV(nfo.ToString());
			#endif

			// Lookup tunnel
			lock(m_tunnels) {
				Tunnel tun = getTunFromPred(new PredTunID(tid));
				if(tun != null) {
					// Inform tunnel of raw data arrival
					tun.processRawData(raw);

					// Try sending this data on to the server
					tun.sendTunnelMsg(true, null);
				}
				else {
					StringBuilder err =  new StringBuilder();
					err.AppendFormat("ClientTunneler::rawDataArrived() - On connection {0}", connID);
					err.AppendFormat(" Unknown tunnel {0}", tid.ToString());
					throw new LookupManagementError(err.ToString());
				}
			}
		}


		// Notification of new UDP 'connection'
		public override void newUDPConnectionArrived(Sock insock)
		{
			////////////////////////////////////////////////////////////////////
			// We currently only support fixed UDP tunnelling, so if we already
			// have a tunnel created we can ignore this 'new connection'.
			////////////////////////////////////////////////////////////////////

			// Do we already have a UDP tunnel to use?
			lock(m_connections) {
				lock(m_tunnels) {
					if(m_tunnels.Count == 0) {
						m_glob.DPRINTI("ClientTunneler::newUDPConnectionArrived()");

						// Create new tunnel config
						TunnelConfig tcfg = createTunnelConfig();

						// Create raw server connection to handle incoming traffic
						Connection rawconn = new RawConnection(m_glob, insock, this,
							tcfg.getTunnelID());

						// Create HTTP client connection to handle tunneled traffic
						HTTPConnection httpconn = new HTTPConnection(m_glob, this);

						// Save connection details
						addConnection(rawconn);
						addConnection(httpconn);

						// Create tunnel
						Tunnel tun = new ClientTunnel(m_glob, tcfg, this);
						tun.addLeftConnection(rawconn);
						tun.addRightConnection(httpconn);

						// Schedule connect out on HTTP side
						ClientSideConfig clientCfg = getClientSideConfig();
						m_connectionMgr.addProxyConnect(httpconn, new Time(0,0),
							clientCfg.getNextHopIP(), Enums.Protocol.TCP, clientCfg.getSSL());

						// Store Tunnel object in our list of active tunnels
						m_tunnels.Add(tun);
					}
				}
			}
		}


		// Check we have received an ack for some sent message
		public override void checkAckReceived()
		{
			lock(m_tunnels) {
				// Check acks for all tunnels
				for(int tit = 0; tit < m_tunnels.Count; ++tit) {
					// Let tunnel check we have received any due acks, & resend if not
					Tunnel tun = (Tunnel)m_tunnels[tit];
					if(tun.checkAckTimeout()) {
						tun.sendTunnelMsg(true, null);
					}
				}
			}

			// Schedule next ack check event
			scheduleEvent(new CheckAckEvent(m_glob, new Time(ACK_CHECK_PERIOD,0), this));
		}


		// Notification of new HTTP message arrival
		public override void messageArrived(uint connID, HTTPMsg msg)
		{
			// Base class
			base.messageArrived(connID, msg);

			lock(m_connections) {
				// Lookup connection
				int cit = getConnectionByID(connID);

				// Check for HTTP error messages
				bool fatalConnectionError = false;
				HTTPResponseMsg respMsg = (HTTPResponseMsg)msg;
				uint respCode = respMsg.getResponseCode();
				if(respCode == 407 || respCode == 401) {
					// Authentication required
					authenticationFailed(connID, respMsg, respCode);
				}
				else if(respCode == 404) {
					// Probably a proxy error
					responseError(connID, respMsg, respCode);
				}
				else {
					// Standard tunneling msg
					try {
						lock(m_tunnels) {
							// Find out which Tunnel this message is from
							HTTPMsgParms parms = msg.getParms();
							TunnelID tid = new TunnelID(parms.getTid());

							// Have we found the tunnel for this msg?
							Tunnel btun = getTunFromPred(new PredTunID(tid));
							if(btun != null && btun is ClientTunnel) {
								ClientTunnel tun = (ClientTunnel)btun;

								// Normal response code?
								if(respCode == 200) {
									// We know we are authenticated
									tun.setUserAuthenticated(true);

									// Reset auth settings
									if(tun.getServerAuthSettings() != null)
										tun.getServerAuthSettings().reset();
									if(tun.getProxySettings() != null)
										tun.getProxySettings().reset();
								}

								// Handle any received tunnel payload
								HTTPConnection httpConn = (HTTPConnection)m_connections[cit];
								tun.processTunnelMsg(msg, httpConn);

								// Does this connection report it is closing?
								if(parms.getConnClose() || parms.getProxyConnClose()) {
									#if DEBUG_TUN
									StringBuilder dbg1 = new StringBuilder();
									dbg1.AppendFormat("Connection id {0} reports it is closing", connID);
									m_glob.DPRINTV(dbg1.ToString());
									#endif

									// Don't use closing connection any more
									connectionClosing(cit);
								}

								// Check reported tunnel state
								if(parms.getState() == Tunnel.TunnelState.CLOSED) {
									// Notification of tunnel closure, schedule tunnel shutdown
									m_glob.DPRINTI("Received tunnel shutdown request from server");
									scheduleEvent(new ShutdownTunnelEvent(m_glob, new Time(0,0), this,
										tid, true));
								}
								else if(parms.getState() == Tunnel.TunnelState.AUTHREQ) {
									// Send back authentication response
									authenticationRequest(tun);
								}
							}
							else {
								// Unknown tunnel
								StringBuilder wrn = new StringBuilder();
								wrn.AppendFormat("ClientTunneler::messageArrived() - Unknown tunnel {0}", tid);
								m_glob.log().warn(wrn.ToString());
								fatalConnectionError = true;
							}
						}
					}
					catch(InvalidID e) {
						m_glob.log().error("Error parsing ID: " + e.Message);
						fatalConnectionError = true;
					}
				}

				// Did something very serious fail?
				if(fatalConnectionError) {
					#if DEBUG_TUN
					StringBuilder dbg2 = new StringBuilder();
					dbg2.AppendFormat("Fatal error processing msg on connection {0}", connID);
					m_glob.DPRINTV(dbg2.ToString());
					#endif

					// No point keeping this connection around, kill it now
					destroyConnectionByIndex(cit);
				}
			}
		}


		// Notification of some new default tunnel parameters
		public override void newTunnelParms(TunnelParms parms, TunnelerID tid)
		{
			m_glob.DPRINTI("ClientTunneler::newTunnelParms() - " + tid);

			// Base class
			base.newTunnelParms(parms, tid);

			lock(m_tunnels) {
				// Save new parms
				ClientTunnelParms cparms = (ClientTunnelParms)parms;
				m_tunParms = cparms;

				#if DEBUG_TUN
				StringBuilder nfo = new StringBuilder();
				nfo.AppendFormat("New ClientTunnelParms: Max URL = {0}", cparms.getURLPayloadSize());
				nfo.AppendFormat(", Max body = {0}", cparms.getBodyPayloadSize());
				m_glob.DPRINTV(nfo.ToString());
				#endif
			}
		}


		// Notification of a poll timeout on a tunnel
		public override void pollTimeout(TunnelID tid)
		{
			lock(m_tunnels) {
				// Lookup tunnel
				Tunnel tun = getTunFromPred(new PredTunID(tid));
				if(tun != null) {
					// Send a poll msg to the server side
					ClientTunnel cliTun = (ClientTunnel)tun;
					cliTun.pollTunnel();
				}
				else {
					StringBuilder err = new StringBuilder();
					err.AppendFormat("ClientTunneler::pollTimeout() - Unknown tunnel {0}", tid);
					m_glob.log().warn(err.ToString());
				}
			}
		}


		// Notification that a tunnel has changed state
		public override void tunnelStateChanged(TunnelID tid, Tunnel.TunnelState state)
		{
			m_glob.DPRINTI("ClientTunneler::tunnelStateChanged() - " + tid.ToString());

			// Has tunnel become CONNECTED?
			if(state == Tunnel.TunnelState.CONNECTED) {
				lock(m_tunnels) {
					// Lookup tunnel
					Tunnel tun = getTunFromPred(new PredTunID(tid));
					if(tun != null) {
						// Do we have a poll scheduled for this tunnel?
						if(!isScheduled<PollTimeoutEvent, TunnelID>(tid)) {
							// Perform first poll of newly connected tunnel
							ClientTunnel cliTun = (ClientTunnel)tun;
							cliTun.pollTunnel();
						}
					}
					else {
						StringBuilder err = new StringBuilder();
						err.AppendFormat("ClientTunneler::tunnelStateChanged() - Unknown tunnel {0}", tid);
						m_glob.log().warn(err.ToString());
					}
				}
			}
		}


		// Notification that we should perform tunnel manangement
		public override void performTunnelManagement()
		{
			lock(m_tunnels) {
				// When did we last check tunneling parameters?
				TimeSpan checkTime = Time.now() - m_lastTunParmsCheckTime;
				if(checkTime.TotalSeconds >= CHECK_TUN_PARMS_PERIOD) {
					m_lastTunParmsCheckTime = Time.now();

					// Do we have completed test tunneling parameters?
					if(m_tunParms.testsCompleted()) {
						// Are our test parameters too old?
						TimeSpan parmAge = Time.now() - m_tunParms.getCreateTime();
						if(parmAge.TotalSeconds >= m_oldTunParmsAge) {
							#if DEBUG_TUN
							StringBuilder dbg = new StringBuilder();
							dbg.AppendFormat("Removing {0} secs old test tunneling parms",
								parmAge.TotalSeconds);
							m_glob.DPRINTV(dbg.ToString());
							#endif

							// Reset old parms and schedule a new test
							m_tunParms.resetTestState();
							checkScheduleServerTest(null);
						}
					}
				}

				// Base class
				base.performTunnelManagement();
			}
		}


		// Launch test server connection
		public virtual void launchServerTestConnection()
		{
			m_glob.DPRINTV("ClientTunneler::launchTestServerConnection()");

			lock(m_connections) {
				lock(m_tunnels) {
					// Create new test tunnel config
					TunnelConfig tcfg = createTunnelConfig();

					// Create new test connections
					DummyRawConnection rawconn = new DummyRawConnection(m_glob, this, tcfg.getTunnelID());
					HTTPConnection httpconn = new HTTPConnection(m_glob, this);
					addConnection(rawconn);
					addConnection(httpconn);

					// Create test tunnel
					Tunnel tun = new TestClientTunnel(m_glob, tcfg, this);
					tun.addLeftConnection(rawconn);
					tun.addRightConnection(httpconn);

					// Schedule connect out to server
					ClientSideConfig clientCfg = getClientSideConfig();
					m_connectionMgr.addProxyConnect(httpconn, new Time(0,0),
						clientCfg.getNextHopIP(), Enums.Protocol.TCP, clientCfg.getSSL());

					// Store Tunnel object in our list of active tunnels
					m_tunnels.Add(tun);
				}
			}
		}


		/***********************/
		/* Protected Interface */
		/***********************/


		/*************/
		/* Constants */
		/*************/

		// Number of connect attempts before giving up
		protected const uint			CLIENT_CONNECT_ATTEMPTS			= 4;

		// Time between reconnection attempts
		protected const uint			CLIENT_RECONNECT_DELAY			= 3;

		// Time to pause (milli-seconds) between preservation reconnects
		protected const int			NC_PRESERVE_RECONNECT_DELAY	= 200;
		protected const int			NP_PRESERVE_RECONNECT_DELAY	= 250;
		protected const int			RESEND_AUTH_DELAY					= 100;

		// Delay (seconds) before relaunching failed test connection
		protected const int			BAD_TEST_LAUNCH_DELAY			= 2;


		/***********/
		/* Methods */
		/***********/

		// Get reconnection attempt parameters
		protected override void getReconnectParms(ref uint maxConnectAttempts,
		ref uint reconnectDelay, uint failedAttempts)
		{
			// Increase reconnect delay with each attempt
			maxConnectAttempts = CLIENT_CONNECT_ATTEMPTS;
			reconnectDelay = failedAttempts * CLIENT_RECONNECT_DELAY;
		}


		// Check whether we should relaunch a server test - Caller holds mutex
		protected override void checkScheduleServerTest(Tunnel tun)
		{
			// Is tunnel testing enabled?
			if(!m_glob.cfg().getTesting())
				return;

			// Got a tunnel to check?
			if(tun != null) {
				if(tun.isTestTunnel()) {
					// Are these test not completed?
					TestClientTunnel tcTun = (TestClientTunnel)tun;
					if(!tcTun.testsCompleted()) {
						// Check we don't already have a test launch scheduled
						if(!isScheduled<LaunchServerTestEvent>()) {
							// Schedule relaunch to complete failed testing
							m_glob.DPRINTV("ClientTunneler::checkScheduleServerTest() - Scheduling relaunch");
							scheduleEvent(new LaunchServerTestEvent(m_glob, new Time(BAD_TEST_LAUNCH_DELAY,0),
								this));
						}
					}
				}
			}
			else {
				// Check we don't already have a test tunnel running
				if(getTunFromPred(new PredTestTun()) == null) {
					// Check we don't already have a test launch scheduled
					if(!isScheduled<LaunchServerTestEvent>()) {
						// Launch new test tunnel
						m_glob.DPRINTV("ClientTunneler::checkScheduleServerTest() - Launching new test");
						scheduleEvent(new LaunchServerTestEvent(m_glob, new Time(0,0), this));
					}
				}
			}
		}


		// Create config for tunnel to server
		protected TunnelConfig createTunnelConfig()
		{
			// Create tunnelID
			ClientSideConfig clientCfg = getClientSideConfig();
			TunnelID tid = new TunnelID(m_glob, clientCfg);

			// Create tunnel config
			return new TunnelConfig(tid, clientCfg.getTunTune(), clientCfg.getProxySettings(),
				clientCfg.getAuthenticator().getDefaultAuthSettings(), clientCfg.getFixedDestIP());
		}


		/********/
		/* Data */
		/********/

		// Tunneling parameters to use
		protected ClientTunnelParms					m_tunParms;

		// Socket tunnel client
		protected SockTunClient							m_sockTunClient;


		/*********************/
		/* Private Interface */
		/*********************/


		/***********/
		/* Methods */
		/***********/

		// Respond to an authentication request from the server
		private void authenticationRequest(Tunnel tun)
		{
			m_glob.DPRINTI("ClientTunneler::authenticationRequest()");

			// Send back authentication response
			tun.sendTunnelMsg(true, null);
		}


		// Notification of an authentication failure
		private void authenticationFailed(uint connID, HTTPResponseMsg respMsg,
		uint respCode)
		{
			#if DEBUG_TUN
			StringBuilder nfo = new StringBuilder();
			nfo.Append("ClientTunneler::authenticationFailed() - ");
			nfo.AppendFormat("Connection id {0}", connID);
			m_glob.DPRINTI(nfo.ToString());
			#endif

			bool proxy = false;

			try {
				// Lookup connection (also lock tunnels to prevent race with tunnel shutdown)
				lock(m_connections) {
					lock(m_tunnels) {
						int cit = getConnectionByID(connID);
						Connection conn = (Connection)m_connections[cit];

						// Lookup carried tunnels (can't rely on having tunnel ID from msg)
						ArrayList tunnels = getAllTunsFromPred(new PredTunConnection(conn));
						for(int tit = 0; tit != tunnels.Count; ++tit) {
							ClientTunnel tun = (ClientTunnel)tunnels[tit];

							try {
								// Reconfigure whatever is necessary for the required auth
								if(respCode == 407) {
									proxy = true;
									setupProxyServerAuthConfig(respMsg.getAuthArgs(),
										tun.getProxySettings(), connID, tun.getTidStr());
								}
								else if(respCode == 401) {
									tun.setUserAuthenticated(false);
									setupProxyServerAuthConfig(respMsg.getAuthArgs(),
										tun.getServerAuthSettings(), connID, tun.getTidStr());
								}

								// Schedule chance to resend messages using authentication
								scheduleEvent(new SendMsgEvent(m_glob, new Time(0,RESEND_AUTH_DELAY),
									connID, true, tun.getTid(), this));
							}
							catch(AuthProcessingError e) {
								m_glob.log().warn("Authentication error: " + e.Message);

								// Fatal tunnel error
								tun.recordConnectionError((proxy? Tunnel.TunnelErrorType.PROXY_AUTH_ERROR :
									Tunnel.TunnelErrorType.SERVER_AUTH_ERROR));
							}
						}
					}
				}
			}
			catch(TunnelError e) {
				m_glob.log().warn("Authentication error: " + e.Message);

				// Schedule fatal connection error
				scheduleConnectionError(connID, (proxy? Tunnel.TunnelErrorType.PROXY_AUTH_ERROR :
					Tunnel.TunnelErrorType.SERVER_AUTH_ERROR));
			}
		}


		// Setup config for required proxy/server authentication
		private void setupProxyServerAuthConfig(Hashtable authDetails, AuthSettings settings,
		uint connID, string tidStr)
		{
			// Pick authentication scheme
			bool triedScheme = false;
			bool basic = false, digest = false, ntlm = false;
			ArrayList basicArgs = null, digestArgs = null, ntlmArgs = null;

			if(authDetails.ContainsKey("Basic")) {
				// Can use basic auth
				basic = true;
				basicArgs = (ArrayList)authDetails["Basic"];
			}
			if(authDetails.ContainsKey("Digest")) {
				// Can use digest auth
				digest = true;
				digestArgs = (ArrayList)authDetails["Digest"];
			}
			if(authDetails.ContainsKey("NTLM")) {
				// Can use NTLM auth
				ntlm = true;
				ntlmArgs = (ArrayList)authDetails["NTLM"];
			}

			// Got nothing to try?
			if(!(basic || digest || ntlm))
				throw new AuthProcessingError("No suitable authentication scheme suggested");

			// Setup for basic authentication
			if(!triedScheme && basic) {
				// Check we haven't already tried this
				if(!settings.triedAuthScheme(AuthSettings.AuthScheme.AUTH_BASIC)) {
					// Try basic authentication
					m_glob.DPRINTI("Trying basic authentication");
					try {
						// Setup basic authentication
						settings.updateAuthScheme(AuthSettings.AuthScheme.AUTH_BASIC, basicArgs);
						triedScheme = true;
					}
					catch(AuthProcessingError e) {
						// Need to try something else
						m_glob.log().info("Basic authentication error for tunnel " + tidStr + ": " + e.Message);
					}
				}
				else {
					// Need to try something else
					m_glob.log().info("Basic authentication failed for tunnel " + tidStr);
				}
			}

			// Setup for digest authentication
			if(!triedScheme && digest) {
				// Check we haven't already tried this
				if(!settings.triedAuthScheme(AuthSettings.AuthScheme.AUTH_DIGEST)) {
					// Try digest authentication
					m_glob.DPRINTI("Trying digest authentication");
					try {
						// Setup digest authentication
						settings.updateAuthScheme(AuthSettings.AuthScheme.AUTH_DIGEST, digestArgs);
						triedScheme = true;
					}
					catch(AuthProcessingError e) {
						// Need to try something else
						m_glob.log().info("Digest authentication error for tunnel " + tidStr + ": " + e.Message);
					}
				}
				else {
					// Need to try something else
					m_glob.log().info("Digest authentication failed for tunnel " + tidStr);
				}
			}

			// Setup for NTLM authentication
			if(!triedScheme && ntlm) {
				// Check we haven't already tried this
				if(!settings.triedAuthScheme(AuthSettings.AuthScheme.AUTH_NTLM)) {
					// Try NTLM authentication
					m_glob.DPRINTI("Trying NTLM authentication");
					try {
						// Setup NTLM authentication
						settings.updateAuthScheme(AuthSettings.AuthScheme.AUTH_NTLM, ntlmArgs);
						triedScheme = true;
					}
					catch(AuthProcessingError e) {
						// Need to try something else
						m_glob.log().info("NTLM authentication error for tunnel " + tidStr + ": " + e.Message);
					}
				}
				else {
					// Need to try something else
					m_glob.log().info("NTLM authentication failed for tunnel " + tidStr);
				}
			}

			// Have we got nothing left to try?
			if(!triedScheme)
				throw new AuthProcessingError("No suitable authentication scheme left to try");
		}


		// Handle a msg response error
		private void responseError(uint connID, HTTPResponseMsg respMsg, uint respCode)
		{
			#if DEBUG_TUN
			StringBuilder nfo = new StringBuilder();
			nfo.Append("ClientTunneler::responseError() - ");
			nfo.AppendFormat("Connection id {0}, response code: {1}", connID, respCode);
			m_glob.DPRINTI(nfo.ToString());
			#endif

			// Replace this connection with a new one and see if that cures the reponse error
			scheduleConnectionError(connID, Tunnel.TunnelErrorType.GENERAL_COMMS_ERROR);
		}


		/********/
		/* Data */
		/********/

		// Age (seconds) at which we retest old tunnel parms
		private int								m_oldTunParmsAge;

	}

}	// namespace

