/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * A convenient place to store all global configuration.
 *
 * $Id: Config.cs,v 1.6 2013/02/17 10:52:29 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */


namespace HTTPTunnel
{
	using System;
	using System.Collections;
	using System.Diagnostics;
	using System.Net.Sockets;
	using System.IO;
	using System.Text;

	public class Config {

		// Constructor
		public Config(Globals glob, string procName)
		{
			m_glob = glob;
			m_procName = procName;
			m_daemon = true;
			m_testing = false;
			m_hostName = "";
			m_trueHostName = "";
			m_debugLevel = Globals.ERROR;
			m_showLog = false;
			m_pid = Process.GetCurrentProcess().Id;
			m_pidFile = "";
			m_realm = "FireDrill";
			m_servers = new ArrayList();
			m_serverConfigs = new ArrayList();
			m_dataLogStem = @"C:\DebugLogs\Win-";
			m_bigEndian = false;
			m_version = "";

			// Default config file name
			string configPath = Environment.GetFolderPath(
				Environment.SpecialFolder.ApplicationData);
			m_configFileName = configPath + Path.DirectorySeparatorChar +
				"FireDrill.cfg";

			// Create pseudo unique local hostname
			m_trueHostName = Environment.MachineName;
			StringBuilder hostname = new StringBuilder();
			hostname.Append(m_trueHostName);

			// Try to ensure pseudeo-uniqueness by appending some random number
			Random rand = new Random();
			hostname.AppendFormat("{0:d4}", rand.Next());
			m_hostName = hostname.ToString();

			// Determine endianness of platform
			m_bigEndian = !BitConverter.IsLittleEndian;

			// Create version identifier
			StringBuilder ver = new StringBuilder();
			ver.AppendFormat("{0} {1}", Globals.LIBVERSION, Globals.PLATNAME);
			m_version = ver.ToString();
		}


		/*************/
		/* Utilities */
		/*************/

		// Serialize to a config file
		public void dumpToFile()
		{
			lock(this) {
				StreamWriter file = null;
				try {
					// Open file for writing
					file = File.CreateText(m_configFileName);
				}
				catch(Exception) {
					throw new ConfigError("Failed to open config file " + m_configFileName +
						" for writing");
				}

				/* Header */

				file.WriteLine("##########################");
				file.WriteLine("## " + m_procName + " config");
				file.WriteLine("## Version " + m_version);
				file.WriteLine("##########################");

				/* Global options */

				file.WriteLine();
				file.WriteLine("#################");
				file.WriteLine("## Global options");
				file.WriteLine("#################");
				file.WriteLine();
				file.WriteLine("[GLOBAL]");

				// Daemon?
				if(!getDaemon()) {
					file.WriteLine("nodaemon");
				}

				// Testing?
				if(getTesting()) {
					file.WriteLine("dotest");
				}

				// Show logger window?
				if(getShowLog()) {
					file.WriteLine("showlog");
				}

				// Debug level
				file.WriteLine("debug " + getDebug());

				// PID file
				string pidfile = getPIDFile();
				if(pidfile.Length > 0) {
					file.WriteLine("pidfile " + pidfile);
				}

				/* Servers config */

				file.WriteLine();

				// Get config for all servers
				for(int sit = 0; sit < m_serverConfigs.Count; ++sit) {
					// Serialize config from this server
					SideConfig cfg = (SideConfig)m_serverConfigs[sit];
					cfg.serialize(file);
				}

				// Close file
				file.Close();
			}
		}


		/**********************/
		/* Accessors/mutators */
		/**********************/

		// Configure server
		public void configureServer(bool ssl, bool sockTun, string cert, string certKey,
		IPAddr listenIP, string firewallIPs, string firewallPorts, bool auth)
		{
			lock(this) {
				m_serverConfigs.Add(new ServerSideConfig(m_glob, ssl, sockTun, cert, certKey,
					listenIP, firewallIPs, firewallPorts, auth));
			}
		}

		// Reconfigure server
		public void reconfigureServer(int no, bool ssl, bool sockTun, string cert,
		string certKey, IPAddr listenIP, string firewallIPs, string firewallPorts,
		bool auth)
		{
			lock(this) {
				// Clear up existing server
				destroyNamedLocalServer(no);

				// Create replacement config
				m_serverConfigs.Insert(no, new ServerSideConfig(m_glob, ssl, sockTun, cert,
					certKey, listenIP, firewallIPs, firewallPorts, auth));

				// Start replacement server
				createNamedLocalServer(no);
			}
		}

		// Configure default server
		public int configureDefaultServer()
		{
			lock(this) {
				// Pick some sensible defaults
				configureServer(false, false, "", "", new IPAddr("", 8080), "", "", false);

				// Return ID for new config
				return getLastConfigID();
			}
		}


		// Configure client
		public void configureClient(Enums.Protocol prot, bool ssl, bool sockTun,
		IPAddr listenIP, IPAddr serverIP, string suser, string spasswd, string dhost,
		uint dport, string tune)
		{
			lock(this) {
				m_serverConfigs.Add(new ClientSideConfig(m_glob, prot, ssl, sockTun,
					listenIP, serverIP, suser, spasswd, dhost, dport, tune));
			}
		}

		public void configureClient(Enums.Protocol prot, bool ssl, bool sockTun,
		IPAddr listenIP, IPAddr serverIP, string suser, string spasswd, string dhost,
		uint dport, string tune, IPAddr proxyIP, string puser, string ppasswd)
		{
			lock(this) {
				m_serverConfigs.Add(new ClientSideConfig(m_glob, prot, ssl, sockTun,
					listenIP, serverIP, suser, spasswd, dhost, dport, tune, proxyIP,
					puser, ppasswd));
			}
		}

		public void configureClient(Enums.Protocol prot, bool ssl, IPAddr listenIP,
		IPAddr serverIP, string suser, string spasswd, string tune)
		{
			lock(this) {
				m_serverConfigs.Add(new ClientSideConfig(m_glob, prot, ssl, listenIP,
					serverIP, suser, spasswd, tune));
			}
		}

		public void configureClient(Enums.Protocol prot, bool ssl, IPAddr listenIP,
		IPAddr serverIP, string suser, string spasswd, string tune, IPAddr proxyIP,
		string puser, string ppasswd)
		{
			lock(this) {
				m_serverConfigs.Add(new ClientSideConfig(m_glob, prot, ssl, listenIP,
					serverIP, suser, spasswd, tune, proxyIP, puser, ppasswd));
			}
		}

		// Reconfigure client
		public void reconfigureClient(int no, Enums.Protocol prot, bool ssl, bool sockTun,
		IPAddr listenIP, IPAddr serverIP, string suser, string spasswd, string dhost,
		uint dport, string tune)
		{
			lock(this) {
				// Clear up existing server
				destroyNamedLocalServer(no);

				// Create replacement config
				m_serverConfigs.Insert(no, new ClientSideConfig(m_glob, prot, ssl, sockTun,
					listenIP, serverIP, suser, spasswd, dhost, dport, tune));

				// Start replacement server
				createNamedLocalServer(no);
			}
		}

		public void reconfigureClient(int no, Enums.Protocol prot, bool ssl, bool sockTun,
		IPAddr listenIP, IPAddr serverIP, string suser, string spasswd, string dhost,
		uint dport, string tune, IPAddr proxyIP, string puser, string ppasswd)
		{
			lock(this) {
				// Clear up existing server
				destroyNamedLocalServer(no);

				// Create replacement config
				m_serverConfigs.Insert(no, new ClientSideConfig(m_glob, prot, ssl, sockTun,
					listenIP, serverIP, suser, spasswd, dhost, dport, tune, proxyIP, puser,
					ppasswd));

				// Start replacement server
				createNamedLocalServer(no);
			}
		}

		public void reconfigureClient(int no, Enums.Protocol prot, bool ssl, IPAddr listenIP,
		IPAddr serverIP, string suser, string spasswd, string tune)
		{
			lock(this) {
				// Clear up existing server
				destroyNamedLocalServer(no);

				// Create replacement config
				m_serverConfigs.Insert(no, new ClientSideConfig(m_glob, prot, ssl, listenIP,
					serverIP, suser, spasswd, tune));

				// Start replacement server
				createNamedLocalServer(no);
			}
		}

		public void reconfigureClient(int no, Enums.Protocol prot, bool ssl, IPAddr listenIP,
		IPAddr serverIP, string suser, string spasswd, string tune, IPAddr proxyIP,
		string puser, string ppasswd)
		{
			lock(this) {
				// Clear up existing server
				destroyNamedLocalServer(no);

				// Create replacement config
				m_serverConfigs.Insert(no, new ClientSideConfig(m_glob, prot, ssl, listenIP,
					serverIP, suser, spasswd, tune, proxyIP, puser, ppasswd));

				// Start replacement server
				createNamedLocalServer(no);
			}
		}

		// Configure default client
		public int configureDefaultClient()
		{
			lock(this) {
				// Pick some sensible defaults
				Random rand = new Random();
				int svrNum = rand.Next(1,3);
				StringBuilder svrAddr = new StringBuilder();
				svrAddr.AppendFormat("tun{0}.fire-drill.com", svrNum);
				configureClient(Enums.Protocol.TCP, false, new IPAddr("", 1080),
					new IPAddr(svrAddr.ToString(), 80), "", "", "default");

				// Return ID for new config
				return getLastConfigID();
			}
		}


		// Setup any additional config required for client tunnelers
		public void additionalClientConfig(TunnelerID tid, ClientConfigBlob blob)
		{
			// No additional configuration currently required for clients
		}

		// Setup any additional config required for server tunnelers
		public void additionalServerConfig(TunnelerID tid, ServerConfigBlob blob)
		{
			lock(this) {
				// Lookup current config for this server
				int no = getConfigIDFromTunnelerID(tid);
				SideConfig cfg = (SideConfig)m_serverConfigs[no];
				if(cfg is ServerSideConfig) {
					ServerSideConfig svrCfg = (ServerSideConfig)cfg;
					// Is this a socket tunnel?
					if(svrCfg.isSockTun()) {
						// Get socket tunnel config blob
						if(blob is SockTunServerConfigBlob) {
						SockTunServerConfigBlob sockTunBlob = (SockTunServerConfigBlob)blob;
							// Set user callback
							svrCfg.setUserCallback(sockTunBlob.getUserCallback());
						}
						else {
							StringBuilder err = new StringBuilder();
							err.AppendFormat("Error setting additional config for server tunneler id {0}",
								tid.ToString());;
							err.Append(": Config says it's a socket tunnel, but blob is of wrong type");
							throw new TunnelError(err.ToString());
						}
					}
				}
				else {
					StringBuilder err = new StringBuilder();
					err.AppendFormat("Error setting additional config for server tunneler id {0}",
						tid.ToString());
					err.Append(": It is not a server");
					throw new TunnelError(err.ToString());
				}
			}
		}


		// Get ID of last configured client/server tunneler
		public TunnelerID getLastTunnelerID()
		{
			lock(this) {
				// Return tunneler ID for most recent config
				int idx = getLastConfigID();
				SideConfig cfg = (SideConfig)m_serverConfigs[idx];
				return cfg.getTunnelerID();
			}
		}


		// Get version identifier
		public string getVersionStr()
		{
			return m_version;
		}

		// Get local host name
		public string getUniqueHostName()
		{
			return m_hostName;
		}
		public string getHostName()
		{
			return m_trueHostName;
		}

		// Get/set daemon flag
		public void setDaemon(bool flag)
		{
			m_daemon = flag;
		}
		public bool getDaemon()
		{
			return m_daemon;
		}

		// Get/set debug level
		public void setDebug(uint debug)
		{
			m_debugLevel = debug;
		}
		public uint getDebug()
		{
			return m_debugLevel;
		}

		// Get/set whether to show logger window
		public void setShowLog(bool flag)
		{
			m_showLog = flag;
			m_glob.log().setShowLogging(m_showLog);
		}
		public bool getShowLog()
		{
			return m_showLog;
		}

		// Get/set tunnel testing flag
		public void setTesting(bool flag, bool tellTunnelers)
		{
			m_testing = flag;

			// Do we need to let active tunnelers know?
			if(tellTunnelers)
				activateTesting();
		}
		public bool getTesting()
		{
			return m_testing;
		}

		// Get our process ID
		public int getPid()
		{
			return m_pid;
		}

		// Get/set PID file name
		public void setPIDFile(string pidfile)
		{
			m_pidFile = pidfile;
		}
		public string getPIDFile()
		{
			return m_pidFile;
		}

		// Get/set the default realm
		public string getRealm()
		{
			return m_realm;
		}
		public void setRealm(string realm)
		{
			m_realm = realm;
		}

		// Get/set data log file name stem
		public string getDataLogFileStem()
		{
			return m_dataLogStem;
		}
		public void setDataLogFileStem(string stem)
		{
			m_dataLogStem = stem;
		}

		// Get/set the config file name
		public string getConfigFileName()
		{
			return m_configFileName;
		}
		public void setConfigFileName(string name)
		{
			m_configFileName = name;
		}

		// Get config for a server
		public SideConfig getServerConfig(int index)
		{
			return (SideConfig)m_serverConfigs[index];
		}

		// Get a server
		public Server getServer(int index)
		{
			return (Server)m_servers[index];
		}


		// Get status strings for servers
		public string[] getServerStatusStrings()
		{
			ArrayList statusStrs = new ArrayList();

			lock(this) {
				for(int sit = 0; sit < m_servers.Count; ++sit) {
					// Build up server status string
					Server svr = (Server)m_servers[sit];
					StringBuilder status = new StringBuilder();

					// Server number
					status.AppendFormat("{0,-3} ", sit + 1);

					// Server description
					status.AppendFormat("{0}", svr.getStatusString());

					statusStrs.Add(status.ToString());
				}
			}

			return (string[])statusStrs.ToArray(typeof(string));
		}


		// Create local servers
		public void createLocalServers()
		{
			lock(this) {
				// Create server for each set of config
				for(int cit = 0; cit < m_serverConfigs.Count; ++cit)
					createNamedLocalServer(cit);
			}
		}

		// Create named local server
		public void createNamedLocalServer(int no)
		{
			lock(this) {
				// Get config for named server
				SideConfig cfg = (SideConfig)m_serverConfigs[no];

				// Create server
				if(cfg.isClient()) {
					// Create client side server
					ClientSideConfig ccfg = (ClientSideConfig)cfg;
					m_servers.Insert(no, new ClientServer(m_glob, ccfg));
				}
				else {
					// Create server side server
					ServerSideConfig scfg = (ServerSideConfig)cfg;
					m_servers.Insert(no, new ServerServer(m_glob, scfg));
				}
			}
		}

		// Create named local server
		public void createNamedLocalServer(TunnelerID tid)
		{
			lock(this) {
				// Get index for named server
				int no = getConfigIDFromTunnelerID(tid);

				// Create named server
				createNamedLocalServer(no);
			}
		}


		// Destroy local servers
		public void destroyLocalServers()
		{
			lock(this) {
				// Stop local servers
				while(m_servers.Count > 0)
					destroyNamedLocalServer(0);
			}
		}


		// Destroy specified local server/config
		public void destroyNamedLocalServer(int no)
		{
			lock(this) {
				// Clear server config
				SideConfig cfg = (SideConfig)m_serverConfigs[no];
				m_serverConfigs.RemoveAt(no);

				// Got a corresponding server for this config?
				if(m_servers.Count > no) {
					Server svr = (Server)m_servers[no];
					if(svr.getConfig() == cfg) {
						// Clear up server
						svr.destroyed();
						m_servers.RemoveAt(no);
					}
				}
			}
		}

		// Destroy specified local server/config
		public void destroyNamedLocalServer(TunnelerID tid)
		{
			lock(this) {
				// Get index for named server
				int no = getConfigIDFromTunnelerID(tid);

				// Destroy named server
				destroyNamedLocalServer(no);
			}
		}


		// Initiate clean shutdown of all local server tunnels
		public void shutdownNamedLocalServerTunnels(int no)
		{
			lock(this) {
				// Got a corresponding server for this config?
				if(m_servers.Count > no) {
					Server svr = (Server)m_servers[no];

					// Initiate shutdown on this servers tunneler
					Tunneler tunneler = svr.getTunneler();
					if(tunneler != null) {
						tunneler.initiateAllTunnelsShutdown();
					}
				}
			}
		}

		// Initiate clean shutdown of all local server tunnels
		public void shutdownNamedLocalServerTunnels(TunnelerID tid)
		{
			lock(this) {
				// Get index for named server
				int no = getConfigIDFromTunnelerID(tid);

				// Destroy named server
				shutdownNamedLocalServerTunnels(no);
			}
		}


		// Have we got at least one client or server configured?
		public bool confCliSvr()
		{
			return (m_serverConfigs.Count > 0);
		}


		// Are we running on a big-endian platform?
		public bool bigEndian()
		{
			return m_bigEndian;
		}


		// Get socket descriptor for a client socket tunnel
		public Sock getCliSockTunSock(TunnelerID tid)
		{
			Sock sock = null;

			lock(this) {
				try {
					// Lookup client tunneler
					int no = getConfigIDFromTunnelerID(tid);
					ClientServer cliSvr = getClientServer(no);

					// Get socket descriptor
					sock = cliSvr.getSockTunSock();
				}
				catch(TunnelError e) {
					StringBuilder err = new StringBuilder();
					err.AppendFormat("Error retrieving socket for client socket tunneler id {0}",
						tid.ToString());
					err.AppendFormat(": {0}", e.Message);
					throw new TunnelError(err.ToString());
				}
			}

			return sock;
		}


		// Get stats for all tunnels carried by the specified tunneler
		public void getTunnelerTunnelStats(TunnelerID tid, ArrayList stats)
		{
			// Clear results
			stats.Clear();

			lock(this) {
				// Lookup tunneler server
				int no = getConfigIDFromTunnelerID(tid);
				Server svr = (Server)m_servers[no];

				// Get stats from server
				svr.getTunnelStats(stats);
			}
		}


		/*********************/
		/* Private Interface */
		/*********************/


		/***********/
		/* Methods */
		/***********/

		// Get ID of last client/server config - Caller holds mutex
		private int getLastConfigID()
		{
			// Most recent config will be last index in list
			if(m_serverConfigs.Count > 0)
				return (m_serverConfigs.Count - 1);

			// No configured clients/servers
			throw new TunnelError("No currently configured clients/servers to get ID for");
		}


		// Get ID of client/server config from tunneler ID - Caller holds mutex
		private int getConfigIDFromTunnelerID(TunnelerID tid)
		{
			// Look for matching tunneler ID in list of configs
			bool foundIt = false;
			int idx;
			for(idx = 0; idx < m_serverConfigs.Count; ++idx) {
				SideConfig cfg = (SideConfig)m_serverConfigs[idx];
				TunnelerID thatTid = cfg.getTunnelerID();
				if(tid == thatTid) {
					foundIt = true;
					break;
				}
			}

			// Not found?
			if(!foundIt) {
				StringBuilder err = new StringBuilder();
				err.AppendFormat("No currently configured tunneler found with ID {0}", tid);
				throw new TunnelError(err.ToString());
			}

			// Return index of matching config
			return idx;
		}


		// Lookup client tunneler with the specified index - Caller holds mutex
		private ClientServer getClientServer(int idx)
		{
			// Check we have a server with this id
			if(m_servers.Count <= idx) {
				StringBuilder err = new StringBuilder();
				err.AppendFormat("No client tunneler found with index {0}", idx);
				throw new TunnelError(err.ToString());
			}

			// Check this is actually a client server/tunneler
			Server svr = (Server)m_servers[idx];
			if(!(svr is ClientServer)) {
				StringBuilder err = new StringBuilder();
				err.AppendFormat("Tunneler at index {0} is not a client tunneler", idx);
				throw new TunnelError(err.ToString());
			}

			ClientServer cliSvr = (ClientServer)svr;
			return cliSvr;
		}

#if false
		// Lookup server tunneler with the specified index - Caller holds mutex
		private ServerServer getServerServer(int idx)
		{
			// Check we have a server with this id
			if(m_servers.Count <= idx) {
				StringBuilder err = new StringBuilder();
				err.AppendFormat("No server tunneler found with index {0}", idx);
				throw new TunnelError(err.ToString());
			}

			// Check this is actually a client server/tunneler
			Server svr = (Server)m_servers[idx];
			if(!(svr is ServerServer)) {
				StringBuilder err = new StringBuilder();
				err.AppendFormat("Tunneler at index {0} is not a server tunneler", idx);
				throw new TunnelError(err.ToString());
			}

			ServerServer svrSvr = (ServerServer)svr;
			return svrSvr;
		}
#endif


		// Tell all client tunnelers about a change to testing
		private void activateTesting()
		{
			lock(this) {
				// Iterate over tunnelers
				for(int idx = 0; idx < m_servers.Count; ++idx) {
					Server svr = (Server)m_servers[idx];

					// Start/stop testing on this tunnel
					svr.getTunneler().setServerTesting();
				}
			}
		}


		/********/
		/* Data */
		/********/

		// Reference to globals
		private Globals						m_glob;

		// Process name to use
		private string							m_procName;

		// Should we daemonise?
		private bool							m_daemon;

		// Should we do tunnel testing?
		private bool							m_testing;

		// Pseudo unique local host identifer
		private string							m_hostName;
		private string							m_trueHostName;

		// Debug level
		private uint							m_debugLevel;
		private bool							m_showLog;

		// Our process ID
		private int								m_pid;
		private string							m_pidFile;

		// Default realm
		private string							m_realm;

		// Config file name
		private string							m_configFileName;

		// Local servers and their config
		private ArrayList						m_servers;
		private ArrayList						m_serverConfigs;

		// Data log file stem
		private string							m_dataLogStem;

		// Big-endian platform flag
		private bool							m_bigEndian;

		// Version identifier
		private string							m_version;

	}

}	// namespace HTTPTunnel

