/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * An event to trigger a socket connection to some server.
 *
 * $Id: ConnectEvent.cs,v 1.2 2013/02/17 10:52:29 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;

	public class ConnectEvent : TimedEvent {

		// Construction
		public ConnectEvent(Globals glob, Time time, Connection conn, ConnectProxy proxy)
		: base(glob, time)
		{
			m_conn = conn;
			m_proxy = proxy;
		}


		// Do the connect
		public override void execute()
		{
			// Assume failure by default
			ConnectResultEvent.ConnectResultType result =
				ConnectResultEvent.ConnectResultType.GENERAL_ERROR;
			Sock sock = null;

			try {
				// Launch the connection
				sock = m_proxy.doConnect();
				result = ConnectResultEvent.ConnectResultType.SUCCESS;
			}
			catch(AsynchronousConnect e) {
				// Connect happening asychronously
				m_glob.log().info("Connect pending: " + e.Message);
				result = ConnectResultEvent.ConnectResultType.ASYNC;
			}
			catch(FirewalledConnectError e) {
				// Connect failed
				m_glob.log().info("Connect failed: " + e.Message);
				result = ConnectResultEvent.ConnectResultType.FIREWALLED;
			}
			catch(ConnectError e) {
				m_glob.log().info("Connect failed: " + e.Message);
				result = ConnectResultEvent.ConnectResultType.GENERAL_ERROR;
			}

			// Do we have a result yet?
			if(result != ConnectResultEvent.ConnectResultType.ASYNC) {
				try {
					// Let connection know result
					m_conn.connectResult(result, sock);
				}
				catch(CommsError e) {
					m_glob.log().info("Error handling connect result: " + e.Message);
				}
			}
		}


		/*************/
		/* Accessors */
		/*************/


		// Get ID of connection error is for
		public uint getConnectionID()
		{
			return m_conn.getID();
		}

		// Get IP address connect failed to
		public IPAddr getAddr()
		{
			return m_proxy.getAddr();
		}


		/***********************/
		/* Protected Interface */
		/***********************/


		/********/
		/* Data */
		/********/

		// Connection to connect out on
		protected Connection					m_conn;

		// ConnectProxy handling this connection attempt
		protected ConnectProxy				m_proxy;
	}

} // namespace

