/* (C) 2007 Richard Mills / The Linux Academy Ltd.
 *
 * Perform a blocking connect for a connection in a new thread.
 *
 * $Id: ConnectProxy.cs,v 1.4 2013/02/17 10:52:29 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Collections;
	using System.Net.Sockets;
	using System.Text;

	public class ConnectProxy : EventReceiver {

		// Construction
		public ConnectProxy(Globals glob, Connection conn, Time time, IPAddr addr,
		Enums.Protocol prot, bool ssl) : base(glob)
		{
			#if DEBUG_TUN
			StringBuilder dbg = new StringBuilder();
			dbg.AppendFormat("ConnectProxy::ConnectProxy() - For connection {0}",
				conn.getID());
			m_glob.DPRINTV(dbg.ToString());
			#endif

			m_conn = conn;
			m_addr = addr;
			m_sockType = prot;
			m_sockSSL = ssl;
			m_connecting = false;
			m_asyncSock = null;
			m_asyncPeriod = 0;

			// Schedule connect event
			scheduleConnect(time);
		}


		// We are no longer required
		public virtual void destroyed()
		{
			#if DEBUG_TUN
			StringBuilder dbg = new StringBuilder();
			dbg.AppendFormat("ConnectProxy::~ConnectProxy() - For connection {0}",
				m_conn.getID());
			m_glob.DPRINTV(dbg.ToString());
			#endif

			// Shutdown event scheduling
			if(!shuttingDown())
				shutdownSchedulers();

			lock(this) {
				// Close asynchronous connecting socket
				if(m_connecting && m_asyncSock != null) {
					m_asyncSock.destroyed();
					m_asyncSock = null;
				}
			}
		}


		/*************/
		/* Accessors */
		/*************/

		// Get address of host we are connecting to
		public IPAddr getAddr()
		{
			return m_addr;
		}


		/*************/
		/* Utilities */
		/*************/

		// Schedule a new connect attempt
		public void scheduleConnect(Time time)
		{
			scheduleEvent(new ConnectEvent(m_glob, time, m_conn, this));
		}


		/**********/
		/* Events */
		/**********/

		// Try to do the connect
		public Sock doConnect()
		{
			lock(this) {
				#if DEBUG_TUN
				StringBuilder nfo = new StringBuilder();
				nfo.Append("ConnectProxy::doConnect() - For connection ");
				nfo.AppendFormat("{0} connecting to {1}", m_conn.getID(), m_addr);
				m_glob.DPRINTI(nfo.ToString());
				#endif

				// Waiting for result of asynchronous connect?
				if(m_connecting) {
					return checkAsyncConnect();
				}
				else {
					return initiateConnect();
				}
			}
		}


		/*********************/
		/* Private Interface */
		/*********************/


		/***********/
		/* Methods */
		/***********/

		// Initiate a connect - Caller holds mutex
		private Sock initiateConnect()
		{
			m_glob.DPRINTV("ConnectProxy::initiateConnect()");

			Sock sock = null;

			// Let connection prepare for connect attempt
			m_conn.connectAttempt(m_addr, m_sockType, m_sockSSL);

			// Setup socket
			try {
				// What type of socket do we require?
				if(m_sockSSL) {
					// Create SSL socket for connect attempt
					sock = new SSLSock(m_glob, false);
				}
				else {
					if(m_sockType == Enums.Protocol.TCP) {
						// Create TCP socket for connect attempt
						sock = new TCPSock(m_glob, false);
					}
					else {
						// Create UDP socket
						sock = new UDPSock(m_glob, false);
					}
				}
			}
			catch(SockError e) {
				throw new ConnectError("Failed to initiate connect to server: " + e.Message);
			}

			// Do the connect
			try {
				sock.connect(m_addr);
			}
			catch(SockError e) {
				// Why did it fail to connect?
				if(e.m_errno == SocketError.WouldBlock) {
					// Asynchronous connect
					m_connecting = true;
					m_asyncSock = sock;
					scaleAsyncPeriod(true);
					scheduleConnect(new Time(0, m_asyncPeriod));
					throw new AsynchronousConnect("Connect will block");
				}
				else {
					// Some other failure
					StringBuilder err = new StringBuilder();
					err.AppendFormat("Socket connect() error: {0} ({1})", e.Message, e.m_errno);
					throw new ConnectError(err.ToString());
				}
			}
			catch(CommsError e) {
				throw new ConnectError("Failed to connect to server: " + e.Message);
			}

			return sock;
		}


		// Check whether an asynchronous connect has completed - Caller holds mutex
		private Sock checkAsyncConnect()
		{
			m_glob.DPRINTV("ConnectProxy::checkAsyncConnect()");

			bool resultAvailable = false;
			bool connected = false;
			Sock sock = null;

			try {
				// Get connect result
				m_asyncSock.getConnectResult(ref resultAvailable, ref connected);
				if(resultAvailable) {
					// Did it succeed or fail?
					if(connected) {
						// Success
						m_connecting = false;
						sock = m_asyncSock;
						m_asyncSock = null;
					}
					else {
						// Failed
						throw new ConnectError("Failed to connect to server");
					}
				}
				else {
					// Connect result not yet available, try again shortly
					scaleAsyncPeriod(false);
					scheduleConnect(new Time(0, m_asyncPeriod));
					throw new AsynchronousConnect("Connect not yet completed");
				}
			}
			catch(SockError e) {
				throw new ConnectError("Failed to connect to server: " + e.Message);
			}

			return sock;
		}


		// Increase asynchronous connect period
		private void scaleAsyncPeriod(bool reset)
		{
			if(reset) {
				m_asyncPeriod = MIN_ASYNC_CHECK_PERIOD;
			}
			else {
				m_asyncPeriod = m_asyncPeriod * 2;
				if(m_asyncPeriod > MAX_ASYNC_CHECK_PERIOD) {
					m_asyncPeriod = MAX_ASYNC_CHECK_PERIOD;
				}
			}
		}


		/*************/
		/* Constants */
		/*************/

		// Time between asynchronous connect checks (milli-seconds)
		private const int			MIN_ASYNC_CHECK_PERIOD		= 20;
		private const int			MAX_ASYNC_CHECK_PERIOD		= 2000;


		/********/
		/* Data */
		/********/

		// The connection we proxy for
		private Connection			m_conn;

		// Address we are connecting to
		private IPAddr					m_addr;

		// Type of connection required
		private Enums.Protocol		m_sockType;
		private bool					m_sockSSL;

		// Asynchronous connection details
		private bool					m_connecting;
		private Sock					m_asyncSock;
		private int						m_asyncPeriod;

	}

}	// namespace

