/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Base class for client and server tunnel controllers.
 *
 * $Id: Connection.cs,v 1.4 2013/02/17 10:52:29 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Collections;
	using System.Text;
	using System.Threading;

	public abstract class Connection {

		/*********/
		/* Types */
		/*********/

		// Connection state changes
		public enum ConnectionState { CONNECTED, SOCKSSETUP };


		/***********/
		/* Methods */
		/***********/

		// Construction
		public Connection(Globals glob, Sock sock, Tunneler tunneler)
		{
			m_glob = glob;
			m_tunneler = tunneler;
			m_closing = false;
			m_connected = false;
			m_connecting = false;
			m_everConnected = false;
			m_active = true;
			m_errored = false;
			m_blocked = false;
			m_firstConnectTime = new DateTime(0);
			m_inactiveTime = Time.now();
			m_sock = sock;
			m_reader = null;
			m_writer = null;
			m_carriedTunnels = new ArrayList();
			m_connectFailedCount = 0;
			m_destIP = new IPAddr("", 0);
			m_destProt = Enums.Protocol.TCP;
			m_destSSL = false;

			// Set connection ID
			m_id = getNextConnectionID();
		}

		public Connection(Globals glob, Tunneler tunneler)
		{
			m_glob = glob;
			m_tunneler = tunneler;
			m_closing = false;
			m_connected = false;
			m_connecting = false;
			m_everConnected = false;
			m_active = true;
			m_errored = false;
			m_blocked = false;
			m_firstConnectTime = new DateTime(0);
			m_inactiveTime = Time.now();
			m_sock = null;
			m_reader = null;
			m_writer = null;
			m_carriedTunnels = new ArrayList();
			m_connectFailedCount = 0;
			m_destIP = new IPAddr("", 0);
			m_destProt = Enums.Protocol.TCP;
			m_destSSL = false;

			// Set connection ID
			m_id = getNextConnectionID();
		}


		// We are no longer required
		public virtual void destroyed()
		{
			lock(this) {
				// Ensure we don;t have any remaining tunnels that think they own us
				setInactive();

				// Tidy up socket/reader/writer
				close();
			}
		}


		/**************/
		/* Predicates */
		/**************/

		// Compare connection first connected times
		public class PredConnectTime : IComparer {
			public int Compare(object o1, object o2) {
				int res = 0;
				Connection c1 = (Connection)o1;
				Connection c2 = (Connection)o2;
				if(c1 != null && c2 != null) {
					if(c1.getFirstConnectTime() < c2.getFirstConnectTime())
						res = -1;
					else if(c1.getFirstConnectTime() > c2.getFirstConnectTime())
						res = 1;
				}
				return res;
			}
		}

		// Compare connection IDs
		public class PredID : IComparer {
			public int Compare(object o1, object o2) {
				int res = 0;
				Connection c1 = (Connection)o1;
				Connection c2 = (Connection)o2;
				if(c1 != null && c2 != null) {
					if(c1.getID() < c2.getID())
						res = -1;
					else if(c1.getID() > c2.getID())
						res = 1;
				}
				return res;
			}
		}


		/*************/
		/* Accessors */
		/*************/

		// Are we shutting down?
		public bool closing()
		{
			return m_closing;
		}

		// Are we connected?
		public bool connected()
		{
			return (m_connected && !m_closing && !m_errored);
		}

		// Are we in the process of connecting?
		public bool connecting()
		{
			return m_connecting;
		}

		// Have we ever been connected?
		public bool everConnected()
		{
			return m_everConnected;
		}

		// Are we not waiting to be shutdown?
		public bool active()
		{
			return m_active;
		}

		// Are we blocked?
		public bool blocked()
		{
			return m_blocked;
		}

		// Get time we first connected
		public DateTime getFirstConnectTime()
		{
			return m_firstConnectTime;
		}

		// Get reference to controlling tunneler
		public Tunneler getTunneler()
		{
			return m_tunneler;
		}

		// Get the count of failed connect attempts
		public uint getConnectFailedCount()
		{
			return m_connectFailedCount;
		}

		// Get connection ID
		public uint getID()
		{
			return m_id;
		}

		// Get destination IP address
		public IPAddr getDestIP()
		{
			return m_destIP;
		}

		// Get destination protocol
		public Enums.Protocol getDestProtocol()
		{
			return m_destProt;
		}

		// Get whether SSL is used to destination
		public bool getDestSSL()
		{
			return m_destSSL;
		}

		// Get socket 
		public Sock getSocket()
		{
			return m_sock;
		}


		/************/
		/* Mutators */
		/************/

		// Increment the count of failed connect attempts
		public void incConnectFailedCount()
		{
			lock(this) {
				m_connectFailedCount++;
			}
		}


		// We are waiting to be shutdown
		public void setInactive()
		{
			// Let all carried tunnels know they are no longer carried by us
			lock(this) {
				m_active = false;
				m_inactiveTime = Time.now();
				while(m_carriedTunnels.Count > 0) {
					// Get & remove tunnel from list
					Tunnel tun = (Tunnel)m_carriedTunnels[0];
					m_carriedTunnels.RemoveAt(0);

					// Let tunnel know
					tun.removeConnection(this);
				}
			}
		}


		// Indicate we have suffered an error and shouldn't be used anymore
		public void setErrored()
		{
			lock(this) {
				m_errored = true;

				// Do we have unprocessed data in our reader?
				if(m_reader != null && m_reader.dataRemaining()) {
					#if DEBUG_TUN
					StringBuilder nfo = new StringBuilder();
					nfo.AppendFormat("Connection {0} has residual read data after error", getID());
					m_glob.DPRINTI(nfo.ToString());
					#endif

					// Let tunneler know we have unprocessed data
					m_tunneler.scheduleConnectionError(getID(), Tunnel.TunnelErrorType.UNPROCESSED_DATA_ERROR);
				}
			}
		}

		// Set whether we are blocked
		public void setBlocked(bool blocked)
		{
			lock(this) {
				m_blocked = blocked;
			}
		}


		/********************/
		/* Public Interface */
		/********************/

		// Prepare for a connect attempt to be made to the server
		public virtual void connectAttempt(IPAddr addr, Enums.Protocol prot, bool ssl)
		{
			lock(this) {
				// Flag we are trying to connect
				m_connecting = true;
				m_destIP = addr;
				m_destProt = prot;
				m_destSSL = ssl;

				// Clear up from any previous connection
				close();
			}
		}


		// Let us know the result of a connection attempt
		public virtual void connectResult(ConnectResultEvent.ConnectResultType result,
		Sock sock)
		{
			#if DEBUG_TUN
			StringBuilder nfo = new StringBuilder();
			nfo.AppendFormat("Connection::connectResult() - connection {0} ", getID());
			nfo.AppendFormat(", result: {0}", result);
			m_glob.DPRINTV(nfo.ToString());
			#endif

			// Let tunneler know connection attempt is over
			m_tunneler.scheduleEvent(new ConnectResultEvent(m_glob, m_tunneler, getID(),
				getDestIP(), getDestProtocol(), getDestSSL(), result));

			// No longer trying to connect
			lock(this) {
				m_connecting = false;

				// Save socket result
				m_sock = sock;

				// Success?
				if(result == ConnectResultEvent.ConnectResultType.SUCCESS) {
					// Create required reader/writer
					createReaderWriter();

					// Let tunneler know we have connected
					m_tunneler.scheduleEvent(new ConnectionStateEvent(m_glob, m_tunneler,
						getID(), ConnectionState.CONNECTED));
				}
			}
		}


		// Attempt to read from our connected socket
		public virtual void read()
		{
			//#if DEBUG_TUN
			//StringBuilder nfo = new StringBuilder();
			//nfo.AppendFormat("Connection::read() - connection {0} ", getID());
			//m_glob.DPRINTV(nfo.ToString());
			//#endif

			lock(this) {
				// Can't read unless connected & active
				if(!connected())
					throw new CommsError("Read attempt while not connected");

				// Instruct reader to do the read
				m_reader.read();
			}
		}


		// Attempt to write to our connected socket
		public virtual void write(MsgParms parms)
		{
			lock(this) {
				// Can't write unless connected
				if(!connected())
					throw new CommsError("Write attempt while not connected");

				// Insruct writer to do the write
				m_writer.write(parms);
			}
		}

		// Attempt to flush any buffered data to our connected socket
		public virtual void flush(bool retry)
		{
			lock(this) {
				// Can't write unless connected
				if(!connected())
					throw new CommsError("Flush attempt while not connected");

				// Instruct writer to flush
				m_writer.flush(retry);
			}
		}


		// Send some special out-of-band data
		public void writeOutOfBand(RawBuff outOfBand)
		{
			m_glob.DPRINTV("Connection::writeOutOfBand()");

			lock(this) {
				// Can't write unless connected
				if(!connected())
					throw new CommsError("Out-of-band write attempt while not connected");

				// Insruct writer to do the write
				m_writer.writeOutOfBand(outOfBand);
			}
		}


		// Add tunnel from list of carried tunnels
		public void addCarriedTunnel(Tunnel tun)
		{
			lock(this) {
				// Check we don't already know
				if(m_carriedTunnels.IndexOf(tun) == -1) {
					// Record new carried tunnel
					m_carriedTunnels.Add(tun);
				}
			}
		}

		// Remove tunnel from list of carried tunnels
		public void removeCarriedTunnel(Tunnel tun)
		{
			lock(this) {
				// Find tunnel
				int index = m_carriedTunnels.IndexOf(tun);
				if(index != -1) {
					m_carriedTunnels.RemoveAt(index);
				}

				// Do we still carry anything?
				if(m_carriedTunnels.Count == 0) {
					// Schedule our shutdown
					m_tunneler.scheduleEvent(new ShutdownConnectionEvent(m_glob, m_id,
						m_tunneler));
				}
			}
		}


		// Tell carried tunnels about some error
		public void tellTunnelsAboutError(Tunnel.TunnelErrorType errType)
		{
			lock(this) {
				// Let all carried tunnels know
				for(int i = 0; i < m_carriedTunnels.Count; ++i) {
					((Tunnel)m_carriedTunnels[i]).recordConnectionError(errType);
				}
			}
		}

		// Tell carried tunnels we have connected
		public void tellTunnelsAboutConnect()
		{
			lock(this) {
				// Let all carried tunnels know
				for(int i = 0; i < m_carriedTunnels.Count; ++i) {
					((Tunnel)m_carriedTunnels[i]).connectionConnected(this);
				}
			}
		}


		// Determine whether this connection can currently be shutdown cleanly
		public bool canBeShutdown()
		{
			lock(this) {
				bool actv = active();
				bool conntng = connecting();
				bool queued = false;
				bool delayed = false;

				// Must first be inactive & not trying to connect
				bool res = !actv && !conntng;

				// Check writer
				if(res && m_writer != null) {
					// Does our writer still have queued data to send?
					queued = !m_writer.flush(false);
					res &= !queued;
				}

				// Leave a small delay after becoming inactive
				if(res) {
					Time delayUntil = new Time(m_inactiveTime);
					delayUntil.addInterval(0, DELAY_CLOSE);
					delayed = !delayUntil.elapsed();
					res &= !delayed;
				}

				#if DEBUG_TUN
				StringBuilder dbg = new StringBuilder();
				dbg.AppendFormat("Connection::canBeShutdown() - connection {0}: {1} ",
					getID(), res);
				if(!res) {
					if(actv) dbg.Append("Actv ");
					if(conntng) dbg.Append("Cntg ");
					if(queued) dbg.Append("Qud ");
					if(delayed) dbg.Append("Dld ");
				}
				m_glob.DPRINTV(dbg.ToString());
				#endif

				return res;
			}
		}


		/***********************/
		/* Protected Interface */
		/***********************/


		/***********/
		/* Methods */
		/***********/

		// Construct reader/writer
		protected abstract void createReaderWriter();


		// Replace reader
		protected void replaceReader(Reader reader)
		{
			lock(this) {
				// Save new reader
				m_reader = reader;
			}
		}

		// Replace writer
		protected void replaceWriter(Writer writer)
		{
			lock(this) {
				// Save new writer
				m_writer = writer;
			}
		}


		// We are now connected
		protected void setConnected()
		{
			lock(this) {
				// Set connected flags etc.
				m_connected = true;
				m_everConnected = true;
				m_firstConnectTime = DateTime.Now;
			}
		}


		/********/
		/* Data */
		/********/

		// Globals
		protected Globals				m_glob;

		// Our controlling tunneler
		protected Tunneler			m_tunneler;

		// Connection state flags
		protected bool					m_closing;
		protected bool					m_connected;
		protected bool					m_connecting;
		protected bool					m_everConnected;
		protected bool					m_active;
		protected bool					m_errored;
		protected bool					m_blocked;

		// Time we first connected
		protected DateTime			m_firstConnectTime;

		// Time we are made inactive
		protected Time					m_inactiveTime;

		// Socket we read/write with
		protected Sock					m_sock;

		// Reader for processing read data
		protected Reader				m_reader;

		// Writer for processing and writing data
		protected Writer				m_writer;

		// A list of tunnels that we carry
		protected ArrayList			m_carriedTunnels;


		/*********************/
		/* Private Interface */
		/*********************/


		/***********/
		/* Methods */
		/***********/

		// Tidy up existing socket connection
		private void close()
		{
			lock(this) {
				// Indicate not connected
				m_connected = false;

				// Free any existing reader/writer
				m_reader = null;
				m_writer = null;

				// Close any existing socket
				if(m_sock != null) {
					m_sock.destroyed();
					m_sock = null;
				}
			}
		}


		// Get next connection ID
		private static uint getNextConnectionID()
		{
			uint id = 0;

			lock(c_idCountMtx) {
				id = c_idCount++;
				if(c_idCount == 0)
					c_idCount = 1;
			}

			return id;
		}


		/*************/
		/* Constants */
		/*************/

		// Delay before allowing connection to be closed
		private const int				DELAY_CLOSE			= 1000;


		/********/
		/* Data */
		/********/

		// Count of various errors
		private uint					m_connectFailedCount;

		// Our destination address
		private IPAddr					m_destIP;
		private Enums.Protocol		m_destProt;
		private bool					m_destSSL;

		// Unique connection ID
		private static uint			c_idCount = 1;
		private static Mutex			c_idCountMtx = new Mutex();
		private uint					m_id;

	}

}  // namespace

