/* (C) 2007 Richard Mills / The Linux Academy Ltd.
 *
 * Manage all blocking Connection operations without requiring a
 * thread per Connection object.
 *
 * $Id: ConnectionManager.cs,v 1.5 2013/02/17 10:52:30 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Collections;
	using System.Net.Sockets;
	using System.Text;
	using System.Threading;

	public class ConnectionManager : EventReceiver {

		// Construction
		public ConnectionManager(Globals glob) : base(glob)
		{
			m_glob.DPRINTV("ConnectionManager::ConnectionManager()");

			m_connections = new Hashtable();
			m_connectProxies = new Hashtable();
			m_pollShutDown = false;
			m_pollShutdownMtx = new Mutex();

			// Launch connection polling thread
			m_pollThread = new Thread(new ThreadStart(pollThread));
			m_pollThread.Start();
		}

		// We are no longer required
		public virtual void destroyed()
		{
			m_glob.DPRINTV("ConnectionManager::~ConnectionManager()");

			// Shutdown event scheduling
			if(!shuttingDown())
				shutdownSchedulers();

			// Take mutexes
			Monitor.Enter(this);
			Monitor.Enter(m_pollShutdownMtx);

			// Indicate to the polling thread we are shutting down
			m_pollShutDown = true;

			// Signal the polling thread to wake up
			Monitor.Pulse(this);
			Monitor.Exit(this);

			// Wait until the scheduler returns
			Monitor.Wait(m_pollShutdownMtx);
			Monitor.Exit(m_pollShutdownMtx);

			lock(this) {
				// Remove any remaining managed connections
				while(m_connections.Count > 0) {
					IDictionaryEnumerator en = m_connections.GetEnumerator();
					if(en.MoveNext())
						delConnection((Connection)en.Value);
				}
			}
		}


		/********************/
		/* Public Interface */
		/********************/


		// Add a new connection to the managed list
		public void addConnection(Connection conn)
		{
			lock(this) {
				// Get connection ID
				uint connID = conn.getID();

				#if DEBUG_TUN
				// Check we don't already know about this connection
				if(m_connections.ContainsKey(connID)) {
					StringBuilder err = new StringBuilder();
					err.Append("ConnectionManager::addConnection() - ");
					err.AppendFormat("Already managing connection ID {0}", connID);
					m_glob.DPRINTE(err.ToString());
				}
				#endif

				// Add connection to managed list
				m_connections[connID] = conn;
			}
		}

		// Remove a connection from the managed list
		public void delConnection(Connection conn)
		{
			lock(this) {
				// Find connection
				uint connID = conn.getID();
				if(m_connections.ContainsKey(connID)) {
					// Do we have an active connect proxy for this connection?
					if(m_connectProxies.ContainsKey(connID)) {
						// Stop connect proxy
						ConnectProxy proxy = (ConnectProxy)m_connectProxies[connID];
						proxy.destroyed();

						// Remove connect proxy from list
						m_connectProxies.Remove(connID);
					}

					// Remove connection from managed list
					m_connections.Remove(connID);
				}
				else {
					StringBuilder wrn = new StringBuilder();
					wrn.Append("ConnectionManager::delConnection() - ");
					wrn.AppendFormat("Unknown connection ID {0}", connID);
					m_glob.log().warn(wrn.ToString());
				}
			}
		}


		// Setup a proxy to connect out for a connection
		public void addProxyConnect(Connection conn, Time time, IPAddr addr, Enums.Protocol prot,
		bool ssl)
		{
			// Get connection ID
			uint connID = conn.getID();

			#if DEBUG_TUN
			StringBuilder dbg = new StringBuilder();
			dbg.AppendFormat("ConnectionManager::addProxyConnect() - For connection {0}", connID);
			m_glob.DPRINTV(dbg.ToString());
			#endif

			lock(this) {
				#if DEBUG_TUN
				// Check we actually know about this connection
				if(!m_connections.ContainsKey(connID)) {
					m_glob.DPRINTE("Connection not found in our managed list");
				}
				#endif

				// Check whether we already have a proxy for this connection
				if(m_connectProxies.ContainsKey(connID)) {
					// Reuse this one
					ConnectProxy proxy = (ConnectProxy)m_connectProxies[connID];
					proxy.scheduleConnect(time);
				}
				else {
					// Create connect proxy
					ConnectProxy proxy = new ConnectProxy(m_glob, conn, time, addr, prot, ssl);
					m_connectProxies[connID] = proxy;
				}
			}
		}

		// Clear up after a completed proxy connect
		public void delProxyConnect(Connection conn)
		{
			// Get connection ID
			uint connID = conn.getID();

			#if DEBUG_TUN
			StringBuilder dbg = new StringBuilder();
			dbg.AppendFormat("ConnectionManager::delProxyConnect() - For connection ", connID);
			m_glob.DPRINTV(dbg.ToString());
			#endif

			lock(this) {
				#if DEBUG_TUN
				// Check we actually know about this connection
				if(!m_connections.ContainsKey(connID)) {
					m_glob.DPRINTE("Connection not found in our managed list");
				}
				// Check we know about proxy for this connection
				if(!m_connectProxies.ContainsKey(connID)) {
					m_glob.DPRINTE("Unknown proxy for this connection");
				}
				#endif

				// Clear up proxy
				ConnectProxy proxy = (ConnectProxy)m_connectProxies[connID];
				proxy.destroyed();
				m_connectProxies.Remove(connID);
			}
		}


		// Display some audit stats
		public void audit()
		{
			m_glob.DPRINTV("ConnectionManager::audit()");

			lock(this) {
				// Number of managed connections
				string nfo = "Number of managed connections: " + m_connections.Count;
				m_glob.DPRINTV(nfo);

				// Number of active connect proxies
				nfo = "Number of active connect proxies: " + m_connectProxies.Count;
				m_glob.DPRINTV(nfo);
			}
		}


		/**********/
		/* Events */
		/**********/


		// Try reading from our managed connections
		public void readConnections()
		{
			lock(this) {
				// How many connected & active Connections should we poll for?
				IDictionaryEnumerator cit = m_connections.GetEnumerator();
				while(cit.MoveNext()) {
					// Worth checking this connection?
					Connection conn = (Connection)cit.Value;
					if(conn.connected() && conn.active() && !conn.blocked()) {
						try {
							// Get this connections socket
							Sock sock = conn.getSocket();

							// Read everything we can from this socket
							bool ready = false, pollError = false, pollHup = false;
							do {
								try {
									// Poll socket
									if(sock != null) {
										ready = sock.pollRead(0, ref pollError, ref pollHup);
									}
								}
								catch(SockError) {
									StringBuilder err = new StringBuilder();
									err.Append("Error polling socket for connection id ");
									err.AppendFormat("{0}", conn.getID());
									throw new CommsError(err.ToString());
								}

								if(ready) {
									// Descriptor ready for reading
									try {
										conn.read();
									}
									catch(NotReadyError e) {
										m_glob.log().info("Socket not really ready for reading: " + e.Message);
										ready = false;
									}

									// Was there also an error condition?
									if(pollError) {
										if(pollHup) {
											StringBuilder err = new StringBuilder();
											err.Append("Poll returned HUP for socket read ");
											err.AppendFormat("for connection id {0}", conn.getID());
											throw new ConnectionClosed(err.ToString());
										}
										else {
											StringBuilder err = new StringBuilder();
											err.Append("Poll returned error for socket read ");
											err.AppendFormat("for connection id {0}", conn.getID());
											throw new CommsError(err.ToString());
										}
									}
								}
							} while(ready);
						}
						catch(BadResponseCodeError e) {
							// Something returned an unexpected HTTP error response code
							m_glob.log().info("Bad HTTP response code: " + e.Message);
							conn.setErrored();
							if(!conn.closing()) {
								// Schedule error reporting event
								Tunneler tunneler = conn.getTunneler();
								tunneler.scheduleConnectionError(conn.getID(),
									Tunnel.TunnelErrorType.BAD_RESPONSE_ERROR);
							}
						}
						catch(ConnectionClosed e) {
							m_glob.log().info("Connection closed: " + e.Message);
							conn.setErrored();
							if(!conn.closing() && conn.active()) {
								// Schedule error reporting event
								Tunneler tunneler = conn.getTunneler();
								tunneler.scheduleConnectionError(conn.getID(),
									Tunnel.TunnelErrorType.CONNECTION_CLOSED_ERROR);
							}
						}
						catch(CommsError e) {
							// A terminal error while reading
							m_glob.log().info("Comms error: " + e.Message);
							conn.setErrored();
							if(!conn.closing() && conn.active()) {
								// Schedule error reporting event
								Tunneler tunneler = conn.getTunneler();
								tunneler.scheduleConnectionError(conn.getID(),
									Tunnel.TunnelErrorType.GENERAL_COMMS_ERROR);
							}
						}
					}
				}
			}
		}


		/*********************/
		/* Private Interface */
		/*********************/


		/*************/
		/* Constants */
		/*************/

		// Time (in miliseconds) between read events
		public const int				POLL_PERIOD			= 5;


		/***********/
		/* Methods */
		/***********/

		// Connection polling thread
		private void pollThread()
		{
			// Take event queue mutex
			Monitor.Enter(this);

			// Run until told to stop
			while(!m_pollShutDown) {
				// Look for connections with active sockets
				if(m_connections.Count > 0 && !isScheduled<ReadEvent>()) {
					ArrayList sockFDs = new ArrayList();
					IDictionaryEnumerator cit = m_connections.GetEnumerator();
					while(cit.MoveNext()) {
						Connection conn = (Connection)cit.Value;
						if(conn.connected() && conn.active() && !conn.blocked()) {
							Sock sock = conn.getSocket();
							if(sock != null)
								sockFDs.Add(sock);
						}
					}

					// Poll sockets
					for(int pit = 0; pit < sockFDs.Count; ++pit) {
						// Check this socket for reading or error
						Sock sock = (Sock)sockFDs[pit];
						try {
							bool pollError = false, pollHup = false;
							if(sock.pollRead(0, ref pollError, ref pollHup)) {
								// Events pending
								scheduleEvent(new ReadEvent(m_glob, new Time(0,0), this));
								break;
							}
						}
						catch(SockError) {
							// Poll error
							m_glob.log().error("ConnectionManager::pollThread() - Error polling socket");
						}
					}
				}

				// Sleep
				TimeSpan sleepTime = new TimeSpan(0,0,0,0,POLL_PERIOD);
				Monitor.Wait(this, sleepTime);
			}

			// Signal we have finished
			Monitor.Enter(m_pollShutdownMtx);
			Monitor.Pulse(m_pollShutdownMtx);
			Monitor.Exit(m_pollShutdownMtx);
			Monitor.Exit(this);

			m_glob.DPRINTV("ConnectionManager::pollThread() - Returning");
		}


		/********/
		/* Data */
		/********/

		// The managed connections
		private Hashtable				m_connections;

		// Active connect proxies
		private Hashtable				m_connectProxies;

		// Polling thread
		private Thread					m_pollThread;
		private bool					m_pollShutDown;
		private Mutex					m_pollShutdownMtx;

	}

}  // namespace

