/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Base class of a TCP/IP socket server to listen for incoming
 * connections.
 *
 * $Id: EventReceiver.cs,v 1.3 2013/03/24 09:55:58 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Collections;
	using System.Text;
	using System.Threading;

	public class EventReceiver {

		// Constructor
		public EventReceiver(Globals glob)
		{
			m_glob = glob;
			m_events = new ArrayList();
			m_eventsRunning = true;
			m_eventsShutdown = new Mutex();

#if false
			instance = g_instance++;
#endif

			// Launch main event scheduler
			m_eventsThread = new Thread(new ThreadStart(processEvents));
			m_eventsThread.Start();
		}


		/********************/
		/* Public Interface */
		/********************/

		// Are we shutting down?
		public bool shuttingDown()
		{
			return (!m_eventsRunning);
		}


		// Schedule a new timed event
		public void scheduleEvent(Event evnt)
		{
			try {
				// Take event queue lock
				Monitor.Enter(m_events);

				// Are we allowed to schedule new events?
				if(m_eventsRunning) {
					// Search for right place for event in queue
					int lower = 0;
					int upper = m_events.Count;
					while(upper > lower) {
						// Compare to event at this index
						int idx = lower + ((upper - lower) / 2);
						Event qevnt = (Event)m_events[idx];

						if(evnt < qevnt) {
							// Either goes here or before this index
							upper = idx;
						}
						else {
							// Goes after this index
							lower = idx + 1;
						}
					}

					// Insert event into queue
					m_events.Insert(lower, evnt);

					// Signal event scheduling thread it has some work
					Monitor.Pulse(m_events);
				}
				else {
					// Can't schedule new event
					m_glob.log().warn("Can't schedule event; we are shutting down");
				}
			}
			finally {
				// Release event queue lock
				Monitor.Exit(m_events);
			}
		}


		// Reschedule an event
		public void rescheduleEvent<T, I>(I identity, Time rtime) where T : IdentifyEvent<I>
		{
			lock(m_events) {
				// Look for matching event
				int i = findEvent<T, I>(identity);
				if(i >= 0) {
					// Cancel this event
					Event evnt = (Event)m_events[i];
					m_events.RemoveAt(i);

					// Reschedule it for new time
					evnt.setRunTime(rtime);
					scheduleEvent(evnt);
				}
			}
		}


		// Cancel events of some type
		public void cancelEvents<T, I>(I identity) where T : IdentifyEvent<I>
		{
			bool keepLooking = false;

			// Look through event queue
			lock(m_events) {
				do {
					keepLooking = false;

					// Look for event
					int i = findEvent<T, I>(identity);
					if(i >= 0) {
						// Cancel this event
						m_events.RemoveAt(i);
						keepLooking = true;
					}
				} while(keepLooking);
			}
		}

		// Cancel all events of some type
		public void cancelAllEvents<T>()
		{
			bool keepLooking = false;

			// Look through event queue
			lock(m_events) {
				do {
					keepLooking = false;

					// Look for event
					int i = findEvent<T>();
					if(i >= 0) {
						// Cancel this event
						m_events.RemoveAt(i);
						keepLooking = true;
					}
				} while(keepLooking);
			}
		}

		// Cancel and return an event of some type
		public T cancelAndGetEvent<T, I>(I identity) where T : IdentifyEvent<I>
		{
			T evnt = default(T);

			// Look through event queue
			lock(m_events) {
				int i = findEvent<T, I>(identity);
				if(i >= 0) {
					// Cancel and return this event
					evnt = (T)m_events[i];
					m_events.RemoveAt(i);
				}
			}

			return evnt;
		}


		// Return whether some event is currently scheduled
		public bool isScheduled<T>()
		{
			bool res = false;

			// Iterate over all queued events
			for(int i = 0; i < m_events.Count; ++i) {
				// Event of the right type?
				if(m_events[i] is T) {
					res = true;
					break;
				}
			}

			return res;
		}

		// Return whether some identified event is currently scheduled
		public bool isScheduled<T, I>(I identity) where T : IdentifyEvent<I>
		{
			bool res = false;

			// Look for plain event
			lock(m_events) {
				res = (findEvent<T, I>(identity) >= 0);
			}

			return res;
		}


		/***********************/
		/* Protected Interface */
		/***********************/

		// Instruct the event scheuling threads to terminate
		protected void shutdownSchedulers()
		{
			// Main scheduler
			shutdownScheduler();
		}


		/********/
		/* Data */
		/********/

		// Reference to the globals
		protected Globals							m_glob;


		/*********************/
		/* Private Interface */
		/*********************/

		// Process events on the event queue - Thread entry point
		private void processEvents()
		{
			// Take event queue mutex
			Monitor.Enter(m_events);

			try {
				// Run until told to stop
				while(m_eventsRunning) {

					// Process next event on the queue
					if(m_events.Count > 0) {
						// Get first event
						Event evnt = (Event)m_events[0];

						// Is it time to run this event?
						Time runTime = evnt.getRunTime();
						if(runTime.elapsed()) {
							// Remove & run the event
							m_events.RemoveAt(0);

							Monitor.Exit(m_events);
							evnt.execute();
							Monitor.Enter(m_events);
						}
						else {
							// Wait until event scheduled time or until new event scheduled
							int waitMillis = runTime.getMillisUntilElapsed();
							Monitor.Wait(m_events, waitMillis);
						}
					}
					else {
						// Should we accept any more events?
						if(m_eventsRunning) {
							// Wait for more events
							Monitor.Wait(m_events);
						}
					}
				}

				// Clear up any still pending events
				m_events.Clear();

				// Signal we have finished
				Monitor.Enter(m_eventsShutdown);
				Monitor.Pulse(m_eventsShutdown);
				Monitor.Exit(m_eventsShutdown);
			}
			catch(Exception e) {
				m_glob.log().error("Error while processing events: " + e.Message);
				m_glob.log().error(e.StackTrace);
			}

			// Ensure we release events lock
			try {
				Monitor.Exit(m_events);
			}
			catch(Exception) {}
		}


		// Shutdown the event schedule
		private void shutdownScheduler()
		{
			// Take event queue mutex
			Monitor.Enter(m_events);

			// Take events running mutex
			Monitor.Enter(m_eventsShutdown);

			// Indicate to the scheduler we are shutting down
			m_eventsRunning = false;

			// Signal the scheduler to empty its event queue
			Monitor.Pulse(m_events);
			Monitor.Exit(m_events);

			// Wait until the scheduler returns
			Monitor.Wait(m_eventsShutdown);
			Monitor.Exit(m_eventsShutdown);
		}


		// Find events of a particular type - Caller holds mutex
		private int findEvent<T, I>(I identity) where T : IdentifyEvent<I>
		{
			// Iterate over all queued events
			int i;
			for(i = 0; i < m_events.Count; ++i) {
				// Event of the right type?
				if(m_events[i] is T) {
					// Test identity
					T evnt = (T)m_events[i];
					if(evnt.identify(identity))
						break;
				}
			}

			// Found one?
			if(i == m_events.Count)
				i = -1;

			return i;
		}

		// Find any event of a particular type - Caller holds mutex
		private int findEvent<T>()
		{
			// Iterate over all queued events
			int i;
			for(i = 0; i < m_events.Count; ++i) {
				// Event of the right type?
				if(m_events[i] is T) {
					break;
				}
			}

			// Found one?
			if(i == m_events.Count)
				i = -1;

			return i;
		}

#if false
		private void printEventDetails(Event evnt, string prefix)
		{
			StringBuilder postfix = new StringBuilder();
			if(evnt is PollTimeoutEvent) {
				PollTimeoutEvent pte = (PollTimeoutEvent)evnt;
				postfix.AppendFormat("for tunnel {0}", pte.getTID().ToString());
			}

			StringBuilder str = new StringBuilder();
			str.AppendFormat("{0} event {1} {2}", prefix, evnt.GetType().Name, postfix.ToString());
			printDetails(str.ToString());
		}

		private void printDetails(string prefix)
		{
			StringBuilder str = new StringBuilder();
			str.AppendFormat("*** EventReceiver {0} - {1}", instance, prefix);
			m_glob.DPRINTV(str.ToString());
		}

		private void printAllEvents(string prefix)
		{
			StringBuilder str = new StringBuilder();
			str.AppendFormat("{0} contents of scheduled events are:", prefix);
			printDetails(str.ToString());

			lock(m_events) {
				for(int i = 0; i < m_events.Count; ++i) {
					Event e = (Event)m_events[i];
					printEventDetails(e, "    ");
				}
			}
		}
#endif


		/********/
		/* Data */
		/********/

		// An event queue
		private ArrayList							m_events;
		private Thread								m_eventsThread;
		private bool								m_eventsRunning;
		private Mutex								m_eventsShutdown;

#if false
		private int instance;
		private static int g_instance = 0;
#endif

	}

}	// namespace


