/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Exceptions thrown by the application.
 *
 * $Id: Exceptions.cs,v 1.4 2009/03/20 11:55:49 rich Exp $
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */


namespace HTTPTunnel
{

	using System;
	using System.Net.Sockets;


	/**************/
	/* Base class */
	/**************/

	public class TunnelError : ApplicationException {
		public TunnelError(string e) : base(e) {}
	}


	/************************/
	/* Configuration errors */
	/************************/

	public class ConfigError : TunnelError {
		public ConfigError(string e) : base(e) {}
	}


	/*************************/
	/* Communications errors */
	/*************************/

	public class CommsError : TunnelError {
		public CommsError(string e) : base(e) {}
	}

	// Server error
	public class ServerError : CommsError {
		public ServerError(string e) : base(e) {}
	}

	// General data reading error
	public class ReadError : CommsError {
		public ReadError(string e) : base(e) {}
	}

	// Connection closed
	public class ConnectionClosed : ReadError {
		public ConnectionClosed(string e) : base(e) {}
	}

	// Received some error response code
	public class BadResponseCodeError : ReadError {
		public BadResponseCodeError(string e, uint respCode) : base(e) { m_respCode = respCode; }
		public uint m_respCode;
	}

	// General data writing error
	public class WriteError : CommsError {
		public WriteError(string e) : base(e) {}
	}

	// Connect failed error
	public class ConnectError : CommsError {
		public ConnectError(string e) : base(e) {}
	}

	// Firewalled connect failed error
	public class FirewalledConnectError : ConnectError {
		public FirewalledConnectError(string e) : base(e) {}
	}

	// Asynchronous connect result
	public class AsynchronousConnect : ConnectError {
		public AsynchronousConnect(string e) : base(e) {}
	}

	// Resending error
	public class ResendError : CommsError {
		public ResendError(string e) : base(e) {}
	}

	// Not ready for read/write error
	public class NotReadyError : CommsError {
		public NotReadyError(string e) : base(e) {}
	}


	/*****************/
	/* Socket errors */
	/*****************/

	public class SockError : TunnelError {
		public SockError(string e, SocketError en) : base(e) { m_errno = en; }
		public SocketError m_errno;
	}


	/*********************/
	/* Processing errors */
	/*********************/

	public class ProcessingError : TunnelError {
		public ProcessingError(string e) : base(e) {}
	}

	// Invalid ID
	public class InvalidID : ProcessingError {
		public InvalidID(string e) : base(e) {}
	}

	// Invalid tunnel ID string
	public class InvalidTunnelID : InvalidID {
		public InvalidTunnelID(string e) : base(e) {}
	}

	// Invalid tunneler ID string
	public class InvalidTunnelerID : InvalidID {
		public InvalidTunnelerID(string e) : base(e) {}
	}

	// No data for the requested operation
	public class NoDataError : ProcessingError {
		public NoDataError(string e) : base(e) {}
	}

	// Authentication processing
	public class AuthProcessingError : ProcessingError {
		public AuthProcessingError(string e) : base(e) {}
	}


	/******************/
	/* Parsing errors */
	/******************/

	public class ParsingError : TunnelError {
		public ParsingError(string e) : base(e) {}
	}

	// Error constructing HTTP message from raw data stream
	public class HTTPMsgParsingError : ParsingError {
		public HTTPMsgParsingError(string e, uint line)
		: base(e) { m_line = line; }
		public uint m_line;
	}

	// Wrong message type parsing error
	public class HTTPMsgWrongMsgType : HTTPMsgParsingError {
		public HTTPMsgWrongMsgType(string e, uint line) : base(e, line) {}
	}

	// Error in message payload content
	public class ContentParsingError : HTTPMsgParsingError {
		public ContentParsingError(string e, uint line) : base(e, line) {}
	}

	// Not enough data in buffer for reported content
	public class NotEnoughDataParsingError : ContentParsingError {
		public NotEnoughDataParsingError(string e, uint reqd, uint recvd, uint line)
		: base(e, line) { m_reqd = reqd; m_recvd = recvd; }
		public uint m_reqd, m_recvd;
	}

	// Reported encoded content length doesn't equal received encoded length
	public class BadContentLengthParsingError : ContentParsingError {
		public BadContentLengthParsingError(string e, uint line) : base(e, line) {}
	}


	/*********************/
	/* Management errors */
	/*********************/

	public class ManagementError : TunnelError {
		public ManagementError(string e) : base(e) {}
	}

	// Failed to find some managed resource
	public class LookupManagementError : ManagementError {
		public LookupManagementError(string e) : base(e) {}
	}


	/*****************/
	/* Memory errors */
	/*****************/

	public class MemoryError : TunnelError {
		public MemoryError(string e) : base(e) {}
	}


	/****************/
	/* Event errors */
	/****************/

	public class EventError : TunnelError {
		public EventError(string e) : base(e) {}
	}


	/********************************/
	/* Server authentication errors */
	/********************************/

	public class AuthError : TunnelError {
		public AuthError(string e) : base(e) {}
	}

	// Error in accessing an authentication database
	public class DBServerAuthError : AuthError {
		public DBServerAuthError(string e) : base(e) {}
	}


	/******************/
	/* General errors */
	/******************/

	public class GeneralError : TunnelError {
		public GeneralError(string e) : base(e) {}
	}

	// Host address info lookup error
	public class AddrLookupError : GeneralError {
		public AddrLookupError(string e) : base(e) {}
	}

}		// namespace HTTPTunnel


