/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Apply firewalling rules for the tunnel server.
 *
 * $Id: Firewall.cs,v 1.2 2013/02/17 10:52:30 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Collections;
	using System.IO;
	using System.Text;

	public class Firewall {

		// Construction
		public Firewall(Globals glob, string ips, string ports)
		{
			m_glob = glob;
			m_denyIPs = new ArrayList();
			m_denyWildcardIPs = new ArrayList();
			m_denyPorts = new ArrayList();

			// Add initial blocked IPs/Ports
			if(ips.Length > 0)
				denyIPs(ips);
			if(ports.Length > 0)
				denyPorts(ports);
		}


		/********************/
		/* Public Interface */
		/********************/

		// Is a requested tunnel destination allowed?
		public bool isAllowed(string dest)
		{
			bool allowed = true;

			// Split address into IP and port
			string[] parms = dest.Split(';');
			if(parms.Length == 2) {
				string ip = parms[0];
				uint port = UInt32.Parse(parms[1]);
				allowed = isAllowed(ip, port);
			}
			else {
				// Deny badly formatted destination
				allowed = false;
			}

			return allowed;
		}

		public bool isAllowed(string ip, uint port)
		{
			bool allowed = true;

			// Port banned?
			if(portBanned(port)) {
				allowed = false;
				StringBuilder nfo = new StringBuilder();
				nfo.AppendFormat("Destination port {0} in denied list", port);
				m_glob.log().info(nfo.ToString());
			}
			else if(ipBanned(ip)) {
				allowed = false;
				StringBuilder nfo = new StringBuilder();
				nfo.AppendFormat("Destination IP address {0} in denied list", ip);
				m_glob.log().info(nfo.ToString());
			}

			return allowed;
		}


		// Add new banned IPs
		public void denyIPs(string dest)
		{
			// Split list
			string[] ips = dest.Split(',');

			lock(this) {
				for(int i = 0; i < ips.Length; ++i) {
					// Basic or wildcarded address?
					string ip = ips[i];
					if(ip.EndsWith("*")) {
						// Trim *
						char[] trim = { '*' };
						ip = ip.TrimEnd(trim);

						// Ignore duplicates
						if(m_denyWildcardIPs.IndexOf(ip) == -1)
							m_denyWildcardIPs.Add(ip);
					}
					else {
						// Resolve IP address
						ip = resolve(ip);
						if(ip.Length > 0) {
							// Ignore duplicates
							if(m_denyIPs.IndexOf(ip) == -1)
								m_denyIPs.Add(ip);
						}
					}
				}
			}
		}

		// Add new banned ports
		public void denyPorts(string dest)
		{
			// Split list
			string[] ports = dest.Split(',');

			lock(this) {
				for(int i = 0; i < ports.Length; ++i) {
					int port = Int32.Parse(ports[i]);
					if(port > 0 && port <= 65535) {
						// Ignore duplicates
						if(m_denyPorts.IndexOf(port) == -1)
							m_denyPorts.Add(port);
					}
				}
			}
		}


		// Serialize to a file
		public void serialize(StreamWriter file)
		{
			lock(this) {
				// Firewalled IPs
				if(m_denyIPs.Count > 0 || m_denyWildcardIPs.Count > 0) {
					file.Write("denyips ");
					bool first = true;
					for(int ipit = 0; ipit < m_denyIPs.Count; ++ipit) {
						if(!first)
							file.Write(",");
						else
							first = true;
						file.Write(m_denyIPs[ipit].ToString());
					}
					for(int wcit = 0; wcit < m_denyWildcardIPs.Count; ++wcit) {
						if(!first)
							file.Write(",");
						else
							first = true;
						file.Write(m_denyWildcardIPs[wcit].ToString() + "*");
					}
					file.WriteLine();
				}

				// Firewalled ports
				if(m_denyPorts.Count > 0) {
					file.Write("denyports ");
					bool first = true;
					for(int pit = 0; pit < m_denyPorts.Count; ++pit) {
						if(!first)
							file.Write(",");
						else
							first = true;
						file.Write(m_denyPorts[pit].ToString());
					}
					file.WriteLine();
				}
			}
		}


		/*********************/
		/* Private Interface */
		/*********************/


		/***********/
		/* Methods */
		/***********/

		// Is an IP address banned?
		private bool ipBanned(string ip)
		{
			bool banned = false;

			// Resolve to numeric address
			ip = resolve(ip);
			if(ip.Length > 0) {
				lock(this) {
					// First check simple list
					if(m_denyIPs.IndexOf(ip) != -1) {
						banned = true;
					}
					else {
						// Check wildcard list
						for(int ipit = 0; ipit < m_denyWildcardIPs.Count; ++ipit) {
							string wcip = (string)m_denyWildcardIPs[ipit];
							ip = ip.Substring(0, wcip.Length);
							if(ip == wcip)
								banned = true;
						}
					}
				}
			}
			else {
				banned = true;
			}

			return banned;
		}

		// Is a port banned?
		private bool portBanned(uint port)
		{
			bool banned = false;

			lock(this) {
				// Is port in banned list?
				banned = (m_denyPorts.IndexOf(port) != -1);
			}

			return banned;
		}

		// Resolve an address to a string IP address
		private string resolve(string address)
		{
			string ip = Utils.resolveIP(address);
			if(ip.Length == 0) {
				m_glob.log().warn("Firewall failed to resolve address " + address);
			}
			return ip;
		}


		/********/
		/* Data */
		/********/

		// Reference to the globals
		private Globals				m_glob;

		// List of denied IP addresses and ports
		private ArrayList				m_denyIPs;
		private ArrayList				m_denyWildcardIPs;
		private ArrayList				m_denyPorts;

	}

}	// namespace

