/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Somewhere to store application "globals".
 *
 * $Id: Globals.cs,v 1.4 2013/02/17 10:52:30 rich Exp $
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */


namespace HTTPTunnel
{
	#if DEBUG_LOG_DATA
	using System.IO;
	using System.Threading;
	#endif

	public class Globals {

		// Construction
		public Globals(string procname)
		{
			m_cfg = new Config(this, procname);

			#if DEBUG_LOG_DATA
			m_logMtx = new Mutex();
			openDebugLogs();
			#endif

			// Create logger
			#if DEBUG_TUN
			m_log = new Logger(this, procname, true);
			#else
			m_log = new Logger(this, procname, false);
			#endif
		}


		// Shutdown
		public void shutdown()
		{
			// Shutdown logger
			m_log.shutdown();
		}


		#if DEBUG_LOG_DATA
		public void logRawRead(RawBuff data)
		{
			logData(data, m_rawReadLogFile, ref m_rawReadChars);
		}
		public void logRawWrite(RawBuff data)
		{
			logData(data, m_rawWriteLogFile, ref m_rawWriteChars);
		}
		public void logPayloadRead(RawBuff data)
		{
			logData(data, m_payloadReadLogFile, ref m_payloadReadChars);
		}
		public void logPayloadWrite(RawBuff data)
		{
			logData(data, m_payloadWriteLogFile, ref m_payloadWriteChars);
		}
		#endif


		/*************/
		/* Accessors */
		/*************/

		public Config cfg()
		{
			return m_cfg;
		}
		public Logger log()
		{
			return m_log;
		}


		/*****************/
		/* Debug Methods */
		/*****************/

		// Debug levels
		public static uint QUIET = 0;
		public static uint CRIT = 1;
		public static uint ERROR = 2;
		public static uint WARN = 3;
		public static uint INFO = 4;
		public static uint VERBOSE = 5;

		// Constants
		public static string LIBVERSION = "1.4.0";
		public static string PLATNAME = "WINDOWS";

		#if DEBUG_TUN
		public void DPRINTV(string text) { m_log.trace(text, VERBOSE); }
		public void DPRINTI(string text) { m_log.trace(text, INFO); }
		public void DPRINTW(string text) { m_log.trace(text, WARN); }
		public void DPRINTE(string text) { m_log.trace(text, ERROR); }
		public void DPRINTC(string text) { m_log.trace(text, CRIT); }
		#else
		public void DPRINTV(string text) {}
		public void DPRINTI(string text) {}
		public void DPRINTW(string text) {}
		public void DPRINTE(string text) {}
		public void DPRINTC(string text) {}
		#endif


		/*********************/
		/* Private Interface */
		/*********************/


		/***********/
		/* Methods */
		/***********/

		#if DEBUG_LOG_DATA
		// Open debug log files
		private void openDebugLogs()
		{
			// Set common filename stem
			string stem = @"C:\DebugLogs\Win-";

			// Initialise char counts
			m_rawReadChars = 0;
			m_rawWriteChars = 0;
			m_payloadReadChars = 0;
			m_payloadWriteChars = 0;

			// Open log files
			string rawReadName = stem + "rawread.log";
			m_rawReadLogFile = new StreamWriter(File.Create(rawReadName));
			string rawWriteName = stem + "rawwrite.log";
			m_rawWriteLogFile = new StreamWriter(File.Create(rawWriteName));
			string payloadReadName = stem + "payloadread.log";
			m_payloadReadLogFile = new StreamWriter(File.Create(payloadReadName));
			string payloadWriteName = stem + "payloadwrite.log";
			m_payloadWriteLogFile = new StreamWriter(File.Create(payloadWriteName));
		}
		#endif

		#if DEBUG_LOG_DATA
		// Log some debug data
		private void logData(RawBuff data, StreamWriter file, ref uint lineChar)
		{
			lock(m_logMtx) {
				// Get data
				byte[] dataPtr = data.getData();

				// Log bytes
				for(int i = 0; i < dataPtr.Length; ++i) {
					file.Write("{0:X2}", dataPtr[i]);

					// Space or newline?
					if((++lineChar % 20) == 0) {
						lineChar = 0;
						file.WriteLine();
					}
					else
						file.Write(" ");
				}

				// Flush
				file.Flush();
			}
		}
		#endif


		/********/
		/* Data */
		/********/

		// The system config
		private Config							m_cfg;

		// A logger
		private Logger							m_log;

		#if DEBUG_LOG_DATA
		// Global log files + Mutex
		private Mutex							m_logMtx;
		private StreamWriter					m_rawReadLogFile;
		private StreamWriter					m_rawWriteLogFile;
		private StreamWriter					m_payloadReadLogFile;
		private StreamWriter					m_payloadWriteLogFile;
		private uint							m_rawReadChars;
		private uint							m_rawWriteChars;
		private uint							m_payloadReadChars;
		private uint							m_payloadWriteChars;
		#endif
	}

}	// namespace
