/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Handle reading/writing for one end of an HTTP tunnel
 * connection.
 *
 * $Id: HTTPConnection.cs,v 1.1 2008/07/25 15:22:38 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Text;

	public class HTTPConnection : Connection {

		// Construction
		public HTTPConnection(Globals glob, Sock sock, Tunneler tunneler)
		: base(glob, sock, tunneler)
		{
			#if DEBUG_TUN
			StringBuilder dbg = new StringBuilder();
			dbg.AppendFormat("HTTPConnection::HTTPConnection(server) - connection {0}", getID());
			m_glob.DPRINTV(dbg.ToString());
			#endif

			m_clientAddr = sock.getConnectedClientAddr();
			m_noMsgsRecvd = 0;
			m_persistent = false;

			// Construct reader/writer
			createReaderWriter();
		}

		public HTTPConnection(Globals glob, Tunneler tunneler)
		: base(glob, tunneler)
		{
			#if DEBUG_TUN
			StringBuilder dbg = new StringBuilder();
			dbg.AppendFormat("HTTPConnection::HTTPConnection(client) - connection {0}", getID());
			m_glob.DPRINTV(dbg.ToString());
			#endif

			m_clientAddr = "";
			m_noMsgsRecvd = 0;
			m_persistent = false;
		}


		// We are no longer required
		public override void destroyed()
		{
			#if DEBUG_TUN
			StringBuilder dbg = new StringBuilder();
			dbg.AppendFormat("HTTPConnection::~HTTPConnection() - connection {0}", getID());
			m_glob.DPRINTV(dbg.ToString());
			#endif

			// Flag we are closing
			lock(this) {
				m_closing = true;
			}

			// Base class
			base.destroyed();
		}


		/********************/
		/* Public Interface */
		/********************/


		/***********/
		/* Methods */
		/***********/

		// Let us know the result of a connection attempt
		public override void connectResult(ConnectResultEvent.ConnectResultType result,
		Sock sock)
		{
			#if DEBUG_TUN
			StringBuilder dbg = new StringBuilder();
			dbg.AppendFormat("HTTPConnection::connectResult() - connection {0}", getID());
			m_glob.DPRINTV(dbg.ToString());
			#endif

			// Base class
			base.connectResult(result, sock);

			// Success?
			if(result == ConnectResultEvent.ConnectResultType.SUCCESS) {
				// Now send initial HTTP GET to setup HTTP connection
				poll();
			}
		}

		// Poll the other end of this connection with an empty GET message
		public void poll()
		{
			lock(this) {
				// Can't poll unless connected & active
				if(connected() && active()) {
					// Only bother polling if we actually currently carry a tunnel
					if(m_carriedTunnels.Count > 0) {
						// Force the sending of a (probably empty) message
						Tunnel tun = (Tunnel)m_carriedTunnels[0];
						m_tunneler.scheduleEvent(new SendMsgEvent(m_glob, new Time(0,0),
							getID(), true, tun.getConfig().getTunnelID(), m_tunneler));
					}
				}
				else {
					throw new CommsError("Error polling server: Not connected or inactive");
				}
			}
		}


		/*************/
		/* Accessors */
		/*************/

		// Get address of connected client at server
		public string getClientAddr()
		{
			return m_clientAddr;
		}

		// Get number of msgs received over us
		public uint getNoMsgsRecvd()
		{
			return m_noMsgsRecvd;
		}

		// Get whether we are persistent
		public bool getPersistent()
		{
			return m_persistent;
		}


		/************/
		/* Mutators */
		/************/

		// Increase count of msgs received over us
		public void msgRecvd()
		{
			lock(this) {
				m_noMsgsRecvd++;
			}
		}

		// We are persistent
		public void setPersistent()
		{
			lock(this) {
				m_persistent = true;
			}
		}


		/***********************/
		/* Protected Interface */
		/***********************/


		/***********/
		/* Methods */
		/***********/

		// Construct reader/writer
		protected override void createReaderWriter()
		{
			lock(this) {
				// Create HTTP reader & writer
				m_reader = new HTTPReader(m_glob, this);
				m_writer = new HTTPWriter(m_glob, this);

				// Only truly connected once reader & writer are created
				setConnected();
			}
		}


		/********/
		/* Data */
		/********/

		// Address of connecting client
		protected string					m_clientAddr;

		// Count of number of HTTP msgs received over us
		protected uint						m_noMsgsRecvd;

		// Are we a persistent connection?
		protected bool						m_persistent;

	}

}	// namespace

