/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate an HTTP get message.
 *
 * $Id: HTTPGetMsg.cs,v 1.1 2008/07/25 15:22:38 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Text;
	using System.Text.RegularExpressions;

	public class HTTPGetMsg : HTTPRequestMsg {

		/********************/
		/* Public Interface */
		/********************/

		// Construction
		public HTTPGetMsg(Globals glob) : base(glob)
		{
			m_uuData = "";
			m_repUULen = 0;
			m_uriEncFields = "";
		}

		public HTTPGetMsg(Globals glob, HTTPMsgParms parms, string uri,
		ClientSideConfig cfg) : base(glob, parms, uri, cfg)
		{
			m_uuData = "";
			m_repUULen = 0;
			m_uriEncFields = "";

			// Construct headers
			uint payloadLen = m_parms.getPayload().getPayloadSize();
			if(payloadLen > 0) {
				// Encode payload data into header
				StringBuilder ourhdr = new StringBuilder();
				packPayload(m_parms.getPayload().getPayload(), ourhdr);
				m_ourHeader += ourhdr.ToString();
			}

			// CONNECT specific fields
			if(m_parms.getState() == Tunnel.TunnelState.CONNECT) {
				// Send tunnel destination details
				m_ourHeader += "HT-Des: " + m_parms.getTunDest() + "\r\n";

				// Send tunnel tuning
				m_ourHeader += "HT-Tun: " + m_parms.getTunTune() + "\r\n";

				// Send client version
				m_ourHeader += "HT-Cvr: " + m_parms.getCliVer() + "\r\n";

				// Also include all our header fields as a test of any proxy filtering
				m_ourHeader += "HT-Dat: \"\"\r\n";
				m_ourHeader += "HT-Dln: 0\r\n";
				m_ourHeader += "HT-Spl: 0\r\n";
			}

			// Construct request line
			StringBuilder req = new StringBuilder();
			req.AppendFormat("GET {0} HTTP/{1}", getRequestURI(), m_parms.getVer());
			m_requestLine = req.ToString();
		}

		public HTTPGetMsg(HTTPGetMsg that) : base(that)
		{
			m_uuData = that.m_uuData;
			m_repUULen = that.m_repUULen;
			m_uriEncFields = that.m_uriEncFields;
		}

		// Construction from a raw read data stream
		public HTTPGetMsg(Globals glob, RawBuff readdata) : base(glob)
		{
			// Construct members
			m_uuData = "";
			m_repUULen = 0;
			m_uriEncFields = "";

			/* Look for basic components of GET message header -
				GET		- message identifier
				/X.htm  - URI
				HTTP/x.y - protocol version
			*/

			uint getStart = 0, getEnd = 0;
			if(!readdata.findString("GET", ref getStart, ref getEnd)) {
				throw new HTTPMsgWrongMsgType("GET message identifier not found",
					m_parsingLine);
			}

			// Extract URI
			uint uriStart1 = 0, uriStart2 = 0, uriEnd = 0;
			if(readdata.findString("/", ref uriStart1, ref uriEnd)) {
				if(readdata.findString(".htm", ref uriStart2, ref uriEnd)) {
					// Get URI
					m_uri = readdata.makeString(uriStart1, uriEnd);
					m_parms.setURI(m_uri);
				}
				else
					throw new HTTPMsgParsingError("End of GET URI not found", m_parsingLine);
			}
			else
				throw new HTTPMsgParsingError("Start of GET URI not found", m_parsingLine);

			// Extract protocol version
			uint httpStart = 0, httpEnd = 0;
			if(!readdata.findString("HTTP/", ref httpStart, ref httpEnd))
				throw new HTTPMsgParsingError("GET protocol id not found", m_parsingLine);

			uint nlStart = 0, nlEnd = 0;
			if(!readdata.findNewLine(ref nlStart, ref nlEnd))
				throw new HTTPMsgParsingError("Expected newline not found", m_parsingLine);
			m_parms.setVer(readdata.makeString(httpEnd, nlStart));

			// Copy entire request line
			m_requestLine = readdata.makeString(getStart, nlStart);

			// Are we passing tunneling data via URI?
			bool dataInURI = false;
			RawBuff reqLineBuff = new RawBuff(m_requestLine);
			if(reqLineBuff.findString("?", ref uriStart1, ref uriEnd)) {
				if(reqLineBuff.findString(" ", uriStart1, ref uriStart2, ref uriEnd)) {
					if(uriEnd - 1 > uriStart1 + 1) {
						// Get encoded tunneling fields
						m_uriEncFields = reqLineBuff.makeString(uriStart1 + 1, uriEnd - 1);
						dataInURI = true;
					}
					else {
						throw new HTTPMsgParsingError("Too little encoded URI data",
							m_parsingLine);
					}
				}
				else {
					throw new HTTPMsgParsingError("End of GET encoded URI data not found",
						m_parsingLine);
				}
			}

			// Discard everything from the first line in the data stream
			readdata.erase(nlEnd);
			m_parsingLine++;

			#if DEBUG_TUN
			StringBuilder nfo = new StringBuilder();
			nfo.AppendFormat("HTTPGetMsg::HTTPGetMsg() - Got URI: {0}", m_uri);
			nfo.AppendFormat(", protocol: {0}", m_parms.getVer());
			m_glob.DPRINTV(nfo.ToString());
			#endif

			// Let base classes look for their headers
			bool eoh = false;
			string hdrline = "";
			parseHdrs(readdata, ref eoh, ref hdrline);

			// Expect final newline to terminate non-content part of message
			if(!readdata.findNewLine(ref nlStart, ref nlEnd)) {
				throw new HTTPMsgParsingError("GET final newline terminator not found",
					m_parsingLine);
			}
			readdata.erase(nlEnd);
			m_parsingLine++;

			// Did we see data passed in the URI?
			if(dataInURI) {
				bool parseURI = true;

				// Was this the first CONNECT msg of a session?
				Tunnel.TunnelState state = m_parms.getState();
				if(state == Tunnel.TunnelState.UNKNOWNSTATE || state == Tunnel.TunnelState.CONNECT) {
					// Remember whether we need to use header tunneling all the time
					parseURI = !allReqHeaders();
					m_parms.setUseHeaderFields(!parseURI);
				}

				if(parseURI) {
					// Try to parse tunneling fields from coded URI
					m_glob.DPRINTV("Trying to parse tunneling fields from URI");
					string decoded = Utils.URLDecode(m_uriEncFields) + "\r\n";
					RawBuff decodedHeaders = new RawBuff(decoded);
					parseHdrs(decodedHeaders, ref eoh, ref hdrline);
				}
			}

			// Now deal with any content
			extractContent(readdata);
		}


		/***********************/
		/* Protected Interface */
		/***********************/


		/***********/
		/* Methods */
		/***********/

		// Parse and extract any UUEncoded message content
		protected override void extractContent(RawBuff readdata)
		{
			m_glob.DPRINTV("HTTPGetMsg::extractContent()");

			// Are we expecting any content data?
			if(m_repUULen > 0) {
				#if DEBUG_TUN
				StringBuilder s1 = new StringBuilder();
				s1.AppendFormat("HTTPGetMsg::extractContent() - Expecting {0}", m_repUULen);
				s1.Append(" bytes of UUEncoded content");
				m_glob.DPRINTV(s1.ToString());
				#endif

				if(m_uuData.Length == m_repUULen) {
					// UUDecode data into our payload buffer
					m_parms.getPayload().getPayload().uudecode(m_uuData);
				}
				else {
					#if DEBUG_TUN
					StringBuilder s2 = new StringBuilder();
					s2.Append("HTTPGetMsg::extractContent() - Wrong UUEncoded data length: ");
					s2.AppendFormat("Reported {0}, but got {1}", m_repUULen, m_uuData.Length);
					m_glob.DPRINTV(s2.ToString());
					#endif

					throw new BadContentLengthParsingError("Incorrect encoded content length",
						m_parsingLine);
				}
			}
		}

		// Attempt to parse headers from a raw byte-stream
		protected override void parseHdrs(RawBuff data, ref bool eoh, ref string hdrline)
		{
			// Look for any GET headers - everything else is handled elsewhere

			bool ok = true;
			eoh = false;

			// If the first thing in the read buffer is a newline, there are no headers
			while(ok && !eoh) {
				// Let base class have a go
				base.parseHdrs(data, ref eoh, ref hdrline);
				if(!eoh) {
					/* Base class didn't like a header, let us try */

					// Is it one of ours?
					if(hdrline.StartsWith("HT-Dat:")) {
						// Some UUEncoded data
						int quotestart = hdrline.IndexOf('\"');
						int quoteend = hdrline.LastIndexOf('\"');
						if(quotestart >= 0 && quoteend >= 0) {
							string packed = hdrline.Substring(quotestart, (quoteend - quotestart) + 1);
							unpackPayload(packed);
							m_seenUUData = true;
							m_lastTunId = "HT-Dat: ";
						}
						else {
							throw new HTTPMsgParsingError("Failed to parse start/end of " +
								"UUEncoded data from get msg", m_parsingLine);
						}
					}
					else if(hdrline.StartsWith("HT-Dln:")) {
						// Length of uuencoded data
						Regex reg = new Regex(@"\S+\s*");
						MatchCollection matches = reg.Matches(hdrline);
						if(matches.Count == 2) {
							m_repUULen = UInt32.Parse(matches[1].ToString());
							m_seenUULen = true;
							m_lastTunId = "HT-Dln: ";
						}
						else {
							throw new HTTPMsgParsingError("Failed to parse UUEncoded length" +
								" from get msg", m_parsingLine);
						}
					}
					else if(hdrline.StartsWith("HT-Des:")) {
						// Get address of destination to tunnel to
						int i = hdrline.IndexOf(' ');
						if(i >= 0) {
							string dest = hdrline.Substring(i + 1);
							m_parms.setTunDest(dest);
							m_seenDest = true;
							m_lastTunId = "HT-Des: ";
						}
						else {
							throw new HTTPMsgParsingError("Failed to parse tunnel destination" +
								" from get msg", m_parsingLine);
						}
					}
					else if(hdrline.StartsWith("HT-Tun:")) {
						// Get tunnel tuning
						int i = hdrline.IndexOf(' ');
						if(i >= 0) {
							string tune = hdrline.Substring(i + 1);
							m_parms.setTunTune(tune);
							m_seenTune = true;
							m_lastTunId = "HT-Tun: ";
						}
						else {
							throw new HTTPMsgParsingError("Failed to parse tunnel tuning" +
								" from get msg", m_parsingLine);
						}
					}
				}
			}
		}

		// Pack payload data to a header string
		protected void packPayload(RawBuff payload, StringBuilder str)
		{
			// Get uuencoded data
			string uue = "";
			uint length = payload.getLength();
			payload.uuencode(ref length, ref uue);

			// If we're packing into header fields, limit the maximum line length
			bool limitLineLength = m_parms.getUseHeaderFields();

			// Write out uuencoded data as header lines
			int bytesToGo = uue.Length;
			int pos = 0;
			do {
				str.Append("HT-Dat: \"");

				if(!limitLineLength || (bytesToGo < MAX_UU_CHARS_LINE)) {
					// Not enough left for a full line
					str.Append(uue.Substring(pos));
					bytesToGo = 0;
				}
				else {
					// Full line
					str.Append(uue.Substring(pos, MAX_UU_CHARS_LINE));
					bytesToGo -= MAX_UU_CHARS_LINE;
					pos += MAX_UU_CHARS_LINE;
				}

				str.Append("\"\r\n");
			} while(bytesToGo > 0);

			// Add encoded data length field
			str.AppendFormat("HT-Dln: {0}\r\n", uue.Length);
		}

		// Unpack payload data from a header string
		protected void unpackPayload(string header)
		{
			int pos = 0;

			// Iterate over encoded data
			while(pos >= 0) {
				// Skip starting "
				pos++;

				// Find terminating "
				int term = header.IndexOf('\"', pos);

				// Extract encoded string
				m_uuData += header.Substring(pos, term - pos);

				// Look for possible next starting "
				pos = header.IndexOf('\"', term + 1);
			}
		}


		/********/
		/* Data */
		/********/

		// Maximum UUEncoded characters per line
		protected const int				MAX_UU_CHARS_LINE = 256;

		// UUEncoded data
		protected string					m_uuData;

		// Reported length of UUEncoded data
		protected uint						m_repUULen;

		// URI encoded header fields
		protected string					m_uriEncFields;

	}

}	// namespace

