/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate an HTTP message.
 *
 * $Id: HTTPMsg.cs,v 1.2 2008/08/07 20:30:29 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Collections;
	using System.Text;
	using System.Text.RegularExpressions;

	public abstract class HTTPMsg {

		/********************/
		/* Public Interface */
		/********************/

		// Construction
		public HTTPMsg(Globals glob)
		{
			m_glob = glob;
			m_parms = new HTTPMsgParms(m_glob);
			m_generalHeader = "";
			m_reqRespHeader = "";
			m_entityHeader = "";
			m_ourHeader = "";
			m_parsingLine = 0;
			m_seenTID = false;
			m_seenSeqNo = false;
			m_seenConn = false;
			m_seenDest = false;
			m_seenUUData = false;
			m_seenUULen = false;
			m_seenAck = false;
			m_seenQuotas = false;
			m_seenCliVer = false;
			m_seenSplit = false;
			m_seenTest = false;
			m_seenTune = false;
			m_seenNotif = false;
			m_authArgs = new Hashtable();
			m_lastTunId = "";
			m_lastTunIdSplit = false;
		}

		public HTTPMsg(Globals glob, HTTPMsgParms parms)
		{
			m_glob = glob;
			m_parms = new HTTPMsgParms(parms);
			m_generalHeader = "";
			m_reqRespHeader = "";
			m_entityHeader = "";
			m_ourHeader = "";
			m_parsingLine = 0;
			m_seenTID = false;
			m_seenSeqNo = false;
			m_seenConn = false;
			m_seenDest = false;
			m_seenUUData = false;
			m_seenUULen = false;
			m_seenAck = false;
			m_seenQuotas = false;
			m_seenCliVer = false;
			m_seenSplit = false;
			m_seenTest = false;
			m_seenTune = false;
			m_seenNotif = false;
			m_authArgs = new Hashtable();
			m_lastTunId = "";
			m_lastTunIdSplit = false;

			// Construct headers
			StringBuilder genhdr = new StringBuilder();
			StringBuilder ourhdr = new StringBuilder();
			genhdr.Append("Cache-Control: no-cache\r\n");
			ourhdr.AppendFormat("HT-Tid: {0}\r\n", m_parms.getTid());
			ourhdr.AppendFormat("HT-Seq: {0}\r\n", m_parms.getPayload().getSeqNo());

			// Include acknowledgments
			if(m_parms.getAck().Length > 0)
				ourhdr.AppendFormat("HT-Ack: {0}\r\n", m_parms.getAck());

			// Split payload indication
			if(m_parms.getPayload().getSplit())
				ourhdr.Append("HT-Spl: 1\r\n");

			// Mark connection state as appropriate
			Tunnel.TunnelState state = m_parms.getState();
			switch(state) {
				case(Tunnel.TunnelState.CONNECT):
				case(Tunnel.TunnelState.AUTHREQ):
				case(Tunnel.TunnelState.RECONNECT):
				case(Tunnel.TunnelState.CLOSING):
				case(Tunnel.TunnelState.CONNECTED): {
					genhdr.Append("Connection: keep-alive\r\n");
					ourhdr.AppendFormat("HT-Cst: {0}\r\n", (uint)state);
					break;
				}
				case(Tunnel.TunnelState.CLOSED): {
					genhdr.Append("Connection: close\r\n");
					ourhdr.AppendFormat("HT-Cst: {0}\r\n", (uint)state);
					break;
				}
				case(Tunnel.TunnelState.UNKNOWNSTATE): {
					m_glob.log().error("Tunnel state is UNKNOWN in HTTPMsg()");
					break;
				}
			}

			// Mark tunnel testing state
			Tunnel.TunnelTestState testState = m_parms.getTestState();
			if(testState != Tunnel.TunnelTestState.NO_TESTING ||
			state == Tunnel.TunnelState.CONNECT) {
				ourhdr.AppendFormat("HT-Tst: {0}\r\n", (uint)testState);
			}

			// Format any notifications
			ArrayList notifs = m_parms.getNotifs();
			for(int nit = 0; nit < notifs.Count; ++nit) {
				ourhdr.AppendFormat("HT-Ntf: {0}\r\n", (string)notifs[nit]);
			}

			// Tidy up
			m_generalHeader = genhdr.ToString();
			m_ourHeader = ourhdr.ToString();
		}

		public HTTPMsg(HTTPMsg that)
		{
			m_glob = that.m_glob;
			m_parms = new HTTPMsgParms(that.m_parms);
			m_generalHeader = that.m_generalHeader;
			m_reqRespHeader = that.m_reqRespHeader;
			m_entityHeader = that.m_entityHeader;
			m_ourHeader = that.m_ourHeader;
			m_parsingLine = that.m_parsingLine;
			m_seenTID = that.m_seenTID;
			m_seenSeqNo = that.m_seenSeqNo;
			m_seenConn = that.m_seenConn;
			m_seenDest = that.m_seenDest;
			m_seenUUData = that.m_seenUUData;
			m_seenUULen = that.m_seenUULen;
			m_seenAck = that.m_seenAck;
			m_seenQuotas = that.m_seenQuotas;
			m_seenCliVer = that.m_seenCliVer;
			m_seenSplit = that.m_seenSplit;
			m_seenTest = that.m_seenTest;
			m_seenTune = that.m_seenTune;
			m_seenNotif = that.m_seenNotif;
			m_authArgs = that.m_authArgs;
			m_lastTunId = that.m_lastTunId;
			m_lastTunIdSplit = that.m_lastTunIdSplit;
		}


		/***********/
		/* Methods */
		/***********/

		// Write ourselves out as a stream of bytes
		public virtual void output(RawBuff dest)
		{
			// Somewhere to build up complete message
			StringBuilder output = new StringBuilder();

			// Get message start line
			string start = getStartLine();
			// Build up complete header
			string header = getHeader();

			#if DEBUG_TUN
			StringBuilder str1 = new StringBuilder();
			str1.Append("HTTPMsg::output() - Sending ");
			str1.AppendFormat("{0} bytes of payload",
				m_parms.getPayload().getPayload().getLength());
			m_glob.DPRINTV(str1.ToString());
			#endif

			// Build up complete message
			output.AppendFormat("{0}\r\n", start);

			if(header.Length > 0)
				output.Append(header);			// Each header line includes its own CRLF

			output.Append("\r\n");

			// Append header part of msg to destination
			dest.append(output.ToString());
		}


		// Get message parameters
		public HTTPMsgParms getParms()
		{
			return m_parms;
		}

		// Get authentication paramaters
		public Hashtable getAuthArgs()
		{
			return m_authArgs;
		}

		// Minimum headers passed as expected?
		public bool allReqHeaders()
		{
			return (m_seenTID && m_seenSeqNo && m_seenConn && m_seenUUData &&
				m_seenUULen && m_seenAck && m_seenCliVer && m_seenSplit &&
				m_seenTest && m_seenTune);
		}
		public bool allRespHeaders()
		{
			return (m_seenTID && m_seenSeqNo && m_seenConn && m_seenAck);
		}


		/***********************/
		/* Protected Interface */
		/***********************/


		/***********/
		/* Methods */
		/***********/

		// Parse and extract any message content
		protected abstract void extractContent(RawBuff readdata);

		// Get start line string
		protected abstract string getStartLine();

		// Attempt to parse headers from a raw byte-stream
		protected virtual void parseHdrs(RawBuff data, ref bool eoh, ref string hdrline)
		{
			/* Look for any general headers - everything else is handled in other classes */

			bool ok = true;
			bool thisSplit = false;
			eoh = false;

			// Parse as much as we can
			while(ok) {
				// We need a complete line to parse
				uint start = 0, end = 0;
				if(data.findNewLine(ref start, ref end, ref thisSplit)) {
					// Is there anything before the newline?
					if(start != 0) {

						// Extract this header field
						hdrline = data.makeString(0, start);
						if(m_lastTunIdSplit) {
							// This is actually a split line
							hdrline = m_lastTunId + hdrline;
						}

						// Clear out parse buff
						data.erase(end);
						if(thisSplit)
							data.eraseWS();
						m_lastTunIdSplit = thisSplit;
						m_parsingLine++;

						// Is it one of ours?
						if(hdrline.StartsWith("HT-Seq:")) {
							// Get msg sequence number
							Regex reg = new Regex(@"\S+\s*");
							MatchCollection matches = reg.Matches(hdrline);
							if(matches.Count == 2) {
								SeqNo seqNo = new SeqNo(UInt32.Parse(matches[1].ToString()));
								m_parms.getPayload().setSeqNo(seqNo);
								m_seenSeqNo = true;
								m_lastTunId = "HT-Seq: ";

								#if DEBUG_TUN
								StringBuilder dbg = new StringBuilder();
								dbg.AppendFormat("HTTPMsg::parseHdrs() - Got sequence no: {0}", seqNo);
								m_glob.DPRINTV(dbg.ToString());
								#endif
							}
							else {
								throw new HTTPMsgParsingError("Failed to parse sequence number " +
									"from msg", m_parsingLine);
							}
						}
						else if(hdrline.StartsWith("HT-Ack:")) {
							// Get list of acknowledgments
							int i = hdrline.IndexOf(' ');
							if(i >= 0) {
								string ack = hdrline.Substring(i + 1);
								m_glob.DPRINTV("HTTPMsg::parseHdrs() - Got ack: " + ack);
								m_parms.setAck(ack);
								m_seenAck = true;
								m_lastTunId = "HT-Ack: ";
							}
							else {
								throw new HTTPMsgParsingError("Failed to parse ack from msg",
									m_parsingLine);
							}
						}
						else if(hdrline.StartsWith("HT-Tid:")) {
							// Get reported tunnel ID
							int i = hdrline.IndexOf(' ');
							if(i >= 0) {
								string tid = hdrline.Substring(i + 1);
								m_glob.DPRINTV("HTTPMsg::parseHdrs() - Got TID: " + tid);
								m_parms.setTid(tid);
								m_seenTID = true;
								m_lastTunId = "HT-Tid: ";
							}
							else {
								throw new HTTPMsgParsingError("Failed to parse TID from msg",
									m_parsingLine);
							}
						}
						else if(hdrline.StartsWith("HT-Cst:")) {
							// Closing?
							Regex reg = new Regex(@"\S+\s*");
							MatchCollection matches = reg.Matches(hdrline);
							if(matches.Count == 2) {
								uint state = UInt32.Parse(matches[1].ToString());
								m_parms.setState((Tunnel.TunnelState)state);
								m_seenConn = true;
								m_lastTunId = "HT-Cst: ";
							}
							else {
								throw new HTTPMsgParsingError("Failed to parse connection state " +
									"from msg", m_parsingLine);
							}
						}
						else if(hdrline.StartsWith("HT-Cvr:")) {
							// Get client version ID
							int i = hdrline.IndexOf(' ');
							if(i >= 0) {
								string text = hdrline.Substring(i + 1);
								m_glob.DPRINTV("HTTPMsg::parseHdrs() - Got client version: " + text);
								m_parms.setCliVer(text);
								m_seenCliVer = true;
								m_lastTunId = "HT-Cvr: ";
							}
							else {
								throw new HTTPMsgParsingError("Failed to parse client version from msg",
									m_parsingLine);
							}
						}
						else if(hdrline.StartsWith("HT-Spl:")) {
							// Get payload split indicator
							Regex reg = new Regex(@"\S+\s*");
							MatchCollection matches = reg.Matches(hdrline);
							if(matches.Count == 2) {
								uint split = UInt32.Parse(matches[1].ToString());
								m_parms.getPayload().setSplit(split != 0);
								m_seenSplit = true;
								m_lastTunId = "HT-Spl: ";
							}
							else {
								throw new HTTPMsgParsingError("Failed to parse split indicator " +
									"from msg", m_parsingLine);
							}
						}
						else if(hdrline.StartsWith("HT-Tst:")) {
							// Get tunnel test state
							Regex reg = new Regex(@"\S+\s*");
							MatchCollection matches = reg.Matches(hdrline);
							if(matches.Count == 2) {
								uint state = UInt32.Parse(matches[1].ToString());
								m_parms.setTestState((Tunnel.TunnelTestState)state);
								m_seenTest = true;
								m_lastTunId = "HT-Tst: ";

								#if DEBUG_TUN
								StringBuilder dbg = new StringBuilder();
								dbg.AppendFormat("HTTPMsg::parseHdrs() - Got test state: {0}", state);
								m_glob.DPRINTV(dbg.ToString());
								#endif
							}
							else {
								throw new HTTPMsgParsingError("Failed to parse test state from msg",
									m_parsingLine);
							}
						}
						else if(hdrline.StartsWith("HT-Ntf:")) {
							// Get notification msg
							int i = hdrline.IndexOf(' ');
							if(i >= 0) {
								string text = hdrline.Substring(i + 1);
								m_glob.DPRINTV("HTTPMsg::parseHdrs() - Got notification: " + text);
								m_parms.addNotif(text);
								m_seenNotif = true;
								m_lastTunId = "HT-Ntf: ";
							}
							else {
								throw new HTTPMsgParsingError("Failed to parse notification from msg",
									m_parsingLine);
							}
						}

						// Is it a standard HTTP header field of interest?
						else if(hdrline.StartsWith("Connection:")) {
							// Get HTTP connection option
							int i = hdrline.IndexOf(' ');
							if(i >= 0) {
								string text = hdrline.Substring(i + 1);
								m_glob.DPRINTV("HTTPMsg::parseHdrs() - Got HTTP connection option: " + text);
								m_parms.setConnOption(text);
							}
							else {
								throw new HTTPMsgParsingError("Failed to parse HTTP connection option from msg",
									m_parsingLine);
							}
						}
						else if(hdrline.StartsWith("Proxy-Connection:")) {
							// Get proxy HTTP connection option
							int i = hdrline.IndexOf(' ');
							if(i >= 0) {
								string text = hdrline.Substring(i + 1);
								m_glob.DPRINTV("HTTPMsg::parseHdrs() - Got proxy HTTP connection option: " + text);
								m_parms.setConnOption(text);
							}
							else {
								throw new HTTPMsgParsingError("Failed to parse proxy HTTP connection option from msg",
									m_parsingLine);
							}
						}

						else {
							// Unknown header
							ok = false;
						}
					}
					else {
						// Newline found at start of line, must be end of headers
						eoh = true;
						break;
					}
				}
				else {
					// Header line terminating new line not found
					throw new HTTPMsgParsingError("Msg header terminating newline not found",
						m_parsingLine);
				}
			}
		}

		// Get header string
		protected string getHeader()
		{
			string header = m_generalHeader + m_reqRespHeader + m_entityHeader;
			if(m_parms.getUseHeaderFields())
				header += m_ourHeader;
			return header;
		}


		/********/
		/* Data */
		/********/

		// Globals
		protected Globals						m_glob;

		// Message parameters
		protected HTTPMsgParms				m_parms;

		// The header fields
		protected string						m_generalHeader;
		protected string						m_reqRespHeader;
		protected string						m_entityHeader;
		protected string						m_ourHeader;

		// Line within header we are parsing
		protected uint							m_parsingLine;

		// Flags to indicate if we've seen particular headers
		protected bool							m_seenTID;
		protected bool							m_seenSeqNo;
		protected bool							m_seenConn;
		protected bool							m_seenDest;
		protected bool							m_seenUUData;
		protected bool							m_seenUULen;
		protected bool							m_seenAck;
		protected bool							m_seenQuotas;
		protected bool							m_seenCliVer;
		protected bool							m_seenSplit;
		protected bool							m_seenTest;
		protected bool							m_seenTune;
		protected bool							m_seenNotif;

		// Authentication arguments
		protected Hashtable					m_authArgs;

		// Split header parameters
		protected string						m_lastTunId;
		protected bool							m_lastTunIdSplit;

	}

}  // namespace


