/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * An HTTP data reader.
 *
 * $Id: HTTPReader.cs,v 1.3 2008/11/22 19:52:11 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Text;

	class HTTPReader : Reader {

		// Construction
		public HTTPReader(Globals glob, Connection conn)
		: base(glob, conn)
		{
			m_glob.DPRINTV("HTTPReader::HTTPReader()");
			m_buff = new RawBuff();
			m_msgErrorCount = 0;
			m_lastRecvdData = 0;
			m_nextReqdData = 0;
			m_lastErrLine = 0;

			// Set message parse state
			if(m_conn.getTunneler().getServer().getConfig().isClient())
				m_parseState = ParseState.RECEIVE;
			else
				m_parseState = ParseState.CONNECT;
		}


		/*************/
		/* Accessors */
		/*************/

		// Do we have unprocessed data remaining?
		public override bool dataRemaining()
		{
			return (m_buff.getLength() > 0);
		}


		/***********************/
		/* Protected Interface */
		/***********************/


		/*********************/
		/* Types & Constants */
		/*********************/

		// Message parsing states
		protected enum ParseState { UNDEFINED, CONNECT, SEND, RECEIVE };

		// Number of message processing errors before we give up
		protected const uint		MAX_PROCESSING_ERRORS		= 25;

		// Msg parsing error levels
		protected const uint		PARSING_ERROR_MINOR			= 1;
		protected const uint		PARSING_ERROR_MAJOR			= 5;
		protected const uint		PARSING_ERROR_CRITICAL		= MAX_PROCESSING_ERRORS;


		/***********/
		/* Methods */
		/***********/

		// Process some newly read HTTP data
		protected override void processData(byte[] data, uint length)
		{
			lock(this) {
				// Append data to parse buffer
				m_buff.append(data, length);

				#if DEBUG_TUN
				StringBuilder err1 = new StringBuilder();
				err1.Append("HTTPReader::processData() - Before parse, ");
				err1.AppendFormat("buff contains {0} bytes", m_buff.getLength());
				m_glob.DPRINTV(err1.ToString());
				#endif

				// Do we have an indication of how much data we're going to need to
				// contruct this HTTP msg?
				if(m_nextReqdData > 0) {
					if(m_buff.getLength() <= m_nextReqdData) {
						#if DEBUG_TUN
						StringBuilder dbg = new StringBuilder();
						dbg.Append("Not enough data to attempt parse, need at least ");
						dbg.AppendFormat("{0}", m_nextReqdData);
						m_glob.DPRINTV(dbg.ToString());
						#endif

						return;
					}
				}

				// Somewhere to copy raw data buffer while attempting HTTP msg construction
				RawBuff parseBuffCopy = new RawBuff(m_buff);

				try {
					// Get tunneler reference
					Tunneler tunneler = m_conn.getTunneler();

					// Try to process as many messages as we can from our current data
					while(parseBuffCopy.getLength() > 0) {

						// Ptr to created message
						HTTPMsg msg = null;

						// What message do we expect next?
						switch(m_parseState) {
							case(ParseState.CONNECT): {
								// Try to parse & create a GET msg
								msg = new HTTPGetMsg(m_glob, parseBuffCopy);
								m_glob.DPRINTV("HTTPReader::processData() - Processed GET connect");

								// Now we just send
								m_parseState = ParseState.SEND;

								break;
							}

							case(ParseState.SEND): {
								// Try to parse & create a GET msg
								msg = new HTTPGetMsg(m_glob, parseBuffCopy);
								m_glob.DPRINTV("HTTPReader::processData() - Processed GET");
								break;
							}

							case(ParseState.RECEIVE): {
								// Try to parse & create a response msg
								HTTPResponseMsg respMsg = new HTTPResponseMsg(m_glob, parseBuffCopy);
								msg = respMsg;
								m_glob.DPRINTV("HTTPReader::processData() - Processed RESPONSE msg");

								// Check response code
								checkMsgResponseCode(respMsg);

								break;
							}

							default: {
								throw new ReadError("Unhandled parse state in HTTPReader");
							}
						}

						/************************************/
						/* Message type agnostic processing */
						/************************************/

						#if DEBUG_LOG_DATA
						// Log read payload data
						m_glob.logPayloadRead(msg.getParms().getPayload().getPayload());
						#endif

						// Schedule message arrival event for tunneler
						tunneler.scheduleEvent(new MsgArrivedEvent(m_glob, m_conn.getID(),
							tunneler, msg));

						// No error in processing, so purge processed data
						m_buff = new RawBuff(parseBuffCopy);
						m_msgErrorCount = 0;
						m_lastRecvdData = 0;
						m_nextReqdData = 0;
					}
				}
				catch(NotEnoughDataParsingError e) {
					#if DEBUG_TUN
					StringBuilder err = new StringBuilder();
					err.AppendFormat("Not enough content to create HTTP msg: {0}", e.Message);
					m_glob.DPRINTV(err.ToString());
					#endif

					// Are we progressing towards having enough data?
					if(!(e.m_recvd > m_lastRecvdData)) {
						// We don't seem to be progressing, treat as a serious error
						if(msgError(PARSING_ERROR_MAJOR)) {
							throw new ReadError("Too many errors in HTTP data");
						}
					}
					m_lastRecvdData = e.m_recvd;
					m_nextReqdData = e.m_reqd;
				}
				catch(BadContentLengthParsingError e) {
					#if DEBUG_TUN
					StringBuilder err = new StringBuilder();
					err.AppendFormat("Bad encoded data length in HTTP msg: {0}", e.Message);
					m_glob.DPRINTV(err.ToString());
					#endif

					// Error in communications
					throw new ReadError("Error processing HTTP data: " + e.Message);
				}
				catch(HTTPMsgParsingError e) {
					#if DEBUG_TUN
					StringBuilder err = new StringBuilder();
					err.AppendFormat("Parsing error for HTTP msg on line {0}: {1}",
						e.m_line, e.Message);
					m_glob.DPRINTV(err.ToString());
					#endif

					// Are we progressing through msg?
					if(!(e.m_line > m_lastErrLine)) {
						// We don't seem to be progressing, treat as serious error
						if(msgError(PARSING_ERROR_MAJOR)) {
							throw new ReadError("Too many errors in HTTP data");
						}
					}
					m_lastErrLine = e.m_line;
				}
				catch(ParsingError e) {
					#if DEBUG_TUN
					StringBuilder err = new StringBuilder();
					err.AppendFormat("General error creating HTTP msg: {0}", e.Message);
					m_glob.DPRINTV(err.ToString());
					#endif

					// Error in communications
					throw new ReadError("Error processing HTTP data: " + e.Message);
				}
				catch(InvalidID e) {
					StringBuilder err = new StringBuilder();
					err.AppendFormat("Error constructing ID: {0}", e.Message);
					m_glob.log().error(err.ToString());

					// Re-throw
					throw new ReadError("Error processing read ID");
				}

				#if DEBUG_TUN
				StringBuilder err2 = new StringBuilder();
				err2.Append("HTTPReader::processData() - After parse, ");
				err2.AppendFormat("buff contains {0} bytes", m_buff.getLength());
				m_glob.DPRINTV(err2.ToString());
				#endif
			}
		}

		// Check a read message's HTTP response code
		protected void checkMsgResponseCode(HTTPResponseMsg msg)
		{
			bool ok = false;
			uint respCode = msg.getResponseCode();

			// Everything OK?
			if(respCode == 200) {
				// Normal response
				ok = true;
			}
			else if(respCode == 401 || respCode == 407) {
				// Authentication required is not necessarily an error
				ok = true;
			}
			else if(respCode == 404) {
				// Probably a proxy connection error, not necessarily an error
				ok = true;
			}

			// Throw exception for bad response codes
			if(!ok) {
				#if DEBUG_TUN
				if(respCode == 500) {
					// Server error response
					m_glob.DPRINTV("Got HTTP error response 500 from server");
				}
				#endif

				// Tell caller
				StringBuilder err = new StringBuilder();
				err.AppendFormat("Got HTTP response code {0}", respCode);
				throw new BadResponseCodeError(err.ToString(), respCode);
			}
		}

		// Deal with an error in HTTP msg processing
		protected bool msgError(uint inc)
		{
			// Increment error count
			m_msgErrorCount += inc;

			// Should we give up?
			if(m_msgErrorCount >= MAX_PROCESSING_ERRORS)
				return true;
			return false;
		}


		/********/
		/* Data */
		/********/

		// A buffer for storing read data
		protected RawBuff					m_buff;

		// What type of HTTP message are we expecting next?
		protected ParseState				m_parseState;

		// Error tracking while trying to parse
		protected uint						m_msgErrorCount;
		protected uint						m_lastRecvdData;
		protected uint						m_nextReqdData;
		protected uint						m_lastErrLine;

	}

}	// namespace

