/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate an HTTP request message.
 *
 * $Id: HTTPRequestMsg.cs,v 1.2 2013/02/17 10:52:30 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Collections;
	using System.Text;

	public abstract class HTTPRequestMsg : HTTPMsg {

		/********************/
		/* Public Interface */
		/********************/

		// Construction
		public HTTPRequestMsg(Globals glob) : base(glob)
		{
			m_uri = "";
			m_requestLine = "";
			m_cfg = null;
		}

		public HTTPRequestMsg(Globals glob, HTTPMsgParms parms, string uri,
		ClientSideConfig cfg) : base(glob, parms)
		{
			m_uri = uri;
			m_requestLine = "";
			m_cfg = cfg;
	
			// Construct headers
			StringBuilder rrhdr = new StringBuilder();

			// Host address
			rrhdr.AppendFormat("Host: {0}", cfg.getResolvedServerIP());
			if(m_cfg.getServerIP().getPort() != 80)
				rrhdr.AppendFormat(":{0}", m_cfg.getServerIP().getPort());
			rrhdr.Append("\r\n");

			// Via proxy?
			if(cfg.viaProxy()) {
				ProxySettings proxySettings = m_parms.getProxySettings();
				if(proxySettings != null)
					rrhdr.Append(proxySettings.getRequestHeader(uri));
			}

			// Server authentication
			if(!m_parms.getUserAuthenticated()) {
				ServerAuthSettings serverSettings = m_parms.getServerAuthSettings();
				rrhdr.Append(serverSettings.getRequestHeader(uri));
			}

			m_reqRespHeader += rrhdr.ToString();
		}

		public HTTPRequestMsg(HTTPRequestMsg that) : base(that)
		{
			m_uri = that.m_uri;
			m_requestLine = that.m_requestLine;
			m_cfg = that.m_cfg;
		}


		/***********************/
		/* Protected Interface */
		/***********************/


		/***********/
		/* Methods */
		/***********/

		// Get start line string
		protected override string getStartLine()
		{
			// Base class has no start line, so just return ours
			return m_requestLine;
		}

		// Attempt to parse headers from a raw byte-stream
		protected override void parseHdrs(RawBuff data, ref bool eoh, ref string hdrline)
		{
			bool ok = true;
			eoh = false;

			// Parse as much as we can
			while(ok && !eoh) {
				// Let base class have a go
				base.parseHdrs(data, ref eoh, ref hdrline);
				if(!eoh) {
					/* Base class didn't like a header, let us try */

					// Is it one of ours?
					if(hdrline.StartsWith("Authorization:")) {
						// Response from client to authenticate itself
						int scmBeg = hdrline.IndexOf(' ');
						if(scmBeg >= 0) {
							// Get authentication scheme used
							string authLine = hdrline.Substring(scmBeg + 1);
							string schemeName = authLine;

							// Get authentication arguments
							ArrayList argList = new ArrayList();
							int scmEnd = authLine.IndexOf(' ');
							if(scmEnd > 0) {
								schemeName = authLine.Substring(0, scmEnd);
								string schemeArgs = authLine.Substring(scmEnd + 1);
								string[] argStrs = schemeArgs.Split(' ');
								for(int i = 0; i < argStrs.Length; ++i) {
									string arg = argStrs[i];
									if(arg.Length > 0) {
										// Strip redundant commas from the end of any parms
										char[] trim = { ',' };
										arg = arg.TrimEnd(trim);
										argList.Add(arg);
									}
								}
							}

							// Save parsed authentication scheme details
							m_authArgs[schemeName] = argList;
						}
						else {
							throw new HTTPMsgParsingError("Failed to parse server auth from response",
								m_parsingLine);
						}
					}

					else {
						// Unknown header
						ok = false;
					}
				}
			}
		}

		// Get URL to include in request
		protected string getRequestURI()
		{
			StringBuilder uri = new StringBuilder();

			// Are we using a proxy?
			if(m_cfg != null && m_cfg.viaProxy()) {
				uri.AppendFormat("http://{0}:{1}", m_cfg.getServerIP().getHost(),
					m_cfg.getServerIP().getPort());
			}

			// Append base uri
			uri.Append(m_uri);

			// Are we using our header fields?
			if(!m_parms.getUseHeaderFields() || m_parms.getState() == Tunnel.TunnelState.CONNECT) {
				// Need to encode tunneling header fields in URI
				uri.AppendFormat("?{0}", Utils.URLEncode(m_ourHeader));
			}

			return uri.ToString();
		}


		/********/
		/* Data */
		/********/

		// URI for message
		protected string						m_uri;

		// Request line
		protected string						m_requestLine;

		// Reference to global client config
		protected ClientSideConfig			m_cfg;

	}

}	// namesapce

