/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate an HTTP response message.
 *
 * $Id: HTTPResponseMsg.cs,v 1.1 2008/07/25 15:22:38 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Collections;
	using System.Text;
	using System.Text.RegularExpressions;

	class HTTPResponseMsg : HTTPMsg {

		/********************/
		/* Public Interface */
		/********************/

		// Construction/destruction
		public HTTPResponseMsg(Globals glob) : base(glob)
		{
			m_responseCode = 200;
			m_responsePhrase = "OK";
			m_responseLine = "";
			m_repContentLen = 0;

			// Construct headers
			StringBuilder enthdr = new StringBuilder();
			enthdr.Append("Content-Length: 0\r\n");
			m_entityHeader += enthdr.ToString();
		}

		public HTTPResponseMsg(Globals glob, HTTPMsgParms parms, ServerSideConfig cfg)
		: base(glob, parms)
		{
			m_responseCode = m_parms.getRespCode();
			m_responsePhrase = m_parms.getRespPhrase();
			m_responseLine = "";
			m_repContentLen = 0;

			// Construct response line
			StringBuilder resp = new StringBuilder();
			resp.AppendFormat("HTTP/{0} {1} {2}", m_parms.getVer(), m_responseCode,
				m_responsePhrase);
			m_responseLine = resp.ToString();

			// Server authentication fields
			if(!m_parms.getUserAuthenticated()) {
				ServerAuthSettings authSettings = m_parms.getServerAuthSettings();
				if(authSettings != null)
					m_reqRespHeader += authSettings.getResponseHeader(m_parms.getURI());
			}

			// AuthReq fields
			if(m_parms.getState() == Tunnel.TunnelState.AUTHREQ) {
				// Quotas
				StringBuilder quotastr = new StringBuilder();
				Quotas quotas = m_parms.getUser().getQuotas();
				quotastr.AppendFormat("HT-Quo: {0}\r\n", quotas.ToString());
				m_ourHeader += quotastr.ToString();
			}

			// Not using headers for tunneling?
			if(!m_parms.getUseHeaderFields()) {
				// Include tunneling headers in payload
				RawBuff ourHeaders = new RawBuff(m_ourHeader);
				ourHeaders.append(0);
				m_parms.getPayload().getPayload().prepend(ourHeaders);
			}

			// Construct headers
			StringBuilder enthdr = new StringBuilder();
			enthdr.AppendFormat("Content-Length: {0}\r\n",
				m_parms.getPayload().getPayload().getLength());
			m_entityHeader += enthdr.ToString();
		}

		public HTTPResponseMsg(HTTPResponseMsg that) : base(that)
		{
			m_responseCode = that.m_responseCode;
			m_responsePhrase = that.m_responsePhrase;
			m_responseLine = that.m_responseLine;
			m_repContentLen = that.m_repContentLen;
		}

		// Construction from a raw read data stream
		public HTTPResponseMsg(Globals glob, RawBuff readdata) : base(glob)
		{
			// Construct members
			m_responseCode = 0;
			m_responsePhrase = "";
			m_responseLine = "";
			m_repContentLen = 0;

			uint start = 0, end = 0;

			/* Look for basic components of response message header -
				HTTP/x.y - protocol version
				int		- response code
				string	- response phrase
			*/

			if(!readdata.findString("HTTP/", ref start, ref end)) {
				throw new HTTPMsgParsingError("Response protocol id not found",
					m_parsingLine);
			}

			uint verStart = end;
			if(!readdata.findString(" ", ref start, ref end)) {
				throw new HTTPMsgParsingError("Space after response protocol id not found",
					m_parsingLine);
			}

			// Extract protocol version
			m_parms.setVer(readdata.makeString(verStart, start));

			uint nlStart = 0, nlEnd = 0;
			if(!readdata.findNewLine(ref nlStart, ref nlEnd))
				throw new HTTPMsgParsingError("Expected newline not found", m_parsingLine);

			// Extract response code and phrase
			string codePhraseStr = readdata.makeString(end, nlStart);
			Regex reg = new Regex(@"\S+\s*");
			MatchCollection matches = reg.Matches(codePhraseStr);
			if(matches.Count < 2) {
				throw new HTTPMsgParsingError("Response code/phrase not found", m_parsingLine);
			}

			// Copy entire repsonse line, and then we've finished with it
			m_responseLine = readdata.makeString(start, nlEnd);
			m_responseCode = UInt32.Parse(matches[0].ToString());
			m_responsePhrase = "";
			for(int i = 1; i < matches.Count; ++i)
				m_responsePhrase += matches[i].ToString();
			readdata.erase(nlEnd);
			m_parsingLine++;

			#if DEBUG_TUN
			StringBuilder nfo = new StringBuilder();
			nfo.AppendFormat("HTTPResponseMsg::HTTPResponseMsg() - Got response code: {0}",
				m_responseCode);
			nfo.AppendFormat(", protocol: {0}", m_parms.getVer());
			m_glob.DPRINTV(nfo.ToString());
			#endif

			// Look for headers
			bool eoh = false;
			string hdrline = "";
			parseHdrs(readdata, ref eoh, ref hdrline);

			// Expect final newline to terminate non-content part of message
			if(!readdata.findNewLine(ref start, ref end)) {
				throw new HTTPMsgParsingError("Response final newline terminator not found",
					m_parsingLine);
			}
			readdata.erase(end);
			m_parsingLine++;

			// Now look for content
			extractContent(readdata);

			// Did we see our headers?
			if(!allRespHeaders()) {
				// Normal response?
				if(m_responseCode == 200) {
					// Remember we seem to be using non-header based tunneling
					m_parms.setUseHeaderFields(false);

					// Look for end of headers in payload marker
					uint headersEnd = 0;
					RawBuff payload = m_parms.getPayload().getPayload();
					if(payload.findChar(0, ref headersEnd)) {
						// Extract headers from payload & try to parse them
						uint headersLength = headersEnd;
						RawBuff payloadHeaders = new RawBuff(payload, headersLength);
						payloadHeaders.append("\r\n");
						parseHdrs(payloadHeaders, ref eoh, ref hdrline);

						// Remove headers from proper payload
						payload.erase(headersLength + 1);
					}
					else {
						throw new HTTPMsgParsingError("End of headers in payload not found",
							m_parsingLine);
					}
				}
			}
		}


		// Write ourselves out as a stream of bytes
		public override void output(RawBuff dest)
		{
			// Base class stuff
			base.output(dest);

			// Write out main payload data
			dest.append(m_parms.getPayload().getPayload());
		}

		// Get our response code
		public uint getResponseCode()
		{
			return m_responseCode;
		}


		/***********************/
		/* Protected Interface */
		/***********************/


		/***********/
		/* Methods */
		/***********/

		// Parse and extract any message content
		protected override void extractContent(RawBuff readdata)
		{
			m_glob.DPRINTV("HTTPResponseMsg::extractContent");

			// Are we expecting any content data?
			if(m_repContentLen > 0) {
				#if DEBUG_TUN
				StringBuilder s1 = new StringBuilder();
				s1.AppendFormat("HTTPResponseMsg::extractContent() - Expecting {0}",
					m_repContentLen);
				s1.Append(" bytes of content");
				m_glob.DPRINTV(s1.ToString());
				#endif

				uint recvdLen = readdata.getLength();
				if(recvdLen >= m_repContentLen) {
					// Copy content data to end of message
					m_parms.getPayload().getPayload().append(readdata, m_repContentLen);

					// Remove extracted data from buffer
					readdata.erase(m_repContentLen);
				}
				else {
					#if DEBUG_TUN
					StringBuilder s2 = new StringBuilder();
					s2.AppendFormat("HTTPResponseMsg::extractContent() - Only {0}", recvdLen);
					s2.Append(" bytes left in buffer");
					m_glob.DPRINTV(s2.ToString());
					#endif

					throw new NotEnoughDataParsingError("Not enough content data", m_repContentLen,
						recvdLen, m_parsingLine);
				}
			}
		}

		// Get start line string
		protected override string getStartLine()
		{
			// Base class has no start line, so just return ours
			return m_responseLine;
		}

		// Attempt to parse headers from a raw byte-stream
		protected override void parseHdrs(RawBuff data, ref bool eoh, ref string hdrline)
		{
			bool ok = true;
			eoh = false;

			// Parse as much as we can
			while(ok && !eoh) {
				// First let base class have a go
				base.parseHdrs(data, ref eoh, ref hdrline);
				if(!eoh) {
					/* Base class didn't like a header, let us try */

					// Is it one of ours?
					if(hdrline.StartsWith("Proxy-Authenticate:") ||
					hdrline.StartsWith("WWW-Authenticate:")) {
						// Request from proxy/server to authenticate ourselves
						int scmBeg = hdrline.IndexOf(' ');
						if(scmBeg >= 0) {
							// Get suggested authentication scheme
							string authLine = hdrline.Substring(scmBeg + 1);
							string schemeName = authLine;

							// Get authentication arguments
							ArrayList argList = new ArrayList();
							int scmEnd = authLine.IndexOf(' ');
							if(scmEnd > 0) {
								schemeName = authLine.Substring(0, scmEnd);
								string schemeArgs = authLine.Substring(scmEnd + 1);
								string[] argStrs = schemeArgs.Split(' ');
								for(int i = 0; i < argStrs.Length; ++i) {
									string arg = argStrs[i];
									if(arg.Length > 0) {
										// Strip redundant commas from the end of any parms
										char[] trim = { ',' };
										arg = arg.TrimEnd(trim);
										argList.Add(arg);
									}
								}
							}

							// Save parsed authentication scheme details
							m_authArgs[schemeName] = argList;
						}
						else {
							throw new HTTPMsgParsingError("Failed to parse proxy/server auth from response",
								m_parsingLine);
						}
					}
					else if(hdrline.StartsWith("HT-Quo:")) {
						// Tunnel quotas
						int i = hdrline.IndexOf(' ');
						if(i >= 0) {
							string quotastr = hdrline.Substring(i + 1);
							m_glob.DPRINTV("HTTPResponseMsg::parseHdrs() - Got quotas: " + quotastr);
							Quotas quotas = new Quotas(quotastr);
							m_parms.setQuotas(quotas);
							m_seenQuotas = true;
							m_lastTunId = "HT-Quo: ";
						}
						else {
							throw new HTTPMsgParsingError("Failed to parse quotas from msg",
								m_parsingLine);
						}
					}
					else if(hdrline.StartsWith("Content-Length:")) {
						// Get reported content length
						Regex reg = new Regex(@"\S+\s*");
						MatchCollection matches = reg.Matches(hdrline);
						if(matches.Count == 2) {
							m_repContentLen = UInt32.Parse(matches[1].ToString());
						}
						else {
							throw new HTTPMsgParsingError("Failed to parse content length " +
								"from msg", m_parsingLine);
						}
					}
				}
			}
		}


		/********/
		/* Data */
		/********/

		// Response code
		protected uint						m_responseCode;

		// Response phrase
		protected string					m_responsePhrase;

		// Response line
		protected string					m_responseLine;

		// Reported message data length
		protected uint						m_repContentLen;

	}

}	// namespace

