/* (C) 2007 Richard Mills / The Linux Academy Ltd.
 *
 * HTTP Tunneler instance. Top level library interface.
 *
 * $Id: HTTPTunneler.cs,v 1.4 2009/03/20 11:55:49 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Collections;
	using System.Net.Sockets;
	using System.Text;
	using System.Threading;
	using System.Windows.Forms;

	public class HTTPTunneler {

		/********************/
		/* Public Interface */
		/********************/


		/**********************************/
		/* HTTPTunneler instance handling */
		/**********************************/

		// Get singleton instance
		public static HTTPTunneler Get()
		{
			// Lock
			lock(c_mtx) {
				// Create singleton instance if required
				if(c_tunneler == null) {
					c_tunneler = new HTTPTunneler();
				}

				// Return instance
				return c_tunneler;
			}
		}

		// Release singleton instance
		public static void Release()
		{
			// Lock
			lock(c_mtx) {
				// Delete singleton instance if required
				if(c_tunneler != null) {
					c_tunneler.destroyed();
					c_tunneler = null;
				}
			}
		}


		// Set name to use when logging
		public static void SetName(string name)
		{
			c_name = name;
		}


		// Get library version
		public static string GetLibVersion()
		{
			StringBuilder ver = new StringBuilder();
			ver.AppendFormat("{0}", Globals.LIBVERSION);
			return ver.ToString();
		}

		// Get library build details
		public static string GetLibBuildName()
		{
			StringBuilder build = new StringBuilder();
			build.AppendFormat("libhttptun version {0} [SSL]", GetLibVersion());

			#if DEBUG_TUN
			build.Append(" [Debug]");
			#endif

			return build.ToString();
		}


		/*************/
		/* Accessors */
		/*************/

		// Get the global config
		public Globals GetGlobals()
		{
			lock(this) {
				return m_glob;
			}
		}


		/*******************************/
		/* Control the HTTP tunnelling */
		/*******************************/

		// Start the background HTTP tunnelling
		public void Start()
		{
			// Take lock
			lock(this) {
				// Check we aren't already running
				if(!m_running) {
					// Create configured tunnel servers from config
					m_glob.cfg().createLocalServers();

					// Indicate we are now running
					m_running = true;
				}
				else {
					throw new TunnelError("HTTPTunneler already started");
				}
			}
		}

		// Stop the background HTTP tunnelling
		public void Stop()
		{
			// Take lock
			lock(this) {
				// Check we are running
				if(m_running) {
					// Destroy configured tunnel servers
					m_glob.cfg().destroyLocalServers();

					// Indicate we are no longer running
					m_running = false;
				}
			}
		}


		/*********************************/
		/* Configure the HTTP tunnelling */
		/*********************************/

		// Configure from default config file
		public void ConfigureFromDefaultFile()
		{
			// Take lock
			lock(this) {
				// Get default config file name
				string dfltFile = m_glob.cfg().getConfigFileName();
				//MessageBox.Show("Cfg file: " + dfltFile);

				// Configure using this file
				ConfigureFromFile(dfltFile);
			}
		}

		// Configure from a file
		public void ConfigureFromFile(string fileName)
		{
			// Take lock
			lock(this) {
				// Parse config from specified file
				OptionsParser parser = new OptionsParser(m_glob);
				parser.parseCfgFile(fileName);

				// Setup debug settings
				#if DEBUG_TUN
				m_glob.cfg().setDaemon(false);
				m_glob.cfg().setDebug(Globals.VERBOSE);
				m_glob.cfg().setShowLog(true);
				#endif
			}
		}

		// Reconfigure from a file
		public void ReconfigureFromFile(string fileName)
		{
			// Take lock
			lock(this) {
				// Stop the tunneling, reread the config, then restart tunneling
				Stop();
				ConfigureFromFile(fileName);
				Start();
			}
		}


		// Regenerate default config file with defaults
		public void RegenerateDefaultCfgFile()
		{
			// Take lock
			lock(this) {
				// Get default config file name
				string dfltFile = m_glob.cfg().getConfigFileName();

				// Recreate this file
				RegenerateCfgFile(dfltFile);
			}
		}

		// Regenerate the named config file with defaults
		public void RegenerateCfgFile(string fileName)
		{
			// Take lock
			lock(this) {
				// Get parse to create default cfg
				OptionsParser parser = new OptionsParser(m_glob);
				parser.createDefaultConfigFile(fileName);
			}
		}


		// Directly configure global settings
		public void ConfigureGlobal(GlobalConfigBlob blob)
		{
			// Take lock
			lock(this) {
				// Parse configuration
				OptionsParser parser = new OptionsParser(m_glob);
				parser.parseBlob(blob);
			}
		}


		// Directly configure a client side tunneler
		public TunnelerID ConfigureClientTunneler(ClientConfigBlob blob)
		{
			// Take lock
			lock(this) {
				// Parse configuration
				OptionsParser parser = new OptionsParser(m_glob);
				parser.parseBlob(blob);
				TunnelerID tid = m_glob.cfg().getLastTunnelerID();

				// Do any additional config required
				m_glob.cfg().additionalClientConfig(tid, blob);

				// Launch newly configured tunneler
				m_glob.cfg().createNamedLocalServer(tid);

				// Return ID of new client tunneler
				return tid;
			}
		}


		// Directly configure a server side tunneler
		public TunnelerID ConfigureServerTunneler(ServerConfigBlob blob)
		{
			// Take lock
			lock(this) {
				// Parse configuration
				OptionsParser parser = new OptionsParser(m_glob);
				parser.parseBlob(blob);
				TunnelerID tid = m_glob.cfg().getLastTunnelerID();

				// Do any additional config required
				m_glob.cfg().additionalServerConfig(tid, blob);

				// Launch newly configured tunneler
				m_glob.cfg().createNamedLocalServer(tid);

				// Return ID of new server tunneler
				return tid;
			}
		}


		// Initiate a clean shutdown on a configured tunneler
		public void ShutdownTunneler(TunnelerID tid)
		{
			// Take lock
			lock(this) {
				// Shutdown named tunneler instance
				m_glob.cfg().shutdownNamedLocalServerTunnels(tid);
			}
		}

		// Remove configured tunneler directly
		public void RemoveTunneler(TunnelerID tid)
		{
			// Take lock
			lock(this) {
				// Destroy named tunneler instance
				m_glob.cfg().destroyNamedLocalServer(tid);
			}
		}


		/********************************/
		/* Socket based HTTP tunnelling */
		/********************************/

		// Launch a HTTP tunneling server socket
		public ServerSockTunHandle LaunchServerSocket(SockTunServerConfigBlob blob)
		{
			// Configure server tunneler
			TunnelerID tid = ConfigureServerTunneler(blob);
			return new ServerSockTunHandle(m_glob, tid);
		}


		// Launch a HTTP tunneling client socket
		public ClientSockTunHandle LaunchClientSocket(SockTunClientConfigBlob blob)
		{
			// Configure client tunneler
			TunnelerID tid = ConfigureClientTunneler(blob);
			return new ClientSockTunHandle(m_glob, tid);
		}


		/***********************************/
		/* Query the HTTP tunnelling state */
		/***********************************/

		// Is the specified tunneler currently active?
		public bool IsTunnelerActive(TunnelerID tid)
		{
			// Fetch stats for the specified tunneler
			ArrayList stats = new ArrayList();
			m_glob.cfg().getTunnelerTunnelStats(tid, stats);

			// If tunneler is active, we'll have some tunnel stats
			return (stats.Count > 0);
		}


		/*********************/
		/* Private Interface */
		/*********************/


		/***********/
		/* Methods */
		/***********/

		// Construction
		private HTTPTunneler()
		{
			m_glob = null;
			m_running = false;

			// Perform initialisation
			initialise();
		}

		// Destruction
		private void destroyed()
		{
			// Take lock
			lock(this) {
				// Stop any active tunneling
				Stop();

				// Perform shutdown
				shutdown();
			}
		}


		// Perform one-off initialisation - Caller holds mutex
		private void initialise()
		{
			// Create globals & default config
			m_glob = new Globals(c_name);
		}

		// Perform one-off shutdown - Caller holds mutex
		private void shutdown()
		{
			// Let globals tidy up
			if(m_glob != null) {
				m_glob.shutdown();
				m_glob = null;
			}
		}


		/********/
		/* Data */
		/********/

		// The global data
		private Globals					m_glob;

		// Flag to indicate whether we have started running
		private bool						m_running;

		// Singleton instance
		private static HTTPTunneler	c_tunneler = null;
		private static Mutex				c_mtx = new Mutex();

		// Name to use when logging
		private static string			c_name = "libhttptun";

	}

}	// namespace

