/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * An HTTP data writer.
 *
 * $Id: HTTPWriter.cs,v 1.1 2008/07/25 15:22:38 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Collections;
	using System.Text;

	public class HTTPWriter : Writer {

		// Construction
		public HTTPWriter(Globals glob, Connection conn) : base(glob, conn)
		{
			m_glob.DPRINTV("HTTPWriter::HTTPWriter()");

			// Client or server?
			if(m_conn.getTunneler().getServer().getConfig().isClient()) {
				m_msgState = MsgState.INIT_CONNECT;
			}
			else {
				m_msgState = MsgState.RESP_CONNECT;
			}
		}


		/***********************/
		/* Protected Interface */
		/***********************/


		/*********/
		/* Types */
		/*********/

		// States for message construction
		protected enum MsgState {
			UNDEFINED, INIT_CONNECT, RESP_CONNECT, RECEIVING, SENDING
		};


		/***********/
		/* Methods */
		/***********/

		// Attempt to process some data to send out
		protected override void processData(MsgParms parms, RawBuff dest)
		{
			#if DEBUG_LOG_DATA
			// Log data
			m_glob.logPayloadWrite(parms.getPayload().getPayload());
			#endif

			// Create HTTP msg
			HTTPMsgParms httpParms = (HTTPMsgParms)parms;
			HTTPMsg msg = createMsg(httpParms);

			// Write message out as a byte stream to destination buffer
			msg.output(dest);
		}

		// Construct appropriate HTTP message to encapsulate some data
		protected HTTPMsg createMsg(HTTPMsgParms parms)
		{
			// Construct HTTP message of appropriate type containing our pending data
			HTTPMsg msg = null;

			if(m_msgState == MsgState.INIT_CONNECT) {
				// Include our version ID in msg to server
				parms.setCliVer(m_glob.cfg().getVersionStr());

				// Create get connect msg
				m_glob.DPRINTV("HTTPWriter::createMsg() - CONNECTING with GET");
				msg = new HTTPGetMsg(m_glob, parms, createRequestURL(),
					m_conn.getTunneler().getClientSideConfig());
				m_msgState = MsgState.SENDING;
			}
			else if(m_msgState == MsgState.RESP_CONNECT) {
				// Create response msg
				m_glob.DPRINTV("HTTPWriter::createMsg() - RESPONDING to connect");
				msg = new HTTPResponseMsg(m_glob, parms,
					m_conn.getTunneler().getServerSideConfig());
				m_msgState = MsgState.RECEIVING;
			}
			else if(m_msgState == MsgState.SENDING) {
				// Create get msg
				m_glob.DPRINTV("HTTPWriter::createMsg() - SENDING with GET");
				msg = new HTTPGetMsg(m_glob, parms, createRequestURL(),
					m_conn.getTunneler().getClientSideConfig());
			}
			else if(m_msgState == MsgState.RECEIVING) {
				// Create response msg
				m_glob.DPRINTV("HTTPWriter::createMsg() - RECEIVING with 200 OK");
				msg = new HTTPResponseMsg(m_glob, parms,
					m_conn.getTunneler().getServerSideConfig());
			}
			else {
				m_glob.DPRINTV("HTTPWriter::createMsg() - Error: Unknown state");
				throw new WriteError("Attempt to get pending data while in unknown state");
			}

			return msg;
		}

		// Create a new URL for some request
		protected string createRequestURL()
		{
			StringBuilder url = new StringBuilder();

			// Increment file string
			lock(c_requestStr) {
				bool carry = false;
				int col = 0;
				do {
					carry = false;

					if(c_requestStr.Count <= col)
						c_requestStr.Add(0);

					int val = (int)c_requestStr[col] + 1;
					if(val > m_alphabet.Length) {
						c_requestStr[col] = 1;
						carry = true;
						col++;
					}
					else {
						c_requestStr[col] = val;
					}
				} while(carry);

				// Build complete URL
				url.Append("/");
				for(col = 0; col < c_requestStr.Count; ++col)
					url.AppendFormat("{0}", m_alphabet[(int)c_requestStr[col] - 1]);
				url.Append(".htm");
			}

			return url.ToString();
		}


		/********/
		/* Data */
		/********/

		// Our current message construction state
		protected MsgState				m_msgState;

		// Static data for generating unique URLS
		protected static ArrayList		c_requestStr = new ArrayList();


		/*********************/
		/* Private Interface */
		/*********************/


		/********/
		/* Data */
		/********/

		private static char[] m_alphabet = { 'a','b','c','d','e','f','g','h','i',
			'j', 'k','l','m','n','o','p','q','r','s','t','u','v','w','x','y','z',
			'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q',
			'R','S','T','U','V','W','X','Y','Z','0','1','2','3','4','5','6','7',
			'8','9' };

	}

}	// namespace

