/* (C) 2013 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate an IP address (actually, a TCP/UDP address).
 *
 * $Id$
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System.Net.Sockets;
	using System.Text;

	public class IPAddr {

		/********************/
		/* Public Interface */
		/********************/


		/***********/
		/* Methods */
		/***********/

		// Construction
		public IPAddr()
		{
			m_host = "";
			m_port = 0;
			m_family = AddrInfo.Family.UNKNOWN_FAMILY;
		}

		public IPAddr(string host, uint port)
		{
			m_host = host;
			m_port = port;
			m_family = AddrInfo.Family.UNKNOWN_FAMILY;

			// Work out address family
			setFamilyFromHost(AddrInfo.Family.IPV4);
		}

		public IPAddr(string host, uint port, AddrInfo.Family family)
		{
			m_host = host;
			m_port = port;
			m_family = AddrInfo.Family.UNKNOWN_FAMILY;

			// Work out address family
			setFamilyFromHost(family);
		}


		// Operators
		public static bool operator==(IPAddr ip1, IPAddr ip2)
		{
			return (ip1.Equals(ip2));
		}
		public static bool operator!=(IPAddr ip1, IPAddr ip2)
		{
			return !(ip1.Equals(ip2));
		}

		// As recommended, also override Equals() & GetHashCode()
		public override bool Equals(object that)
		{
			bool res = false;
			if(that != null && that is IPAddr) {
				IPAddr ip = (IPAddr)that;
				res = (
					getHost() == ip.getHost() &&
					getPort() == ip.getPort() &&
					getFamily() == ip.getFamily()
				);
			}

			return res;
		}
		public override int GetHashCode()
		{
			return ToString().GetHashCode();
		}

		// Create a representation of ourselves as a string
		public override string ToString()
		{
			StringBuilder str = new StringBuilder();

			if(getHost().Length == 0)
				str.Append("<No Host>");
			else
				str.Append(getHost());

			str.AppendFormat(";{0} ", getPort());

			if(getFamily() == AddrInfo.Family.IPV4)
				str.Append("(IPV4)");
			else if(getFamily() == AddrInfo.Family.IPV6)
				str.Append("(IPV6)");
			else
				str.Append("(Unknown)");

			return str.ToString();
		}


		/*************/
		/* Accessors */
		/*************/

		// Get host
		public string getHost()
		{
			return m_host;
		}

		// Get port
		public uint getPort()
		{
			return m_port;
		}

		// Get address family
		public AddrInfo.Family getFamily()
		{
			return m_family;
		}

		// Get address family as a string
		public string getFamilyStr()
		{
			return AddrInfo.getFamilyStr(m_family);
		}


		/************/
		/* Mutators */
		/************/

		// Set port
		public void setPort(uint port)
		{
			m_port = port;
		}

		// Set host
		public void setHost(string host)
		{
			m_host = host;

			// Work out address family
			setFamilyFromHost(AddrInfo.Family.IPV4);
		}


		/*********************/
		/* Private Interface */
		/*********************/


		/***********/
		/* Methods */
		/***********/

		// Set address family from host
		private void setFamilyFromHost(AddrInfo.Family defaultFamily)
		{
			// Set default for when we can't work it out
			m_family = defaultFamily;

			// Try and work it out from resolved address info
			if(m_host.Length != 0 && m_host != "-") {
				try {
					AddrInfo addrInfo = Utils.getAddrInfo(m_host);
					m_family = addrInfo.getFamily();
				}
				catch(AddrLookupError) {}
			}
		}


		/********/
		/* Data */
		/********/

		// Address details
		private string					m_host;
		private uint					m_port;
		private AddrInfo.Family		m_family;

	}

}	// namespace

