/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Perform logging to either stderr or syslog.
 *
 * $Id: Logger.cs,v 1.5 2010/06/29 14:01:36 rich Exp $
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Text;

	public class Logger {

		// Construction 
		public Logger(Globals glob, string pname, bool time)
		{
			m_glob = glob;
			m_procName = pname;
			m_time = time;
			m_logForm = null;
			m_showLogForm = false;
			m_logToConsole = false;

			#if DEBUG_TUN
			m_logToConsole = true;
			#endif
		}


		/*************/
		/* Accessors */
		/*************/

		// Get whether the logging window is shown
		public bool getShowLogging()
		{
			return m_showLogForm;
		}


		/************/
		/* Mutators */
		/************/

		// Shutdown
		public void shutdown()
		{
			lock(this) {
				// Close logger form
				if(m_logForm != null) {
					if(!m_logForm.IsDisposed) {
						m_logForm.Close();
					}
					else {
						m_logForm.shutdown();
					}
					m_logForm = null;
				}
			}
		}


		// Show/hide logging window
		public void setShowLogging(bool show)
		{
			lock(this) {
				if(show) {
					// Need to show logging window?
					if(!m_showLogForm) {
						m_showLogForm = true;
						m_logToConsole = false;
						m_logForm = new LoggerForm(m_glob);
						m_logForm.Show();
					}
				}
				else {
					// Need to hide logging window
					if(m_showLogForm) {
						m_showLogForm = false;
						m_logForm.Close();
						m_logForm = null;

						#if DEBUG_TUN
						m_logToConsole = true;
						#endif
					}
				}
			}
		}


		/*******************/
		/* Logging methods */
		/*******************/

		// Log an error
		public void error(string msg)
		{
			// Check debug level
			if(getDebugLevel() >= Globals.ERROR) {
				// Build string
				StringBuilder line = new StringBuilder();
				lock(this) {
					if(m_time)
						line.AppendFormat("{0}: ", getTimeStr());

					// Tag message as an error
					line.AppendFormat("{0} ERROR: {1}", m_procName, msg);
					logLine(line);
				}
			}
		}

		// Log a warning
		public void warn(string msg)
		{
			// Check debug level
			if(getDebugLevel() >= Globals.WARN) {
				// Build string
				StringBuilder line = new StringBuilder();
				lock(this) {
					if(m_time)
						line.AppendFormat("{0}: ", getTimeStr());

					// Tag message as a warning
					line.AppendFormat("{0} WARN: {1}", m_procName, msg);
					logLine(line);
				}
			}
		}

		// Log an informational message
		public void info(string msg)
		{
			// Check debug level
			if(getDebugLevel() >= Globals.INFO) {
				// Build string
				StringBuilder line = new StringBuilder();
				lock(this) {
					if(m_time)
						line.AppendFormat("{0}: ", getTimeStr());

					// Tag message as an informational
					line.AppendFormat("{0} INFO: {1}", m_procName, msg);
					logLine(line);
				}
			}
		}


		/*******************/
		/* Tracing methods */
		/*******************/

		// Log a new trace message
		public void trace(string msg, uint level)
		{
			// Check debug level
			if(getDebugLevel() >= level) {
				// Build string
				StringBuilder line = new StringBuilder();
				lock(this) {
					if(m_time)
						line.AppendFormat("{0}: ", getTimeStr());

					// Trace message
					line.Append(msg);
					logLine(line);
				}
			}
		}


		/************************/
		/* Notification methods */
		/************************/

		// Log a new notification message
		public void notif(string msg)
		{
			// Build string
			StringBuilder line = new StringBuilder();
			lock(this) {
				if(m_time)
					line.AppendFormat("{0}: ", getTimeStr());

				// Tag message as a notification
				line.AppendFormat("{0} NOTIF: {1}", m_procName, msg);
				logLine(line);
			}
		}


		/*********************/
		/* Private Interface */
		/*********************/


		/***********/
		/* Methods */
		/***********/

		// Get current debug level
		private uint getDebugLevel()
		{
			#if DEBUG_TUN
			uint level = Globals.VERBOSE;
			#else
			uint level = Globals.QUIET;
			#endif

			if(m_glob != null)
				level = m_glob.cfg().getDebug();

			return level;
		}


		// Get time as a tring for including in logging
		private string getTimeStr()
		{
			return DateTime.Now.TimeOfDay.ToString();
		}


		// Log a line - Caller holds mutex
		private void logLine(StringBuilder lineStr)
		{
			if(m_showLogForm && m_logForm != null && !m_logForm.IsDisposed) {
				m_logForm.newLogMsg(lineStr.ToString());
			}
			else if(m_logToConsole) {
				System.Console.WriteLine(lineStr.ToString());
			}
		}


		/********/
		/* Data */
		/********/

		// The process name we were started with
		private string							m_procName;

		// Whether we include timestamps in logging
		private bool							m_time;

		// Reference to globals
		private Globals						m_glob;

		// Form to log to
		private LoggerForm					m_logForm;
		private bool							m_showLogForm;
		private bool							m_logToConsole;

	}

}		// namespace HTTPTunnel

