using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Threading;
using System.Windows.Forms;

namespace HTTPTunnel
{
	using System.IO;

	public partial class LoggerForm : Form
	{
		public LoggerForm(Globals glob)
		{
			m_glob = glob;
			m_updateThread = null;
			m_updateMtx = new Mutex();
			m_shutDownMtx = new Mutex();
			m_shutDown = false;
			m_logStr = new StringBuilder();
			m_gotNewMsg = false;
			m_logPause = false;
			m_logLevel = glob.cfg().getDebug();

			// Initialise full GUI
			InitializeComponent();
			initialSetupGUI();

			// Launch update thread
			m_updateThread = new Thread(new ThreadStart(updateDisplay));
			m_updateThread.Start();
		}


		/********************/
		/* Public Interface */
		/********************/


		/***********/
		/* Methods */
		/***********/

		// Perform cleanup & shutdown
		public void shutdown()
		{
			// Take locks
			Monitor.Enter(m_updateMtx);
			Monitor.Enter(m_shutDownMtx);

			// Signal we are no longer running
			m_shutDown = true;

			// Wakeup server thread & wait for it to finish
			if(m_updateThread != null) {
				Monitor.Pulse(m_updateMtx);
				Monitor.Exit(m_updateMtx);
				Monitor.Wait(m_shutDownMtx);
				Monitor.Exit(m_shutDownMtx);
				m_updateThread.Join();
				m_updateThread = null;
			}
			else {
				Monitor.Exit(m_updateMtx);
				Monitor.Exit(m_shutDownMtx);
			}
		}


		// Queue a message for logging
		public void newLogMsg(string msg)
		{
			// Take mutex
			Monitor.Enter(m_updateMtx);

			// Check we are still running
			if(!m_shutDown) {
				m_logStr.AppendLine(msg);
				m_gotNewMsg = true;
			}

			// Release mutex
			Monitor.Exit(m_updateMtx);
		}


		/*********************/
		/* Private Interface */
		/*********************/


		/*************/
		/* Constants */
		/*************/

		// Time between log window updates
		private const int					LOG_UPDATE_TIME			= 200;


		/***********/
		/* Methods */
		/***********/

		// Setup initial GUI values
		private void initialSetupGUI()
		{
			// Show current debug level
			changeLogLevel((int)m_logLevel);
		}


		// Change the selected log level
		delegate void LogLevelDelegate(int level);
		private void changeLogLevel(int level)
		{
			if(logLevelCombo.InvokeRequired) {
				LogLevelDelegate dlg = new LogLevelDelegate(changeLogLevel);
				this.BeginInvoke(dlg, new object[] {level});
			}
			else {
				logLevelCombo.SelectedIndex = level;
			}
		}


		// Setup GUI from tunnel state - Caller holds mutex
		private void setupGUI()
		{
			// Show any changed log config values
			if(m_glob.cfg().getDebug() != m_logLevel) {
				m_logLevel = m_glob.cfg().getDebug();
				initialSetupGUI();
			}
				
			// Display new log msgs
			if(m_gotNewMsg && !m_logPause) {
				logText(m_logStr.ToString());
				m_logStr = new StringBuilder();
				m_gotNewMsg = false;
			}
		}


		// Display some log text
		delegate void LogTextDelegate(string text);
		private void logText(string text)
		{
			if(text.Length > 0) {
				if(loggerFormTextBox.InvokeRequired) {
					LogTextDelegate dlg = new LogTextDelegate(logText);
					this.BeginInvoke(dlg, new object[] {text});
				}
				else {
					loggerFormTextBox.AppendText(text);
				}
			}
		}


		// Run periodic updates for the display - Thread entry point
		private void updateDisplay()
		{
			// Take mutex
			Monitor.Enter(m_updateMtx);

			while(!m_shutDown) {
				// Sleep a while
				TimeSpan pollTime = new TimeSpan(0,0,0,0,LOG_UPDATE_TIME);
				Monitor.Wait(m_updateMtx, pollTime);

				// Update display
				if(!m_shutDown) {
					setupGUI();
				}
			}

			// Signal we have finished
			Monitor.Exit(m_updateMtx);
			Monitor.Enter(m_shutDownMtx);
			Monitor.Pulse(m_shutDownMtx);
			Monitor.Exit(m_shutDownMtx);
		}


		/********/
		/* Data */
		/*********/

		// Globals
		private Globals				m_glob;

		// Update thread
		private Thread					m_updateThread;
		private Mutex					m_updateMtx;
		private Mutex					m_shutDownMtx;
		private bool					m_shutDown;

		// Messages awaiting display
		private StringBuilder		m_logStr;
		private bool					m_gotNewMsg;

		// Flag to control log display pausing
		private bool					m_logPause;

		// Logging config values
		private uint					m_logLevel;


		/***************/
		/* Form Events */
		/***************/


		private void LoggerForm_FormClosing(object sender, FormClosingEventArgs e)
		{
			// Shutdown
			shutdown();
		}

		private void logLevelCombo_SelectedIndexChanged(object sender, EventArgs e)
		{
			// Take lock
			Monitor.Enter(m_updateMtx);
	
			// Set new log level
			m_logLevel = (uint)logLevelCombo.SelectedIndex;
			m_glob.cfg().setDebug(m_logLevel);

			// Release lock
			Monitor.Exit(m_updateMtx);
		}

		private void logPauseButton_Click(object sender, EventArgs e)
		{
			// Take lock
			Monitor.Enter(m_updateMtx);
	
			// Toggle pausing
			m_logPause = !m_logPause;
			logPauseButton.Text = (m_logPause? "UnPause" : "Pause");

			// Release lock
			Monitor.Exit(m_updateMtx);
		}

		private void logClearButton_Click(object sender, EventArgs e)
		{
			// Take lock
			Monitor.Enter(m_updateMtx);

			// Clear logger buffer
			loggerFormTextBox.Clear();
	
			// Release lock
			Monitor.Exit(m_updateMtx);
		}

		private void logSaveButton_Click(object sender, EventArgs e)
		{
			// Create dialog to get file to save to
			if(saveLogFileDialog.ShowDialog() == DialogResult.OK) {
				// Take lock
				Monitor.Enter(m_updateMtx);

				// Save log
				Stream stream = saveLogFileDialog.OpenFile();
				if(stream != null) {
					StreamWriter logFile = new StreamWriter(stream);
					logFile.Write(loggerFormTextBox.Text);
					logFile.Close();
				}

				Monitor.Exit(m_updateMtx);
			}
		}


	}
}
