/* (C) 2008 Richard Mills / The Linux Academy Ltd.
 *
 * Handle an NTLM authentication session.
 *
 * $Id: NTLM.cs,v 1.1 2008/07/25 15:22:38 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Collections;
	using System.Text;

	public class NTLM {

		// Construction
		public NTLM(Globals glob)
		{
			m_glob = glob;
			m_state = NTLMState.NTLM_NEG;
			m_dataBlock = new RawBuff();
			m_chalDomain = "";
			m_chalFlags = 0;
			m_chalChallenge = new RawBuff();
			m_chalContext = new RawBuff();
			m_chalTargets = new ArrayList();
		}


		/*************/
		/* Utilities */
		/*************/

		// Create next NTLM message
		public bool createNextMsg(RawBuff buff, string user, string domain, string passwd)
		{
			m_glob.DPRINTV("NTLM::createNextMsg()");

			bool finished = false;

			// Clear buffers to start
			buff.clear();
			m_dataBlock.clear();

			// What type of msg is next?
			switch(m_state) {
				case(NTLMState.NTLM_NEG): {
					// Create negotiate msg
					createNegotiateMsg(buff, domain);
					m_state = NTLMState.NTLM_CHAL;
					break;
				}
				case(NTLMState.NTLM_CHAL): {
					throw new AuthProcessingError("Can't create unsupported NTLM challenge msg");
				}
				case(NTLMState.NTLM_AUTH): {
					// Create auth msg
					createAuthMsg(buff, user, domain, passwd);
					finished = true;
					break;
				}
				default: {
					throw new AuthProcessingError("Can't create unknown NTLM msg type");
				}
			}

			return finished;
		}


		// Parse received NTLM message
		public void parseRecvdMsg(string msg)
		{
			m_glob.DPRINTV("NTLM::parseRecvdMsg()");

			// Base64 decode msg
			RawBuff rawMsg = Utils.Base64Decode(msg);
			RawBuff origMsg = new RawBuff(rawMsg);

			// NTLMSSP signature
			string sig = decOem(rawMsg, 7);
			rawMsg.erase(1);
			if(sig != "NTLMSSP")
				throw new AuthProcessingError("Missing NTLMSSP signature in NTLM msg");
			m_glob.DPRINTV("Parsed NTLM signature");

			// Msg type
			uint msgType = decLong(rawMsg);
			if(msgType != Convert.ToUInt32(NTLMState.NTLM_CHAL)) {
				StringBuilder err = new StringBuilder();
				err.AppendFormat("Error parsing unknown NTLM msg type {0}", msgType);
				throw new AuthProcessingError(err.ToString());
			}
			#if DEBUG_TUN
			StringBuilder nfo1 = new StringBuilder();
			nfo1.AppendFormat("Parsed NTLM msg type: {0}", ((NTLMState)msgType).ToString());
			m_glob.DPRINTV(nfo1.ToString());
			#endif

			// Target domain
			SecurityBuff domainBuff = decSecBuff(rawMsg, origMsg);

			// Flags
			m_chalFlags = decLong(rawMsg);
			#if DEBUG_TUN
			StringBuilder nfo2 = new StringBuilder();
			nfo2.AppendFormat("Parsed NTLM CHALLENGE flags: {0:x8}", m_chalFlags);
			m_glob.DPRINTV(nfo2.ToString());
			#endif

			// Decode target domain
			if((m_chalFlags & FLAG_UNICODE) != 0)
				m_chalDomain = decUnicode(domainBuff);
			else
				m_chalDomain = decOem(domainBuff);
			m_glob.DPRINTV("Parsed NTLM CHALLENGE target domain: " + m_chalDomain);

			// Challenge
			if(rawMsg.getLength() < 8)
				throw new AuthProcessingError("Missing challenge in NTLM msg");
			m_chalChallenge = new RawBuff(rawMsg, 8);
			rawMsg.erase(8);
			m_glob.DPRINTV("Parsed NTLM CHALLENGE challenge");

			// Context
			if(rawMsg.getLength() < 8)
				throw new AuthProcessingError("Missing context in NTLM msg");
			m_chalContext = new RawBuff(rawMsg, 8);
			rawMsg.erase(8);
			m_glob.DPRINTV("Parsed NTLM CHALLENGE context");

			// Target address list
			m_chalTargets.Clear();
			SecurityBuff addrListBuff = decSecBuff(rawMsg, origMsg);
			RawBuff addrList = new RawBuff(addrListBuff.getData());
			NTLMTarget.TargetType tgtType = NTLMTarget.TargetType.TGT_NONE;
			do {
				// Type & length
				uint uitgtType = decShort(addrList);
				if(uitgtType >= Convert.ToUInt32(NTLMTarget.TargetType.TGT_NONE) &&
				uitgtType < Convert.ToUInt32(NTLMTarget.TargetType.TGT_UNKNOWN))
					tgtType = (NTLMTarget.TargetType)uitgtType;
				else
					tgtType = NTLMTarget.TargetType.TGT_UNKNOWN;
				uint tgtLen = decShort(addrList);

				// End of list?
				if(tgtType != NTLMTarget.TargetType.TGT_NONE) {
					// Name
					string tgtName = decUnicode(addrList, tgtLen);

					// Add target to list
					m_chalTargets.Add(new NTLMTarget(tgtType, tgtName));
				}
			} while(tgtType != NTLMTarget.TargetType.TGT_NONE);
			#if DEBUG_TUN
			m_glob.DPRINTV("Parsed NTLM CHALLENGE targets:");
			for(int tit = 0; tit != m_chalTargets.Count; ++tit) {
				NTLMTarget tgt = (NTLMTarget)m_chalTargets[tit];
				StringBuilder nfo3 = new StringBuilder();
				nfo3.AppendFormat("  Type: {0}, Name: {1}", tgt.m_type, tgt.m_name);
				m_glob.DPRINTV(nfo3.ToString());
			}
			#endif

			// Update state
			m_state = NTLMState.NTLM_AUTH;
		}


		// Reset our state
		public void reset()
		{
			// Reset/clear data
			m_state = NTLMState.NTLM_NEG;
			m_dataBlock.clear();
			m_chalDomain = "";
			m_chalFlags = 0;
			m_chalChallenge.clear();
			m_chalContext.clear();
			m_chalTargets.Clear();
		}


		/*********************/
		/* Private Interface */
		/*********************/


		/*********/
		/* Types */
		/*********/

		// NTLM states
		private enum NTLMState { NTLM_NEG = 1, NTLM_CHAL, NTLM_AUTH };

		// Target
		private class NTLMTarget {
			public enum TargetType {
				TGT_NONE = 0, TGT_SVR, TGT_DMN, TGT_FQDNS, TGT_DNS, TGT_UNKNOWN
			};
			public NTLMTarget(TargetType ttype, string name)
			{
				m_type = ttype;
				m_name = name;
			}
			public TargetType		m_type;
			public string			m_name;
		}


		/*************/
		/* Constants */
		/*************/

		// Flags
		private const uint		FLAG_UNICODE		= 0x00000001;
		private const uint		FLAG_OEM				= 0x00000002;
		private const uint		FLAG_NTLM2			= 0x00080000;


		/***********/
		/* Methods */
		/***********/

		// Create NTLM negotiate message
		private void createNegotiateMsg(RawBuff buff, string domain)
		{
			m_glob.DPRINTV("NTLM::createNegotiateMsg()");

			try {
				// NTLMSSP signature
				buff.append("NTLMSSP");
				buff.append(0x00);

				// Msg type
				buff.append(encLong((uint)NTLMState.NTLM_NEG));

				// Flags
				buff.append(encLong(0x00088207));

				// Workstation domain & name
				string hostname = m_glob.cfg().getHostName();
				m_glob.DPRINTV("Domain name: " + domain);
				m_glob.DPRINTV("Workstation name: " + hostname);
				RawBuff wsDomainBuff = encOem(domain);
				RawBuff wsNameBuff = encOem(hostname);
				encSecBuff(wsDomainBuff, buff, 32);
				encSecBuff(wsNameBuff, buff, 32);

				// Data block
				buff.append(m_dataBlock);
			}
			catch(TunnelError e) {
				throw new AuthProcessingError("Error creating NTLM negotiate msg: " + e.Message);
			}
		}


		// Create NTLM auth message
		private void createAuthMsg(RawBuff buff, string user, string domain, string passwd)
		{
			m_glob.DPRINTV("NTLM::createAuthMsg()");

			try {
				// NTLMSSP signature
				buff.append("NTLMSSP");
				buff.append(0x00);

				// Msg type
				buff.append(encLong((uint)NTLMState.NTLM_AUTH));

				// Create LM and NTLM response fields
				byte[] lmResponse = new byte[24];
				byte[] ntlmResponse = new byte[24];
				if((m_chalFlags & FLAG_NTLM2) != 0) {
					// Use NTLM2
					m_glob.DPRINTV("Creating NTLM2 session response");
					getNTLM2SessionResponse(passwd, m_chalChallenge.getData(), lmResponse, ntlmResponse);
				}
				else {
					// Plain NTLM
					m_glob.DPRINTV("Creating NTLM response");
					getLMResponse(passwd, m_chalChallenge.getData(), lmResponse);
					getNTLMResponse(passwd, m_chalChallenge.getData(), ntlmResponse);
				}

				#if DEBUG_TUN
				m_glob.DPRINTV("Challenge: " + Utils.ToHex(m_chalChallenge.getData()));
				m_glob.DPRINTV("LM Response: " + Utils.ToHex(lmResponse));
				m_glob.DPRINTV("NTLM Response: " + Utils.ToHex(ntlmResponse));
				#endif

				// LM response
				RawBuff lmBuff = new RawBuff(lmResponse);
				encSecBuff(lmBuff, buff, 64);

				// NTLM response
				RawBuff ntlmBuff = new RawBuff(ntlmResponse);
				encSecBuff(ntlmBuff, buff, 64);

				// Domain name
				m_glob.DPRINTV("Domain name: " + domain);
				if((m_chalFlags & FLAG_UNICODE) != 0) {
					RawBuff domainName = encUnicode(domain);
					encSecBuff(domainName, buff, 64);
				}
				else {
					RawBuff domainName = encOem(domain);
					encSecBuff(domainName, buff, 64);
				}

				// User name
				m_glob.DPRINTV("User name: " + user);
				if((m_chalFlags & FLAG_UNICODE) != 0) {
					RawBuff userName = encUnicode(user);
					encSecBuff(userName, buff, 64);
				}
				else {
					RawBuff userName = encOem(user);
					encSecBuff(userName, buff, 64);
				}

				// Host name
				string host = m_glob.cfg().getHostName();
				m_glob.DPRINTV("Host name: " + host);
				if((m_chalFlags & FLAG_UNICODE) != 0) {
					RawBuff hostName = encUnicode(host);
					encSecBuff(hostName, buff, 64);
				}
				else {
					RawBuff hostName = encOem(host);
					encSecBuff(hostName, buff, 64);
				}

				// Session key
				buff.append(encLong(0));
				buff.append(encLong(0));

				// Flags
				buff.append(encLong(m_chalFlags));

				// Data block
				buff.append(m_dataBlock);
			}
			catch(TunnelError e) {
				throw new AuthProcessingError("Error creating NTLM auth msg: " + e.Message);
			}
		}


		// Get response for NTLM LM authentication
		private void getLMResponse(string passwd, byte[] challenge, byte[] response)
		{
			// Create LM hash
			byte[] hash = new byte[16];
			createLMHash(passwd, hash);

			// Create response
			createLMResponse(hash, challenge, response);
		}

		// Get response for NTLM authentication
		private void getNTLMResponse(string passwd, byte[] challenge, byte[] response)
		{
			// Create NTLM hash
			byte[] hash = new byte[16];
			createNTLMHash(passwd, hash);

			// Create response
			createLMResponse(hash, challenge, response);
		}

		// Get response for NTLM2 session authentication
		private void getNTLM2SessionResponse(string passwd, byte[] challenge,
		byte[] clientNonce, byte[] response)
		{
			// Create client nonce
			createNTLMClientNonce(clientNonce);

			// Create NTLM2 session hash
			byte[] sessionNonce = new byte[16];
			Array.Copy(challenge, 0, sessionNonce, 0, 8);
			Array.Copy(clientNonce, 0, sessionNonce, 8, 8);
			byte[] sessionHash = new byte[16];
			AuthDigest.MD5Digest(sessionNonce, sessionHash);

			// Create NTLM hash
			byte[] hash = new byte[16];
			createNTLMHash(passwd, hash);

			// Create response
			createLMResponse(hash, sessionHash, response);
		}

		// Create response for NTLM LM authentication
		private void createLMResponse(byte[] hash, byte[] challenge, byte[] response)
		{
			// Pad hash to 21 bytes
			byte[] padHash = new byte[21];
			Array.Clear(padHash, 0, 21);
			Array.Copy(hash, padHash, 16);

			// Split padded hash into 7 byte chunks
			byte[] loPadHash = new byte[7];
			byte[] midPadHash = new byte[7];
			byte[] hiPadHash = new byte[7];
			Array.Copy(padHash, 0, loPadHash, 0, 7);
			Array.Copy(padHash, 7, midPadHash, 0, 7);
			Array.Copy(padHash, 14, hiPadHash, 0, 7);

			// Create DES keys from padded hash
			byte[] loKey = new byte[8];
			byte[] midKey = new byte[8];
			byte[] hiKey = new byte[8];
			Utils.CreateDESKey(loPadHash, loKey);
			Utils.CreateDESKey(midPadHash, midKey);
			Utils.CreateDESKey(hiPadHash, hiKey);

			// Encrypt challenge with DES keys
			byte[] loResp = new byte[8];
			byte[] midResp = new byte[8];
			byte[] hiResp = new byte[8];
			Utils.DESCrypt(challenge, loKey, loResp, true);
			Utils.DESCrypt(challenge, midKey, midResp, true);
			Utils.DESCrypt(challenge, hiKey, hiResp, true);

			// Construct final response
			Array.Copy(loResp, 0, response, 0, 8);
			Array.Copy(midResp, 0, response, 8, 8);
			Array.Copy(hiResp, 0, response, 16, 8);
		}

		// Create the hash for NTLM LM authentication
		private void createLMHash(string passwd, byte[] hash)
		{
			// Convert passwd to 14 bytes uppercase
			byte[] keyBytes = new byte[14];
			Array.Clear(keyBytes, 0, 14);
			int length = passwd.Length;
			if(length > 14)
				length = 14;
			for(int i = 0; i < length; ++i)
				keyBytes[i] = (byte)(Char.ToUpper(passwd[i]));

			// Split key into 7 byte chunks
			byte[] loKeyBytes = new byte[7];
			byte[] hiKeyBytes = new byte[7];
			Array.Copy(keyBytes, 0, loKeyBytes, 0, 7);
			Array.Copy(keyBytes, 7, hiKeyBytes, 0, 7);

			// Create DES keys from passwd
			byte[] loKey = new byte[8];
			byte[] hiKey = new byte[8];
			Utils.CreateDESKey(loKeyBytes, loKey);
			Utils.CreateDESKey(hiKeyBytes, hiKey);

			// Encrypt constant with DES keys
			byte[] loHash = new byte[8];
			byte[] hiHash = new byte[8];
			Utils.DESCrypt(Encoding.UTF8.GetBytes("KGS!@#$%"), loKey, loHash, true);
			Utils.DESCrypt(Encoding.UTF8.GetBytes("KGS!@#$%"), hiKey, hiHash, true);

			// Create final combined hash
			Array.Copy(loHash, 0, hash, 0, 8);
			Array.Copy(hiHash, 0, hash, 8, 8);
		}

		// Create the hash for NTLM authentication
		private void createNTLMHash(string passwd, byte[] hash)
		{
			// Convert passwd to Unicode
			int length = passwd.Length;
			byte[] keyBytes = new byte[length * 2];
			Array.Clear(keyBytes, 0, keyBytes.Length);
			for(int i = 0; i < length; ++i) {
				keyBytes[i*2] = (byte)passwd[i];
				keyBytes[i*2 + 1] = 0;
			}

			// Generate MD4 hash of passwd
			AuthDigest.MD4Digest(keyBytes, hash);
		}

		// Create random 8 byte (padded to 24 byte) client nonce
		private void createNTLMClientNonce(byte[] nonce)
		{
			// 8 random bytes
			Random rand = new Random();
			for(int i = 0; i < 8; ++i)
				nonce[i] = (byte)rand.Next(256);

			// Pad to 24 bytes
			Array.Clear(nonce, 8, 24 - 8);
		}


		/********************/
		/* Encoding Methods */
		/********************/

		// Encode a value as a little-endian unsigned short
		private RawBuff encShort(uint val)
		{
			RawBuff usht = new RawBuff();

			// Convert 2 bytes
			byte b1 = (byte)(val & 0xFF);
			byte b2 = (byte)((val & 0xFF00) >> 8);
			usht.append(b1);
			usht.append(b2);

			return usht;
		}

		// Encode a value as a little-endian unsigned long
		private RawBuff encLong(uint val)
		{
			RawBuff ulng = new RawBuff();

			// Convert 4 bytes
			byte b1 = (byte)(val & 0xFF);
			byte b2 = (byte)((val & 0xFF00) >> 8);
			byte b3 = (byte)((val & 0xFF0000) >> 16);
			byte b4 = (byte)((val & 0xFF000000) >> 24);
			ulng.append(b1);
			ulng.append(b2);
			ulng.append(b3);
			ulng.append(b4);

			return ulng;
		}

		// Encode a string as unicode
		private RawBuff encUnicode(string str)
		{
			RawBuff ucode = new RawBuff();

			// Convert each char from original string
			for(int i = 0; i < str.Length; ++i) {
				ucode.append((byte)str[i]);
				ucode.append(0x00);
			}

			return ucode;
		}

		// Encode a string as OEM
		private RawBuff encOem(string str)
		{
			RawBuff oem = new RawBuff();

			// Convert each char from original string
			for(int i = 0; i < str.Length; ++i) {
				char ch = Char.ToUpper(str[i]);
				oem.append((byte)ch);
			}

			return oem;
		}

		// Append to a buffer in security buffer format
		private void encSecBuff(RawBuff data, RawBuff buff, uint constLen)
		{
			// Append data to data block
			uint dataLen = data.getLength();
			uint dataStart = constLen + m_dataBlock.getLength();
			m_dataBlock.append(data);

			// Append security buffer details
			buff.append(encShort(dataLen));
			buff.append(encShort(dataLen));
			buff.append(encLong(dataStart));
		}


		/********************/
		/* Decoding Methods */
		/********************/

		// Decode a little-endian unsigned short
		private uint decShort(RawBuff buff)
		{
			uint val = 0;

			// Length check
			if(buff.getLength() >= 2) {
				// Convert 2 bytes
				uint b1 = (uint)buff[0];
				uint b2 = (uint)buff[1];
				val = b1 | (b2 << 8);

				// Clear out buffer
				buff.erase(2);
			}
			else {
				throw new AuthProcessingError("Insufficient NTLM data for decoding short");
			}

			return val;
		}

		// Decode a little-endian unsigned long
		private uint decLong(RawBuff buff)
		{
			uint val = 0;

			// Length check
			if(buff.getLength() >= 4) {
				// Convert 4 bytes
				uint b1 = (uint)buff[0];
				uint b2 = (uint)buff[1];
				uint b3 = (uint)buff[2];
				uint b4 = (uint)buff[3];
				val = b1 | (b2 << 8) | (b3 << 16) | (b4 << 24);

				// Clear out buffer
				buff.erase(4);
			}
			else {
				throw new AuthProcessingError("Insufficient NTLM data for decoding long");
			}

			return val;
		}

		// Decode a unicode string
		private string decUnicode(RawBuff buff, uint len)
		{
			StringBuilder val = new StringBuilder();

			// Length check
			if(buff.getLength() >= len) {
				// Should be even number of bytes
				if(len % 2 == 0) {
					// Decode each char
					for(int i = 0; i < len / 2; ++i) {
						val.Append((char)buff[i*2]);
					}

					// Clear out buffer
					buff.erase(len);
				}
				else {
					throw new AuthProcessingError("Odd length for decoding unicode");
				}
			}
			else {
				throw new AuthProcessingError("Insufficient NTLM data for decoding unicode");
			}

			return val.ToString();
		}

		// Decode a unicode string from a security buffer
		private string decUnicode(SecurityBuff buff)
		{
			RawBuff rawBuff = buff.getData();
			return decUnicode(rawBuff, rawBuff.getLength());
		}

		// Decode an OEM string
		private string decOem(RawBuff buff, uint len)
		{
			StringBuilder val = new StringBuilder();

			// Length check
			if(buff.getLength() >= len) {
				// Decode each char
				for(int i = 0; i < len; ++i) {
					val.Append((char)buff[i]);
				}

				// Clear out buffer
				buff.erase(len);
			}
			else {
				throw new AuthProcessingError("Insufficient NTLM data for decoding OEM");
			}

			return val.ToString();
		}

		// Decode an OEM string from a security buffer
		private string decOem(SecurityBuff buff)
		{
			RawBuff rawBuff = buff.getData();
			return decOem(rawBuff, rawBuff.getLength());
		}

		// Decode a security buffer
		private SecurityBuff decSecBuff(RawBuff buff, RawBuff origMsg)
		{
			uint dataLen = 0;
			uint buffLen = 0;
			uint buffStart = 0;
			RawBuff data = new RawBuff(origMsg);

			// Check minimum length
			if(buff.getLength() >= 8) {
				// Decode security buffer parms
				dataLen = decShort(buff);
				buffLen = decShort(buff);
				buffStart = decLong(buff);

				// Get pointer to actual data
				if(origMsg.getLength() >= (buffStart + buffLen)) {
					data.erase(buffStart);
				}
				else {
					throw new AuthProcessingError("Insufficient NTLM data for decoding secbuff");
				}
			}
			else {
				throw new AuthProcessingError("Insufficient NTLM data for decoding secbuff");
			}

			return new SecurityBuff(dataLen, buffLen, buffStart, data);
		}


		/********/
		/* Data */
		/********/

		// Reference to globals
		private Globals					m_glob;

		// Current state
		private NTLMState					m_state;

		// Data block under construction
		private RawBuff					m_dataBlock;

		// Challenge msg fields
		private string						m_chalDomain;
		private uint						m_chalFlags;
		private RawBuff					m_chalChallenge;
		private RawBuff					m_chalContext;
		private ArrayList					m_chalTargets;

	}

}	// namespace

