/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Parse options from either the command line or a config file.
 *
 * $Id: OptionsParser.cs,v 1.4 2013/02/17 10:52:30 rich Exp $
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.IO;
	using System.Text;

	public class OptionsParser {

		// Construction
		public OptionsParser(Globals glob)
		{
			m_glob = glob;
			m_localIP = "";
			m_localPort = 0;
			m_localFamily = AddrInfo.Family.IPV4;
			m_sslCert = "";
			m_sslCertKey = "";
			m_cliFixed = false;
			m_cliSOCKS = false;
			m_cliProxy = false;
			m_cliSocketTun = false;
			m_cliProtocol = Enums.Protocol.TCP;
			m_cliServerIP = "";
			m_cliServerPort = 0;
			m_cliSSL = false;
			m_cliProxyIP = "";
			m_cliProxyPort = 0;
			m_cliProxyUser = "";
			m_cliProxyPasswd = "";
			m_cliDestIP = "";
			m_cliDestPort = 0;
			m_cliAuthUser = "";
			m_cliAuthPasswd = "";
			m_cliTune = "";
			m_svrFirewallIPs = "";
			m_svrFirewallPorts = "";
			m_svrAuth = false;
			m_svrSSL = false;
			m_svrSocketTun = false;

			// Initialise parsing state
			sectionFinished(false, false, false);
		}


		/********************/
		/* Public Interface */
		/********************/

		// Parse the contents of a config file
		public void parseCfgFile(string name)
		{
			StreamReader file = null;
			try {
				// Open file
				file = File.OpenText(name);
			}
			catch(Exception) {
				// Try creating default config file
				createDefaultConfigFile(name);

				// Try reading it again
				try {
					file = File.OpenText(name);
				}
				catch(Exception) {
					throw new ConfigError("Failed to open config file " + name);
				}
			}

			int lineNo = 0;
			string line;
			bool inGlobal = false, inClient = false, inServer = false;

			// Process line at a time
			while((line = file.ReadLine()) != null) {
				lineNo++;

				if(line.Length <= 1 || line[0] == '#') {
					// Comment line or blank line
					continue;
				}

				try {
					// Start of new section?
					if(line[0] == '[') {
						// What type of section?
						if(line == "[GLOBAL]") {
							// Old section possibly finished, new one starting
							sectionFinished(inGlobal, inClient, inServer);
							inGlobal = true;
							inClient = false;
							inServer = false;
						}
						else if(line == "[CLIENT]") {
							// Old section possibly finished, new one starting
							sectionFinished(inGlobal, inClient, inServer);
							inGlobal = false;
							inClient = true;
							inServer = false;
						}
						else if(line == "[SERVER]") {
							// Old section possibly finished, new one starting
							sectionFinished(inGlobal, inClient, inServer);
							inGlobal = false;
							inClient = false;
							inServer = true;
						}
					}
					else {
						// Line within existing section
						if(inGlobal)
							parseGlobal(line);
						else if(inClient)
							parseClient(line);
						else if(inServer)
							parseServer(line);
					}
				}
				catch(ConfigError e) {
					file.Close();

					// Rethrow
					StringBuilder err = new StringBuilder();
					err.AppendFormat("Error in config file at line {0}: {1}", lineNo, e.Message);
					throw new ConfigError(err.ToString());
				}
			}

			// Old section possibly finished
			sectionFinished(inGlobal, inClient, inServer);

			// Close file
			file.Close();
		}


		// Parse some config blob
		public void parseBlob(ConfigBlob blob)
		{
			// Get details from blob
			string blobType = blob.getType();
			string blobCfg = blob.getConfig();

			// What type of config are we parsing?
			bool global = (blobType == "GLOBAL");
			bool client = (blobType == "CLIENT");
			bool server = (blobType == "SERVER");

			// Parse config line at a time
			StringReader cfgStream = new StringReader(blobCfg);
			string line;
			while((line = cfgStream.ReadLine()) != null) {
				try {
					if(global)
						parseGlobal(line);
					else if(client)
						parseClient(line);
					else if(server)
						parseServer(line);
				}
				catch(ConfigError e) {
					// Rethrow
					StringBuilder err = new StringBuilder();
					err.AppendFormat("Error in {0} config blob: {1}", blobType, e.Message);
					throw new ConfigError(err.ToString());
				}
			}

			// Finished parsing
			sectionFinished(global, client, server);
		}


		// Create default config file
		public void createDefaultConfigFile(string name)
		{
			StreamWriter file = null;
			try {
				// Open file
				file = File.CreateText(name);
			}
			catch(Exception) {
				throw new ConfigError("Failed to create default config file " + name);
			}

			// Write sensible default config
			file.WriteLine("###########################");
			file.WriteLine("## Default FireDrill config");
			file.WriteLine("###########################");
			file.WriteLine();
			file.WriteLine("#################");
			file.WriteLine("## Global options");
			file.WriteLine("#################");
			file.WriteLine();
			file.WriteLine("[GLOBAL]");
			file.WriteLine("debug 2");
			file.WriteLine();
			file.WriteLine();
			file.WriteLine("#######################");
			file.WriteLine("## Client configuration");
			file.WriteLine("#######################");
			file.WriteLine();
			file.WriteLine("[CLIENT]");
			file.WriteLine("socks");
			file.WriteLine("listen 1080");

			// Pick one of the tunnel servers to default to
			Random rand = new Random();
			int svrNum = rand.Next(1,3);
			StringBuilder svrAddr = new StringBuilder();
			svrAddr.AppendFormat("tun{0}.fire-drill.com", svrNum);
			file.WriteLine("tunsvr " + svrAddr.ToString() + " 80");

			// Close file
			file.Close();
		}


		/*********************/
		/* Private Interface */
		/*********************/


		/***********/
		/* Methods */
		/***********/


		// Check and get numeric value
		private uint getNumeric(string from, string err)
		{
			if(isNumeric(from)) {
				return UInt32.Parse(from);
			}
			else {
				throw new ConfigError("Error in parameters for " + err);
			}
		}


		// Check a string is numeric
		private bool isNumeric(string str)
		{
			bool res = true;

			for(int i = 0; i < str.Length; ++i) {
				if(!Char.IsDigit(str[i])) {
					res = false;
					break;
				}
			}

			return res;
		}


		// Parse global section line
		private void parseGlobal(string line)
		{
			bool ok = false;

			// Split
			string[] parms = line.Split(' ');

			// At least one parm?
			if(parms.Length >= 1) {
				// At least 2 parms?
				if(parms.Length >= 2) {
					if(parms[0] == "debug") {
						// Set debug level
						ok = true;
						uint debug = (uint)getNumeric(parms[1], "debug");
						m_glob.cfg().setDebug(debug);
					}
					else if(parms[0] == "pidfile") {
						// Set PID file name
						ok = true;
						m_glob.cfg().setPIDFile(parms[1]);
					}
					else if(parms[0] == "datafilestem") {
						// Data log file name stem
						ok = true;
						m_glob.cfg().setDataLogFileStem(parms[1]);
					}
				}
				else {
					// 1 Arg
					if(parms[0] == "nodaemon") {
						// Set daemon flag
						ok = true;
						m_glob.cfg().setDaemon(false);
					}
					else if(parms[0] == "dotest") {
						// Set tunnel testing flag
						ok = true;
						m_glob.cfg().setTesting(true, false);
					}
					else if(parms[0] == "showlog") {
						// Show logger window
						ok = true;
						m_glob.cfg().setShowLog(true);
					}
				}
			}

			// Recognised ok?
			if(!ok)
				throw new ConfigError("Unknown GLOBAL config option: " + line);
		}


		// Parse client section line
		private void parseClient(string line)
		{
			bool ok = false;

			// Split
			string[] parms = line.Split(' ');

			// At least one parm?
			if(parms.Length >= 1) {
				// At least 2 parms?
				if(parms.Length >= 2) {
					// At least 3 parms?
					if(parms.Length >= 3) {
						// At least 4 parms?
						if(parms.Length >= 4) {
							// 4 args
							if(parms[0] == "fixed") {
								// Address of fixed tunnel destination
								ok = true;
								m_cliFixed = true;
								m_cliSOCKS = false;
								m_cliDestIP = parms[1];
								m_cliDestPort = getNumeric(parms[2], "fixed");
								if(parms[3] == "udp")
									m_cliProtocol = Enums.Protocol.UDP;
								else
									m_cliProtocol = Enums.Protocol.TCP;
							}
						}
						else {
							// 3 args
							if(parms[0] == "tunsvr") {
								// Address of server end of tunnel
								ok = true;
								m_cliServerIP = parms[1];
								m_cliServerPort = getNumeric(parms[2], "tunsvr");
							}
							else if(parms[0] == "fixed") {
								// Address of fixed tunnel destination
								ok = true;
								m_cliFixed = true;
								m_cliSOCKS = false;
								m_cliDestIP = parms[1];
								m_cliDestPort = getNumeric(parms[2], "fixed");
								m_cliProtocol = Enums.Protocol.TCP;
							}
							else if(parms[0] == "proxy") {
								// Address of proxy to use
								ok = true;
								m_cliProxy = true;
								m_cliProxyIP = parms[1];
								m_cliProxyPort = getNumeric(parms[2], "proxy");
							}
							else if(parms[0] == "proxyauth") {
								// Authorisation to use at proxy
								ok = true;
								m_cliProxyUser = parms[1];
								m_cliProxyPasswd = parms[2];
							}
							else if(parms[0] == "cliauth") {
								// Client authentication user/passwd
								ok = true;
								m_cliAuthUser = parms[1];
								m_cliAuthPasswd = parms[2];
							}
							else if(parms[0] == "listen") {
								if(isNumeric(parms[2])) {
									// Local listening address & port
									ok = true;
									m_localIP = parms[1];
									m_localPort = getNumeric(parms[2], "listen");
								}
								else {
									// Local listening port and IP version
									string ipv = parms[2].ToUpper();
									if(ipv == "IPV4" || ipv == "IPV6") {
										ok = true;
										m_localPort = getNumeric(parms[1], "listen");
										m_localFamily = (ipv == "IPV4")? AddrInfo.Family.IPV4 : AddrInfo.Family.IPV6;
									}
								}
							}
						}
					}
					else {
						// 2 args
						if(parms[0] == "listen") {
							// Local listening port
							ok = true;
							m_localPort = getNumeric(parms[1], "listen");
						}
						else if(parms[0] == "tunsvr") {
							// Address of server end of tunnel
							ok = true;
							m_cliServerIP = parms[1];
							m_cliServerPort = 80;
						}
						else if(parms[0] == "tune") {
							// Tune tunnelling behaviour
							if(parms[1] == "bursty" || parms[1] == "realtime" || parms[1] == "default") {
								ok = true;
								m_cliTune = parms[1];
							}
						}
					}
				}
				else {
					// 1 arg
					if(parms[0] == "socks") {
						// SOCKS client
						ok = true;
						m_cliSOCKS = true;
						m_cliFixed = false;
					}
					else if(parms[0] == "ssl") {
						// Use https
						ok = true;
						m_cliSSL = true;
					}
					else if(parms[0] == "socktun") {
						// Socket tunneler
						ok = true;
						m_cliSocketTun = true;
					}
				}
			}

			// Recognised ok?
			if(!ok)
				throw new ConfigError("Unknown CLIENT config option: " + line);
		}


		// Parse server section line
		private void parseServer(string line)
		{
			bool ok = false;

			// Split
			string[] parms = line.Split(' ');

			// At least one parm?
			if(parms.Length >= 1) {
				// At least 2 parms?
				if(parms.Length >= 2) {
					// At least 3 parms?
					if(parms.Length >= 3) {
						// 3 args
						if(parms[0] == "listen") {
							if(isNumeric(parms[2])) {
								// Local listening address & port
								ok = true;
								m_localIP = parms[1];
								m_localPort = getNumeric(parms[2], "listen");
							}
							else {
								// Local listening port and IP version
								string ipv = parms[2].ToUpper();
								if(ipv == "IPV4" || ipv == "IPV6") {
									ok = true;
									m_localPort = getNumeric(parms[1], "listen");
									m_localFamily = (ipv == "IPV4")? AddrInfo.Family.IPV4 : AddrInfo.Family.IPV6;
								}
							}
						}
						else if(parms[0] == "ssl") {
							// Use https
							ok = true;
							m_svrSSL = true;
							m_sslCert = parms[1];
							m_sslCertKey = parms[2];
						}
					}
					else {
						// 2 args
						if(parms[0] == "listen") {
							// Local listening port
							ok = true;
							m_localPort = getNumeric(parms[1], "listen");
						}
						else if(parms[0] == "denyips") {
							// Firewalled IPs
							ok = true;
							m_svrFirewallIPs = parms[1];
						}
						else if(parms[0] == "denyports") {
							// Firewalled ports
							ok = true;
							m_svrFirewallPorts = parms[1];
						}
					}
				}
				else {
					// 1 arg
					if(parms[0] == "svrauth") {
						// Use authentication at the server
						ok = true;
						m_svrAuth = true;
					}
					else if(parms[0] == "socktun") {
						// Socket tunneler
						ok = true;
						m_svrSocketTun = true;
					}
				}
			}

			// Recognised ok?
			if(!ok)
				throw new ConfigError("Unknown SERVER config option: " + line);
		}


		// Use parsed options for a section to create runtime config
		private void sectionFinished(bool global, bool client, bool server)
		{
			// What section has finished?
			if(global) {
				// Nothing for now
			}
			else if(client) {
				// Construct client config
				if(m_cliSOCKS && m_cliProxy) {
					// SOCKS client, via proxy
					if(gotValidLocal() && m_cliServerIP != "" && m_cliServerPort != 0 &&
					m_cliProxyIP != "" && m_cliProxyPort != 0) {
						m_glob.cfg().configureClient(m_cliProtocol, m_cliSSL, new IPAddr(m_localIP,
							m_localPort, m_localFamily), new IPAddr(m_cliServerIP, m_cliServerPort),
							m_cliAuthUser, m_cliAuthPasswd, m_cliTune, new IPAddr(m_cliProxyIP,
							m_cliProxyPort), m_cliProxyUser, m_cliProxyPasswd);
					}
					else {
						throw new ConfigError("Incomplete CLIENT SOCKS (via proxy) config");
					}
				}
				else if(m_cliSOCKS && !m_cliProxy) {
					// SOCKS client, no proxy
					if(gotValidLocal() && m_cliServerIP != "" && m_cliServerPort != 0) {
						m_glob.cfg().configureClient(m_cliProtocol, m_cliSSL, new IPAddr(m_localIP,
							m_localPort, m_localFamily), new IPAddr(m_cliServerIP, m_cliServerPort),
							m_cliAuthUser, m_cliAuthPasswd, m_cliTune);
					}
					else {
						throw new ConfigError("Incomplete CLIENT SOCKS config");
					}
				}
				else if(m_cliFixed && m_cliProxy) {
					// Direct client, via proxy
					if(gotValidLocal() && m_cliServerIP != "" && m_cliServerPort != 0 &&
					gotValidDest() && m_cliProxyIP != "" && m_cliProxyPort != 0) {
						m_glob.cfg().configureClient(m_cliProtocol, m_cliSSL, m_cliSocketTun,
							new IPAddr(m_localIP, m_localPort, m_localFamily), new IPAddr(m_cliServerIP,
							m_cliServerPort), m_cliAuthUser, m_cliAuthPasswd, m_cliDestIP, m_cliDestPort,
							m_cliTune, new IPAddr(m_cliProxyIP, m_cliProxyPort), m_cliProxyUser,
							m_cliProxyPasswd);
					}
					else {
						throw new ConfigError("Incomplete CLIENT fixed (via proxy) config");
					}
				}
				else if(m_cliFixed && !m_cliProxy) {
					// Direct client, no proxy
					if(gotValidLocal() && m_cliServerIP != "" && m_cliServerPort != 0 &&
					gotValidDest()) {
						m_glob.cfg().configureClient(m_cliProtocol, m_cliSSL, m_cliSocketTun,
							new IPAddr(m_localIP, m_localPort, m_localFamily), new IPAddr(m_cliServerIP,
							m_cliServerPort), m_cliAuthUser, m_cliAuthPasswd, m_cliDestIP, m_cliDestPort,
							m_cliTune);
					}
					else {
						throw new ConfigError("Incomplete CLIENT fixed config");
					}
				}
				else {
					throw new ConfigError("You must specify the CLIENT type as one of socks/fixed");
				}
			}
			else if(server) {
				// Server
				if(m_localPort != 0) {
					m_glob.cfg().configureServer(m_svrSSL, m_svrSocketTun, m_sslCert, m_sslCertKey,
						new IPAddr(m_localIP, m_localPort, m_localFamily), m_svrFirewallIPs,
						m_svrFirewallPorts, m_svrAuth);
				}
				else {
					throw new ConfigError("Incomplete SERVER config");
				}
			}

			// Reset all parsing variables
			m_localIP = "";
			m_localPort = 0;
			m_localFamily = AddrInfo.Family.IPV4;
			m_sslCert = "";
			m_sslCertKey = "";
			m_cliFixed = false;
			m_cliSOCKS = false;
			m_cliProxy = false;
			m_cliSocketTun = false;
			m_cliProtocol = Enums.Protocol.TCP;
			m_cliServerIP = "";
			m_cliServerPort = 0;
			m_cliSSL = false;
			m_cliProxyIP = "";
			m_cliProxyPort = 0;
			m_cliProxyUser = "";
			m_cliProxyPasswd = "";
			m_cliDestIP = "";
			m_cliDestPort = 0;
			m_cliAuthUser = "";
			m_cliAuthPasswd = "";
			m_cliTune = "";
			m_svrFirewallIPs = "";
			m_svrFirewallPorts = "";
			m_svrAuth = false;
			m_svrSSL = false;
			m_svrSocketTun = false;
		}


		// Check we have a valid local listening address
		private bool gotValidLocal()
		{
			// Must either have a valid port, or the address is the special case "-"
			return (m_localPort != 0 || m_localIP == "-");
		}

		// Check we have a valid destination address
		private bool gotValidDest()
		{
			// Must have valid IP and port, or the address is the special case "-"
			return ((m_cliDestIP != "" && m_cliDestPort != 0) || m_cliDestIP == "-");
		}


		/********/
		/* Data */
		/********/

		// Globals
		private Globals				m_glob;

		// Parsed general config
		private string					m_localIP;
		private uint					m_localPort;
		private AddrInfo.Family		m_localFamily;
		private string					m_sslCert;
		private string					m_sslCertKey;

		// Parsed client config
		private bool					m_cliFixed;
		private bool					m_cliSOCKS;
		private bool					m_cliProxy;
		private bool					m_cliSocketTun;
		private Enums.Protocol		m_cliProtocol;
		private string					m_cliServerIP;
		private uint					m_cliServerPort;
		private bool					m_cliSSL;
		private string					m_cliProxyIP;
		private uint					m_cliProxyPort;
		private string					m_cliProxyUser;
		private string					m_cliProxyPasswd;
		private string					m_cliDestIP;
		private uint					m_cliDestPort;
		private string					m_cliAuthUser;
		private string					m_cliAuthPasswd;
		private string					m_cliTune;

		// Parsed server config
		private string					m_svrFirewallIPs;
		private string					m_svrFirewallPorts;
		private bool					m_svrAuth;
		private bool					m_svrSSL;
		private bool					m_svrSocketTun;

	}

}	// namespace

