/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Base class for client and server tunnel controllers.
 *
 * $Id: RawBuff.cs,v 1.1 2008/07/25 15:22:38 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Collections;
	using System.Text;

	public class RawBuff : IEnumerable {

		// Construction
		public RawBuff()
		{
			m_data = new byte[0];
		}

		public RawBuff(uint length)
		{
			m_data = new byte[length];
		}

		public RawBuff(byte[] data)
		{
			m_data = (byte[])data.Clone();
		}

		public RawBuff(byte[] data, uint length)
		{
			// Range check
			if(length > data.Length)
				length = (uint)data.Length;

			// Copy data array
			m_data = new byte[length];
			Array.Copy(data, m_data, length);
		}

		public RawBuff(string str)
		{
			// Convert string to byte[]
			m_data = new byte[str.Length];
			for(int i = 0; i < str.Length; ++i) {
				m_data[i] = (byte)str[i];
			}
		}

		public RawBuff(RawBuff that)
		{
			// Copy data array
			m_data = new byte[that.m_data.Length];
			Array.Copy(that.m_data, m_data, that.m_data.Length);
		}

		public RawBuff(RawBuff that, uint length)
		{
			// Range check
			if(length > that.getLength())
				length = that.getLength();

			// Copy data array
			m_data = new byte[length];
			Array.Copy(that.m_data, m_data, (int)length);
		}


		// Indexer
		public byte this[int pos]
		{
			get { return m_data[pos]; }
			set { m_data[pos] = value; }
		}

		public IEnumerator GetEnumerator()
		{
			return m_data.GetEnumerator();
		}

		// ToString
		public override string ToString()
		{
			uint length = getLength();
			return ToString(ref length);
		}

		public string ToString(ref uint length)
		{
			// Range check length
			length = (length <= getLength())? length : getLength();

			// Create string builder to hold result
			StringBuilder res = new StringBuilder((int)length);

			// Append each byte in raw buffer
			for(uint i = 0; i < length; ++i) {
				res.Append((char)m_data[i]);
			}

			return res.ToString();
		}


		/*************/
		/* Accessors */
		/*************/

		// Get length of data
		public uint getLength()
		{
			return (uint)m_data.Length;
		}

		// Get entire byte array
		public byte[] getData()
		{
			return m_data;
		}


		/************/
		/* Mutators */
		/************/

		// Empty ourselves
		public void clear()
		{
			m_data = new byte[0];
		}

		// Erase some number of bytes from the start of our data
		public void erase(uint length)
		{
			// Do nothing?
			if(length == 0)
				return;

			// Erase everything?
			if(length >= getLength()) {
				clear();
			}
			else {
				// How much to keep?
				uint keepLength = getLength() - length;

				// Replace our data with new cropped array
				byte[] newBuff = new byte[keepLength];
				Array.Copy(m_data, (int)length, newBuff, 0, (int)keepLength);
				m_data = newBuff;
			}
		}

		// Erase whitespace from our start
		public void eraseWS()
		{
			// Calculate number of whitespace chars are at our start
			uint wsCount = 0;
			while(wsCount < getLength()) {
				char c = (char)m_data[wsCount];
				if(c != ' ' && c != '\t')
					break;
				wsCount++;
			}

			// Erase whitespace
			if(wsCount > 0)
				erase(wsCount);
		}


		// Append some data
		public void append(string str)
		{
			RawBuff buff = new RawBuff(str);
			append(buff);
		}

		public void append(RawBuff data)
		{
			append(data, data.getLength());
		}

		public void append(RawBuff data, uint length)
		{
			// Check range
			if(length > data.getLength())
				length = data.getLength();

			// Create new array large enough for all data
			byte[] newBuff = new byte[getLength() + length];

			// Copy our data to start of new buff
			Array.Copy(m_data, 0, newBuff, 0, (int)getLength());

			// Copy new data to end of new buff
			Array.Copy(data.m_data, 0, newBuff, (int)getLength(), (int)length);

			// Replace our buff with new buff
			m_data = newBuff;
		}

		public void append(byte[] data, uint length)
		{
			// Create array with just data we want to append
			byte[] dataCopy = new byte[length];
			Array.Copy(data, dataCopy, length);

			// Append
			RawBuff buff = new RawBuff(dataCopy);
			append(buff);
		}

		public void append(byte data)
		{
			// Create new array large enough for all data
			byte[] newBuff = new byte[getLength() + 1];

			// Copy our data to start of new buff
			Array.Copy(m_data, 0, newBuff, 0, (int)getLength());

			// Copy new byte to end of buff
			newBuff[newBuff.Length - 1] = data;

			// Replace our buff with new buff
			m_data = newBuff;
		}


		// Prepend some data
		public void prepend(RawBuff data)
		{
			// Prepend all data
			prepend(data, data.getLength());
		}

		public void prepend(RawBuff data, uint length)
		{
			// Range check
			if(length > data.getLength())
				length = data.getLength();

			if(length > 0) {
				// Create new array large enough for all data
				byte[] newBuff = new byte[getLength() + length];

				// Copy new data to start of new buff
				Array.Copy(data.m_data, 0, newBuff, 0, (int)length);

				// Copy our data to end of new buff
				Array.Copy(m_data, 0, newBuff, (int)length, (int)getLength());

				// Replace our buff with new buff
				m_data = newBuff;
			}
		}


		/*************/
		/* Utilities */
		/*************/

		// Find the next newline in some raw data
		public bool findNewLine(ref uint start, ref uint end)
		{
			// Try looking for "\r\n" or "\n"
			return (findString("\r\n", ref start, ref end) || findString("\n", ref start, ref end));
		}

		// Find the next newline in some raw data, also checking for split lines
		public bool findNewLine(ref uint start, ref uint end, ref bool split)
		{
			split = false;

			// Try looking for "\r\n"
			bool res = findString("\r\n", ref start, ref end);
			if(!res) {
				// Try looking for just "\n"
				res = findString("\n", ref start, ref end);
				if(!res) {
					// Try looking for lines joined with ","
					res = findString(",", ref start, ref end);
					if(res)
						split = true;
				}
			}

			return res;
		}


		// Look for some string within ourselves
		public bool findString(string str, ref uint start, ref uint end)
		{
			return findString(str, 0, ref start, ref end);
		}

		// Look for some string within ourselves, beginning at some offset
		public bool findString(string str, uint begin, ref uint start, ref uint end)
		{
			// Range check beginning
			uint startChar = begin;
			if(startChar >= getLength())
				startChar = getLength() - 1;

			bool res = false;
			uint index = startChar;
			int strLen = str.Length;
			int buffLen = m_data.Length;
			byte firstChar = (byte)str[0];

			// Keep going while there is room in our buffer for the string
			while((buffLen - index) >= strLen) {
				// Find first character of string
				int indexOfRes = Array.IndexOf(m_data, firstChar, (int)index);

				if(indexOfRes >= 0) {
					// Room for complete string?
					uint firstCharPos = (uint)indexOfRes;
					if(firstCharPos + strLen <= buffLen) {
						// Look for complete string
						bool foundIt = true;
						for(uint i = 0; i < strLen; ++i) {
							uint pos = firstCharPos + i;
							if(m_data[pos] != str[(int)i]) {
								// Chars don't match
								foundIt = false;
								break;
							}
						}

						// Found it?
						if(foundIt) {
							res = true;
							start = firstCharPos;
							end = start + (uint)strLen;
							break;
						}
						else {
							// Keep looking from here
							index = firstCharPos + 1;
						}
					}
					else {
						// String not found
						break;
					}
				}
				else {
					// String not found
					break;
				}
			}

			return res;
		}


		// Look for first occurance of some character
		public bool findChar(byte ch, ref uint pos)
		{
			bool res = false;
			int index = Array.IndexOf(m_data, ch);
			if(index != -1) {
				pos = (uint)index;
				res = true;
			}

			return res;
		}


		// Make the longest string we can from our raw data
		public string makeString(uint start, uint end)
		{
			// Range check
			if(end > getLength())
				end = getLength();
			if(start > end)
				start = end;

			// Build up string
			StringBuilder str = new StringBuilder((int)getLength());
			for(uint i = start; i < end; ++i) {
				// Stop if we see a NULL terminator
				if(m_data[i] == 0)
					break;
				str.Append((char)m_data[i]);
			}

			return str.ToString();
		}

		// Uuencode some of our data
		public void uuencode(ref uint length, ref string res)
		{
			res = Utils.ModBase64Encode(ToString(ref length));
			length = (uint)res.Length;
		}

		// Uudecode some string into ourselves
		public void uudecode(string uuencoded)
		{
			RawBuff uud = new RawBuff(Utils.ModBase64Decode(uuencoded));
			m_data = uud.m_data;
		}

		// Base64 encode some of our data
		public void base64encode(ref uint length, ref string res)
		{
			res = Utils.Base64Encode(ToString(ref length));
			length = (uint)res.Length;
		}


		/*********************/
		/* Private Interface */
		/*********************/


		/********/
		/* Data */
		/********/

		// Our raw data buffer
		private byte[]				m_data;

	}

}  // namespace


