/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Handle reading/writing for one end of a raw tunnel
 * connection.
 *
 * $Id: RawConnection.cs,v 1.2 2013/02/17 10:52:30 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Text;

	public class RawConnection : Connection {

		// Construction
		public RawConnection(Globals glob, Sock sock, Tunneler tunneler, TunnelID tid)
		: base(glob, sock, tunneler)
		{
			#if DEBUG_TUN
			StringBuilder dbg = new StringBuilder();
			dbg.AppendFormat("RawConnection::RawConnection(server) - connection {0}", getID());
			m_glob.DPRINTV(dbg.ToString());
			#endif

			m_tid = tid;

			// Construct reader/writer
			createReaderWriter();
		}

		public RawConnection(Globals glob, Tunneler tunneler, TunnelID tid)
		: base(glob, tunneler)
		{
			#if DEBUG_TUN
			StringBuilder dbg = new StringBuilder();
			dbg.AppendFormat("RawConnection::RawConnection(client) - connection {0}", getID());
			m_glob.DPRINTV(dbg.ToString());
			#endif

			m_tid = tid;
		}


		// We are no longer required
		public override void destroyed()
		{
			#if DEBUG_TUN
			StringBuilder dbg = new StringBuilder();
			dbg.AppendFormat("RawConnection::~RawConnection() - connection {0}", getID());
			m_glob.DPRINTV(dbg.ToString());
			#endif

			// Flag we are closing
			lock(this) {
				m_closing = true;
			}

			// Base class
			base.destroyed();
		}


		// Prepare for a connect attempt to be made to the server
		public override void connectAttempt(IPAddr addr, Enums.Protocol prot, bool ssl)
		{
			// Check to see if connection is allowed
			if(!m_tunneler.getServerSideConfig().getFirewall().isAllowed(addr.getHost(), addr.getPort())) {
				#if DEBUG_TUN
				StringBuilder wrn = new StringBuilder();
				wrn.Append("RawConnection::connectAttempt() - Refused connect to ");
				wrn.AppendFormat("host {0} port {1}", addr.getHost(), addr.getPort());
				m_glob.log().warn(wrn.ToString());
				#endif

				throw new FirewalledConnectError("Connect destination is firewalled");
			}

			// Let base class take it from here
			base.connectAttempt(addr, prot, ssl);
		}


		/***********************/
		/* Protected Interface */
		/***********************/


		/***********/
		/* Methods */
		/***********/

		// Construct reader/writer
		protected override void createReaderWriter()
		{
			lock(this) {
				// Create raw reader & writer
				m_reader = new RawReader(m_glob, this, m_tid);
				m_writer = new RawWriter(m_glob, this);

				// Only truly connected once reader & writer are constructed
				setConnected();
			}
		}


		/********/
		/* Data */
		/********/

		// ID of tunnel we work for
		protected TunnelID					m_tid;

	}

}	// namespace

