/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Tunnel controller for client end of tunnel acting as a
 * SOCKS proxy.
 *
 * $Id: SOCKSClientTunneler.cs,v 1.3 2013/02/17 10:53:18 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Collections;
	using System.Text;

	public class SOCKSClientTunneler : ClientTunneler {

		// Construction
		public SOCKSClientTunneler(Globals glob, Server server) : base(glob, server)
		{
			m_glob.DPRINTV("SOCKSClientTunneler::SOCKSClientTunneler()");
		}

		// Destruction
		public override void destroyed()
		{
			m_glob.DPRINTV("SOCKSClientTunneler::~SOCKSClientTunneler()");

			// Shutdown event scheduling
			if(!shuttingDown())
				shutdownSchedulers();

			// Base class
			base.destroyed();
		}


		/**********/
		/* Events */
		/**********/

		// Notification of a new socket connection
		public override void newConnectionArrived(Sock insock)
		{
			m_glob.DPRINTI("SOCKSClientTunneler::newConnectionArrived()");

			/////////////////////////////////////////////////////////////////////
			// In SOCKSClientTunneler a new connection always means a new tunnel,
			// but we don't want to create that tunnel until we have completed
			// The SOCKS negotiation.
			/////////////////////////////////////////////////////////////////////

			// Do our users quotas allow another tunnel?
			Authenticator authenticator = m_server.getConfig().getAuthenticator();
			string userName = authenticator.getDefaultUserName();
			bool userQuotaOk = authenticator.getUserCanCreateTunnel(userName);
			if(userQuotaOk) {
				// Create SOCKS connection to handle incoming data
				lock(m_connections) {
					Connection rawconn = new SOCKSConnection(m_glob, insock, this);
					addConnection(rawconn);
				}
			}
			else {
				StringBuilder err = new StringBuilder();
				err.AppendFormat("Tunnel creation quota exceeded for user {0}", userName);
				throw new CommsError(err.ToString());
			}
		}


		// Notification that a connection has changed state
		public override void connectionStateChange(uint connID,
			Connection.ConnectionState state)
		{
			// Let base class know
			base.connectionStateChange(connID, state);

			// Something for us?
			if(state == Connection.ConnectionState.SOCKSSETUP) {

				///////////////////////////////////////////////////////////////
				// In SOCKSClientTunneler this means that the SOCKS connection
				// has completed negotiation.
				///////////////////////////////////////////////////////////////

				try {
					// Lookup connection
					lock(m_connections) {
						int cit = getConnectionByID(connID);
						SOCKSConnection socksconn = (SOCKSConnection)m_connections[cit];

						// Create HTTP client connection to handle tunneled traffic
						HTTPConnection httpconn = new HTTPConnection(m_glob, this);
						addConnection(httpconn);

						// Create new tunnel config
						TunnelConfig tcfg = createTunnelConfig(socksconn);

						// Let SOCKS connection know about new tunnels ID
						socksconn.setTunnelID(tcfg.getTunnelID());

						// Create tunnel
						Tunnel tun = new ClientTunnel(m_glob, tcfg, this);
						tun.addLeftConnection(socksconn);
						tun.addRightConnection(httpconn);

						// Schedule connect out on HTTP side
						ClientSideConfig clientCfg = getClientSideConfig();
						m_connectionMgr.addProxyConnect(httpconn, new Time(0,0),
							clientCfg.getNextHopIP(), Enums.Protocol.TCP, clientCfg.getSSL());

						// Store Tunnel object in our list of active tunnels
						lock(m_tunnels) {
							m_tunnels.Add(tun);
						}
					}
				}
				catch(LookupManagementError e) {
					m_glob.log().warn("Error looking up connection: " + e.Message);
				}
			}
		}


		// Notification that a tunnel has changed state
		public override void tunnelStateChanged(TunnelID tid, Tunnel.TunnelState state)
		{
			m_glob.DPRINTI("SOCKSClientTunneler::tunnelStateChanged() - " + tid.ToString());

			// Let base class know
			base.tunnelStateChanged(tid, state);

			// Lookup tunnel
			lock(m_connections) {
				lock(m_tunnels) {
					Tunnel tun = getTunFromPred(new PredTunID(tid));
					if(tun != null) {
						// Check this isn't a test tunnel
						if(!tun.isTestTunnel()) {
							// Lookup SOCKS connection
							ArrayList sel = new ArrayList();
							sel.Add(new Tunnel.SelLeftConn());
							SOCKSConnection socksconn = (SOCKSConnection)tun.getConnFromSel(sel);
							if(socksconn != null) {
								try {
									// Let connection know
									if(state == Tunnel.TunnelState.CONNECTED) {
										socksconn.tunnelConnected(true);
									}
								}
								catch(CommsError e) {
									// Log error
									m_glob.log().error(
										"Error processing tunnel state change notification: " +
										e.Message);

									// Report error on this connection
									if(!socksconn.closing()) {
										scheduleConnectionError(socksconn.getID(),
											Tunnel.TunnelErrorType.GENERAL_COMMS_ERROR);
									}
								}
							}
							else {
								// Abandon this tunnel
								m_glob.DPRINTI("Failed to lookup SOCKS connection on tunnel");
								scheduleEvent(new ShutdownTunnelEvent(m_glob, new Time(0,0),
									this, tid, false));
							}
						}
					}
					else {
						StringBuilder wrn = new StringBuilder();
						wrn.Append("State change notification for unknown tunnel ");
						wrn.AppendFormat("{0}", tid.ToString());
						m_glob.log().warn(wrn.ToString());
					}
				}
			}
		}


		/***********************/
		/* Protected Interface */
		/***********************/


		/***********/
		/* Methods */
		/***********/

		// Create config for tunnel to server
		protected TunnelConfig createTunnelConfig(SOCKSConnection inConn)
		{
			// Create tunnelID
			ClientSideConfig clientCfg = getClientSideConfig();
			TunnelID tid = new TunnelID(m_glob, clientCfg);

			// Get destination IP & port from SOCKS connection
			string destIP = inConn.getSOCKSDestIP();
			uint destPort = inConn.getSOCKSDestPort();
			IPAddr destAddr = new IPAddr(destIP, destPort);

			// Create tunnel config
			return new TunnelConfig(tid, clientCfg.getTunTune(), clientCfg.getProxySettings(),
				clientCfg.getAuthenticator().getDefaultAuthSettings(), destAddr);
		}

	}

}	// namespace


