/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Handle reading/writing for one end of a raw tunnel
 * SOCKS connection.
 *
 * $Id: SOCKSConnection.cs,v 1.2 2013/02/17 10:53:18 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Text;

	public class SOCKSConnection : Connection {

		// Construction
		public SOCKSConnection(Globals glob, Sock sock, Tunneler tunneler)
		: base(glob, sock, tunneler)
		{
			#if DEBUG_TUN
			StringBuilder dbg = new StringBuilder();
			dbg.AppendFormat("SOCKSConnection::SOCKSConnection(server) - connection {0}", getID());
			m_glob.DPRINTV(dbg.ToString());
			#endif

			// Construct reader/writer
			createReaderWriter();
		}


		// We are no longer required
		public override void destroyed()
		{
			#if DEBUG_TUN
			StringBuilder dbg = new StringBuilder();
			dbg.AppendFormat("SOCKSConnection::~SOCKSConnection() - connection {0}", getID());
			m_glob.DPRINTV(dbg.ToString());
			#endif

			// Flag we are closing
			lock(this) {
				m_closing = true;
			}

			// Base class
			base.destroyed();
		}


		/********************/
		/* Public Interface */
		/********************/


		/***********/
		/* Methods */
		/***********/

		// Get the IP address of the SOCKS destination
		public string getSOCKSDestIP()
		{
			SOCKSReader reader = (SOCKSReader)m_reader;
			return reader.getDestIP();
		}

		// Get the port of the SOCKS destination
		public uint getSOCKSDestPort()
		{
			SOCKSReader reader = (SOCKSReader)m_reader;
			return reader.getDestPort();
		}


		// Tell our SOCKS reader what tunnel ID they are working for
		public void setTunnelID(TunnelID tid)
		{
			lock(this) {
				SOCKSReader reader = (SOCKSReader)m_reader;
				reader.setTunnelID(tid);
			}
		}

		// Notification that the tunnel we carry has been established
		public void tunnelConnected(bool tunOk)
		{
			lock(this) {
				SOCKSReader reader = (SOCKSReader)m_reader;
				reader.tunnelConnected(tunOk);
			}
		}


		/***********************/
		/* Protected Interface */
		/***********************/


		/***********/
		/* Methods */
		/***********/

		// Construct reader/writer
		protected override void createReaderWriter()
		{
			lock(this) {
				// Create HTTP reader & writer
				m_reader = new SOCKSReader(m_glob, this);
				m_writer = new RawWriter(m_glob, this);

				// Only truly connected once reader & writer are created
				setConnected();
			}
		}

	}

}	// namespace

