/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * A SOCKS data reader.
 *
 * $Id: SOCKSReader.cs,v 1.2 2013/02/17 10:53:18 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Collections;
	using System.Collections.Generic;
	using System.Text;

	public class SOCKSReader : RawReader {

		// Construction
		public SOCKSReader(Globals glob, Connection conn)
		: base(glob, conn)
		{
			m_glob.DPRINTV("SOCKSReader::SOCKSReader()");

			m_state = SOCKSState.SOCKS_VERSION;
			m_buff = new RawBuff();
			m_version = 0;
			m_destIP = "";
			m_destPort = 0;
			m_atype = 0x01;
		}


		/********************/
		/* Public Interface */
		/********************/


		/*************/
		/* Accessors */
		/*************/

		// Get the IP address of the SOCKS destination
		public string getDestIP()
		{
			return m_destIP;
		}

		// Get the port of the SOCKS destination
		public uint getDestPort()
		{
			return m_destPort;
		}


		/*************/
		/* Utilities */
		/*************/

		// Notification of whether our tunnel connected ok
		public void tunnelConnected(bool tunOk)
		{
			// Are we waiting for this notification?
			if(m_state == SOCKSState.SOCKS_REQUEST_RSP) {
				socksRequestRsp(tunOk);
			}
		}


		/***********************/
		/* Protected Interface */
		/***********************/


		/*********/
		/* Types */
		/*********/

		// SOCKS parsing states
		protected enum SOCKSState {
			SOCKS_VERSION, SOCKS_METHOD, SOCKS_REQUEST_RCV, SOCKS_REQUEST_RSP,
			SOCKS_CONNECTED
		};


		/***********/
		/* Methods */
		/***********/

		// Process some newly read raw data
		protected override void processData(byte[] data, uint length)
		{
			m_glob.DPRINTV("SOCKSReader::processData()");

			// What SOCKS protocol stage are we in?
			switch(m_state) {
				case(SOCKSState.SOCKS_VERSION): {
					socksVersion(data, length);
					break;
				}
				case(SOCKSState.SOCKS_REQUEST_RCV): {
					socksRequestRcv(data, length);
					break;
				}
				case(SOCKSState.SOCKS_REQUEST_RSP): {
					throw new ReadError("SOCKS protocol error: Read in unexpected state");
				}
				case(SOCKSState.SOCKS_CONNECTED): {
					// Let base class handle it
					base.processData(data, length);
					break;
				}
				default: {
					throw new ReadError("Unknown SOCKS protocol state");
				}
			}
		}


		/*********************/
		/* Private Interface */
		/*********************/


		/***********/
		/* Methods */
		/***********/

		// Parse & respond to a SOCKS version message
		private void socksVersion(byte[] data, uint length)
		{
			m_glob.DPRINTV("SOCKSReader::socksVersion()");

			// Append read data to our buffer
			lock(this) {
				m_buff.append(data, length);

				// SOCKS version message is at least 2 bytes
				if(m_buff.getLength() >= 2) {
					m_version = m_buff[0];

					#if DEBUG_TUN
					StringBuilder nfo1 = new StringBuilder();
					nfo1.AppendFormat("SOCKSReader::socksVersion() - Version {0}", m_version);
					m_glob.DPRINTI(nfo1.ToString());
					#endif

					if(m_version == 5) {
						// SOCKS 5
						byte nmethods = m_buff[1];
						int expectedMsgLen = 2 + nmethods;

						// Got enough data for entire message?
						if(m_buff.getLength() == expectedMsgLen) {
							bool ok = false;

							for(int i = 0; i < nmethods; ++i) {
								// Get suggested method
								byte method = m_buff[2 + i];

								#if DEBUG_TUN
								StringBuilder nfo2 = new StringBuilder();
								nfo2.AppendFormat("SOCKSReader::socksVersion() - Method {0}",
									 method);
								m_glob.DPRINTI(nfo2.ToString());
								#endif

								// We currently only support "no authentication"
								if(method == 0) {
									// Build and send SOCKS response msg
									byte[] response = { 0x05, 0x00 };
									socksSendMsg(response);

									// We're all done here
									ok = true;
									break;
								}
							}

							// Responded ok?
							if(ok) {
								// Clear out message buffer
								m_buff.clear();

								// Advance state
								m_state = SOCKSState.SOCKS_REQUEST_RCV;
							}
							else {
								// No suitable method suggested
								byte[] response = { 0x05, 0xFF };
								socksSendMsg(response);

								// Give up on this connection
								throw new ReadError("SOCKS protocol error: Bad SOCKS5 method received");
							}
						}
						else {
							// Too much data?
							if(m_buff.getLength() > expectedMsgLen) {
								// SOCKS protocol error, give up
								throw new ReadError("SOCKS protocol error: Bad SOCKS5 request length");
							}
						}
					}
					else if(m_version == 4) {
						// SOCKS4
						int index = 1;
						byte command = m_buff[index++];

						#if DEBUG_TUN
						StringBuilder nfo2 = new StringBuilder();
						nfo2.AppendFormat("SOCKSReader::socksVersion() - Command {0}", command);
						m_glob.DPRINTI(nfo2.ToString());
						#endif

						// Got enough data for entire message?
						if(m_buff.getLength() >= 8) {
							bool ok = false;

							// Only currently support Connect
							if(command == 1) {
								// Get port
								uint port = m_buff[index++];
								port <<= 8;
								port |= m_buff[index++];

								// IPv4 - 4 bytes
								StringBuilder address = new StringBuilder();
								for(int i = 0; i < 4-1; ++i)
									address.AppendFormat("{0}.", (int)m_buff[index++]);
								address.AppendFormat("{0}", (int)m_buff[index++]);

								// Save address & port
								m_destIP = address.ToString();
								m_destPort = port;

								#if DEBUG_TUN
								StringBuilder nfo3 = new StringBuilder();
								nfo3.AppendFormat("SOCKS4 request address {0}", m_destIP);
								nfo3.AppendFormat(", port {0}", m_destPort);
								m_glob.DPRINTV(nfo3.ToString());
								#endif

								// Valid IP address?
								if(m_destIP.StartsWith("0.0.0.")) {
									// Skip over any username
									while(m_buff[index] != 0)
										index++;
									index++;

									// Look for SOCKS4a domain name
									StringBuilder newAddress = new StringBuilder();
									while(m_buff[index] != 0)
										newAddress.AppendFormat("{0}", (char)m_buff[index++]);
									m_destIP = newAddress.ToString();

									#if DEBUG_TUN
									StringBuilder nfo4 = new StringBuilder();
									nfo4.AppendFormat("SOCKS4a request address {0}", m_destIP);
									nfo4.AppendFormat(", port {0}", m_destPort);
									m_glob.DPRINTV(nfo4.ToString());
									#endif
								}

								ok = true;
							}

							if(ok) {
								// Clear out message buffer
								m_buff.clear();

								// Advance state
								m_state = SOCKSState.SOCKS_REQUEST_RSP;

								// Let tunneler try to establish tunnel to destination
								Tunneler tunneler = m_conn.getTunneler();
								tunneler.scheduleEvent(new ConnectionStateEvent(m_glob, tunneler,
									m_conn.getID(), Connection.ConnectionState.SOCKSSETUP));
							}
							else {
								// No suitable command
								List<byte> response = new List<byte>();
								response.Add(0x00);
								response.Add(0x5b);
								for(int i = 0; i < 2+4; i++)
									response.Add(0x00);
								socksSendMsg(response.ToArray());

								// Give up on this connection
								throw new ReadError("SOCKS protocol error: Bad SOCKS4 command received");
							}
						}
						else {
							// SOCKS protocol error, give up
							throw new ReadError("SOCKS protocol error: Bad SOCKS4 request length");
						}
					}
				}
			}
		}


		// Parse & respond to a SOCKS5 request message
		private void socksRequestRcv(byte[] data, uint length)
		{
			m_glob.DPRINTV("SOCKSReader::socksRequestRcv()");

			// Append read data to our buffer
			lock(this) {
				m_buff.append(data, length);

				// Got minimal header?
				if(m_buff.getLength() >= 4) {
					int index = 0;
					byte version = m_buff[index++];
					byte cmd = m_buff[index++];
					index++;
					m_atype = m_buff[index++];

					#if DEBUG_TUN
					StringBuilder nfo1 = new StringBuilder();
					nfo1.Append("SOCKSReader::socksRequest() - Version ");
					nfo1.AppendFormat("{0}, Cmd {1}, AType {2}", version, cmd, m_atype);
					m_glob.DPRINTI(nfo1.ToString());
					#endif

					// Supported command?
					if(cmd != 0x01) {
						// Currently only support TCP connect
						throw new ReadError("SOCKS protocol error: Unsupported request command");
					}

					// Get address
					StringBuilder address = new StringBuilder();
					switch(m_atype) {
						case(0x01): {
							// IPv4 - 4 bytes
							for(int i = 0; i < 4-1; ++i)
								address.AppendFormat("{0}.", (int)m_buff[index++]);
							address.AppendFormat("{0}", (int)m_buff[index++]);
							break;
						}
						case(0x03): {
							// FQDN
							byte len = m_buff[index++];
							for(int i = 0; i < len; ++i)
								address.AppendFormat("{0}", (char)m_buff[index++]);
							break;
						}
						default: {
							throw new ReadError("SOCKS protocol error: Bad address type in request");
						}
					}

					// Get port
					uint port = m_buff[index++];
					port <<= 8;
					port |= m_buff[index];

					// Save address & port
					m_destIP = address.ToString();
					m_destPort = port;

					StringBuilder nfo2 = new StringBuilder();
					nfo2.AppendFormat("SOCKS5 request address {0}, port {1}", m_destIP,
						m_destPort);
					m_glob.DPRINTV(nfo2.ToString());

					// Finished with msg buff
					m_buff.clear();

					// Advance state
					m_state = SOCKSState.SOCKS_REQUEST_RSP;

					// Let tunneler try to establish tunnel to destination
					Tunneler tunneler = m_conn.getTunneler();
					tunneler.scheduleEvent(new ConnectionStateEvent(m_glob, tunneler,
						m_conn.getID(), Connection.ConnectionState.SOCKSSETUP));
				}
			}
		}


		// Send back a SOCKS connect response message
		private void socksRequestRsp(bool ok)
		{
			m_glob.DPRINTV("SOCKSReader::socksRequestRsp()");

			lock(this) {
				// Advance state
				m_state = SOCKSState.SOCKS_CONNECTED;

				// SOCKS response msg
				List<byte> response = new List<byte>();

				if(m_version == 5) {
					// Basic header fields
					response.Add(0x05);
					if(ok)
						response.Add(0x00);
					else
						response.Add(0x01);
					response.Add(0x00);
					response.Add(0x01);

					// IPv4 address
					response.Add(0x00);
					response.Add(0x00);
					response.Add(0x00);
					response.Add(0x00);

					// Port
					response.Add(0x00);
					response.Add(0x00);
				}
				else if(m_version == 4) {
					// Basic header fields
					response.Add(0x00);
					if(ok)
						response.Add(0x5a);
					else
						response.Add(0x5b);

					// Port/IPv4 address
					for(int i = 0; i < 2+4; i++)
						response.Add(0x00);
				}

				// Send response out of band
				socksSendMsg(response.ToArray());
			}
		}


		// Send a SOCKS message
		private void socksSendMsg(byte[] msg)
		{
			try {
				// Build and send SOCKS msg
				RawBuff oob = new RawBuff(msg);
				m_conn.writeOutOfBand(oob);
			}
			catch(TunnelError e) {
				// Rethrow
				throw new ReadError("Failed to send SOCKS msg: " + e.Message);
			}
		}


		/********/
		/* Data */
		/********/

		// Current SOCKS parsing state
		private SOCKSState					m_state;

		// A buffer for storing read data
		private RawBuff						m_buff;

		// Fields from parsed SOCKS address
		private byte							m_version;
		private string							m_destIP;
		private uint							m_destPort;
		private byte							m_atype;

	}

}	// namespace

