/* (C) 2007 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate an SSL socket.
 *
 * $Id: SSLSock.cs,v 1.7 2013/02/17 10:53:18 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{

	using System;
	using System.Reflection;
	using System.Net;
	using System.Net.Security;
	using System.Net.Sockets;
	using System.Resources;
	using System.Security.Authentication;
	using System.Security.Cryptography.X509Certificates;
	using System.Text;
	using System.Threading;

	public class SSLSock : Sock {

		// Construction
		public SSLSock(Globals glob, bool blocking) : base(glob, blocking)
		{
			m_sslStream = null;
			m_netStream = null;
			m_readThread = null;
			m_readMtx = new Mutex();
			m_shutDownMtx = new Mutex();
			m_shutDown = false;
			m_cert = "";
			m_certKey = "";
			m_connectResultAvailable = false;
			m_connectConnected = false;
			m_connectAuthenticated = false;
			m_readBuff = new RawBuff();
			m_rawReadBuff = new byte[RAW_READ_BUF_SIZE];
			m_readError = false;
			m_readClosed = false;
			m_writeBuff = null;
			m_writeError = false;
		}

		public SSLSock(Globals glob, string cert, string key, bool blocking)
		: base(glob, blocking)
		{
			m_sslStream = null;
			m_netStream = null;
			m_readThread = null;
			m_readMtx = new Mutex();
			m_shutDownMtx = new Mutex();
			m_shutDown = false;
			m_cert = cert;
			m_certKey = key;
			m_connectResultAvailable = false;
			m_connectConnected = false;
			m_connectAuthenticated = false;
			m_readBuff = new RawBuff();
			m_rawReadBuff = new byte[RAW_READ_BUF_SIZE];
			m_readError = false;
			m_readClosed = false;
			m_writeBuff = null;
			m_writeError = false;
		}


		// We are no longer required
		public override void destroyed()
		{
			m_glob.DPRINTV("SSLSock::~SSLSock()");

			lock(this) {
				try {
					// Close SSL socket
					close();
				}
				catch(SockError e) {
					m_glob.log().info(e.Message);
				}
			}

			// Take locks
			Monitor.Enter(m_readMtx);
			Monitor.Enter(m_shutDownMtx);

			// Signal we are no longer running
			m_shutDown = true;

			// Wakeup server thread & wait for it to finish
			if(m_readThread != null) {
				Monitor.Pulse(m_readMtx);
				Monitor.Exit(m_readMtx);
				Monitor.Wait(m_shutDownMtx);
				Monitor.Exit(m_shutDownMtx);
				m_readThread.Join();
				m_readThread = null;
			}
			else {
				Monitor.Exit(m_readMtx);
				Monitor.Exit(m_shutDownMtx);
			}

			// Base class
			base.destroyed();
		}


		/********************/
		/* Public Interface */
		/********************/

		// Wait for a new incoming TCP connection
		public override void newConnection(Tunneler tunneler)
		{
			lock(this) {
				// Socket ready to accept
				Sock connectedSock = accept();

				try {
					// Tell tunneler
					if(tunneler != null)
						tunneler.newConnectionArrived(connectedSock);
					else
						throw new SockError("No tunneler to handle new SSL connection", 0);
				}
				catch(TunnelError e) {
					m_glob.log().info(e.Message);
				}
			}
		}


		// Set socket listening
		public override void listen()
		{
		}


		// Wait for a new incoming connection
		public override Sock accept()
		{
			return null;
		}


		// Establish connection out
		public override void connect(IPAddr ipaddr)
		{
			// Create underlying socket
			createSockFD(ipaddr, SocketType.Stream, ProtocolType.Tcp);

			// Get address details
			string addr = ipaddr.getHost();
			uint port = ipaddr.getPort();

			lock(this) {
				// Lookup destination address
				IPAddress destaddr = null;
				if(!Utils.getIPByName(addr, ref destaddr)) {
					throw new ConnectError("Failed to resolve SSL server address " + addr);
				}

				try {
					// Try the connect
					IPEndPoint ipEndPoint = new IPEndPoint(destaddr, (int)port);
					m_fd.BeginConnect(ipEndPoint, new AsyncCallback(this.connectCallback), null);
				}
				catch(SocketException e) {
					throw new SockError("Failed to connect to SSL server: " + e.Message,
						e.SocketErrorCode);
				}
				catch(Exception e) {
					throw new SockError("Failed to connect to SSL server: " + e.Message, 0);
				}

				// We haven't connected yet
				throw new SockError("Not yet connected to SSL server", SocketError.WouldBlock);
			}
		}


		// Read from socket
		public override int read(byte[] buff)
		{
			int noread = 0;

			lock(this) {
				// Got data waiting to be read?
				noread = (int)m_readBuff.getLength();
				if(noread > 0) {
					// Don't overflow return buffer
					if(noread > buff.Length)
						noread = buff.Length;

					// Return contents of read buffer
					Array.Copy(m_readBuff.getData(), buff, noread);
					m_readBuff.erase((uint)noread);
				}
				else {
					// Check for EOF
					if(m_readClosed) {
						throw new ConnectionClosed("Other end closed socket");
					}
				}
			}

			return noread;
		}


		// Write to socket
		public override int write(RawBuff buff)
		{
			int nowritten = 0;

			lock(this) {
				// Check we don;t have a write in progress
				if(m_writeBuff != null) {
					throw new SockError("Can't write to SSL socket, write already in progress", 0);
				}

				// Copy data to our buffer
				m_writeBuff = (byte[])buff.getData().Clone();
				nowritten = m_writeBuff.Length;

				try {
					// Kick off asynchronous write
					m_sslStream.BeginWrite(m_writeBuff, 0, m_writeBuff.Length,
						new AsyncCallback(this.writeCallback), null);
				}
				catch(Exception e) {
					m_writeError = true;
					throw new SockError("Write error to SSL socket: " + e.Message, 0);
				}
			}

			return nowritten;
		}


		// Get the result of a non-blocking connect attempt
		public override void getConnectResult(ref bool resultAvailable, ref bool connected)
		{
			lock(this) {
				resultAvailable = m_connectResultAvailable;
				connected = m_connectConnected && m_connectAuthenticated;
			}
		}


		// Bind socket to an address
		public override void bind(IPAddr ipaddr)
		{
			// Create underlying socket
			createSockFD(ipaddr, SocketType.Stream, ProtocolType.Tcp);
		}


		// Poll socket for reading
		public override bool pollRead(int timeout, ref bool err, ref bool hup)
		{
			bool ready = false;
			err = hup = false;

			lock(this) {
				// Got error condition?
				if(m_readError || m_readClosed) {
					ready = true;
					err = m_readError;
				}
				else {
					// Got data for reading?
					if(m_readBuff.getLength() > 0) {
						ready = true;
					}
				}
			}

			return ready;
		}


		// Poll socket for writing
		public override bool pollWrite(int timeout, ref bool err)
		{
			bool ready = false;
			err = false;

			lock(this) {
				// Got error condition?
				if(m_writeError) {
					ready = true;
					err = true;
				}
				else {
					// Can't write while we have an existing write in progress
					if(m_connectAuthenticated && m_writeBuff == null) {
						ready = true;
					}
				}
			}

			return ready;
		}


		/***********************/
		/* Protected Interface */
		/***********************/


		/***********/
		/* Methods */
		/***********/

		// Close SSL session
		protected override void close()
		{
			m_glob.DPRINTV("SSLSock::close()");

			lock(this) {
				// Close SSL socket
				if(m_sslStream != null) {
					m_sslStream.Close();
					m_sslStream = null;
				}
			}
		}


		// Setup blocking/nonblocking mode
		protected override void blocknonblock()
		{
			// We can't set non-blocking mode on a NetworkStream
		}


		/*********************/
		/* Private Interface */
		/*********************/


		/*************/
		/* Constants */
		/*************/

		// Reading thread poll interval
		private const int				READ_INTERVAL			= 20;

		// Size of read buffer
		private const int				RAW_READ_BUF_SIZE		= 65536;


		/***********/
		/* Methods */
		/***********/

		// Do the reading from the SSL stream - Thread entry point
		private void doReading()
		{
			// Take mutex
			Monitor.Enter(m_readMtx);

			while(!m_shutDown) {
				// Sleep a while
				TimeSpan pollTime = new TimeSpan(0,0,0,0,READ_INTERVAL);
				Monitor.Wait(m_readMtx, pollTime);

				// Read any pending data
				if(!m_shutDown) {
					lock(this) {
						// Can we read yet?
						bool canRead = true;
						try {
							canRead = m_netStream.DataAvailable;
						}
						catch(Exception) {}

						if(canRead) {
							try {
								// Read data into buffer
								int noread = m_sslStream.Read(m_rawReadBuff, 0, m_rawReadBuff.Length);
								if(noread > 0) {
									// Append newly read data to buffer
									m_readBuff.append(m_rawReadBuff, (uint)noread);
								}
								else {
									// Connection closed
									m_readError = true;
									m_readClosed  =true;
								}
							}
							catch(Exception e) {
								m_glob.log().info("SSL read failed: " + e.Message);
								m_readError = true;
							}
						}
					}
				}
			}

			// Signal we have finished
			Monitor.Exit(m_readMtx);
			Monitor.Enter(m_shutDownMtx);
			Monitor.Pulse(m_shutDownMtx);
			Monitor.Exit(m_shutDownMtx);
		}


		// Connect callback
		private void connectCallback(IAsyncResult ar)
		{
			lock(this) {
				try {
					// End the connect attempt
					m_fd.EndConnect(ar);

					// We've connected ok
					m_glob.DPRINTI("SSL connect() succeeded");
					m_connectConnected = true;

					// Create Streams to handle reading/writing
					m_netStream = new NetworkStream(m_fd, true);
					m_sslStream = new SslStream(m_netStream, false,
						new RemoteCertificateValidationCallback(this.certValidationCallback));

					// Kick off SSL authentication
					m_sslStream.BeginAuthenticateAsClient("TheLinuxAcademy",
						new AsyncCallback(this.clientAuthCallback), null);
				}
				catch(Exception e) {
					m_glob.log().info("SSL connect failed: " + e.Message);

					// Store failed connect result
					m_glob.DPRINTI("SSL connect() failed");
					m_connectResultAvailable = true;
					m_connectConnected = false;
					m_connectAuthenticated = false;
				}
			}
		}

		// Certificate validation callback
		private bool certValidationCallback(object sender, X509Certificate cert,
			X509Chain chain, SslPolicyErrors sslPolicyErrors)
		{
			// Allways validate
			return true;
		}

		// Client authentication callback
		private void clientAuthCallback(IAsyncResult ar)
		{
			lock(this) {
				try {
					// End the authentication attempt
					m_sslStream.EndAuthenticateAsClient(ar);

					// We've authenticated ok
					m_glob.DPRINTI("SSL client authenticate() succeeded");
					m_connectResultAvailable = true;
					m_connectAuthenticated = true;

					// Launch reading thread
					m_readThread = new Thread(new ThreadStart(doReading));
					m_readThread.Start();
				}
				catch(Exception e) {
					m_glob.log().info("SSL client authenticate failed: " + e.Message);

					// Store failed connect result
					m_glob.DPRINTI("SSL client authenticate() failed");
					m_connectResultAvailable = true;
					m_connectAuthenticated = false;
				}
			}
		}

		// Write callback
		private void writeCallback(IAsyncResult ar)
		{
			lock(this) {
				try {
					// Get write result
					m_sslStream.EndWrite(ar);

					// We're ready for another write
					m_writeBuff = null;
				}
				catch(Exception e) {
					m_glob.log().info("SSL write failed: " + e.Message);
					m_writeError = true;
				}
			}
		}


		/********/
		/* Data */
		/********/

		// SSL Stream
		private SslStream				m_sslStream;
		private NetworkStream		m_netStream;

		// Reading thread
		private Thread					m_readThread;
		private Mutex					m_readMtx;
		private Mutex					m_shutDownMtx;
		private bool					m_shutDown;

		// Certificate/key files
		private string					m_cert;
		private string					m_certKey;

		// Connect results
		private bool					m_connectResultAvailable;
		private bool					m_connectConnected;
		private bool					m_connectAuthenticated;

		// Read buffer
		private RawBuff				m_readBuff;
		private byte[]					m_rawReadBuff;
		private bool					m_readError;
		private bool					m_readClosed;

		// Write buffer
		private byte[]					m_writeBuff;
		private bool					m_writeError;

	}

}	// namespace

