/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Base class of a TCP/IP socket server to listen for incoming
 * connections.
 *
 * $Id: Server.cs,v 1.4 2013/02/17 10:53:18 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Collections;
	using System.Net.Sockets;
	using System.Text;
	using System.Threading;

	abstract public class Server {

		// Construction
		public Server(Globals glob, SideConfig serverConfig)
		{
			m_glob = glob;
			m_thread = null;
			m_serverConfig = serverConfig;
			m_sock = null;
			m_tunneler = null;
			m_shutDownMtx = new Mutex();
			m_shutDown = false;
			m_serverThreadStartMtx = new Mutex();
			m_serverThreadContMtx = new Mutex();
			m_serverMtx = new Mutex();
		}

		// Destruction
		public virtual void destroyed()
		{
			// Take mutexes
			Monitor.Enter(m_serverMtx);
			Monitor.Enter(m_shutDownMtx);

			// Signal we are no longer running
			m_shutDown = true;

			// Wakeup server thread & wait for it to finish
			if(m_thread != null) {
				Monitor.Pulse(m_serverMtx);
				Monitor.Exit(m_serverMtx);
				Monitor.Wait(m_shutDownMtx);
				Monitor.Exit(m_shutDownMtx);
				m_thread.Join();
				m_thread = null;
			}
			else {
				Monitor.Exit(m_serverMtx);
				Monitor.Exit(m_shutDownMtx);
			}

			lock(this) {
				// Finished with listening socket
				if(m_sock != null) {
					m_sock.destroyed();
					m_sock = null;
				}

				// Finished with tunneler
				if(m_tunneler != null) {
					m_tunneler.destroyed();
					m_tunneler = null;
				}
			}
		}


		/*************/
		/* Accessors */
		/*************/

		// Get config settings
		public SideConfig getConfig()
		{
			return m_serverConfig;
		}


		// Get status string
		public virtual string getStatusString()
		{
			StringBuilder status = new StringBuilder();

			// Protocol
			string ssl = "";
			if(m_serverConfig.getSSL())
				ssl = " (SSL)";
			status.AppendFormat("[Protocol {0}{1}] ", m_serverConfig.getProtocol(), ssl);

			// Address to listen on
			IPAddr listenIP = m_serverConfig.getListenIP();
			string host = listenIP.getHost();
			if(host.Length == 0)
				host = "*";
			status.AppendFormat("[Listen {0}:{1}]", host, listenIP.getPort());

			return status.ToString();
		}


		// Get tunneler
		public Tunneler getTunneler()
		{
			return m_tunneler;
		}


		// Get tunnel stats for our tunnelers carried tunnels
		public void getTunnelStats(ArrayList stats)
		{
			lock(this) {
				// Get IDs for active tunnels
				ArrayList tids = new ArrayList();
				m_tunneler.getActiveTunnelIDs(tids);
				for(int tit = 0; tit < tids.Count; ++tit) {
					// Get & copy stats for this tunnel
					TunnelStats ts = m_tunneler.getTunnelStats((TunnelID)tids[tit]);
					if(ts != null)
						stats.Add(ts);
				}
			}
		}


		/***********************/
		/* Protected interface */
		/***********************/


		/***********/
		/* Methods */
		/***********/

		// Wait for arriving connections (never returns)
		protected void startServer()
		{
			// Notify launching thread we have started
			Monitor.Enter(m_serverThreadContMtx);
			Monitor.Enter(m_serverThreadStartMtx);
			Monitor.Pulse(m_serverThreadStartMtx);
			Monitor.Exit(m_serverThreadStartMtx);

			// Wait for permission to start accepting connections
			Monitor.Wait(m_serverThreadContMtx);
			Monitor.Exit(m_serverThreadContMtx);
			m_glob.DPRINTI("Server::startServer() - Waiting for connection");

			// Take mutex
			Monitor.Enter(m_serverMtx);

			while(!m_shutDown) {
				try {
					// Poll socket until we think it's worth trying to accept()
					bool pollError = false, pollHup = false;
					if(m_sock.pollRead(0, ref pollError, ref pollHup)) {
						// Socket connection pending
						m_sock.newConnection(m_tunneler);
					}
					else {
						// Poll timeout, wait a while before retrying
						TimeSpan pollTime = new TimeSpan(0,0,0,0,POLL_TIMEOUT);
						Monitor.Wait(m_serverMtx, pollTime);
					}
				}
				catch(SockError) {
					// General error, just carry on
				}
			}

			// Signal we have finished
			Monitor.Exit(m_serverMtx);
			Monitor.Enter(m_shutDownMtx);
			Monitor.Pulse(m_shutDownMtx);
			Monitor.Exit(m_shutDownMtx);

			m_glob.DPRINTI("Server closing down");
		}


		// Configure our server socket
		protected void configureSocket(bool ssl)
		{
			// Create listening socket
			if(ssl) {
				// Use SSL
				if(m_serverConfig.isServer()) {
					// Need both SSL certificate & private key at server
					ServerSideConfig scfg = (ServerSideConfig)m_serverConfig;
					m_sock = new SSLSock(m_glob, scfg.getCertificate(),
						scfg.getCertificateKey(), false);
				}
				else {
					// Don't require anything at client
					m_sock = new SSLSock(m_glob, false);
				}
			}
			else {
				// Unencypted comms
				if(m_serverConfig.getProtocol() == Enums.Protocol.TCP) {
					m_sock = new TCPSock(m_glob, false);
				}
				else {
					m_sock = new UDPSock(m_glob, false);
				}
			}

			// Get address & port to bind server on
			IPAddr listenIPAddr = m_serverConfig.getListenIP();
			if(listenIPAddr.getHost() == "-") {
				// Pick our own port to bind to on loopback
				listenIPAddr = new IPAddr("127.0.0.1", 0);
			}

			// Bind to requested address
			m_sock.bind(listenIPAddr);
			m_sock.listen();

			// Update listening address & port incase they've changed
			m_serverConfig.setListenIP(listenIPAddr);

			// Launch server
			m_thread = new Thread(new ThreadStart(startServer));
			m_thread.Start();
		}


		/*************/
		/* Constants */
		/*************/

		// Time to wait in poll()
		protected const int						POLL_TIMEOUT	= 100;


		/********/
		/* Data */
		/********/

		// Reference to globals
		protected Globals							m_glob;

		// Our main thread
		protected Thread							m_thread;

		// Our server config
		protected SideConfig						m_serverConfig;

		// Socket to listen for incoming connections
		protected Sock								m_sock;

		// A Tunneler to handle tunneling
		protected Tunneler						m_tunneler;

		// Mutex & flag to control shutdown
		protected Mutex							m_shutDownMtx;
		protected bool								m_shutDown;

		// Mutexes for synchronising with server thread start
		protected Mutex							m_serverThreadStartMtx;
		protected Mutex							m_serverThreadContMtx;

		// Mutex
		protected Mutex							m_serverMtx;

	}

}	// namespace

