/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate settings for server authentication.
 *
 * $Id: ServerAuthSettings.cs,v 1.1 2008/07/25 15:22:38 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Text;

	public class ServerAuthSettings : AuthSettings, ICloneable {

		// Construction
		public ServerAuthSettings(Globals glob) : base(glob)
		{
		}

		protected ServerAuthSettings(ServerAuthSettings that) : base(that)
		{
		}

		// Clone
		public object Clone()
		{
			ServerAuthSettings clone = new ServerAuthSettings(this);
			return clone;
		}


		/************/
		/* Mutators */
		/************/

		// Setup server side authentication settings
		public void setServerSideAuth(AuthScheme scheme, string realm)
		{
			m_authScheme = scheme;
			m_realm = realm;

			// What authentication scheme?
			if(m_authScheme == AuthScheme.AUTH_DIGEST) {
				// Create initial digest CNonce
				m_digestCNonceCount = 1;
				StringBuilder cnonce = new StringBuilder();
				Random rand = new Random();
				cnonce.AppendFormat("{0:d8}", rand.Next());
				m_digestCNonce = cnonce.ToString();
			}
			else {
				// Unsupported authentication scheme at the server
				throw new AuthProcessingError("Unsupported server authentication scheme");
			}
		}


		/*************/
		/* Utilities */
		/*************/

		// Construct tunneling server response header
		public override string getResponseHeader(string uri)
		{
			StringBuilder res = new StringBuilder();

			// Using authorization?
			if(m_authScheme == AuthScheme.AUTH_BASIC) {
				// Create basic authentication fields
				res.AppendFormat("WWW-Authenticate: Basic realm=\"{0}\"\r\n", m_realm);

				// Remember trying basic authentication
				finishedAuthScheme(AuthScheme.AUTH_BASIC);
			}
			else if(m_authScheme == AuthScheme.AUTH_DIGEST) {
				// Set digest authentication fields
				res.AppendFormat("WWW-Authenticate: Digest realm=\"{0}\", ", m_realm);
				res.AppendFormat("nonce=\"{0}\", ", m_digestNonce);
				res.AppendFormat("qop=\"{0}\", ", m_digestQop);
				res.Append("stale=false");
				res.Append("\r\n");

				// Remember trying digest authentication
				finishedAuthScheme(AuthScheme.AUTH_DIGEST);
			}

			return res.ToString();
		}


		// Clear and reset any exisiting authentication scheme details
		public override void reset()
		{
			// Reset current scheme
			m_authScheme = AuthScheme.AUTH_NONE;

			// Clear tried schemes
			m_triedSchemes.Clear();
		}


		/***********************/
		/* Protected Interface */
		/***********************/


		/***********/
		/* Methods */
		/***********/

		// Get authorisation request header field identifier
		protected override string getAuthRequestHeaderID()
		{
			return "Authorization";
		}

	}

}	// namespace

