/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Interface for authenticating a user at the server, and
 * tracking/controling resource usage.
 *
 * $Id: ServerAuthenticator.cs,v 1.1 2008/07/25 15:22:38 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System.Collections;

	abstract public class ServerAuthenticator : Authenticator {

		// Construction
		public ServerAuthenticator(Globals glob) : base(glob)
		{
		}


		/*************/
		/* Accessors */
		/*************/

		// Get whether user is authenticated for some tunnel
		public override bool getUserAuthenticated(string name, Tunnel tun)
		{
			bool res = true;

			lock(this) {
				// Lookup user in cached details
				User user = getUser(name);
				if(user != null) {
					// Is user authenticated for this tunnel?
					res = user.getAuthenticated(tun);
				}
			}

			return res;
		}


		/*************/
		/* Utilities */
		/*************/

		// Log details about a new tunnel connection
		public virtual uint logNewConnection(ServerTunnel tun)
		{
			// Do nothing by default
			return 0;
		}

		// Log details about a closing tunnel
		public virtual void logCloseTunnel(ServerTunnel tun)
		{
			// Do nothing by default
		}


		/***********************/
		/* Protected Interface */
		/***********************/


		/***********/
		/* Methods */
		/***********/

		// Create a new tunnel user
		protected override User createUser(string name, bool keepCached)
		{
			// Create user un-authenticated
			Quotas quotas = new Quotas();
			User user = new User(m_glob, false, name, "", quotas, keepCached);
			return user;
		}


		// Check a digest authentication response
		protected bool checkDigestAuthResponse(string user, string passwd,
		ServerAuthSettings settings, ArrayList authArgs)
		{
			bool res = false;

			// Get required authentication parameters
			string realm = "", nonce = "", uri = "", alg = "", qop = "";
			string cnonce = "", nc = "", response = "";
			if(Utils.getParmValue(authArgs, "realm", ref realm) &&
			Utils.getParmValue(authArgs, "nonce", ref nonce) &&
			Utils.getParmValue(authArgs, "uri", ref uri) &&
			Utils.getParmValue(authArgs, "algorithm", ref alg) &&
			Utils.getParmValue(authArgs, "qop", ref qop) &&
			Utils.getParmValue(authArgs, "cnonce", ref cnonce) &&
			Utils.getParmValue(authArgs, "nc", ref nc) &&
			Utils.getParmValue(authArgs, "response", ref response)) {
				// Calculate expected response & compare
				string calcresponse = AuthDigest.MD5DigestResponse(alg, user,
					passwd, realm, nonce, cnonce, nc, qop, uri);
				res = (response == calcresponse);
			}

			return res;
		}

	}

}	// namespace

