/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * A convenient place to store all server side global configuration.
 *
 * $Id: ServerSideConfig.cs,v 1.2 2013/02/17 10:53:18 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.IO;

	public class ServerSideConfig : SideConfig {

		// Construction
		public ServerSideConfig(Globals glob, bool ssl, bool sockTun, string cert,
		string certKey, IPAddr listenIP, string firewallIPs, string firewallPorts,
		bool auth) : base(glob, Enums.Protocol.TCP, ssl, sockTun, listenIP)
		{
			m_useAuth = auth;
			m_cert = cert;
			m_certKey = certKey;
			m_userCallback = null;

			// Create firewall
			m_firewall = new Firewall(glob, firewallIPs, firewallPorts);
		}


		/*************/
		/* Accessors */
		/*************/

		// Server config?
		public override bool isServer()
		{
			return true;
		}

		// Get firewall
		public Firewall getFirewall()
		{
			return m_firewall;
		}

		// Get SSL certificate
		public string getCertificate()
		{
			return m_cert;
		}

		// Get SSL private key
		public string getCertificateKey()
		{
			return m_certKey;
		}

		// Get user callback
		public SockTunServer.SockTunCallback getUserCallback()
		{
			return m_userCallback;
		}


		/************/
		/* Mutators */
		/************/

		// Turn authentication on/off
		public void setAuthenticate(bool auth)
		{
			m_useAuth = auth;
		}

		// Set user callback
		public void setUserCallback(SockTunServer.SockTunCallback cb)
		{
			m_userCallback = cb;
		}


		/*************/
		/* Utilities */
		/*************/

		// Create authenticator
		public void createAuthenticator()
		{
			// Create default authenticator
			m_authenticator = new DefaultServerAuthenticator(m_glob);
		}


		// Serialize to a file
		public override void serialize(StreamWriter file)
		{
			// Server header
			file.WriteLine("#######################");
			file.WriteLine("## Server configuration");
			file.WriteLine("#######################");
			file.WriteLine();
			file.WriteLine("[SERVER]");

			// Local listening address
			file.Write("listen ");
			string host = m_listenIP.getHost();
			if(host.Length > 0) {
				file.Write(host + " ");
			}
			file.Write(m_listenIP.getPort().ToString());
			if(host.Length == 0) {
				file.Write(" " + m_listenIP.getFamilyStr());
			}
			file.WriteLine();

			// Server authentication
			if(m_useAuth) {
				file.WriteLine("svrauth");
			}

			// SSL certificate & key
			if(m_ssl) {
				file.WriteLine("ssl " + getCertificate() + " " + getCertificateKey());
			}

			// Firewall settings
			m_firewall.serialize(file);

			file.WriteLine();
		}


		/***********************/
		/* Protected Interface */
		/***********************/


		/********/
		/* Data */
		/********/

		// Server authenticator
		protected bool										m_useAuth;

		// Firewall
		protected Firewall								m_firewall;

		// SSL certificate & private key files
		protected string									m_cert;
		protected string									m_certKey;

		// Callback for user socket tunnel notifications
		protected SockTunServer.SockTunCallback	m_userCallback;

	}

}	// namespace

