/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Server side Tunnel.
 *
 * $Id: ServerTunnel.cs,v 1.4 2013/02/17 10:53:18 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Collections;
	using System.Text;

	public class ServerTunnel : Tunnel {

		// Construction
		public ServerTunnel(Globals glob, TunnelConfig tcfg, ServerTunneler tunneler)
		: base(glob, tcfg, tunneler, Direction.LEFT, TunnelState.AUTHREQ)
		{
			m_logKey = 0;
			m_authFailureCount = 0;
			m_cliVerStr = "";

			m_glob.DPRINTI("ServerTunnel::ServerTunnel()");
		}

		// Destruction
		public override void destroyed()
		{
			#if DEBUG_TUN
			StringBuilder str = new StringBuilder();
			str.AppendFormat("ServerTunnel::~ServerTunnel() - Destroying tunnel {0}",
				m_tid.ToString());
			m_glob.DPRINTI(str.ToString());
			#endif

			lock(this) {
				// Tell user they own one less tunnel
				ServerAuthenticator authenticator = getServerAuthenticator();
				authenticator.userTunnelClosed(m_user, this);

				// Log tunnel closed
				authenticator.logCloseTunnel(this);

				// Base class
				base.destroyed();
			}
		}


		/*************/
		/* Accessors */
		/*************/

		// Get source addresses of connections for this tunnel
		public override string getSrcConnectionAddresses()
		{
			StringBuilder res = new StringBuilder();
			lock(this) {
				// Get addresses for left side connections
				bool comma = false;
				for(int cit = 0; cit < m_leftConn.Count; ++cit) {
					// Should only be HTTPConnections in server side tunnel
					if(m_leftConn[cit] is HTTPConnection) {
						HTTPConnection conn = (HTTPConnection)m_leftConn[cit];

						// Do we have a client address?
						string cliAddr = conn.getClientAddr();
						if(cliAddr.Length > 0) {
							if(comma)
								res.Append(",");
							else
								comma = true;
							res.Append(cliAddr);
						}
					}
				}
			}

			return res.ToString();
		}

		// Get destination addresses of connections for this tunnel
		public override string getDstConnectionAddresses()
		{
			StringBuilder res = new StringBuilder();
			lock(this) {
				// Get addresses for left side connections
				bool comma = false;
				for(int cit = 0; cit < m_rightConn.Count; ++cit) {
					// Should only be RawConnections in server side tunnel
					if(m_leftConn[cit] is RawConnection) {
						RawConnection conn = (RawConnection)m_rightConn[cit];

						// Do we have a destination address?
						string destAddr = conn.getDestIP().getHost();
						uint destPort = conn.getDestIP().getPort();
						if(destPort > 0) {
							if(comma)
								res.Append(",");
							else
								comma = true;
							res.AppendFormat("{0};{1}", destAddr, destPort);
						}
					}
				}
			}

			return res.ToString();
		}


		// Get logging key
		public uint getLogKey()
		{
			return m_logKey;
		}

		// Get client version identifier
		public string getClientVersionStr()
		{
			return m_cliVerStr;
		}


		/************/
		/* Mutators */
		/************/

		// A connection that carries us has connected
		public override void connectionConnected(Connection conn)
		{
			lock(this) {
				// Base class
				base.connectionConnected(conn);

				// Maybe log this new connection
				if(m_tunnelState == TunnelState.CONNECTED) {
					ServerAuthenticator authenticator = getServerAuthenticator();
					m_logKey = authenticator.logNewConnection(this);
				}
			}
		}


		/********************/
		/* Public Interface */
		/********************/

		// Process some newly arrived tunnel msg from one of our connections
		public override void processTunnelMsg(HTTPMsg msg, HTTPConnection httpConn)
		{
			m_glob.DPRINTV("ServerTunnel::processTunnelMsg() " + m_tid.ToString());

			lock(this) {
				// Do we need to try and authenticate client?
				if(!userAuthenticated()) {
					m_glob.DPRINTV("Client not authenticated");
					authenticateClient(msg.getAuthArgs());
				}

				// Is the client authenticated now?
				if(userAuthenticated()) {
					// Base class
					base.processTunnelMsg(msg, httpConn);
				}
			}
		}


		/***********************/
		/* Protected Interface */
		/***********************/


		// Create complete HTTP msg parameters for sending down tunnel - Caller hold mutex
		protected override HTTPMsgParms makeTunnelMsg(HTTPMsgParms recvd, bool retry)
		{
			// Get basic parameters from base class
			HTTPMsgParms parms = base.makeTunnelMsg(recvd, retry);

			// Do we need to authenticate the client?
			if(!userAuthenticated()) {
				parms.setRespCodePhrase(401, "Unauthorized");
			}

			// Is the client version ok?
			if(recvd.getState() == TunnelState.CONNECT) {
				m_cliVerStr = recvd.getCliVer();

				// Split version fields
				string[] verFields = m_cliVerStr.Split(' ');
				string cliVerNum = verFields[0];

				// Split version number
				string[] verNumFields = cliVerNum.Split('.');
				if(verNumFields.Length == 3) {
					uint cliMajor = UInt32.Parse(verNumFields[0]);
					uint cliMinor = UInt32.Parse(verNumFields[1]);
					uint cliPoint = UInt32.Parse(verNumFields[2]);

					// Check version fields
					bool verOk = false;
					if(cliMajor > MIN_CLIVER_MAJOR) {
						verOk = true;
					}
					else if(cliMajor == MIN_CLIVER_MAJOR) {
						if(cliMinor > MIN_CLIVER_MINOR) {
							verOk = true;
						}
						else if(cliMinor == MIN_CLIVER_MINOR) {
							if(cliPoint >= MIN_CLIVER_POINT) {
								verOk = true;
							}
						}
					}

					if(!verOk) {
						// Client too old, so reject them
						StringBuilder nfo = new StringBuilder();
						nfo.AppendFormat("Client version {0} is too old, rejecting them", cliVerNum);
						m_glob.log().info(nfo.ToString());

						// Tell client they are forbidden
						parms.setRespCodePhrase(403, "Forbidden");

						// Send notification back to client
						StringBuilder ntf = new StringBuilder();
						ntf.AppendFormat("Your client software version ({0}) is no longer supported:",
							cliVerNum);
						addSendNotifMsg(ntf.ToString());
						addSendNotifMsg("Try upgrading your client software.");
					}
				}
				else {
					// Reject client
					StringBuilder wrn = new StringBuilder();
					wrn.AppendFormat("Bad format for client version identifier: {0}", m_cliVerStr);
					m_glob.log().warn(wrn.ToString());

					// Tell client they are forbidden
					parms.setRespCodePhrase(403, "Forbidden");
				}
			}

			return parms;
		}


		// Successfully sent a new message down the tunnel - Caller hold mutex
		protected override void sentNewTunnelMsg(HTTPConnection conn, HTTPMsgParms parms)
		{
			// Only remember if the sent msg is part of a normal authenticated tunnel session
			bool remember = userAuthenticated();
			base.sentNewTunnelMsg(conn, parms, remember);
		}


		// Determine whether a multi-ack implies we should resend a msg - Caller holds mutex
		protected override bool checkForMultiAckResendRequest(SentMsgParms sentMsgParms)
		{
			bool ackCountExceeded = false;

			// Update ack count
			sentMsgParms.m_ackCount++;

			/* The client may send several msgs over a persistant connection with
			 * the same ack'd sequence number. So don't consider this a resend request
			 * unless they've exceeded the maximum possible unacked count.
			 */

			if(sentMsgParms.m_ackCount >= m_maxUnackedMsgs) {
				m_glob.DPRINTV("ServerTunnel::checkForMultiAckResendRequest() - Multi-Ack resend request");
				ackCountExceeded = true;
				sentMsgParms.m_ackCount = 0;
			}

			return ackCountExceeded;
		}


		// Can we send a new tunnel msg? - Caller holds mutex
		protected override bool clearToSendNewMsg(HTTPConnection conn)
		{
			bool sendNew = false;

			try {
				// Check for unacked management msgs and consider resending them early
				checkUnackedMgmtMsgs();

				// New message or resend of old message?
				sendNew = !resendUnackedMsgs(conn);
			}
			catch(ResendError e) {
				// Resending error
				m_glob.log().warn("Error resending msg: " + e.Message);
				recordConnectionError(TunnelErrorType.MSG_RESEND_ERROR);
			}

			return sendNew;
		}


		// Check whether any connections need to be blocked/unblocked - Caller holds mutex
		protected override void checkBlockedConnections()
		{
			// Check raw connections
			ArrayList filt = new ArrayList();
			filt.Add(new SelRightConn());
			ArrayList conns = new ArrayList();
			getConnsFromSel(filt, conns);

			// Do we want to block or unblock?
			bool unblock = false;
			bool block = (m_rawDataQueueSize >= MAX_RAW_DATA_FILL);
			if(!block) {
				// Only unblock if we're well below full
				uint buffFill = (uint)(((float)m_rawDataQueueSize / (float)MAX_RAW_DATA_FILL)
					* 100);
				unblock = (buffFill <= RAW_DATA_UNBLOCK_LEVEL);
			}

			// Check each connection
			for(int cit = 0; cit != conns.Count; ++cit) {
				// Get connection
				Connection conn = (Connection)conns[cit];

				// Need to change blocked state?
				if(block)
					conn.setBlocked(true);
				else if(unblock)
					conn.setBlocked(false);
			}
		}


		// Advance our internal state machine - Caller holds mutex
		protected override void advanceState(HTTPMsgParms parms)
		{
			// Respond to the reported state from this msg
			switch(parms.getState()) {
				case(TunnelState.CONNECT): {
					// Now we need to authenticate
					setState(TunnelState.AUTHREQ);
					break;
				}
				case(TunnelState.RECONNECT): {
					// Reconnected
					setState(TunnelState.CONNECTED);
					break;
				}
				case(TunnelState.CLOSING): {
					// Closing
					setState(TunnelState.CLOSING);
					break;
				}
				case(TunnelState.CLOSED): {
					// Closed
					setState(TunnelState.CLOSED);
					break;
				}
				case(TunnelState.CONNECTED): {
					// Stay connected
					setState(TunnelState.CONNECTED);
					break;
				}
				case(TunnelState.AUTHREQ):
				case(TunnelState.UNKNOWNSTATE):
				default: {
					StringBuilder err = new StringBuilder();
					err.AppendFormat("Unexpected ServerTunnel reported msg state: {0}",
						parms.getState());
					m_glob.log().error(err.ToString());
					break;
				}
			}
		}

		// The tunnel state has changed - Caller holds mutex
		protected override void stateChanged(TunnelState oldState)
		{
			// Base class
			base.stateChanged(oldState);

			// Log if we've become CONNECTED
			if(m_tunnelState == TunnelState.CONNECTED) {
				ServerAuthenticator authenticator = getServerAuthenticator();
				m_logKey = authenticator.logNewConnection(this);
			}
		}


		// Get maximum allowable payload size - Caller holds mutex
		protected override uint getMaxPayloadSize(bool useHeader)
		{
			// Get payload size from server tunnel parameters
			ServerTunneler svrTunneler = (ServerTunneler)m_tunneler;
			ServerTunnelParms tunParms = svrTunneler.getTunParms(m_tid.getTunnelerID());
			return tunParms.getMaxPayloadSize(useHeader);
		}


		// Get server side authenticator
		protected virtual ServerAuthenticator getServerAuthenticator()
		{
			// Get base authenticator
			Authenticator auth = getAuthenticator();

			// Convert to ServerAuthenticator and return
			ServerAuthenticator svrAuth = (ServerAuthenticator)auth;
			return svrAuth;
		}


		/*********************/
		/* Private Interface */
		/*********************/


		/*************/
		/* Constants */
		/*************/

		// Minimum client version we allow to connect
		private const uint			MIN_CLIVER_MAJOR		= 1;
		private const uint			MIN_CLIVER_MINOR		= 4;
		private const uint			MIN_CLIVER_POINT		= 0;

		// Maximum no. of authentication failures we allow
		private const uint			MAX_AUTH_FAILURES		= 2;


		/***********/
		/* Methods */
		/***********/

		// Try to authenticate a client - Caller holds mutex
		private void authenticateClient(Hashtable authDetails)
		{
			// Get authentication arguments for a suitable scheme
			ArrayList authArgs = new ArrayList();
			IDictionaryEnumerator ait = authDetails.GetEnumerator();
			if(ait.MoveNext())
				authArgs = (ArrayList)ait.Value;

			// Get supplied user name
			string user = "";
			if(Utils.getParmValue(authArgs, "username", ref user)) {
				// Use supplied user name
				m_user = user;
			}

			// First check quotas for number of tunnels
			Authenticator authenticator = getAuthenticator();
			if(authenticator.getUserCanCreateTunnel(m_user) || isTestTunnel()) {
				// Try to authenticate
				authenticator.authenticate(m_user, this, authArgs);
				if(userAuthenticated()) {
					m_glob.DPRINTV("Client authentication OK");
					m_authFailureCount = 0;

					// Let controlling tunneler know
					m_tunneler.scheduleEvent(new AuthSuccessEvent(m_glob, m_tunneler, m_tid));
				}
				else {
					m_glob.DPRINTV("Client authentication failed");
					m_authFailureCount++;

					// Give up if the client keeps failing to authenticate
					if(m_authFailureCount >= MAX_AUTH_FAILURES) {
						// Record fatal error
						recordConnectionError(TunnelErrorType.CLIENT_AUTH_FAILED_ERROR);

						// Send notification back to client
						if(m_user == authenticator.getDefaultUserName()) {
							addSendNotifMsg("No authentication details given to server:");
							addSendNotifMsg("Please check your configuration.");
						}
						else {
							addSendNotifMsg("Authentication for user '" + m_user + "' failed:");
							addSendNotifMsg("Please check your configured user name and password.");
						}
					}
				}
			}
			else {
				// User trying to exceed tunnel number quota
				StringBuilder nfo = new StringBuilder();
				nfo.AppendFormat("User {0} exceeding tunnel count quota, ", m_user);
				nfo.Append("not allowing them to authenticate");
				m_glob.log().info(nfo.ToString());
				recordConnectionError(TunnelErrorType.FATAL_QUOTAS_ERROR);

				// Send notification back to client
				addSendNotifMsg("Authentication of new tunnel for user '" + m_user + "' failed:");
				addSendNotifMsg("Maximum active tunnel quota exceeded.");
			}
		}


		// Is this tunnel user authenticated? - Caller holds mutex
		private bool userAuthenticated()
		{
			Authenticator authenticator = getAuthenticator();
			return authenticator.getUserAuthenticated(m_user, this);
		}


		/********/
		/* Data */
		/********/

		// Key to use for logging
		private uint					m_logKey;

		// Count of consecutive client authentication failures
		private uint					m_authFailureCount;

		// Client version identifier
		private string					m_cliVerStr;

	}

}	// namespace

