/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Tunnel controller for server end of tunnel.
 *
 * $Id: ServerTunneler.cs,v 1.4 2013/02/17 10:53:19 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Collections;
	using System.Text;

	public class ServerTunneler : Tunneler {

		// Construction
		public ServerTunneler(Globals glob, Server server) : base(glob, server)
		{
			m_glob.DPRINTV("ServerTunneler::ServerTunneler()");

			m_tunParms = new Hashtable();
			m_tcpSockTunServer = null;
			m_udpSockTunServer = null;

			try {
				// Create authenticator
				ServerSideConfig serverCfg = getServerSideConfig();
				serverCfg.createAuthenticator();

				// See if we need a socket tunnel server
				if(serverCfg.isSockTun()) {
					m_tcpSockTunServer = new TCPSockTunServer(m_glob, serverCfg.getUserCallback());
					m_udpSockTunServer = new UDPSockTunServer(m_glob, serverCfg.getUserCallback());
				}
			}
			catch(TunnelError e) {
				// Clear up and rethrow
				destroyed();
				throw e;
			}
		}

		// Destruction
		public override void destroyed()
		{
			m_glob.DPRINTV("ServerTunneler::~ServerTunneler()");

			// Shutdown event scheduling
			if(!shuttingDown())
				shutdownSchedulers();

			// Shutdown any socket tunnel server
			if(m_tcpSockTunServer != null) {
				m_tcpSockTunServer.destroyed();
				m_tcpSockTunServer = null;
			}
			if(m_udpSockTunServer != null) {
				m_udpSockTunServer.destroyed();
				m_udpSockTunServer = null;
			}

			// Base class
			base.destroyed();
		}


		/*************/
		/* Accessors */
		/*************/

		// Get reference to current tunneling parameters
		public ServerTunnelParms getTunParms(TunnelerID tunnelerID)
		{
			// Assume default parms unless we find better
			ServerTunnelParms parms = TestServerTunnel.getDefaultTunParms();

			// Lookup parms for the specified client tunneler ID
			lock(m_tunnels) {
				if(m_tunParms.ContainsKey(tunnelerID)) {
					// Got parameters for specified tunneler ID
					parms = (ServerTunnelParms)m_tunParms[tunnelerID];
				}
			}

			return parms;
		}


		/**********/
		/* Events */
		/**********/

		// Notification of a new socket connection
		public override void newConnectionArrived(Sock insock)
		{
			m_glob.DPRINTI("ServerTunneler::newConnectionArrived()");

			//////////////////////////////////////////////////////////////////
			// In ServerTunneler we can't really do much until we are told
			// whether this is a connection from a new tunnel, or a new pipe
			// within an existing tunnel.
			//////////////////////////////////////////////////////////////////

			// Lock connection details
			lock(m_connections) {
				// Create HTTP connection to handle incoming traffic
				HTTPConnection httpconn = new HTTPConnection(m_glob, insock, this);
				addConnection(httpconn);
			}
		}


		// Notification that a tunnel has insufficient left/right connections to be viable
		public override void unviableTunnel(TunnelID tid)
		{
			m_glob.DPRINTI("ServerTunneler::unviableTunnel() - " + tid);

			// Lookup tunnel
			lock(m_tunnels) {
				Tunnel tun = getTunFromPred(new PredTunID(tid));
				if(tun != null) {
					// If tunnel is already closed don't bother checking any further
					if(!tun.isClosed()) {
						// Get details about state of tunnel
						int leftConns = tun.noLeftConnections();
						int rightConns = tun.noRightConnections();

						// Check whether our left side warrants preserving
						bool leftPreserve = true;
						if(leftConns == 0) {
							/* In the server, no remaining left connections means that if
							 * we want to try and preserve this tunnel we should wait for
							 * a re-connection from the client.
							 */

							// Do we want to try and preserve this tunnel?
							if(!tun.errorLevelExceeded(Tunnel.Direction.LEFT)) {
								// Consider preserving left side
								m_glob.DPRINTI(tid + " - No lconns, not erred: Left preserve yes");
								leftPreserve = true;
							}
							else {
								// Don't consider preserving left side
								m_glob.DPRINTI(tid + " - No lconns, erred: Left preserve no");
								leftPreserve = false;
							}
						}

						// Check whether our right side warrants preserving
						bool rightPreserve = true;
						if(rightConns == 0) {
							if(!tun.everConnected(Tunnel.Direction.RIGHT)) {
								// Never been connected right, so might want to let it try
								if(!tun.errorLevelExceeded()) {
									// Haven't failed to connect right, so let it try
									m_glob.DPRINTI(tid + " - No rconns, not erred: Right preserve yes");
									rightPreserve = true;
								}
								else {
									// Maybe failed to connect right, don't try
									m_glob.DPRINTI(tid + " - No rconns, erred: Right preserve no");
									rightPreserve = false;
								}
							}
							else {
								// Is this tunnel in the process of closing?
								if(tun.isClosing()) {
									// Give us a chance to finish closing
									m_glob.DPRINTI(tid + " - No rconns, closing: Right preserve yes");
									rightPreserve = true;
								}
								else {
									// Have been connected right and not closing, so don't preserve
									m_glob.DPRINTI(tid + " - No rconns, not closing: Right preserve no");
									rightPreserve = false;
								}
							}
						}

						// Both sides agree to preserve?
						if(leftPreserve && rightPreserve) {
							m_glob.DPRINTI(tid + " - Trying to preserve");
							scheduleEvent(new ShutdownTunnelEvent(m_glob,
								new Time(TUNNEL_RECONNECT_TIMEOUT,0), this, tid, false));
						}
						else {
							m_glob.DPRINTI(tid + " - Not preserving");
							scheduleEvent(new ShutdownTunnelEvent(m_glob, new Time(0,0), this, tid, false));
						}
					}
					else {
						m_glob.DPRINTI(tid + " - Already closed, not preserving further");
					}
				}
				else {
					StringBuilder err = new StringBuilder();
					err.AppendFormat("ServerTunneler::unviableTunnel() - Unknown tunnel {0}", tid);
					m_glob.log().warn(err.ToString());
				}
			}
		}


		// Notification of new tunneled/raw data arrival
		public override void rawDataArrived(uint connID, TunnelID tid, RawBuff raw)
		{
			#if DEBUG_TUN
			StringBuilder nfo = new StringBuilder();
			nfo.AppendFormat("ServerTunneler::rawDataArrived() - {0} bytes ", raw.getLength());
			nfo.AppendFormat("on connection {0}", connID);
			m_glob.DPRINTV(nfo.ToString());
			#endif

			// Lookup tunnel
			lock(m_tunnels) {
				Tunnel tun = getTunFromPred(new PredTunID(tid));
				if(tun != null) {
					// Inform tunnel of raw data arrival
					tun.processRawData(raw);
				}
				else {
					StringBuilder err = new StringBuilder();
					err.Append("ServerTunneler::rawDataArrived() - ");
					err.AppendFormat("On connection {0}", connID);
					err.AppendFormat(" Unknown tunnel {0}", tid.ToString());
					throw new LookupManagementError(err.ToString());
				}
			}
		}


		// Check we have received an ack for some sent message
		public override void checkAckReceived()
		{
			lock(m_tunnels) {
				// Check acks for all tunnels
				for(int tit = 0; tit < m_tunnels.Count; ++tit) {
					// Let tunnel check we have received any due acks, & resend if not
					Tunnel tun = (Tunnel)m_tunnels[tit];
					tun.checkAckTimeout();
				}
			}

			// Schedule next ack check event
			scheduleEvent(new CheckAckEvent(m_glob, new Time(ACK_CHECK_PERIOD,0), this));
		}


		// Notification of new HTTP message arrival
		public override void messageArrived(uint connID, HTTPMsg msg)
		{
			// Base class
			base.messageArrived(connID, msg);

			lock(m_connections) {
				// Lookup connection
				int cit = getConnectionByID(connID);
				Connection conn = (Connection)m_connections[cit];
				HTTPConnection httpConn = (HTTPConnection)conn;

				// Process msg arrival
				bool fatalConnectionError = false;
				try {
					lock(m_tunnels) {
						// Find out which Tunnel this message is from
						HTTPMsgParms parms = msg.getParms();
						Tunnel.TunnelState state = parms.getState();
						TunnelID tid = new TunnelID(parms.getTid());
						Tunnel tun = getTunFromPred(new PredTunID(tid));

						// Do we already know about the tunnel this msg claims it is from?
						if(tun != null) {
							// Inform tunnel of payload arrival
							tun.processTunnelMsg(msg, httpConn);

							// What state does this message report the tunnel as?
							if(state == Tunnel.TunnelState.CONNECT) {
								// A new pipe connection within this tunnel
								m_glob.DPRINTI("Got new pipe connection within tunnel");
								newTunnelPipeConnectionRequest(tun, conn);
							}
							else if(state == Tunnel.TunnelState.CLOSED) {
								// Schedule tunnel shutdown
								m_glob.DPRINTI("Received tunnel shutdown request from client");
								scheduleEvent(new ShutdownTunnelEvent(m_glob, new Time(0,0),
									this, tid, true));
							}
							else if(state == Tunnel.TunnelState.CLOSING) {
								// Treat like a normal connected tunnel for now
								m_glob.DPRINTV("Peer reports tunnel as CLOSING");
							}

							// Check for reconnection events
							if(state == Tunnel.TunnelState.RECONNECT) {
								// Reconnection for existing tunnel
								m_glob.DPRINTI("Got explicit reconnection request from client");
								tunnelReconnectionRequest(tun, conn, state);
							}
							else {
								// Check for new connection from proxy
								if(!tun.ownsConnection(conn)) {
									// Treat like a reconnection
									m_glob.DPRINTI("Got reconnection request, probably from proxy");
									tunnelReconnectionRequest(tun, conn, state);
								}
							}

							// Try sending back any queued data + acks to the client
							tun.respondToTunnelMsg(conn, parms);
						}

						else {
							// Unknown tunnel ID; what state does this message report the tunnel as?
							if(state == Tunnel.TunnelState.CONNECT) {
								// Connection for a new tunnel
								tun = newTunnelConnectionRequest(conn, tid, parms);
								if(tun != null) {
									// Inform tunnel of payload arrival, & send back any queued data/acks
									tun.processTunnelMsg(msg, httpConn);
									tun.respondToTunnelMsg(conn, parms);
								}
								else {
									// Failed to create tunnel
									fatalConnectionError = true;
								}
							}
							else {
								StringBuilder wrn = new StringBuilder();
								wrn.AppendFormat("Unexpected state {0} in msg for unknown tunnel {1}", state, tid);
								m_glob.log().warn(wrn.ToString());
								fatalConnectionError = true;
							}
						}

						// Does this connection report it is closing?
						if(!fatalConnectionError && (parms.getConnClose() || parms.getProxyConnClose())) {
							#if DEBUG_TUN
							StringBuilder dbg1 = new StringBuilder();
							dbg1.AppendFormat("Connection id {0} reports it is closing", connID);
							m_glob.DPRINTV(dbg1.ToString());
							#endif

							// Don't use closing connection any more
							connectionClosing(cit);
						}
					}
				}
				catch(InvalidID e) {
					m_glob.log().error("Error parsing ID: " + e.Message);
					fatalConnectionError = true;
				}

				// Did something very serious fail?
				if(fatalConnectionError) {
					#if DEBUG_TUN
					StringBuilder dbg2 = new StringBuilder();
					dbg2.AppendFormat("Fatal error processing msg on connection {0}", connID);
					m_glob.DPRINTV(dbg2.ToString());
					#endif

					// No point keeping this connection around, kill it now
					sendFatalErrorResponse(conn, msg);
					cit = getConnectionByID(connID);
					destroyConnectionByIndex(cit);
				}
			}
		}


		// Notification of some new default tunnel parameters
		public override void newTunnelParms(TunnelParms parms, TunnelerID tid)
		{
			m_glob.DPRINTI("ServerTunneler::newTunnelParms() - " + tid);

			// Base class
			base.newTunnelParms(parms, tid);

			lock(m_tunnels) {
				// Are these brand new parms or an update to existing parms?
				ServerTunnelParms sparms = (ServerTunnelParms)parms;
				if(m_tunParms.ContainsKey(tid)) {
					// We already have parms for this client tunneler
					m_tunParms[tid] = sparms;

					#if DEBUG_TUN
					StringBuilder nfo = new StringBuilder();
					nfo.AppendFormat("New ServerTunnelParms for existing tunneler {0}", tid);
					nfo.AppendFormat(" : Max = {0}", sparms.getMaxPayloadSize(true));
					m_glob.DPRINTV(nfo.ToString());
					#endif
				}
				else {
					// Save parms from new client tunneler
					m_tunParms[tid] = sparms;

					#if DEBUG_TUN
					StringBuilder nfo = new StringBuilder();
					nfo.AppendFormat("ServerTunnelParms for new tunneler {0}", tid);
					nfo.AppendFormat(" : Max = {0}", sparms.getMaxPayloadSize(true));
					m_glob.DPRINTV(nfo.ToString());
					#endif
				}
			}
		}


		// Notification that authentication has succeeded for a tunnel
		public override void authenticationSuccess(TunnelID tid)
		{
			m_glob.DPRINTI("ServerTunneler::authenticationSuccess() - " + tid.ToString());

			// Lookup tunnel
			lock(m_connections) {
				lock(m_tunnels) {
					Tunnel tun = getTunFromPred(new PredTunID(tid));
					if(tun != null) {
						/* Now we have authenticated the connecting client, setup raw connection
						 * out to tunnel destination.
						 */

						// How we handle raw connection side depends on whether this is a test tunnel
						if(tun.isTestTunnel()) {
							// Create a dummy raw connection that never actually does anything
							RawConnection rawconn = new DummyRawConnection(m_glob, this, tid);
							addConnection(rawconn);
							tun.addRightConnection(rawconn);
						}
						else {
							// Create proper raw connection
							RawConnection rawconn = new RawConnection(m_glob, this, tid);
							addConnection(rawconn);
							tun.addRightConnection(rawconn);

							// Address of server to connect to
							TunnelConfig tcfg = tun.getConfig();
							IPAddr svrIP = tcfg.getDestIP();
							Enums.Protocol svrProtocol = tcfg.getDestProtocol();

							// Are we for a socket tunnel?
							ServerSideConfig serverCfg = getServerSideConfig();
							if(serverCfg.isSockTun()) {
								// What type of socket tunnel do we need to use?
								if(svrProtocol == Enums.Protocol.TCP) {
									// Get address of TCP socket tunnel server to connect to
									svrIP = m_tcpSockTunServer.getListenIP();
								}
								else {
									// Get address of UDP socket tunnel server to connect to
									svrIP = m_udpSockTunServer.getListenIP();
								}

								#if DEBUG_TUN
								StringBuilder dbg = new StringBuilder();
								dbg.Append("ServerTunneler connecting to local ");
								if(svrProtocol == Enums.Protocol.TCP)
									dbg.Append("TCP");
								else
									dbg.Append("UDP");
								dbg.AppendFormat(" SockTunServer at {0}", svrIP);
								m_glob.DPRINTV(dbg.ToString());
								#endif
							}

							// Schedule initial connect out raw side
							m_connectionMgr.addProxyConnect(rawconn, new Time(0,0), svrIP,
								svrProtocol, false);
						}
					}
					else {
						StringBuilder err = new StringBuilder();
						err.Append("ServerTunneler::authenticationSuccess() - ");
						err.AppendFormat("Unknown tunnel {0}", tid);
						m_glob.log().warn(err.ToString());
					}
				}
			}
		}


		// Notification that we should perform tunnel manangement
		public override void performTunnelManagement()
		{
			lock(m_tunnels) {
				// When did we last check tunneling parameters?
				TimeSpan checkTime = Time.now() - m_lastTunParmsCheckTime;
				if(checkTime.TotalSeconds >= CHECK_TUN_PARMS_PERIOD) {
					m_lastTunParmsCheckTime = Time.now();

					#if DEBUG_TUN
					if(m_tunParms.Count > 0) {
						StringBuilder dbg1 = new StringBuilder();
						dbg1.Append("ServerTunneler::performTunnelManagement() - Managing ");
						dbg1.AppendFormat("{0} test tunneling parms", m_tunParms.Count);
						m_glob.DPRINTV(dbg1.ToString());
					}
					#endif

					// Check our store of tunneling parameters for old details
					bool foundOne = false;
					do {
						foundOne = false;
						IDictionaryEnumerator pit = m_tunParms.GetEnumerator();
						while(pit.MoveNext()) {
							// Are these parms old enough to consider removing?
							ServerTunnelParms parms = (ServerTunnelParms)pit.Value;
							TimeSpan parmAge = Time.now() - parms.getCreateTime();
							if(parmAge.TotalSeconds >= EXPIRE_OLD_TUN_PARMS_AGE) {
								// Do we have an active test tunnel trying to update them?
								TunnelerID tid = (TunnelerID)pit.Key;
								if(getTunFromPred(new PredTestTunID(tid)) == null) {
									#if DEBUG_TUN
									StringBuilder dbg2 = new StringBuilder();
									dbg2.AppendFormat("Removing {0} secs old test ", parmAge.TotalSeconds);
									dbg2.AppendFormat("tunneling parms for {0}", tid);
									m_glob.DPRINTV(dbg2.ToString());
									#endif

									// Remove old parms
									m_tunParms.Remove(tid);
									foundOne = true;
									break;
								}
							}
						}
					} while(foundOne);
				}

				// Base class
				base.performTunnelManagement();
			}
		}


		/***********************/
		/* Protected Interface */
		/***********************/


		/*************/
		/* Constants */
		/*************/

		// Number of connect attempts before giving up
		protected const uint				SERVER_CONNECT_ATTEMPTS		= 3;

		// Time between reconnection attempts
		protected const uint				SERVER_RECONNECT_DELAY		= 1;


		/***********/
		/* Methods */
		/***********/

		// Get reconnection attempt parameters
		protected override void getReconnectParms(ref uint maxConnectAttempts,
		ref uint reconnectDelay, uint failedAttempts)
		{
			// Keep reconnect delay constant at the server
			maxConnectAttempts = SERVER_CONNECT_ATTEMPTS;
			reconnectDelay = SERVER_RECONNECT_DELAY;
		}


		/********/
		/* Data */
		/********/

		// Tunneling parameters to use for each connected client
		protected Hashtable							m_tunParms;

		// Socket tunnel server
		protected TCPSockTunServer					m_tcpSockTunServer;
		protected UDPSockTunServer					m_udpSockTunServer;


		/*********************/
		/* Private Interface */
		/*********************/


		/*************/
		/* Constants */
		/*************/

		// Timeout for client reconnecting a broken tunnel
		private const int			TUNNEL_RECONNECT_TIMEOUT		= 30;

		// Control when we expire old tunnel parms
		private const int			EXPIRE_OLD_TUN_PARMS_AGE		= OLD_TUN_PARMS_AGE * 5;


		/***********/
		/* Methods */
		/***********/

		// Create a new carried tunnel in response to a received message - Caller holds mutexes
		private Tunnel newTunnelConnectionRequest(Connection conn, TunnelID tid,
		HTTPMsgParms parms)
		{
			#if DEBUG_TUN
			StringBuilder nfo = new StringBuilder();
			nfo.AppendFormat("ServerTunneler::newTunnelConnectionRequest() - {0}", tid);
			m_glob.DPRINTI(nfo.ToString());
			#endif

			// Create tunnel config
			ServerSideConfig serverCfg = getServerSideConfig();
			TunnelConfig tcfg = new TunnelConfig(tid, parms.getTunTune(), parms.getTunDest(),
				serverCfg.getAuthenticator().getDefaultAuthSettings());

			// Create tunnel
			ServerTunnel tun = null;
			if(parms.getTestState() != Tunnel.TunnelTestState.NO_TESTING) {
				// Request for new test tunnel
				checkNewTestTunnelConnection(tid.getTunnelerID());
				tun = new TestServerTunnel(m_glob, tcfg, this);
			}
			else {
				// Plain tunnel
				tun = new ServerTunnel(m_glob, tcfg, this);
			}

			// Add incoming connection
			tun.addLeftConnection(conn);

			// Store tunnel object in our list of active tunnels
			m_tunnels.Add(tun);

			return tun;
		}

		// Reconfigure an existing tunnel in response to a reconnection message
		private void tunnelReconnectionRequest(Tunnel tun, Connection conn,
		Tunnel.TunnelState reportedState)
		{
			m_glob.DPRINTI("ServerTunneler::tunnelReconnectionRequest() - " +
				tun.getTidStr());

			// Let tunnel know it has a new incoming connection
			tun.addLeftConnection(conn);

			// Is it worth keeping this tunnel?
			if(reportedState != Tunnel.TunnelState.CLOSED) {
				// Cancel any pending timeout
				cancelEvents<ShutdownTunnelEvent, TunnelID>(tun.getTid());
			}
		}

		// Reconfigure an existing tunnel in response to a new pipe connection
		private void newTunnelPipeConnectionRequest(Tunnel tun, Connection conn)
		{
			m_glob.DPRINTI("ServerTunneler::newTunnelPipeConnectionRequest() - " +
				tun.getTidStr());

			// Check we don't already know about this connection
			if(!tun.ownsConnection(conn)) {
				// Let tunnel know it has a new connection to manage
				tun.addLeftConnection(conn);
			}
		}


		// Send back a tunnel error response to a client
		private void sendFatalErrorResponse(Connection conn, HTTPMsg msg)
		{
			m_glob.DPRINTV("ServerTunneler::sendFatalErrorResponse()");

			try {
				// Create suitable msg parameters for response
				HTTPMsgParms recvd = msg.getParms();
				ErrorMsgParms respParms = new ErrorMsgParms(m_glob, recvd);

				// Send error response back to client
				conn.write(respParms);
			}
			catch(TunnelError e) {
				m_glob.log().warn("Failed to send error response msg: " + e.Message);
			}
		}


		// Check if a new test tunnel connection should clear out any old results - Caller holds mutex
		private void checkNewTestTunnelConnection(TunnelerID tunnelerID)
		{
			// Do we already have test parameters for the connecting tunneler?
			if(m_tunParms.Contains(tunnelerID)) {
				// If the existing test parms are completed, let this new test replace them
				ServerTunnelParms parms = (ServerTunnelParms)m_tunParms[tunnelerID];
				if(parms.testsCompleted()) {
					// Reset parms test state
					m_glob.DPRINTV("Retesting tunneling parms for " + tunnelerID);
					parms.resetTestState();
				}
				else {
					// Do we have a test already running for the connecting tunneler?
					Tunnel oldTun = getTunFromPred(new PredTestTunID(tunnelerID));
					if(oldTun != null) {
						m_glob.DPRINTV("Got replacement tunneling test connection for " + tunnelerID);

						// Shutdown existing test tunnel and let this one replace it
						scheduleEvent(new ShutdownTunnelEvent(m_glob, new Time(0,0), this,
							oldTun.getTid(), true));
					}
				}
			}
		}

	}


}	// namespace

