/* (C) 2007 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate a TCP/UDP socket.
 *
 * $Id: Sock.cs,v 1.5 2013/02/17 10:53:19 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{

	using System;
	using System.Net;
	using System.Net.Sockets;
	using System.Text;
	using System.Threading;

	public abstract class Sock {

		// Construction
		public Sock(Globals glob, bool blocking)
		{
			m_glob = glob;
			m_fd = null;
			m_connectedClientAddr = "";
			m_blocking = blocking;

			/* Construct unconnected */

			// Set socket ID
			m_id = getNextSockID();
		}

		public Sock(Globals glob, Socket fd, string cliAddr)
		{
			m_glob = glob;
			m_fd = fd;
			m_connectedClientAddr = cliAddr;
			m_blocking = false;

			/* Construct already connected */

			// Set socket ID
			m_id = getNextSockID();
		}


		// We are no longer required
		public virtual void destroyed()
		{
		}


		/*************/
		/* Accessors */
		/*************/

		// Get address of connected client
		public string getConnectedClientAddr()
		{
			return m_connectedClientAddr;
		}


		// Get underlying socket descriptor
		public Socket getFD()
		{
			return m_fd;
		}


		/************/
		/* Mutators */
		/************/

		// Release and return our underlying socket descriptor
		public Socket release()
		{
			Socket sockfd = null;

			lock(this) {
				// Check we have a socket to release
				if(m_fd != null) {
					sockfd = m_fd;
					m_fd = null;
				}
				else {
					throw new SockError("Sock::release() - No socket to release", 0);
				}
			}

			return sockfd;
		}


		// Set blocking mode for socket
		public void setBlocking(bool blocking)
		{
			lock(this) {
				// Set flag
				m_blocking = blocking;

				// Set blocking state
				blocknonblock();
			}
		}


		/********************/
		/* Public Interface */
		/********************/

		// Wait for a new incoming TCP connection
		public abstract void newConnection(Tunneler tunneler);

		// Set socket listening
		public abstract void listen();

		// Wait for a new incoming connection
		public abstract Sock accept();

		// Establish connection out
		public abstract void connect(IPAddr ipaddr);

		// Read from socket
		public abstract int read(byte[] buff);

		// Write to socket
		public abstract int write(RawBuff buff);

		// Get the result of a non-blocking connect attempt
		public abstract void getConnectResult(ref bool resultAvailable, ref bool connected);


		// Bind socket to an address
		public virtual void bind(IPAddr ipaddr)
		{
			// Get address details
			string bindaddr = ipaddr.getHost();
			uint bindport = ipaddr.getPort();

			lock(this) {
				// Get local address to bind to
				IPAddress addr = System.Net.IPAddress.Any;
				if(bindaddr.Length > 0) {
					// Resolve bind address
					if(!Utils.getIPByName(bindaddr, ref addr)) {
						throw new SockError("Unable to resolve bind address " + bindaddr, 0);
					}
				}

				try {
					// Reuse address
					m_fd.SetSocketOption(SocketOptionLevel.Socket, SocketOptionName.ReuseAddress, 1);

					// Bind
					IPEndPoint serverAddr = new IPEndPoint(addr, (int)bindport);
					m_fd.Bind(serverAddr);
				}
				catch(SocketException e) {
					StringBuilder err = new StringBuilder();
					err.AppendFormat("Failed to bind socket to address {0} : {1}", ipaddr, e.Message);
					throw new SockError(err.ToString(), e.SocketErrorCode);
				}
				catch(Exception e) {
					StringBuilder err = new StringBuilder();
					err.AppendFormat("Failed to bind socket to address {0} : {1}", ipaddr, e.Message);
					throw new SockError(err.ToString(), 0);
				}

				// Did we let the OS pick an ephemeral port?
				if(bindport == 0) {
					// Find out what port we bound to
					ipaddr.setPort((uint)Utils.getSockPort(m_fd));
				}
			}
		}


		// Poll socket for reading
		public virtual bool pollRead(int timeout, ref bool err, ref bool hup)
		{
			bool ready = false;
			err = hup = false;

			lock(this) {
				// Do the poll
				try {
					ready = m_fd.Poll(timeout, SelectMode.SelectRead);
					err = m_fd.Poll(0, SelectMode.SelectError);
				}
				catch(Exception) {
					throw new SockError("Error polling socket", 0);
				}
			}

			return ready;
		}


		// Poll socket for writing
		public virtual bool pollWrite(int timeout, ref bool err)
		{
			bool ready = false;
			err = false;

			lock(this) {
				// Do the poll
				try {
					ready = m_fd.Poll(timeout, SelectMode.SelectWrite);
					err = m_fd.Poll(0, SelectMode.SelectError);
				}
				catch(Exception) {
					throw new SockError("Error polling socket", 0);
				}
			}

			return ready;
		}


		/***********************/
		/* Protected Interface */
		/***********************/


		/***********/
		/* Methods */
		/***********/

		// Create socket descriptor for use with some address
		protected void createSockFD(IPAddr addr, SocketType stype, ProtocolType ptype)
		{
			Socket sockfd = null;

			lock(this) {
				// Do we already have a socket?
				if(m_fd != null) {
					throw new SockError("Sock::createSockFD() - We already have a socket", 0);
				}

				try {
					// What IP version?
					switch(addr.getFamily()) {
						case AddrInfo.Family.IPV4:
							sockfd = new Socket(AddressFamily.InterNetwork, stype, ptype);
							break;

						case AddrInfo.Family.IPV6:
							sockfd = new Socket(AddressFamily.InterNetworkV6, stype, ptype);
							break;

						default:
							throw new SockError("Sock::createSockFD() - Unsupported protocol", 0);
					}
				}
				catch(SocketException e) {
					StringBuilder err = new StringBuilder();
					err.Append("Sock::createSockFD() - Failed to create new ");
					switch(stype) {
						case(SocketType.Stream):
							err.Append("TCP socket descriptor");
							break;
						case(SocketType.Dgram):
							err.Append("UDP socket descriptor");
							break;
						default:
							err.Append("Unknown socket descriptor");
							break;
					}
					err.AppendFormat(": {0}", e.Message);
					throw new SockError(err.ToString(), 0);
				}

				// Save socket descriptor
				m_fd = sockfd;

				// Set blocking mode
				blocknonblock();
			}
		}


		// Close socket
		protected virtual void close()
		{
			m_glob.DPRINTV("Sock::close()");

			lock(this) {
				if(m_fd != null) {
					m_fd.Close();
					m_fd = null;
				}
			}
		}


		// Setup blocking mode for this socket
		protected virtual void blocknonblock()
		{
			m_glob.DPRINTV("Sock::blocknonblock()");

			lock(this) {
				// Turn on blocking mode for socket
				try {
					m_fd.Blocking = m_blocking;
				}
				catch(Exception e) {
					throw new SockError("Unable to set socket blocking flag: " + e.Message, 0);
				}
			}
		}


		/********/
		/* Data */
		/********/

		// Globals
		protected Globals				m_glob;

		// Socket descriptor
		protected Socket				m_fd;

		// Address of connected client
		protected string				m_connectedClientAddr;

		// Should we be blocking or non-blocking?
		protected bool					m_blocking;


		/*********************/
		/* Private Interface */
		/*********************/


		/***********/
		/* Methods */
		/***********/

		// Get next socket ID
		private static uint getNextSockID()
		{
			uint id = 0;

			lock(c_idCountMtx) {
				id = c_idCount++;
				if(c_idCount == 0)
					c_idCount = 1;
			}

			return id;
		}


		/********/
		/* Data */
		/********/

		// Unique ID
		private static uint			c_idCount = 1;
		private static Mutex			c_idCountMtx = new Mutex();
		private uint					m_id;

	}

}	// namespace

