/* (C) 2007 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate a TCP socket.
 *
 * $Id: TCPSock.cs,v 1.5 2013/02/17 10:53:19 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{

	using System;
	using System.Net;
	using System.Net.Sockets;
	using System.Text;

	public class TCPSock : Sock {

		// Construction
		public TCPSock(Globals glob, bool blocking) : base(glob, blocking)
		{
		}

		public TCPSock(Globals glob, Socket fd, string cliAddr)
		: base(glob, fd, cliAddr)
		{
		}


		// We are no longer required
		public override void destroyed()
		{
			m_glob.DPRINTV("TCPSock::~TCPSock()");

			lock(this) {
				try {
					// Close socket descriptor
					close();
				}
				catch(SockError e) {
					m_glob.log().info(e.Message);
				}
			}

			// Base class
			base.destroyed();
		}


		/********************/
		/* Public Interface */
		/********************/

		// Wait for a new incoming TCP connection
		public override void newConnection(Tunneler tunneler)
		{
			lock(this) {
				// Socket ready to accept
				Sock connectedSock = accept();

				try {
					// Tell tunneler
					if(tunneler != null)
						tunneler.newConnectionArrived(connectedSock);
					else
						throw new SockError("No tunneler to handle new TCP connection", 0);
				}
				catch(TunnelError e) {
					m_glob.log().info(e.Message);
					connectedSock.destroyed();
				}
			}
		}


		// Set socket listening
		public override void listen()
		{
			lock(this) {
				// Set listening
				try {
					m_fd.Listen(5);
				}
				catch(SocketException e) {
					throw new SockError("Failed to activate listening TCP socket: " + e.Message,
						e.SocketErrorCode);
				}
				catch(Exception e) {
					throw new SockError("Failed to activate listening TCP socket: " + e.Message, 0);
				}
			}
		}


		// Wait for a new incoming connection
		public override Sock accept()
		{
			Socket connfd = null;
			string clientAddr = "";

			lock(this) {
				try {
					// Accept connection
					connfd = m_fd.Accept();
					if(connfd.Connected) {
						// Find client address
						IPEndPoint clientEndPoint = (IPEndPoint)connfd.RemoteEndPoint;
						clientAddr = clientEndPoint.Address.ToString();
						m_glob.DPRINTI("TCPSock::accept() - Got new connection from " + clientAddr);
					}
					else {
						throw new SockError("Accept returned non-connected socket", 0);
					}
				}
				catch(SocketException e) {
					throw new SockError("Failed to accept TCP connection: " + e.Message,
						e.SocketErrorCode);
				}
				catch(Exception e) {
					throw new SockError("Failed to accept TCP connection: " + e.Message, 0);
				}
			}

			// Create and return new TCPSock to handle this connected socket
			return new TCPSock(m_glob, connfd, clientAddr);
		}


		// Establish connection out
		public override void connect(IPAddr ipaddr)
		{
			// Create underlying socket
			createSockFD(ipaddr, SocketType.Stream, ProtocolType.Tcp);

			// Get address details
			string addr = ipaddr.getHost();
			uint port = ipaddr.getPort();

			lock(this) {
				// Lookup destination address
				IPAddress destaddr = null;
				if(!Utils.getIPByName(addr, ref destaddr)) {
					throw new ConnectError("Failed to resolve TCP server address " + addr);
				}

				// Try the connect
				try {
					IPEndPoint ipEndPoint = new IPEndPoint(destaddr, (int)port);
					m_fd.Connect(ipEndPoint);
				}
				catch(SocketException e) {
					StringBuilder err = new StringBuilder();
					err.AppendFormat("Failed to connect to TCP server {0} : {1}", ipaddr, e.Message);
					throw new SockError(err.ToString(), e.SocketErrorCode);
				}
				catch(Exception e) {
					StringBuilder err = new StringBuilder();
					err.AppendFormat("Failed to connect to TCP server {0} : {1}", ipaddr, e.Message);
					throw new SockError(err.ToString(), 0);
				}
			}
		}


		// Read from socket
		public override int read(byte[] buff)
		{
			int noread = 0;

			lock(this) {
				// Try the read
				try {
					noread = m_fd.Receive(buff);
				}
				catch(SocketException e) {
					throw new SockError("Read error from TCP socket: " + e.Message,
						e.SocketErrorCode);
				}
				catch(Exception e) {
					throw new SockError("Read error from TCP socket: " + e.Message, 0);
				}
			}

			// Check for EOF
			if(noread == 0) {
				throw new ConnectionClosed("Other end closed socket");
			}

			return noread;
		}


		// Write to socket
		public override int write(RawBuff buff)
		{
			int nowritten = 0;

			lock(this) {
				// Try the write
				try {
					nowritten = m_fd.Send(buff.getData());
				}
				catch(SocketException e) {
					throw new SockError("Write error to TCP socket: " + e.Message,
						e.SocketErrorCode);
				}
				catch(Exception e) {
					throw new SockError("Write error to TCP socket: " + e.Message, 0);
				}
			}

			return nowritten;
		}


		// Get the result of a non-blocking connect attempt
		public override void getConnectResult(ref bool resultAvailable, ref bool connected)
		{
			resultAvailable = false;
			connected = false;

			lock(this) {
				// Poll for connect result
				bool ready = false;
				bool pollError = false;
				ready = pollWrite(0, ref pollError);

				// Get connect result
				resultAvailable = (ready || pollError);
				if(resultAvailable) {
					connected = (ready && !pollError);
				}
			}
		}


		// Bind socket to an address
		public override void bind(IPAddr ipaddr)
		{
			// Create underlying socket
			createSockFD(ipaddr, SocketType.Stream, ProtocolType.Tcp);

			// Do the bind
			base.bind(ipaddr);
		}

	}

}	// namespace

