/* (C) 2007 Richard Mills / The Linux Academy Ltd.
 *
 * Client side Tunnel for testing.
 *
 * $Id: TestClientTunnel.cs,v 1.2 2008/08/07 20:30:29 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Text;

	public class TestClientTunnel : ClientTunnel, TestTunnel {

		// Construction
		public TestClientTunnel(Globals glob, TunnelConfig tcfg, ClientTunneler tunneler)
		: base(glob, tcfg, tunneler)
		{
			m_glob.DPRINTI("TestClientTunnel::TestClientTunnel()");

			m_testType = ClientTestType.CTEST_IDLE;
			m_goodURLPayloadSize = 0;
			m_goodBodyPayloadSize = 0;
			m_currURLPayloadSize = 0;
			m_currBodyPayloadSize = 0;

			// Setup initial test parameters
			ClientTunnelParms tunParms = tunneler.getTunParms();
			m_tunnelTestState = tunParms.getTestState();
			if(m_tunnelTestState == TunnelTestState.BEGIN_TESTING) {
				m_currURLPayloadSize = MIN_TEST_URL_PAYLOAD;
				m_currBodyPayloadSize = MIN_TEST_BODY_PAYLOAD;
			}
			else {
				m_currURLPayloadSize = tunParms.getURLPayloadSize();
				m_currBodyPayloadSize = tunParms.getBodyPayloadSize();
				m_goodURLPayloadSize = m_currURLPayloadSize;
				m_goodBodyPayloadSize = m_currBodyPayloadSize;
			}

			#if DEBUG_TUN
			StringBuilder dbg = new StringBuilder();
			dbg.AppendFormat("TestClientTunnel created in state: {0}", getTestState());
			dbg.AppendFormat(", URL payload size: {0}", m_currURLPayloadSize);
			dbg.AppendFormat(", Body payload size: {0}", m_currBodyPayloadSize);
			m_glob.DPRINTV(dbg.ToString());
			#endif
		}

		// Destruction
		public override void destroyed()
		{
			#if DEBUG_TUN
			StringBuilder str = new StringBuilder();
			str.AppendFormat("TestClientTunnel::~TestClientTunnel() - {0}", m_tid);
			m_glob.DPRINTI(str.ToString());
			#endif

			// Base class
			base.destroyed();
		}


		/*************/
		/* Accessors */
		/*************/

		// Get default tunneling parameters
		public static ClientTunnelParms getDefaultTunParms()
		{
			return new ClientTunnelParms(TunnelTestState.BEGIN_TESTING, DEFAULT_URL_PAYLOAD,
				DEFAULT_BODY_PAYLOAD);
		}


		// Get whether we are a test tunnel
		public override bool isTestTunnel()
		{
			return true;
		}


		// Get whether we have completed these tests
		public bool testsCompleted()
		{
			bool completed = false;

			lock(this) {
				// Only completed once both we and the server side have finished testing
				if(getTestState() == TunnelTestState.COMPLETED_TESTING) {
					completed = true;
				}
			}

			return completed;
		}


		/************/
		/* Mutators */
		/************/

		// Record connection error
		public override void recordConnectionError(TunnelErrorType errType)
		{
			lock(this) {
				// Base class
				base.recordConnectionError(errType);

				// Does this error indicate a problem with our testing?
				switch(errType) {
					case(TunnelErrorType.BAD_RESPONSE_ERROR): {
						// Assume our tests broke something
						nextTest(false);
						break;
					}
					case(TunnelErrorType.GENERAL_COMMS_ERROR): {
						// Are we in the process of testing?
						if(getTestState() == TunnelTestState.CLIENT_TESTING) {
							// Assume our tests broke something
							nextTest(false);
						}
						break;
					}
					case(TunnelErrorType.UNPROCESSED_DATA_ERROR): {
						// Assume our tests broke something
						nextTest(false);
						break;
					}
					default: {
						break;
					}
				}
			}
		}


		/******************************/
		/* Tunnel Protected Interface */
		/******************************/


		/***********/
		/* Methods */
		/***********/

		// Create payload for tunneled msg to send to the destination - Caller holds mutex
		protected override void makeTunnelPayload(ref RawBuff data, ref bool split, uint max)
		{
			uint payloadLen = 0;

			// Are we currently testing?
			if(getTestState() == TunnelTestState.CLIENT_TESTING && m_testType != ClientTestType.CTEST_IDLE) {
				// What are we currently testing?
				if(m_testType == ClientTestType.CTEST_URLMAXSIZE) {
					payloadLen = m_currURLPayloadSize;
				}
				else if(m_testType == ClientTestType.CTEST_BODYMAXSIZE) {
					payloadLen = m_currBodyPayloadSize;
				}

				// Create some payload data of the required length
				if(payloadLen > 0) {
					data = new RawBuff(payloadLen);
				}
			}
		}


		// Check message parms before resending - Caller holds mutex
		protected override void checkResendMsgParms(HTTPMsgParms parms)
		{
			// We don't want to resend any potentially problematic payload
			Payload payload = parms.getPayload();
			payload.getPayload().clear();
		}


		// Advance our internal state machine - Caller holds mutex
		protected override void advanceState(HTTPMsgParms parms)
		{
			// Let base class update our state
			base.advanceState(parms);

			// Are we waiting to start testing?
			if(getTestState() == TunnelTestState.BEGIN_TESTING) {
				// Once we become CONNECTED we can begin testing
				if(getState() == TunnelState.CONNECTED) {
					// We're now fully up and ready to begin testing
					setTestState(TunnelTestState.CLIENT_TESTING);

					// What type of testing should we do?
					if(!m_tunCfg.getUseHeaders()) {
						// Need to test URL payload size
						m_glob.DPRINTV("TestClientTunnel now testing max URL payload size");
						m_testType = ClientTestType.CTEST_URLMAXSIZE;
					}
					else {
						// Need to test msg body payload size
						m_glob.DPRINTV("TestClientTunnel now testing max body payload size");
						m_testType = ClientTestType.CTEST_BODYMAXSIZE;
					}
				}
			}
			else {
				// Does server report later testing state than us?
				TunnelTestState svrState = parms.getTestState();
				if(svrState > getTestState()) {
					// Track server testing state
					setTestState(svrState);

					// Testing completed?
					if(getTestState() == TunnelTestState.COMPLETED_TESTING) {
						m_glob.DPRINTV("Both tunnel ends have completed testing");

						// Send final results to controlling tunneler
						sendTestResults();

						// We've finished with this test tunnel for now
						m_tunneler.scheduleEvent(new ShutdownTunnelEvent(m_glob, new Time(0,0),
							m_tunneler, m_tid, false));
					}
				}
			}
		}


		// A sent message has been acknowledged - Caller holds mutex
		protected override void msgAcked(SentMsgParms parms)
		{
			// Base class
			base.msgAcked(parms);

			// Are we currently testing?
			if(getTestState() == TunnelTestState.CLIENT_TESTING) {
				// Get details for acked msg
				HTTPMsgParms msgParms = parms.m_parms;
				uint payloadLen = msgParms.getPayload().getPayloadSize();
				if(payloadLen > 0) {
					// What type of testing are we doing?
					if(m_testType == ClientTestType.CTEST_URLMAXSIZE) {
						// Record new good baseline
						m_goodURLPayloadSize = payloadLen;
					}
					else if(m_testType == ClientTestType.CTEST_BODYMAXSIZE) {
						// Record new good baseline
						m_goodBodyPayloadSize = payloadLen;
					}
				}

				// Consider moving tests on
				nextTest(true);
			}
		}


		// Record we need to resend a msg - Caller holds mutex
		protected override void needToResendMsg(SeqNo seqNo, bool msgLost)
		{
			// Base class
			base.needToResendMsg(seqNo, msgLost);

			// Is this a lost msg?
			if(msgLost) {
				// Are we currently testing?
				if(getTestState() == TunnelTestState.CLIENT_TESTING) {
					#if DEBUG_TUN
					StringBuilder nfo = new StringBuilder();
					nfo.AppendFormat("TestClientTunnel::needToResendMsg() - Resend of seq no {0}", seqNo);
					nfo.Append(" indicates test failed");
					m_glob.DPRINTI(nfo.ToString());
					#endif

					// Assume our tests broke something
					nextTest(false);
				}
			}
		}


		/************************/
		/* TestTunnel Interface */
		/************************/


		/***********/
		/* Methods */
		/***********/

		// Move testing on - Caller holds mutex
		public void nextTest(bool ok)
		{
			// Are we actually currently testing?
			if(getTestState() == TunnelTestState.CLIENT_TESTING) {
				#if DEBUG_TUN
				StringBuilder nfo = new StringBuilder();
				nfo.Append("TestClientTunnel::nextTest() - Test ");
				if(ok) nfo.Append("passed");
				else nfo.Append("failed");
				nfo.AppendFormat(", URL payload size: {0}", m_currURLPayloadSize);
				nfo.AppendFormat(", Body payload size: {0}", m_currBodyPayloadSize);
				m_glob.DPRINTV(nfo.ToString());
				#endif

				// What are we currently testing?
				if(m_testType == ClientTestType.CTEST_URLMAXSIZE) {
					// Did test pass?
					if(ok) {
						// Do we want to try another test?
						if(m_currURLPayloadSize < MAX_TEST_URL_PAYLOAD) {
							// Increase size of tested URL payload
							m_currURLPayloadSize = m_currURLPayloadSize + MIN_TEST_URL_PAYLOAD;
						}
						else {
							// We've finished this test, move state on
							m_glob.DPRINTV("TestClientTunnel finished our testing");
							setTestState(TunnelTestState.SERVER_TESTING);
						}
					}
					else {
						// Test failed; Use last known good value and move on
						m_glob.DPRINTV("TestClientTunnel finished our testing");
						setTestState(TunnelTestState.SERVER_TESTING);
					}
				}

				else if(m_testType == ClientTestType.CTEST_BODYMAXSIZE) {
					// Did test pass?
					if(ok) {
						// Do we want to try another test?
						if(m_currBodyPayloadSize < MAX_TEST_BODY_PAYLOAD) {
							// Increase size of tested body payload
							m_currBodyPayloadSize = m_currBodyPayloadSize + MIN_TEST_BODY_PAYLOAD;
						}
						else {
							// We've finished this test, move state on
							m_glob.DPRINTV("TestClientTunnel finished our testing");
							setTestState(TunnelTestState.SERVER_TESTING);
						}
					}
					else {
						// Test failed; Use last known good value and move on
						m_glob.DPRINTV("TestClientTunnel finished our testing");
						setTestState(TunnelTestState.SERVER_TESTING);
					}
				}

				// Send latest results to controlling tunneler
				sendTestResults();
			}
		}


		// Inform controlling tunneler of our test results - Caller holds mutex
		public void sendTestResults()
		{
			// Collect results and pass onto tunneler
			ClientTunnelParms parms = new ClientTunnelParms(getTestState(),
				m_goodURLPayloadSize, m_goodBodyPayloadSize);
			m_tunneler.scheduleEvent(new NewTunnelParmsEvent(m_glob, m_tunneler, parms,
				m_tid.getTunnelerID()));
		}


		/*********************/
		/* Private Interface */
		/*********************/


		/*********/
		/* Types */
		/*********/

		// Test states
		private enum ClientTestType { CTEST_IDLE, CTEST_URLMAXSIZE, CTEST_BODYMAXSIZE };


		/*************/
		/* Constants */
		/*************/

		// Maximum/minimum payload length for created HTTP messages
		private const uint			MIN_TEST_BODY_PAYLOAD		= 512;
		private const uint			MAX_TEST_BODY_PAYLOAD		= MIN_TEST_BODY_PAYLOAD * 32;
		private const uint			MIN_TEST_URL_PAYLOAD			= 128;
		private const uint			MAX_TEST_URL_PAYLOAD			= MIN_TEST_URL_PAYLOAD * 32;

		// Default payload length for created HTTP messages
		private const uint			DEFAULT_BODY_PAYLOAD			= 8192;
		private const uint			DEFAULT_URL_PAYLOAD			= 1024;


		/********/
		/* Data */
		/********/

		// What type of testing are we doing?
		private ClientTestType				m_testType;

		// Last known good test parameters
		private uint							m_goodURLPayloadSize;
		private uint							m_goodBodyPayloadSize;

		// Current test parameters
		private uint							m_currURLPayloadSize;
		private uint							m_currBodyPayloadSize;


	}

}	// namespace

