/* (C) 2007 Richard Mills / The Linux Academy Ltd.
 *
 * Server side Tunnel for testing.
 *
 * $Id: TestServerTunnel.cs,v 1.3 2013/03/31 12:53:55 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Text;

	public class TestServerTunnel : ServerTunnel, TestTunnel {

		/*************/
		/* Constants */
		/*************/

		// Minimum size of maximum payload
		public const uint		MIN_MAX_PAYLOAD_SIZE				= MIN_TEST_PAYLOAD;


		/***********/
		/* Methods */
		/***********/

		// Construction
		public TestServerTunnel(Globals glob, TunnelConfig tcfg, ServerTunneler tunneler)
		: base(glob, tcfg, tunneler)
		{
			m_glob.DPRINTI("TestServerTunnel::TestServerTunnel()");

			m_testType = ServerTestType.STEST_IDLE;
			m_goodPayloadSize = 0;
			m_currPayloadSize = 0;

			// Setup initial test parameters
			ServerTunnelParms tunParms = tunneler.getTunParms(m_tid.getTunnelerID());
			m_tunnelTestState = tunParms.getTestState();
			if(m_tunnelTestState == TunnelTestState.BEGIN_TESTING) {
				m_currPayloadSize = MIN_TEST_PAYLOAD;
			}
			else {
				m_currPayloadSize = tunParms.getMaxPayloadSize(true);
				m_goodPayloadSize = m_currPayloadSize;
			}

			#if DEBUG_TUN
			StringBuilder dbg = new StringBuilder();
			dbg.AppendFormat("TestServerTunnel created in state: {0}", getTestState());
			dbg.AppendFormat(", Payload size: {0}", m_currPayloadSize);
			m_glob.DPRINTV(dbg.ToString());
			#endif
		}

		// Destruction
		public override void destroyed()
		{
			#if DEBUG_TUN
			StringBuilder str = new StringBuilder();
			str.AppendFormat("TestServerTunnel::~TestServerTunnel() - {0}", m_tid);
			m_glob.DPRINTI(str.ToString());
			#endif

			// Base class
			base.destroyed();
		}


		/*************/
		/* Accessors */
		/*************/

		// Get default tunneling parameters
		public static ServerTunnelParms getDefaultTunParms()
		{
			return new ServerTunnelParms(TunnelTestState.BEGIN_TESTING, DEFAULT_PAYLOAD);
		}


		// Get whether we are a test tunnel
		public override bool isTestTunnel()
		{
			return true;
		}


		/************/
		/* Mutators */
		/************/

		// Record connection error
		public override void recordConnectionError(TunnelErrorType errType)
		{
			lock(this) {
				// Base class
				base.recordConnectionError(errType);

				// Does this error indicate a problem with our testing?
				switch(errType) {
					case(TunnelErrorType.GENERAL_COMMS_ERROR): {
						// Are we in the process of testing?
						if(getTestState() == TunnelTestState.SERVER_TESTING) {
							// Assume our tests broke something
							nextTest(false);
						}
						break;
					}
					default: {
						break;
					}
				}
			}
		}


		/******************************/
		/* Tunnel Protected Interface */
		/******************************/


		/***********/
		/* Methods */
		/***********/

		// Create payload for tunneled msg to send to the destination - Caller holds mutex
		protected override void makeTunnelPayload(ref RawBuff data, ref bool split, uint max)
		{
			uint payloadLen = 0;

			// Are we currently testing?
			if(getTestState() == TunnelTestState.SERVER_TESTING) {
				// What are we currently testing?
				if(m_testType == ServerTestType.STEST_PAYLOADMAXSIZE) {
					payloadLen = m_currPayloadSize;
				}

				// Create some payload data of the required length
				if(payloadLen > 0) {
					data = new RawBuff(payloadLen);
				}
			}
		}


		// Check message parms before resending - Caller holds mutex
		protected override void checkResendMsgParms(HTTPMsgParms parms)
		{
			// We don't want to resend any potentially problematic payload
			Payload payload = parms.getPayload();
			payload.getPayload().clear();
		}


		// Advance our internal state machine - Caller holds mutex
		protected override void advanceState(HTTPMsgParms parms)
		{
			// Let base class update our state
			base.advanceState(parms);

			// Does client report later testing state than us?
			TunnelTestState cliState = parms.getTestState();
			if(cliState > getTestState()) {
				// Track client testing state
				setTestState(cliState);

				// Has client just told us it's our turn to start testing?
				if(cliState == TunnelTestState.SERVER_TESTING) {
					// What type of testing should we do?
					m_glob.DPRINTV("TestServerTunnel now testing max payload size");
					m_testType = ServerTestType.STEST_PAYLOADMAXSIZE;
				}
			}
		}


		// A sent message has been acknowledged - Caller holds mutex
		protected override void msgAcked(SentMsgParms parms)
		{
			// Base class
			base.msgAcked(parms);

			// Are we currently testing?
			if(getTestState() == TunnelTestState.SERVER_TESTING) {
				// Get details for acked msg
				HTTPMsgParms msgParms = parms.m_parms;
				uint payloadLen = msgParms.getPayload().getPayloadSize();
				if(payloadLen > 0) {
					// What type of testing are we doing?
					if(m_testType == ServerTestType.STEST_PAYLOADMAXSIZE) {
						// Record new good baseline
						m_goodPayloadSize = payloadLen;
					}
				}

				// Consider moving tests on
				nextTest(true);
			}
		}


		// Record we need to resend a msg - Caller holds mutex
		protected override void needToResendMsg(SeqNo seqNo, bool msgLost)
		{
			// Base class
			base.needToResendMsg(seqNo, msgLost);

			// Is this a lost msg?
			if(msgLost) {
				// Are we currently testing?
				if(getTestState() == TunnelTestState.SERVER_TESTING) {
					#if DEBUG_TUN
					StringBuilder nfo = new StringBuilder();
					nfo.AppendFormat("TestServerTunnel::needToResendMsg() - Resend of seq no {0}", seqNo);
					nfo.Append(" indicates test failed");
					m_glob.DPRINTI(nfo.ToString());
					#endif

					// Assume our tests broke something
					nextTest(false);
				}
			}
		}


		/************************/
		/* TestTunnel Interface */
		/************************/


		/***********/
		/* Methods */
		/***********/

		// Move testing on - Caller holds mutex
		public void nextTest(bool ok)
		{
			// Are we actually currently testing?
			if(getTestState() == TunnelTestState.SERVER_TESTING) {
				#if DEBUG_TUN
				StringBuilder nfo = new StringBuilder();
				nfo.Append("TestServerTunnel::nextTest() - Test ");
				if(ok) nfo.Append("passed");
				else nfo.Append("failed");
				nfo.AppendFormat(", Payload size: {0}", m_currPayloadSize);
				m_glob.DPRINTV(nfo.ToString());
				#endif

				// What are we currently testing?
				if(m_testType == ServerTestType.STEST_PAYLOADMAXSIZE) {
					// Did test pass?
					if(ok) {
						// Do we want to try another test?
						if(m_currPayloadSize < MAX_TEST_PAYLOAD) {
							// Increase size of tested payload
							m_currPayloadSize = m_currPayloadSize + MIN_TEST_PAYLOAD;
						}
						else {
							// We've finished this test, move state on
							m_glob.DPRINTV("TestServerTunnel finished our testing");
							setTestState(TunnelTestState.COMPLETED_TESTING);
						}
					}
					else {
						// Test failed; Use last known good value and move on
						m_glob.DPRINTV("TestServerTunnel finished our testing");
						setTestState(TunnelTestState.COMPLETED_TESTING);
					}
				}

				// Send latest results to controlling tunneler
				sendTestResults();
			}
		}


		// Inform controlling tunneler of our test results - Caller holds mutex
		public void sendTestResults()
		{
			// Collect results and pass onto tunneler
			ServerTunnelParms parms = new ServerTunnelParms(getTestState(), m_goodPayloadSize);
			m_tunneler.scheduleEvent(new NewTunnelParmsEvent(m_glob, m_tunneler, parms,
				m_tid.getTunnelerID()));
		}


		/*********************/
		/* Private Interface */
		/*********************/


		/*********/
		/* Types */
		/*********/

		// Test states
		private enum ServerTestType { STEST_IDLE, STEST_PAYLOADMAXSIZE };


		/*************/
		/* Constants */
		/*************/

		// Maximum/minimum payload length for created HTTP messages
		private const uint			MIN_TEST_PAYLOAD				= 1024;
		private const uint			MAX_TEST_PAYLOAD				= MIN_TEST_PAYLOAD * 32;

		// Default payload length for created HTTP messages
		private const uint			DEFAULT_PAYLOAD				= 16384;


		/********/
		/* Data */
		/********/

		// What type of testing are we doing?
		private ServerTestType				m_testType;

		// Last known good test parameters
		private uint							m_goodPayloadSize;

		// Current test parameters
		private uint							m_currPayloadSize;


	}

}	// namespace

