/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Represent a time in seconds & milli-seconds.
 *
 * $Id: Time.cs,v 1.2 2013/03/24 09:55:58 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;

	public class Time : IComparable {

		// Construction
		public Time(int secs, int millis)
		{
			// Convert relative time to absolute time
			TimeSpan relative = new TimeSpan(0, 0, 0, secs, millis);
			m_timeVal = DateTime.Now + relative;
		}

		public Time(Time that)
		{
			m_timeVal = that.m_timeVal;
		}


		/*************************/
		/* IComparable Interface */
		/*************************/

		// Compare based on time
		public int CompareTo(Object that)
		{
			Time tm = (Time)that;
			return (m_timeVal.CompareTo(tm.m_timeVal));
		}


		/********************/
		/* Public Interface */
		/********************/

		// Create a representation of ourselves as a string
		public override string ToString()
		{
			string val = "Ticks: " + m_timeVal.Ticks;
			return val;
		}

		// Equality/inequality
		public static bool operator==(Time t1, Time t2)
		{
			return (t1.Equals(t2));
		}
		public static bool operator!=(Time t1, Time t2)
		{
			return !(t1.Equals(t2));
		}

		// As recommended, also override Equals() & GetHashCode()
		public override bool Equals(object that)
		{
			bool res = false;
			if(that != null && that is Time) {
				Time tm = (Time)that;
				res = (m_timeVal == tm.m_timeVal);
			}

			return res;
		}
		public override int GetHashCode()
		{
			return ToString().GetHashCode();
		}


		// Subtraction
		public static TimeSpan operator-(Time t1, Time t2)
		{
			return (t1.m_timeVal - t2.m_timeVal);
		}


		/*************/
		/* Accessors */
		/*************/

		// Get current time
		public static Time now()
		{
			return new Time(0,0);
		}

		// Get raw internal time val
		public DateTime getTimeVal()
		{
			return m_timeVal;
		}

		// Get seconds component
		public int getSecs()
		{
			return (m_timeVal.Second);
		}

		// Get milli-seconds component
		public int getMillis()
		{
			return (m_timeVal.Millisecond);
		}

		// Get total milli-seconds
		public int getTotalMillis()
		{
			TimeSpan ts = m_timeVal.TimeOfDay;
			int tot = (int)ts.TotalMilliseconds;
			return tot;
		}


		/************/
		/* Mutators */
		/************/

		// Subtract some fixed interval
		public void subInterval(int secs, int millis)
		{
			// Create appropriate TimeSpan
			TimeSpan subts = new TimeSpan(0, 0, 0, secs, millis);

			// Subtract TimeSpan
			m_timeVal = m_timeVal - subts;
		}

		// Add some fixed interval
		public void addInterval(int secs, int millis)
		{
			// Create appropriate TimeSpan
			TimeSpan addts = new TimeSpan(0, 0, 0, secs, millis);

			// Add TimeSpan
			m_timeVal = m_timeVal + addts;
		}
	

		/*************/
		/* Utilities */
		/*************/

		// Has this time elapsed?
		public bool elapsed()
		{
			return (getMillisUntilElapsed() == 0);
		}

		// Get milli-seconds (rounded up) until this time has elapsed
		public int getMillisUntilElapsed()
		{
			DateTime now = DateTime.Now;
			if(m_timeVal > now) {
				TimeSpan ts = m_timeVal - now;
				return (int)Math.Ceiling(ts.TotalMilliseconds);
			}
			else {
				return 0;
			}
		}


		/*********************/
		/* Private Interface */
		/*********************/

		/********/
		/* Data */
		/********/

		// Time we represent
		private DateTime					m_timeVal;

	}

}	// namespace

