/* (C) 2008 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate tunable parameters for a tunnel.
 *
 * $Id: TuneSettings.cs,v 1.2 2009/02/27 11:55:10 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Text;

	public class TuneSettings {

		/*************/
		/* Constants */
		/*************/

		// Absolute maximum no. of oustanding unacked msgs
		public const uint			MAX_UNACKED_MSGS			= 5;


		/***********/
		/* Methods */
		/***********/

		// Construction
		public TuneSettings()
		{
			m_tune = "";
			m_tunedMaxUnackedMsgs = 0;
			m_tunedFastPollTimeout = 0;
			m_tunedMediumPollTimeout = 0;
			m_tunedSlowPollTimeout = 0;
			m_tunedMaxPollTimeout = 0;

			// Setup defaults
			setup();
		}

		public TuneSettings(string tune)
		{
			m_tune = tune;
			m_tunedMaxUnackedMsgs = 0;
			m_tunedFastPollTimeout = 0;
			m_tunedMediumPollTimeout = 0;
			m_tunedSlowPollTimeout = 0;
			m_tunedMaxPollTimeout = 0;

			// Setup defaults
			setup();
		}


		// Create a representation of ourselves as a string
		public override string ToString()
		{
			StringBuilder str = new StringBuilder();
			str.AppendFormat("{0}", getTuneDesc());
			return str.ToString();
		}


		/*************/
		/* Accessors */
		/*************/

		// Get tuning description
		public string getTuneDesc()
		{
			return m_tune;
		}

		// Get maximum no. of unacked msgs
		public uint getMaxUnackedMsgs()
		{
			return m_tunedMaxUnackedMsgs;
		}

		// Get fast poll timeout
		public uint getFastPollTimeout()
		{
			return m_tunedFastPollTimeout;
		}

		// Get medium poll timeout
		public uint getMediumPollTimeout()
		{
			return m_tunedMediumPollTimeout;
		}

		// Get slow poll timeout
		public uint getSlowPollTimeout()
		{
			return m_tunedSlowPollTimeout;
		}

		// Get maximum poll timeout
		public uint getMaxPollTimeout()
		{
			return m_tunedMaxPollTimeout;
		}


		/*********************/
		/* Private Interface */
		/*********************/


		/*************/
		/* Constants */
		/*************/

		// Absolute maximum poll timeout (milli-seconds)
		private const uint		MAX_POLL_TIMEOUT					= 2500;

		// Maximum numbers of outstanding unacked msgs
		private const uint		DFLT_MAX_UNACKED_MSGS			= 1;
		private const uint		BRST_MAX_UNACKED_MSGS			= 1;
		private const uint		RLTM_MAX_UNACKED_MSGS			= 1;

		// Time between poll timeout events (milli-seconds)
		private const uint		DFLT_FAST_POLL_TIMEOUT			= 200;
		private const uint		DFLT_MEDIUM_POLL_TIMEOUT		= 250;
		private const uint		DFLT_SLOW_POLL_TIMEOUT			= 500;
		private const uint		DFLT_MAX_POLL_TIMEOUT			= MAX_POLL_TIMEOUT;

		private const uint		BRST_FAST_POLL_TIMEOUT			= 250;
		private const uint		BRST_MEDIUM_POLL_TIMEOUT		= 500;
		private const uint		BRST_SLOW_POLL_TIMEOUT			= 2000;
		private const uint		BRST_MAX_POLL_TIMEOUT			= MAX_POLL_TIMEOUT;

		private const uint		RLTM_FAST_POLL_TIMEOUT			= 50;
		private const uint		RLTM_MEDIUM_POLL_TIMEOUT		= 100;
		private const uint		RLTM_SLOW_POLL_TIMEOUT			= 200;
		private const uint		RLTM_MAX_POLL_TIMEOUT			= 500;


		/***********/
		/* Methods */
		/***********/

		// Set tuning settings based on the tuning description
		private void setup()
		{
			// What tuning is required?
			if(m_tune == "bursty") {
				// Bursty tuning
				m_tunedMaxUnackedMsgs = BRST_MAX_UNACKED_MSGS;
				m_tunedFastPollTimeout = BRST_FAST_POLL_TIMEOUT;
				m_tunedMediumPollTimeout = BRST_MEDIUM_POLL_TIMEOUT;
				m_tunedSlowPollTimeout = BRST_SLOW_POLL_TIMEOUT;
				m_tunedMaxPollTimeout = BRST_MAX_POLL_TIMEOUT;
			}
			else if(m_tune == "realtime") {
				// Realtime tuning
				m_tunedMaxUnackedMsgs = RLTM_MAX_UNACKED_MSGS;
				m_tunedFastPollTimeout = RLTM_FAST_POLL_TIMEOUT;
				m_tunedMediumPollTimeout = RLTM_MEDIUM_POLL_TIMEOUT;
				m_tunedSlowPollTimeout = RLTM_SLOW_POLL_TIMEOUT;
				m_tunedMaxPollTimeout = RLTM_MAX_POLL_TIMEOUT;
			}
			else {
				// Default tuning
				m_tunedMaxUnackedMsgs = DFLT_MAX_UNACKED_MSGS;
				m_tunedFastPollTimeout = DFLT_FAST_POLL_TIMEOUT;
				m_tunedMediumPollTimeout = DFLT_MEDIUM_POLL_TIMEOUT;
				m_tunedSlowPollTimeout = DFLT_SLOW_POLL_TIMEOUT;
				m_tunedMaxPollTimeout = DFLT_MAX_POLL_TIMEOUT;
			}
		}


		/********/
		/* Data */
		/********/

		// Tuning description
		private string						m_tune;

		// Maximum no. of unacked msgs
		private uint						m_tunedMaxUnackedMsgs;

		// Poll timeouts
		private uint						m_tunedFastPollTimeout;
		private uint						m_tunedMediumPollTimeout;
		private uint						m_tunedSlowPollTimeout;
		private uint						m_tunedMaxPollTimeout;

	}

}	// namespace

