/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Somewhere to store all information about a unique Tunnel.
 *
 * $Id: Tunnel.cs,v 1.5 2009/05/08 13:44:37 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Collections;
	using System.Text;

	public abstract class Tunnel {

		/*********/
		/* Types */
		/*********/

		// Left and right
		public enum Direction { LEFT, RIGHT, UNKNOWNDIR }

		// Tunnel states
		public enum TunnelState {
			CONNECT, AUTHREQ, RECONNECT, CLOSING, CLOSED, CONNECTED, UNKNOWNSTATE
		}

		// Tunnel testing states
		public enum TunnelTestState {
			NO_TESTING, BEGIN_TESTING, CLIENT_TESTING, SERVER_TESTING, COMPLETED_TESTING
		}

		// Enumerate all communication error types
		public enum TunnelErrorType {
			GENERAL_COMMS_ERROR, PROXY_AUTH_ERROR, SERVER_AUTH_ERROR,
			BAD_RESPONSE_ERROR, MSG_RESEND_ERROR, MAX_RESEND_ERROR,
			CONNECTION_CLOSED_ERROR, UNVIABLE_ERROR, QUOTAS_EXCEEDED_ERROR,
			FATAL_QUOTAS_ERROR, SERVER_CONNECT_ERROR, CLIENT_AUTH_FAILED_ERROR,
			UNPROCESSED_DATA_ERROR
		}


		/***********/
		/* Methods */
		/***********/

		// Construction
		public Tunnel(Globals glob, TunnelConfig tcfg, Tunneler tunneler,
		Direction tunDir, TunnelState state)
		{
			m_glob = glob;
			m_tunneler = tunneler;
			m_leftConn = new ArrayList();
			m_rightConn = new ArrayList();
			m_tunCfg = tcfg;
			m_tid = m_tunCfg.getTunnelID();
			m_minAckTimeout = MIN_ACK_TIMEOUT;
			m_maxAckTimeout = MAX_ACK_TIMEOUT;
			m_ackTimeout = 0;
			m_ackRTTs = new ArrayList();
			m_avgAckRTT = 0;
			m_biasAvgAckRTT = 0;
			m_user = "";
			m_tunnelState = state;
			m_tunnelTestState = TunnelTestState.NO_TESTING;
			m_unackedMsgQueue = new ArrayList();
			m_lastTunnelAnySendTime = Time.now();
			m_tunnelDataQueue = new ArrayList();
			m_rawDataQueue = new ArrayList();
			m_rawDataQueueSize = 0;
			m_maxUnackedMsgs = 1;
			m_lastRecvAckSeqNo = new SeqNo(1);
			m_lastRecvAckResent = false;
			m_resendQueue = new ArrayList();
			m_oldConns = new ArrayList();
			m_nextSendSeqNo = new SeqNo(1);
			m_nextRecvSeqNo = new SeqNo(1);
			m_errorCount = 0;
			m_leftEverConnected = false;
			m_rightEverConnected = false;
			m_tunnelDirection = tunDir;
			m_shutdownScheduled = false;
			m_lastTunnelSendTime = DateTime.Now;
			m_sentCloseMsg = false;
			m_lastTunnelConnectedSendTime = Time.now();
			m_lastRawRecvTime = Time.now();
			m_lastTunRecvTime = Time.now();
			m_rawBytesIn = 0;
			m_rawBytesOut = 0;
			m_httpBytesIn = 0;
			m_httpBytesOut = 0;
			m_sendNotifMsgs = new ArrayList();
			m_recvNotifMsgs = new ArrayList();

			#if DEBUG_TUN
			StringBuilder nfo = new StringBuilder();
			nfo.AppendFormat("Tunnel::Tunnel() - Creating tunnel {0}", m_tid.ToString());
			m_glob.DPRINTI(nfo.ToString());
			#endif

			// Get tuned settings
			m_maxUnackedMsgs = m_tunCfg.getTuning().getMaxUnackedMsgs();

			// Set initial ack timeout as midway between min & max
			m_ackTimeout = m_minAckTimeout + ((m_maxAckTimeout - m_minAckTimeout) / 2);

			// Start with default username
			Authenticator authenticator = getAuthenticator();
			m_user = authenticator.getDefaultUserName();

			// What type of data are we tunneling?
			if(m_tid.getProtocol() == Enums.Protocol.UDP) {
				// Start running UDP silent timeout events
				m_tunneler.scheduleEvent(new UDPSilentEvent(m_glob, m_tunneler, m_tid));
			}
		}

		// Destruction
		public virtual void destroyed()
		{
			m_glob.DPRINTI("Tunnel::~Tunnel()");

			lock(this) {
				// Tidy up all remaining connections
				for(int i = 0; i < m_leftConn.Count; ++i) {
					// Tell connection it no longer carries us
					Connection conn = (Connection)m_leftConn[i];
					conn.removeCarriedTunnel(this);
				}
				for(int i = 0; i < m_rightConn.Count; ++i) {
					// Tell connection it no longer carries us
					Connection conn = (Connection)m_rightConn[i];
					conn.removeCarriedTunnel(this);
				}
			}
		}


		/*************/
		/* Selectors */
		/*************/

		// Base class
		public abstract class Selector {
			public Selector() {}
			public abstract void select(ArrayList sel, Tunnel tun);
		}

		// Select only left side connections
		public class SelLeftConn : Selector {
			public SelLeftConn() {}
			public override void select(ArrayList sel, Tunnel tun) {
				// Remove all right side connections
				bool keepGoing = false;
				do {
					keepGoing = false;
					for(int i = 0; i < sel.Count; ++i) {
						// Right side connection here?
						Connection conn = (Connection)sel[i];
						if(tun.rightConnection(conn)) {
							// Remove this connection, & restart scanning
							sel.RemoveAt(i);
							keepGoing = true;
							break;
						}
					}
				} while(keepGoing);
			}
		}

		// Select only right side connections
		public class SelRightConn : Selector {
			public SelRightConn() {}
			public override void select(ArrayList sel, Tunnel tun) {
				// Remove all left side connections
				bool keepGoing = false;
				do {
					keepGoing = false;
					for(int i = 0; i < sel.Count; ++i) {
						// Left side connection here?
						Connection conn = (Connection)sel[i];
						if(tun.leftConnection(conn)) {
							// Remove this connection, & restart scanning
							sel.RemoveAt(i);
							keepGoing = true;
							break;
						}
					}
				} while(keepGoing);
			}
		}

		// Sort according to most recently first connected connection
		public class SelRecentConn : Selector {
			public SelRecentConn() {}
			public override void select(ArrayList sel, Tunnel tun) {
				// Sort according to connection time, then reverse
				sel.Sort(new Connection.PredConnectTime());
				sel.Reverse();
			}
		}

		// Exclude the specified connection
		public class SelNotConn : Selector {
			public SelNotConn(Connection conn) { m_conn = conn; }
			public override void select(ArrayList sel, Tunnel tun) {
				// Remove occurances of the specified connection
				int index = -1;
				do {
					index = sel.IndexOf(m_conn);
					if(index >= 0)
						sel.RemoveAt(index);
				} while(index >= 0);
			}
			private Connection m_conn;
		}

		// Select only connected connections
		public class SelConnectedConn : Selector {
			public SelConnectedConn() {}
			public override void select(ArrayList sel, Tunnel tun) {
				// Remove all non-connected connections
				bool keepGoing = false;
				do {
					keepGoing = false;
					for(int i = 0; i < sel.Count; ++i) {
						// Non-connected connection here?
						Connection conn = (Connection)sel[i];
						if(!conn.connected()) {
							// Remove this connection, & restart scanning
							sel.RemoveAt(i);
							keepGoing = true;
							break;
						}
					}
				} while(keepGoing);
			}
		}


		/*************/
		/* Accessors */
		/*************/

		// Get our config
		public TunnelConfig getConfig()
		{
			return m_tunCfg;
		}

		// Get our proxy settings for modification
		public ProxySettings getProxySettings()
		{
			return m_tunCfg.getProxySettings();
		}

		// Get our serverauth settings for modification
		public ServerAuthSettings getServerAuthSettings()
		{
			return m_tunCfg.getServerAuthSettings();
		}


		// Get our tunnel ID
		public TunnelID getTid()
		{
			return m_tid;
		}

		// Get our tunnel ID as a string
		public string getTidStr()
		{
			return m_tid.ToString();
		}

		// Get our owning tunneler
		public Tunneler getTunneler()
		{
			return m_tunneler;
		}

		// Does the specified connection belong to this Tunnel?
		public bool ownsConnection(Connection conn)
		{
			lock(this) {
				// Need to check each managed list
				return (leftConnection(conn) || rightConnection(conn));
			}
		}

		// Is the specified connection from the left or right side?
		public bool leftConnection(Connection conn)
		{
			lock(this) {
				return m_leftConn.Contains(conn);
			}
		}
		public bool rightConnection(Connection conn)
		{
			lock(this) {
				return m_rightConn.Contains(conn);
			}
		}

		// Get the number of left/right side connections for this tunnel
		public int noLeftConnections()
		{
			lock(this) {
				return m_leftConn.Count;
			}
		}
		public int noRightConnections()
		{
			lock(this) {
				return m_rightConn.Count;
			}
		}


		// Get whether we've ever been connected in some direction
		public bool everConnected(Direction dir)
		{
			lock(this) {
				return ((dir == Direction.LEFT)? m_leftEverConnected : m_rightEverConnected);
			}
		}

		// Get whether we are closing
		public bool isClosing()
		{
			lock(this) {
				return (m_tunnelState == TunnelState.CLOSING || m_tunnelState == TunnelState.CLOSED);
			}
		}

		// Get whether we are closed
		public bool isClosed()
		{
			lock(this) {
				return (m_tunnelState == TunnelState.CLOSED);
			}
		}


		// Get the number of connections matching the provided criteria
		public int noConnFromSel(ArrayList crit)
		{
			lock(this) {
				// Get all matching connections
				ArrayList allConns = new ArrayList();
				getConnsFromSel(crit, allConns);

				// Return count
				return allConns.Count;
			}
		}


		// Get a single connection matching the provided criteria
		public Connection getConnFromSel(ArrayList crit)
		{
			// Get all matching connections
			ArrayList allConns = new ArrayList();
			getConnsFromSel(crit, allConns);

			// Resulting connection is the one at the head of the remaining list
			if(allConns.Count > 0)
				return (Connection)allConns[0];

			return null;
		}

		// Get all connections matching the provided criteria
		public void getConnsFromSel(ArrayList crit, ArrayList res)
		{
			// Build list of all connections
			res.Clear();
			lock(this) {
				for(int i = 0; i < m_leftConn.Count; ++i)
					res.Add(m_leftConn[i]);
				for(int i = 0; i < m_rightConn.Count; ++i)
					res.Add(m_rightConn[i]);
			}

			// Apply all selection criteria
			for(int s = 0; s < crit.Count; ++s) {
				Selector sel = (Selector)crit[s];
				sel.select(res, this);
			}
		}

		// Get name of owning user
		public string getUserName()
		{
			return m_user;
		}

		// Get owning user details
		public User getUserDetails()
		{
			lock(this) {
				// Lookup user details and get a copy
				Authenticator authenticator = getAuthenticator();
				return authenticator.getUserDetails(m_user);
			}
		}

		// Get source addresses of connections for this tunnel
		public virtual string getSrcConnectionAddresses()
		{
			// Only know in the server
			return "Unknown";
		}

		// Get destination addresses of connections for this tunnel
		public virtual string getDstConnectionAddresses()
		{
			// Only know in the server
			return "Unknown";
		}

		// Get our current state
		public TunnelState getState()
		{
			return m_tunnelState;
		}

		// Get current tunnel test state
		public TunnelTestState getTestState()
		{
			return m_tunnelTestState;
		}

		// Have we not received any raw data for a suspiciously long time?
		public bool rawRecvTimeout()
		{
			lock(this) {
				TimeSpan quiet = Time.now() - m_lastRawRecvTime;
				return (quiet.TotalSeconds >= MAX_NO_RECV_TIME);
			}
		}

		// Have we not received any tunnel data for a suspiciously long time?
		public bool tunRecvTimeout()
		{
			lock(this) {
				TimeSpan quiet = Time.now() - m_lastTunRecvTime;
				return (quiet.TotalSeconds >= MAX_NO_RECV_TIME);
			}
		}

		// Get whether we are a test tunnel
		public virtual bool isTestTunnel()
		{
			return false;
		}

		// Get raw bytes sent/received
		public uint getRawBytesIn()
		{
			return m_rawBytesIn;
		}
		public uint getRawBytesOut()
		{
			return m_rawBytesOut;
		}

		// Get HTTP bytes sent/received
		public uint getHTTPBytesIn()
		{
			return m_httpBytesIn;
		}
		public uint getHTTPBytesOut()
		{
			return m_httpBytesOut;
		}

		// Get notification msgs
		public ArrayList getSendNotifMsgs()
		{
			lock(this) {
				// Return copy and clear our list
				ArrayList notifs = (ArrayList)m_sendNotifMsgs.Clone();
				m_sendNotifMsgs.Clear();
				return notifs;
			}
		}
		public ArrayList getRecvNotifMsgs()
		{
			lock(this) {
				return (ArrayList)m_recvNotifMsgs.Clone();
			}
		}

		// Are we persistent?
		public virtual bool isPersistent()
		{
			// Assume we are
			return true;
		}


		/************/
		/* Mutators */
		/************/

		// Add connection to list of left side connections
		public void addLeftConnection(Connection conn)
		{
			lock(this) {
				// Already know about this connection?
				if(!m_leftConn.Contains(conn)) {
					// Don't re-add old removed connections
					if(!oldConnection(conn)) {
						m_leftConn.Add(conn);

						#if DEBUG_TUN
						StringBuilder dbg = new StringBuilder();
						dbg.AppendFormat("Tunnel::addLeftConnection() - connection {0}",
							conn.getID());
						m_glob.DPRINTV(dbg.ToString());
						#endif
					}
				}

				// Connected?
				m_leftEverConnected |= conn.everConnected();
			}

			// Also inform connection about us
			conn.addCarriedTunnel(this);
		}

		// Add connection to list of right side connections
		public void addRightConnection(Connection conn)
		{
			lock(this) {
				// Already know about this connection?
				if(!m_rightConn.Contains(conn)) {
					// Don't re-add old removed connections
					if(!oldConnection(conn)) {
						m_rightConn.Add(conn);

						#if DEBUG_TUN
						StringBuilder dbg = new StringBuilder();
						dbg.AppendFormat("Tunnel::addRightConnection() - connection {0}",
							conn.getID());
						m_glob.DPRINTV(dbg.ToString());
						#endif
					}
				}

				// Connected?
				m_rightEverConnected |= conn.everConnected();
			}

			// Also inform connection about us
			conn.addCarriedTunnel(this);
		}


		// Record connection error
		public virtual void recordConnectionError(TunnelErrorType errType)
		{
			// How much to increase error level by
			uint errLevelInc = 0;

			// Do we consider this to be an error from the other tunnel end?
			bool otherEndError = false;

			// What type of error?
			switch(errType) {
				case(TunnelErrorType.GENERAL_COMMS_ERROR): {
					errLevelInc = 5;
					break;
				}
				case(TunnelErrorType.PROXY_AUTH_ERROR):
				case(TunnelErrorType.SERVER_AUTH_ERROR): {
					// Fatal error
					errLevelInc = FATAL_ERROR_COUNT;
					otherEndError = true;
					break;
				}
				case(TunnelErrorType.BAD_RESPONSE_ERROR): {
					// Fatal error
					errLevelInc = FATAL_ERROR_COUNT;
					otherEndError = true;
					break;
				}
				case(TunnelErrorType.MSG_RESEND_ERROR): {
					// Not an error until we exceed maximum resend count
					break;
				}
				case(TunnelErrorType.MAX_RESEND_ERROR): {
					errLevelInc = FATAL_ERROR_COUNT;
					break;
				}
				case(TunnelErrorType.CONNECTION_CLOSED_ERROR): {
					// Not necessarily an error
					break;
				}
				case(TunnelErrorType.UNVIABLE_ERROR): {
					// Are we still managing to send data down the tunnel?
					TimeSpan connSendDiff = Time.now() - m_lastTunnelConnectedSendTime;
					TimeSpan anySendDiff = Time.now() - m_lastTunnelAnySendTime;
					if(connSendDiff.TotalSeconds >= UNVIABLE_ERROR_TIME) {
						errLevelInc = 2;
						if(anySendDiff.TotalSeconds >= UNVIABLE_ERROR_TIME)
							errLevelInc += 5;
					}
					break;
				}
				case(TunnelErrorType.QUOTAS_EXCEEDED_ERROR): {
					errLevelInc = 2;
					break;
				}
				case(TunnelErrorType.FATAL_QUOTAS_ERROR): {
					errLevelInc = FATAL_ERROR_COUNT;
					break;
				}
				case(TunnelErrorType.SERVER_CONNECT_ERROR): {
					errLevelInc = FATAL_ERROR_COUNT;
					break;
				}
				case(TunnelErrorType.CLIENT_AUTH_FAILED_ERROR): {
					errLevelInc = FATAL_ERROR_COUNT;
					break;
				}
				case(TunnelErrorType.UNPROCESSED_DATA_ERROR): {
					// Not necessarily an error
					break;
				}
				default: {
					m_glob.log().warn("Unknown tunnel error type");
					break;
				}
			}

			// Increase error count
			lock(this) {
				m_errorCount += errLevelInc;

				#if DEBUG_TUN
				StringBuilder dbg = new StringBuilder();
				dbg.AppendFormat("Tunnel::recordConnectionError() - {0} type {1} count {2}",
					getTidStr(), errType, m_errorCount);
				m_glob.DPRINTV(dbg.ToString());
				#endif

				// Are we too broken to continue?
				if(errorLevelExceeded() && !m_shutdownScheduled) {
					// Schedule our own shutdown
					m_glob.DPRINTV("Fatal error level exceeded, scheduling our own shutdown");
					m_tunneler.scheduleEvent(new ShutdownTunnelEvent(m_glob, new Time(0,0),
						m_tunneler, m_tid, otherEndError));

					// Set flag so we don't schedule repeated shutdowns
					m_shutdownScheduled = true;
				}
			}
		}


		// Remove closing connection from this tunnel
		public virtual void removeConnection(Connection conn)
		{
			#if DEBUG_TUN
			StringBuilder dbg = new StringBuilder();
			dbg.AppendFormat("Tunnel::removeConnection() - connection {0}", conn.getID());
			if(leftConnection(conn)) dbg.Append(" left");
			else if(rightConnection(conn)) dbg.Append(" right");
			m_glob.DPRINTV(dbg.ToString());
			#endif

			bool removed = false;

			lock(this) {
				// Left connection?
				int index = m_leftConn.IndexOf(conn);
				if(index >= 0) {
					// Remove it
					m_leftConn.RemoveAt(index);
					removed = true;

					// Are we still a viable tunnel?
					if(m_leftConn.Count == 0) {
						m_glob.DPRINTI(getTidStr() + " - Tunnel no longer viable");
						recordConnectionError(TunnelErrorType.UNVIABLE_ERROR);
						m_tunneler.scheduleEvent(new TunnelUnviableEvent(m_glob, m_tunneler,
							m_tid));
					}
				}
				else {
					// Right connection?
					index = m_rightConn.IndexOf(conn);
					if(index >= 0) {
						// Remove it
						m_rightConn.RemoveAt(index);
						removed = true;

						// Are we still a viable tunnel?
						if(m_rightConn.Count == 0) {
							m_glob.DPRINTI(getTidStr() + " - Tunnel no longer viable");
							recordConnectionError(TunnelErrorType.UNVIABLE_ERROR);
							m_tunneler.scheduleEvent(new TunnelUnviableEvent(m_glob, m_tunneler,
								m_tid));
						}
					}
				}

				// Remember connections we've dropped
				if(removed && !oldConnection(conn)) {
					m_oldConns.Add(conn.getID());
				}
			}
		}


		// A connection that carries us has connected
		public virtual void connectionConnected(Connection conn)
		{
			#if DEBUG_TUN
			StringBuilder dbg = new StringBuilder();
			dbg.AppendFormat("Tunnel::connectionConnected() - connection {0}", conn.getID());
			m_glob.DPRINTV(dbg.ToString());
			#endif

			lock(this) {
				// Remember that we have connected
				if(leftConnection(conn))
					m_leftEverConnected = true;
				else if(rightConnection(conn))
					m_rightEverConnected = true;
			}
		}


		// An attempt to connect out somewhere failed and has been abandoned
		public virtual void connectAttemptFailed(Connection conn)
		{
			#if DEBUG_TUN
			StringBuilder dbg = new StringBuilder();
			dbg.AppendFormat("Tunnel::connectAttemptFailed() - connection {0}", conn.getID());
			m_glob.DPRINTV(dbg.ToString());
			#endif

			lock(this) {
				// Record failure to connect to server
				recordConnectionError(TunnelErrorType.SERVER_CONNECT_ERROR);
			}
		}


		// Decrease leaky bucket error count
		public void decreaseErrorCount()
		{
			lock(this) {
				#if DEBUG_TUN
				uint oldCount = m_errorCount;
				#endif

				// Once we've reached the fatal error level, don't leak anymore
				if(m_errorCount < FATAL_ERROR_COUNT) {
					// Leak, but check for going negative
					if(m_errorCount > ERROR_COUNT_DECREASE)
						m_errorCount = m_errorCount - ERROR_COUNT_DECREASE;
					else
						m_errorCount = 0;
				}

				#if DEBUG_TUN
				if(oldCount > 0) {
					StringBuilder nfo = new StringBuilder();
					nfo.AppendFormat("{0} old errorCount: {1} new: {2}", getTidStr(),
						oldCount, m_errorCount);
					m_glob.DPRINTV(nfo.ToString());
				}
				#endif
			}
		}

		// State change notifications
		public void closing()
		{
			lock(this) {
				setState(TunnelState.CLOSING);

				// Allow any backlog of unacked msgs to clear
				m_maxUnackedMsgs = 1;
			}
		}
		public void reconnecting()
		{
			lock(this) {
				setState(TunnelState.RECONNECT);
			}
		}


		// Add a new notification msg for sending
		public void addSendNotifMsg(string msg)
		{
			lock(this) {
				m_sendNotifMsgs.Add(msg);
			}
		}


		/*************/
		/* Utilities */
		/*************/

		// Process some newly arrived tunnel msg from one of our connections
		public virtual void processTunnelMsg(HTTPMsg msg, HTTPConnection httpConn)
		{
			// Get msg parameters
			HTTPMsgParms parms = msg.getParms();

			#if DEBUG_TUN
			StringBuilder dbg = new StringBuilder();
			dbg.AppendFormat("Processing msg seq no: {0} ", parms.getPayload().getSeqNo());
			dbg.AppendFormat("state: {0}", parms.getState().ToString());
			if(parms.getAck().Length > 0)
				dbg.AppendFormat(" ack: {0}", parms.getAck());
			m_glob.DPRINTV(dbg.ToString());
			#endif

			// Get payload sequence number
			Payload payload = parms.getPayload();
			SeqNo seqno = payload.getSeqNo();

			lock(this) {
				// If we actually have any payload data, update data received time
				uint payloadSize = payload.getPayloadSize();
				if(payloadSize > 0) {
					m_httpBytesIn += payloadSize;
					m_lastTunRecvTime = Time.now();
				}

				// Check this isn't a duplicate msg
				if(!duplicateTunnelMsg(seqno)) {
					// If this is not a test tunnel, check quotas
					if(!isTestTunnel()) {
						Authenticator authenticator = getAuthenticator();
						if(!authenticator.getUserCanRecvBytes(m_user, payloadSize)) {
							m_glob.DPRINTV("Quota exceeded for received msg");
							recordConnectionError(TunnelErrorType.QUOTAS_EXCEEDED_ERROR);
						}
					}

					// Insert payload into queue & resort
					m_tunnelDataQueue.Add(payload);
					m_tunnelDataQueue.Sort();
				}

				// Process ack from this msg
				string ack = parms.getAck();
				if(ack.Length > 0)
					processAck(ack);

				// Check whether we are using header fields for tunneling
				if(!parms.getUseHeaderFields()) {
					m_tunCfg.setUseHeaders(false);
				}

				// Track any received notification msgs
				processRecvNotifs(parms);

				// Advance our state machine according to this msg
				advanceState(parms);

				// Try sending any received data on to the destination
				sendRawMsgs();
			}
		}


		// Process some newly extracted raw data from one of our connections
		public void processRawData(RawBuff raw)
		{
			lock(this) {
				// Got any data to save?
				uint rawLen = raw.getLength();
				if(rawLen > 0) {
					// Append data to queue
					m_rawDataQueue.Add(raw);
					m_rawDataQueueSize += rawLen;
					m_rawBytesIn += rawLen;

					// Check if we need to block any connections
					checkBlockedConnections();
				}

				// Update data received time
				m_lastRawRecvTime = Time.now();
			}
		}

		// Determine whether we have experienced too many errors to continue
		public bool errorLevelExceeded(Direction dir)
		{
			lock(this) {
				// Give up if we failed to connect out in the indicated direction
				bool res = !everConnected(dir);

				if(!res) {
					// Check direction agnostic errors
					res = errorLevelExceeded();
				}
				else {
					m_glob.DPRINTV("Too many errors; tunnel never connected");
				}

				return res;
			}
		}

		public bool errorLevelExceeded()
		{
			// Leaky error count exceeded?
			bool res = (m_errorCount >= FATAL_ERROR_COUNT);
			return res;
		}


		// Send a message down the tunnel
		public void sendTunnelMsg(bool retry, Connection conn)
		{
			m_glob.DPRINTV("Tunnel::sendTunnelMsg() - " + m_tid.ToString());

			lock(this) {
				// Check supplied connection for suitability
				HTTPConnection httpconn = (HTTPConnection)getTunnelConnection(conn);
				if(httpconn != null) {
					// Clear to send new message?
					if(clearToSendNewMsg(httpconn)) {
						try {
							// Build parameters for this message
							HTTPMsgParms parms = makeTunnelMsg(null, retry);

							// Try to send it (takes ownership of parms)
							sendTunnelMsgWithParms(httpconn, parms);
						}
						catch(TunnelError e) {
							m_glob.log().warn("Failed to send new tunnel msg: " + e.Message);
						}
					}
				}
				else {
					m_glob.DPRINTI("No suitable connection to send msg down found");
				}
			}
		}


		// Send a response to a tunnel message
		public void respondToTunnelMsg(Connection conn, HTTPMsgParms recvd)
		{
			m_glob.DPRINTV("Tunnel::respondToTunnelMsg() - " + m_tid.ToString());

			lock(this) {
				// Check supplied connection for suitability
				HTTPConnection httpconn = (HTTPConnection)getTunnelConnection(conn);
				if(httpconn != null) {
					// Clear to send new message?
					if(clearToSendNewMsg(httpconn)) {
						try {
							// Build parameters for this message
							HTTPMsgParms parms = makeTunnelMsg(recvd, true);

							// Try to send it (takes ownership of parms)
							sendTunnelMsgWithParms(httpconn, parms);
						}
						catch(TunnelError e) {
							m_glob.log().error("Failed to respond to tunnel msg: " + e.Message);
						}
					}
				}
				else {
					m_glob.DPRINTI("No suitable connection to send msg down found");
				}
			}
		}


		// Send out as many raw messages as we have queued tunnel payload for
		public void sendRawMsgs()
		{
			m_glob.DPRINTV("Tunnel::sendRawMsgs()");

			// Get a connection in this tunnel going the right way
			ArrayList filt = new ArrayList();
			if(m_tunnelDirection == Direction.LEFT)
				filt.Add(new SelRightConn());
			else
				filt.Add(new SelLeftConn());
			filt.Add(new SelConnectedConn());

			lock(this) {
				Connection conn = getConnFromSel(filt);
				try {
					// Send as much queued tunneled data as we can
					bool keepGoing = true;
					while(keepGoing) {
						// Get next raw message payload
						Payload raw = makeRawPayload();
						if(raw != null) {
							// Create message parameters
							MsgParms parms = new MsgParms(m_glob, getTidStr(), raw, false);

							// Do we have a connection to send this out on?
							if(conn != null) {
								#if DEBUG_TUN
								StringBuilder dbg = new StringBuilder();
								dbg.AppendFormat("Sending raw msg down connection id {0}",
									conn.getID());
								m_glob.DPRINTV(dbg.ToString());
								#endif

								// Write data to connection
								conn.write(parms);

								// Successfully sent message
								sentRawPayload(true);
							}
							else {
								m_glob.DPRINTI("No suitable connection to send msg down found");
								keepGoing = false;

								// If we're just waiting to close down there's no point keeping payload
								if(m_tunnelState == TunnelState.CLOSING) {
									sentRawPayload(false);
								}
							}
						}
						else {
							// No more tunneled data to send
							keepGoing = false;
						}
					}
				}
				catch(WriteError e) {
					// Error on socket, schedule error on this connection
					m_glob.DPRINTW("Write failure sending raw msg: " + e.Message);
					if(!conn.closing()) {
						m_tunneler.scheduleConnectionError(conn.getID(), TunnelErrorType.GENERAL_COMMS_ERROR);
					}
				}
				catch(TunnelError e) {
					// Failed to send raw msg
					m_glob.DPRINTW("General failure sending raw msg: " + e.Message);
				}
			}
		}


		// Check timeout on ack for some sent message
		public bool checkAckTimeout()
		{
			bool res = false;

			lock(this) {
				// If this tunnel has been silent for too long, start worrying
				TimeSpan quietTime = Time.now() - m_lastTunnelAnySendTime;
				if(quietTime.TotalSeconds > MAX_SILENCE_TIME) {
					m_glob.DPRINTV("Tunnel::checkAckTimeout() - " + getTidStr() + " is very quiet");
					recordConnectionError(TunnelErrorType.GENERAL_COMMS_ERROR);
				}

				// Iterate over unacked msg queue
				for(int pit = 0; pit < m_unackedMsgQueue.Count; ++pit) {
					// How long has this msg been unacknowledged?
					SentMsgParms sentMsgParms = (SentMsgParms)m_unackedMsgQueue[pit];
					TimeSpan unackTime = Time.now() - sentMsgParms.m_lastSentTime;
					if(unackTime.TotalSeconds >= m_ackTimeout) {
						// Ack timeout exceeded for this msg
						SeqNo seqNo = sentMsgParms.m_parms.getPayload().getSeqNo();

						#if DEBUG_TUN
						StringBuilder dbg = new StringBuilder();
						dbg.AppendFormat("Tunnel::checkAckTimeout() - Timeout ({0}) ", m_ackTimeout);
						dbg.AppendFormat("exceeded for msg seq no {0}", seqNo);
						m_glob.DPRINTV(dbg.ToString());
						#endif

						// Remember to resend this msg
						needToResendMsg(seqNo, true);
						res = true;
					}
				}
			}

			return res;
		}


		// Are we in a state where we can be shutdown?
		public bool canBeShutdown()
		{
			bool res = false;

			#if DEBUG_TUN
			StringBuilder cbstr = new StringBuilder();
			#endif

			lock(this) {
				if(errorLevelExceeded(Direction.LEFT) || errorLevelExceeded(Direction.RIGHT)) {
					// Tunnel is errored, so close
					res = true;

					#if DEBUG_TUN
					cbstr.Append("Errored");
					#endif
				}
				else if(m_sentCloseMsg) {
					// We've sent final close msg to other end, so close
					res = true;

					#if DEBUG_TUN
					cbstr.Append("SentClose");
					#endif
				}
				else if((DateTime.Now - m_lastTunnelSendTime).TotalSeconds >= MAX_ACK_TIMEOUT) {
					// Time since last msg sent exceeds limit, so close
					res = true;

					#if DEBUG_TUN
					cbstr.Append("SendTimeout");
					#endif
				}
			}

			#if DEBUG_TUN
			StringBuilder dbg = new StringBuilder();
			dbg.AppendFormat("Tunnel::canBeShutdown() - {0}: {1}", getTidStr(), res);
			if(res) dbg.AppendFormat(" {0}", cbstr.ToString());
			m_glob.DPRINTV(dbg.ToString());
			#endif

			return res;
		}


		/***********************/
		/* Protected Interface */
		/***********************/


		/*********/
		/* Types */
		/*********/

		// HTTPMsgParms for a sent message, plus other details
		protected class SentMsgParms {
			// Construction/destruction
			public SentMsgParms(HTTPMsgParms parms, uint connID)
			{
				m_parms = parms;
				m_connID = connID;
				m_firstSentTime = Time.now();
				m_lastSentTime = Time.now();
				m_ackCount = 0;
				m_resentCount = 0;
			}

			// Data
			public HTTPMsgParms			m_parms;
			public uint						m_connID;
			public Time						m_firstSentTime;
			public Time						m_lastSentTime;
			public uint						m_ackCount;
			public uint						m_resentCount;
		}


		/*************/
		/* Constants */
		/*************/

		// Maximum no. of times we will try resending a msg
		protected const int		MAX_MSG_RESEND					= 3;

		// Raw data buffer fill threshhold
		protected const uint		MAX_RAW_DATA_FILL				= 1024 * 1024;
		protected const uint		RAW_DATA_UNBLOCK_LEVEL		= 90;


		/***********/
		/* Methods */
		/***********/

		// Create complete HTTP msg parameters for sending down tunnel - Caller holds mutex
		protected virtual HTTPMsgParms makeTunnelMsg(HTTPMsgParms recvd, bool retry)
		{
			// Get maximum payload size
			bool useHeader = m_tunCfg.getUseHeaders();
			uint maxPayload = getMaxPayloadSize(useHeader);

			// Tunnel state we will report in this msg
			TunnelState reportedState = m_tunnelState;

			// Get authenticator
			Authenticator authenticator = getAuthenticator();

			// Can we send any actual payload yet?
			if(m_tunnelState == TunnelState.CONNECTED || m_tunnelState == TunnelState.RECONNECT ||
			m_tunnelState == TunnelState.CLOSING) {
				// If this is not a test tunnel, check quotas
				if(!isTestTunnel()) {
					// Check quotas for allowed payload size
					uint quotaMaxPayload = authenticator.getUserBytesCanSend(m_user);
					if(quotaMaxPayload < maxPayload)
						maxPayload = quotaMaxPayload;
				}
			}
			else {
				// Can't send any payload just now
				maxPayload = 0;
			}

			// Get some queued raw data to make up payload
			RawBuff payloadData = new RawBuff();
			bool splitPayload = false;
			makeTunnelPayload(ref payloadData, ref splitPayload, maxPayload);

			// Create actual payload
			Payload payload = new Payload(m_nextSendSeqNo, payloadData, splitPayload);

			// Are we trying to shutdown?
			if(m_tunnelState == TunnelState.CLOSING) {
				// Do we still have raw data to send over the tunnel?
				bool dataToSend = (m_rawDataQueueSize > 0);

				// Do we still have outstanding unacked msgs?
				bool unackedMsgs = (m_unackedMsgQueue.Count > 0);

				// Is this the final msg we will send before closing?
				if(!dataToSend && !unackedMsgs) {
					setState(TunnelState.CLOSED);
					reportedState = TunnelState.CLOSED;
				}
				else {
					// Don't let the other side know just yet
					reportedState = TunnelState.CONNECTED;

					#if DEBUG_TUN
					StringBuilder dbg = new StringBuilder();
					dbg.Append("Not moving to CLOSED: Remaining raw data len = ");
					dbg.AppendFormat("{0}, Unacked msg count = {1}", m_rawDataQueueSize,
						m_unackedMsgQueue.Count);
					m_glob.DPRINTV(dbg.ToString());
					#endif
				}
			}

			// Make msg parms
			User userDetails = authenticator.getUserDetails(m_user);
			HTTPMsgParms parms = new HTTPMsgParms(m_glob, this, payload, retry, userDetails,
				reportedState, m_tunCfg.getHTTPVersion(), m_nextRecvSeqNo.ToString(),
				m_tunCfg.getDestAddress(), useHeader);

			// Do we have a URI to respond to?
			if(recvd != null) {
				parms.setURI(recvd.getURI());
			}

			return parms;
		}


		// Create payload for tunneled msg to send to the destination - Caller holds mutex
		protected virtual void makeTunnelPayload(ref RawBuff data, ref bool split, uint max)
		{
			split = false;

			// Iterate over queued raw packets
			for(int pit = 0; pit < m_rawDataQueue.Count; ++pit) {
				// Get next queued raw data packet
				RawBuff queuedData = (RawBuff)m_rawDataQueue[pit];

				// Check type of queued packets
				if(m_tid.getProtocol() == Enums.Protocol.UDP) {
					// Maintain UDP datagram boundaries
					data.append(queuedData, max);
					if(data.getLength() > 0 && data.getLength() < queuedData.getLength())
						split = true;
					break;
				}
				else {
					// Pack TCP data upto the maximum
					uint appendBytes = max - data.getLength();
					data.append(queuedData, appendBytes);
					if(data.getLength() == max) {
						// Reached maximum size
						break;
					}
				}
			}
		}


		// Successfully sent a message down the tunnel - Caller hold mutex
		protected abstract void sentNewTunnelMsg(HTTPConnection conn, HTTPMsgParms msgParms);

		// Successfully sent a message down the tunnel - Caller hold mutex
		protected void sentNewTunnelMsg(HTTPConnection conn, HTTPMsgParms msgParms,
		bool remember)
		{
			// Free sent raw data
			sentTunnelPayload(msgParms);

			// Update quota tracking
			uint payloadLen = msgParms.getPayload().getPayloadSize();
			m_httpBytesOut += payloadLen;
			if(!isTestTunnel()) {
				Authenticator authenticator = getAuthenticator();
				authenticator.setUserSentBytes(m_user, payloadLen);
			}

			// Update last sent tunnel msg time
			m_lastTunnelSendTime = DateTime.Now;
			tunnelAnyMsgSent();
			TunnelState msgTunState = msgParms.getState();
			if(msgTunState == TunnelState.CONNECTED || msgTunState == TunnelState.RECONNECT ||
			msgTunState == TunnelState.CLOSING) {
				m_lastTunnelConnectedSendTime = Time.now();
			}

			// Do we want to remember about this?
			if(remember) {
				// Did this msg contain a final close for this tunnel?
				if(msgTunState == TunnelState.CLOSED) {
					m_sentCloseMsg = true;

					// We won't be requiring an ack for the final close
					//delete msgParms;
				}
				else {
					// Add paramaters to queue of paramaters waiting for an ack
					m_unackedMsgQueue.Add(new SentMsgParms(msgParms, conn.getID()));

					// Was this a RECONNECT?
					if(msgTunState == TunnelState.RECONNECT) {
						// Move straight into CONNECTED state
						m_glob.DPRINTI("Moving straight from RECONNECT to CONNECTED");
						setState(TunnelState.CONNECTED);
					}
				}

				// Increment next sent sequence number
				incSendSeqNo();
			}
			else {
				// We're going to pretend this msg hasn't been sent at all
				//delete msgParms;
			}
		}


		// Determine whether an ack implies we should resend a msg - Caller holds mutex
		protected virtual void checkForResendRequest(SeqNo ackNo)
		{
			bool ackCountExceeded = false;
			bool consecutiveAck = false;

			// Is the msg still unacked?
			for(int i = 0; i < m_unackedMsgQueue.Count; ++i) {
				SentMsgParms sentMsgParms = (SentMsgParms)m_unackedMsgQueue[i];
				if(sentMsgParms.m_parms.getPayload().getSeqNo() == ackNo) {
					// Is the sender acking the same sequence number as last time?
					if(ackNo == m_lastRecvAckSeqNo) {
						// Check for multi-ack resend request
						ackCountExceeded = checkForMultiAckResendRequest(sentMsgParms);
					}
					else if(m_lastRecvAckResent) {
						// The sender seems to be requesting a resend of a subsequent unacked msg
						m_glob.DPRINTV("Tunnel::checkForResendRequest() - Consecutive-Ack resend request");
						consecutiveAck = true;
					}

					break;
				}
			}

			// Do we want to resend?
			if(ackCountExceeded || consecutiveAck) {
				m_lastRecvAckResent = true;
				needToResendMsg(ackNo, true);
			}
			else {
				m_lastRecvAckResent = false;
			}

			// Remember for next time
			m_lastRecvAckSeqNo = ackNo;
		}

		// Determine whether a multi-ack implies we should resend a msg
		protected abstract bool checkForMultiAckResendRequest(SentMsgParms sentMsgParms);


		// Check message parms before resending - Caller holds mutex
		protected virtual void checkResendMsgParms(HTTPMsgParms parms)
		{
			// Nothing to do here
		}


		// Get authenticator
		protected virtual Authenticator getAuthenticator()
		{
			return m_tunneler.getServer().getConfig().getAuthenticator();
		}


		// Have we got a stable HTTP connection to other end?
		protected virtual bool httpStable()
		{
			return true;
		}


		// Can we send a new tunnel msg?
		protected abstract bool clearToSendNewMsg(HTTPConnection conn);

		// Check whether any connections need to be blocked/unblocked
		protected abstract void checkBlockedConnections();


		// Advance our internal state machine
		protected abstract void advanceState(HTTPMsgParms parms);

		// The tunnel state has changed - Caller holds mutex
		protected virtual void stateChanged(TunnelState oldState)
		{
			#if DEBUG_TUN
			StringBuilder nfo = new StringBuilder();
			nfo.AppendFormat("Tunnel moved from state {0} to state {1}", oldState, m_tunnelState);
			m_glob.DPRINTI(nfo.ToString());
			#endif

			// Let controlling tunneler know
			m_tunneler.scheduleEvent(new TunnelStateEvent(m_glob, m_tunneler, m_tid,
				m_tunnelState));
		}


		// Get maximum allowable payload size
		protected abstract uint getMaxPayloadSize(bool useHeader);


		// A sent message has been acknowledged - Caller holds mutex
		protected virtual void msgAcked(SentMsgParms parms)
		{
			// Nothing more to do in base class
		}


		// Something has been sent down the tunnel - Caller holds mutex
		protected virtual void tunnelAnyMsgSent()
		{
			// Update any data sent time
			m_lastTunnelAnySendTime = Time.now();
		}


		// Record we need to resend a msg - Caller holds mutex
		protected virtual void needToResendMsg(SeqNo seqNo, bool msgLost)
		{
			#if DEBUG_TUN
			StringBuilder nfo = new StringBuilder();
			nfo.AppendFormat("Tunnel::needToResendMsg() - Need to resend seq no {0} ", seqNo);
			nfo.AppendFormat("for tunnel {0}", m_tid);
			m_glob.DPRINTI(nfo.ToString());
			#endif

			// Check we haven't already recorded this msg
			int pos = m_resendQueue.IndexOf(seqNo);
			if(pos < 0) {
				// Record msg in resend queue
				m_resendQueue.Add(seqNo);
				m_resendQueue.Sort();
			}
		}


		// Look for unacked management msgs we want to try resending - Caller holds mutex
		protected void checkUnackedMgmtMsgs()
		{
			// Check all currently unacked msgs
			for(int sit = 0; sit < m_unackedMsgQueue.Count; ++sit) {
				// Get msg details
				SentMsgParms sentMsgParms = (SentMsgParms)m_unackedMsgQueue[sit];
				HTTPMsgParms msgParms = sentMsgParms.m_parms;

				// If the HTTP session is not yet stable, allow resending of mgmt msgs early
				if(isMgmtMsg(msgParms) && !httpStable()) {
					// Resend this mgmt msg
					SeqNo seqNo = msgParms.getPayload().getSeqNo();

					#if DEBUG_TUN
					StringBuilder nfo = new StringBuilder();
					nfo.AppendFormat("Allowing resend of unacked mgmt msg seq no {0} ", seqNo);
					nfo.AppendFormat("for tunnel {0}", m_tid);
					m_glob.DPRINTI(nfo.ToString());
					#endif

					// Update server auth settings incase they have changed
					msgParms.setServerAuthSettings(getServerAuthSettings());
					Authenticator authenticator = getAuthenticator();
					msgParms.setUser(authenticator.getUserDetails(m_user));

					// Remember to resend this msg
					needToResendMsg(seqNo, false);
				}
			}
		}

		// Resend all flagged unacknowledged msgs - Caller holds mutex
		protected bool resendUnackedMsgs(Connection conn)
		{
			bool resentMsg = false;

			// Got any msgs to resend?
			while(m_resendQueue.Count > 0) {
				// Get sequence number of next msg to resend
				SeqNo seqNo = (SeqNo)m_resendQueue[0];

				#if DEBUG_TUN
				StringBuilder dbg = new StringBuilder();
				dbg.Append("Tunnel::resendUnackedMsgs() - Considering resend of seq no ");
				dbg.AppendFormat("{0} for {1} on connection {2}", seqNo, m_tid.ToString(),
					conn.getID());
				m_glob.DPRINTV(dbg.ToString());
				#endif

				// Look for msg paramaters in unacked msg queue
				SentMsgParms sentMsgParms = null;
				bool stillOutstanding = false;
				for(int i = 0; i < m_unackedMsgQueue.Count; ++i) {
					sentMsgParms = (SentMsgParms)m_unackedMsgQueue[i];
					if(sentMsgParms.m_parms.getPayload().getSeqNo() == seqNo) {
						stillOutstanding = true;
						break;
					}
				}

				// Ack still outstanding?
				if(stillOutstanding) {
					// Try to resend msg
					HTTPMsgParms msgParms = sentMsgParms.m_parms;
					resentMsg = true;

					// Check we haven't already tried to resend this msg too many times
					if(sentMsgParms.m_resentCount >= MAX_MSG_RESEND) {
						recordConnectionError(TunnelErrorType.MAX_RESEND_ERROR);
						throw new ResendError("Resend count exceeded for msg");
					}

					// Update connection msg sent on
					sentMsgParms.m_connID = conn.getID();

					// Update acks to current value
					msgParms.setAck(m_nextRecvSeqNo.ToString());

					// Check message parms before resending
					checkResendMsgParms(msgParms);

					#if DEBUG_TUN
					StringBuilder dbg2 = new StringBuilder();
					dbg2.Append("Tunnel::resendUnackedMsgs() - Attempting resend with ack: ");
					dbg2.AppendFormat("{0}", msgParms.getAck());
					m_glob.DPRINTV(dbg2.ToString());
					#endif

					bool resentOk = false;

					try {
						// Record this as an error on this tunnel
						if(httpStable())
							recordConnectionError(TunnelErrorType.MSG_RESEND_ERROR);

						// Update msg resent count
						sentMsgParms.m_resentCount++;

						// Write data to connection
						conn.write(msgParms);
						resentOk = true;

						// Remove resent sequence number from queue
						m_resendQueue.RemoveAt(0);

						// Update sent times
						tunnelAnyMsgSent();
						sentMsgParms.m_lastSentTime = Time.now();
					}
					catch(CommsError e) {
						// Comms error resending msg
						m_glob.log().info("CommsError resending msg: " + e.Message);
						if(!conn.closing()) {
							// Schedule error reporting event for connection
							m_tunneler.scheduleConnectionError(conn.getID(), TunnelErrorType.GENERAL_COMMS_ERROR);
						}
					}
					catch(TunnelError e) {
						// General error resending msg
						m_glob.log().info("Error resending msg: " + e.Message);
					}

					// Error resending?
					if(!resentOk) {
						// Let caller know
						throw new ResendError("An error occured while resending msg");
					}
				}
				else {
					// Fatal error
					recordConnectionError(TunnelErrorType.MAX_RESEND_ERROR);
					throw new ResendError("Can't find msg parameters in unacked msg list");
				}
			}

			return resentMsg;
		}

		// Is some msg a management msg?
		protected bool isMgmtMsg(HTTPMsgParms parms)
		{
			TunnelState msgState = parms.getState();
			return (msgState == TunnelState.CONNECT || msgState == TunnelState.AUTHREQ ||
				msgState == TunnelState.RECONNECT);
		}


		// Set our current state - Caller holds mutex
		protected void setState(TunnelState newState)
		{
			// Remember starting state
			TunnelState oldState = m_tunnelState;

			// Once we are closing or closed, restrict further state changes
			if(oldState == TunnelState.CLOSED || oldState == TunnelState.CLOSING) {
				// Allow us to go from closing to closed
				if(oldState == TunnelState.CLOSING && newState == TunnelState.CLOSED)
					m_tunnelState = TunnelState.CLOSED;
			}
			else {
				// Set new state
				m_tunnelState = newState;
			}

			// Tell us if we changed state
			if(oldState != m_tunnelState) {
				stateChanged(oldState);
			}
		}

		// Set our current test state - Caller holds mutex
		protected void setTestState(TunnelTestState newState)
		{
			// Remember starting state
			TunnelTestState oldState = m_tunnelTestState;

			// Set new state
			m_tunnelTestState = newState;

			// Tell us if we changed state
			if(oldState != m_tunnelTestState) {
				#if DEBUG_TUN
				StringBuilder nfo = new StringBuilder();
				nfo.AppendFormat("Tunnel moved from test state {0} to test state {1}",
					oldState, m_tunnelTestState);
				m_glob.DPRINTI(nfo.ToString());
				#endif
			}
		}


		// Get no. of unacked msgs on a connection - Caller holds mutex
		protected uint noUnackedMsgsOnConn(uint connID)
		{
			uint no = 0;

			// Scan unacked msgs list
			for(int sit = 0; sit < m_unackedMsgQueue.Count; ++sit) {
				SentMsgParms parms = (SentMsgParms)m_unackedMsgQueue[sit];
				if(parms.m_connID == connID)
					no++;
			}

			return no;
		}


		/********/
		/* Data */
		/********/

		// Global data
		protected Globals						m_glob;

		// Owning tunneler
		protected Tunneler					m_tunneler;

		// Lists of connections that comprise this Tunnel
		protected ArrayList					m_leftConn;
		protected ArrayList					m_rightConn;

		// Configuration for this Tunnel
		protected TunnelConfig				m_tunCfg;
		protected TunnelID					m_tid;

		// Min/max and current ack timeout values
		protected int							m_minAckTimeout;
		protected int							m_maxAckTimeout;
		protected int							m_ackTimeout;

		// Current ack timeout values
		protected ArrayList					m_ackRTTs;
		protected int							m_avgAckRTT;
		protected int							m_biasAvgAckRTT;

		// User of this tunnel
		protected string						m_user;

		// Current tunnel states
		protected TunnelState				m_tunnelState;
		protected TunnelTestState			m_tunnelTestState;

		// Queue of parameters for sent but unacknowledged messages
		protected ArrayList					m_unackedMsgQueue;

		// Time we last sent ANY msg down the tunnel
		protected Time							m_lastTunnelAnySendTime;

		// Queues of received tunnel/raw data
		protected ArrayList					m_tunnelDataQueue;
		protected ArrayList					m_rawDataQueue;
		protected uint							m_rawDataQueueSize;

		// Maximum no. of outstanding unacked msgs
		protected uint							m_maxUnackedMsgs;

		// Last received ack sequence no.
		protected SeqNo						m_lastRecvAckSeqNo;
		protected bool							m_lastRecvAckResent;

		// Queue of sequence nos. for messages we must resend
		protected ArrayList					m_resendQueue;


		/*********************/
		/* Private Interface */
		/*********************/


		/***********/
		/* Methods */
		/***********/

		// Increment next send packet sequence numbers - Caller holds mutex
		private void incSendSeqNo()
		{
			m_nextSendSeqNo.inc();
		}

		// Increment next received packet sequence numbers - Caller holds mutex
		private void incRecvSeqNo()
		{
			// Always increment
			m_nextRecvSeqNo.inc();

			#if DEBUG_TUN
			if(m_tunnelDataQueue.Count >= m_nextRecvSeqNo.getMaxSeqNo() - 5) {
				StringBuilder wrn = new StringBuilder();
				wrn.AppendFormat("Tunnel data queue size approaching seqno limit: {0},{1}",
					m_tunnelDataQueue.Count, m_nextRecvSeqNo.getMaxSeqNo());
				m_glob.DPRINTW(wrn.ToString());
			}
			#endif

			#if DEBUG_TUN
			StringBuilder dbg1 = new StringBuilder();
			dbg1.AppendFormat("Tunnel::incSendSeqNo() - Tunnel data queue size: {0}",
				m_tunnelDataQueue.Count);
			m_glob.DPRINTV(dbg1.ToString());
			if(m_tunnelDataQueue.Count > 0) {
				StringBuilder dbg2 = new StringBuilder();
				dbg2.AppendFormat("SeqNo for first item in queue: {0}",
					((Payload)m_tunnelDataQueue[0]).getSeqNo());
				m_glob.DPRINTV(dbg2.ToString());
			}
			#endif

			// Have we already received a msg with this sequence no.?
			for(int i = 0; i < m_tunnelDataQueue.Count; ++i) {
				Payload pld = (Payload)m_tunnelDataQueue[i];
				if(pld.getSeqNo() != m_nextRecvSeqNo) {
					// Still waiting for msg with this sequence no.
					break;
				}

				// Already received this msg, so increment again
				m_nextRecvSeqNo.inc();
			}

			#if DEBUG_TUN
			StringBuilder dbg3 = new StringBuilder();
			dbg3.AppendFormat("Next recv/ack seq no = {0}", m_nextRecvSeqNo);
			m_glob.DPRINTV(dbg3.ToString());
			#endif
		}


		// Clear up raw packet queue after sending tunneled msg - Caller holds mutex
		private void sentTunnelPayload(HTTPMsgParms msgParms)
		{
			// Get size of sent payload
			uint payloadLen = msgParms.getPayload().getPayloadSize();

			// Iterate over queued raw packets
			while(payloadLen > 0 && m_rawDataQueue.Count > 0) {
				// Get next queued raw data packet
				RawBuff queuedData = (RawBuff)m_rawDataQueue[0];
				uint queuedDataLen = queuedData.getLength();

				// Whole of this packet sent?
				if(queuedDataLen <= payloadLen) {
					// Keep track of bytes erased
					payloadLen -= queuedDataLen;
					m_rawDataQueueSize -= queuedDataLen;

					// Erase this packet from queue
					m_rawDataQueue.RemoveAt(0);
				}
				else {
					// Partial packet sent
					queuedData.erase(payloadLen);
					m_rawDataQueueSize -= payloadLen;
					payloadLen = 0;
				}
			}

			// Check if we need to unblock any connections
			checkBlockedConnections();
		}


		// Create raw data for sending to the destination - Caller hold mutex
		private Payload makeRawPayload()
		{
			Payload raw = null;

			// Got any payload data we can make the raw msg from?
			if(m_tunnelDataQueue.Count > 0) {
				// Are we missing any payload?
				Payload payload = (Payload)m_tunnelDataQueue[0];
				if(m_nextRecvSeqNo >= payload.getSeqNo()) {

					// Got split packets to deal with?
					if(payload.getSplit()) {
						bool buildSplit = false;
						// Do we have all payloads for split packet?
						for(int pit = 0; pit < m_tunnelDataQueue.Count; ++pit) {
							payload = (Payload)m_tunnelDataQueue[pit];
							if(m_nextRecvSeqNo >= payload.getSeqNo()) {
								// Member of split set
								payload.setSent(true);

								// Final packet from split set?
								if(!payload.getSplit()) {
									// Got all splits, can build payload
									buildSplit = true;
									break;
								}
							}
							else {
								// Missing packet from split payload
								break;
							}
						}

						// Can we go ahead and recombine split packets?
						if(buildSplit) {
							RawBuff recombined = new RawBuff();
							for(int pit = 0; pit < m_tunnelDataQueue.Count; ++pit) {
								payload = (Payload)m_tunnelDataQueue[pit];
								if(payload.getSent()) {
									recombined.append(payload.getPayload());
								}
								else {
									break;
								}
							}

							raw = new Payload(new SeqNo(0), recombined, false);
						}
					}

					else {
						// Single non-split payload, so just copy it
						raw = new Payload(payload);
						payload.setSent(true);
					}
				}
			}

			return raw;
		}


		// Successfully sent raw data to the destination - Caller holds mutex
		private void sentRawPayload(bool sentOk)
		{
			// Free all sent payload data
			while(m_tunnelDataQueue.Count > 0) {
				Payload payload = (Payload)m_tunnelDataQueue[0];
				if(payload.getSent()) {
					// Count raw bytes sent
					if(sentOk)
						m_rawBytesOut += payload.getPayloadSize();

					// Free payload data
					m_tunnelDataQueue.RemoveAt(0);
				}
				else {
					// No more sent payload packets
					break;
				}
			}
		}


		// Process an acknowledgement - Mutex held by caller
		private void processAck(string ack)
		{
			// Convert to numeric sequence number
			SeqNo seqNo = new SeqNo(UInt32.Parse(ack));

			// This acknowledges everything less than that sequence number
			bool gotIt;
			do {
				gotIt = false;
				int pit;
				SentMsgParms sentMsgParms = null;
				for(pit = 0; pit < m_unackedMsgQueue.Count; pit++) {
					sentMsgParms = (SentMsgParms)m_unackedMsgQueue[pit];
					if(sentMsgParms.m_parms.getPayload().getSeqNo() < seqNo) {
						gotIt = true;
						break;
					}
				}

				// Got it?
				if(gotIt) {
					// Acknowledges this msg
					SeqNo acked = sentMsgParms.m_parms.getPayload().getSeqNo();

					// Store round trip time for this ack
					TimeSpan rtt = Time.now() - sentMsgParms.m_firstSentTime;
					if(m_ackRTTs.Count > ACKRTT_AVG_LENGTH)
						m_ackRTTs.RemoveAt(0);
					m_ackRTTs.Add(rtt);

					#if DEBUG_TUN
					StringBuilder nfo1 = new StringBuilder();
					nfo1.AppendFormat("This ack rtt: {0}", (int)rtt.TotalMilliseconds);
					m_glob.DPRINTV(nfo1.ToString());
					#endif

					// Calculate average round trip time for acks
					calculateAvgAckRTT();

					// Update ack timeout
					#if DEBUG_TUN
					int oldAckTimeout = m_ackTimeout;
					#endif

					m_ackTimeout = (m_avgAckRTT * 5) / 1000;
					if(m_ackTimeout < m_minAckTimeout)
						m_ackTimeout = m_minAckTimeout;
					else if(m_ackTimeout > m_maxAckTimeout)
						m_ackTimeout = m_maxAckTimeout;

					#if DEBUG_TUN
					if(m_ackTimeout != oldAckTimeout) {
						StringBuilder nfo2 = new StringBuilder();
						nfo2.AppendFormat("Ack timeout for tunnel {0} ", m_tid);
						nfo2.AppendFormat("was {0}, now {1}", oldAckTimeout, m_ackTimeout);
						m_glob.DPRINTV(nfo2.ToString());
					}
					#endif

					// Let anyone interested know
					msgAcked(sentMsgParms);

					// Free up sent parms
					m_unackedMsgQueue.RemoveAt(pit);

					// Is this a late arrival for a msg we've flagged as needing a resend?
					int rit = m_resendQueue.IndexOf(acked);
					if(rit >= 0) {
						// No longer need to resend
						m_resendQueue.RemoveAt(rit);
					}
				}
			} while(gotIt);

			// Does it look like the sender is requesting we resend an unacked msg?
			checkForResendRequest(seqNo);
		}


		// Calculate ack-rtt averages - Caller holds mutex
		private void calculateAvgAckRTT()
		{
			// How many times do we have to average?
			int noRTTs = m_ackRTTs.Count;

			// Initialise totals
			double tot = 0;
			double biastot = 0;
			const double biasmin = 0.2;
			const double biasmax = 2.0 - biasmin;
			const double biasdiff = biasmax - biasmin;

			// Calculate ack round trip time totals
			for(int i = 0; i < noRTTs; ++i) {
				// Simple total
				double rtt = ((TimeSpan)m_ackRTTs[i]).TotalMilliseconds;
				tot += rtt;

				// Got enough times to calculate biased avg?
				if(noRTTs > 2) {
					double bias = biasmin + ((biasdiff / (noRTTs - 1)) * i);
					biastot += rtt * bias;
				}
				else {
					biastot = tot;
				}
			}

			// Calculate averages
			m_avgAckRTT = (int)(tot / noRTTs);
			m_biasAvgAckRTT = (int)(biastot / noRTTs);

			#if DEBUG_TUN
			StringBuilder nfo = new StringBuilder();
			nfo.AppendFormat("Tunnel::calculateAvgAckRTT() - Average: {0}", m_avgAckRTT);
			nfo.AppendFormat(", Biased: {0}", m_biasAvgAckRTT);
			m_glob.DPRINTV(nfo.ToString());
			#endif
		}


		// Check some new payload isn't from a duplicate message
		private bool duplicateTunnelMsg(SeqNo seqNo)
		{
			// Assume it is not a duplicate until proved otherwise
			bool res = false;

			// Is this payload out of sequence?
			if(seqNo != m_nextRecvSeqNo) {
				// If it's earlier then it must be a duplicate
				if(seqNo < m_nextRecvSeqNo) {
					res = true;
				}
				else {
					// Look for a duplicate in the queue of received msgs
					for(int i = 0; i < m_tunnelDataQueue.Count; ++i) {
						Payload payload = (Payload)m_tunnelDataQueue[i];
						if(payload.getSeqNo() == seqNo) {
							// Found a duplicate
							res = true;
							break;
						}
					}
				}

				#if DEBUG_TUN
				StringBuilder dbg = new StringBuilder();
				dbg.AppendFormat("Got out of sequence msg: {0}, expected next {1}", seqNo,
					m_nextRecvSeqNo);
				if(res) dbg.Append(" (Duplicate)");
				m_glob.DPRINTV(dbg.ToString());
				#endif
			}
			else {
				// Increase next expected received sequence no.
				incRecvSeqNo();
			}

			return res;
		}

		// Check & get a suitable connection for sending a tunnel msg out of
		// Caller holds mutex
		private Connection getTunnelConnection(Connection conn)
		{
			Connection res = conn;

			// Already supplied a candidate connection to use?
			if(res != null) {
				// Check supplied connection is actually one of ours going the right way
				if(m_tunnelDirection == Direction.LEFT) {
					// Check connection
					if(!leftConnection(res) || !conn.connected()) {
						#if DEBUG_TUN
						StringBuilder dbg = new StringBuilder();
						dbg.Append("Tunnel::getTunnelConnection() - Overriding connection ");
						dbg.AppendFormat("{0} ", conn.getID());
						if(!leftConnection(res))
							dbg.Append(" not-left");
						if(!conn.connected())
							dbg.Append(" not-connected");
						m_glob.DPRINTI(dbg.ToString());
						#endif

						res = null;
					}
				}
				else {
					// Check connection
					if(!rightConnection(res) || !conn.connected()) {
						#if DEBUG_TUN
						StringBuilder dbg = new StringBuilder();
						dbg.Append("Tunnel::getTunnelConnection() - Overriding connection ");
						dbg.AppendFormat("{0} ", conn.getID());
						if(!rightConnection(res))
							dbg.Append(" not-right");
						if(!conn.connected())
							dbg.Append(" not-connected");
						m_glob.DPRINTI(dbg.ToString());
						#endif

						res = null;
					}
				}
			}

			// Do we need to look a connection up?
			if(res == null) {
				// Get a recent connection in this tunnel going the right way
				ArrayList filt = new ArrayList();
				if(m_tunnelDirection == Direction.LEFT)
					filt.Add(new SelLeftConn());
				else
					filt.Add(new SelRightConn());
				filt.Add(new SelConnectedConn());
				filt.Add(new SelRecentConn());
				res = getConnFromSel(filt);

				#if DEBUG_TUN
				if(res != null) {
					StringBuilder nfo = new StringBuilder();
					nfo.AppendFormat("Selected connection id {0}", res.getID());
					m_glob.DPRINTV(nfo.ToString());
				}
				#endif
			}

			return res;
		}


		// Is a connection an old removed one? - Caller holds mutex
		private bool oldConnection(Connection conn)
		{
			// Is this an old connection?
			int ocit = m_oldConns.IndexOf(conn.getID());
			bool old = (ocit != -1);

			// Trim list of old connections if it's getting too large
			if(m_oldConns.Count > MAX_OLD_CONNS_LENGTH) {
				m_oldConns.RemoveRange(0, MAX_OLD_CONNS_LENGTH / 2);
			}

			return old;
		}


		// Send a tunnel msg with the given paramaters - Caller holds mutex
		private void sendTunnelMsgWithParms(HTTPConnection conn, HTTPMsgParms parms)
		{
			bool sentOk = false;

			try {
				#if DEBUG_TUN
				StringBuilder dbg = new StringBuilder();
				dbg.Append("Sending HTTP msg seq no: ");
				dbg.AppendFormat("{0} ", parms.getPayload().getSeqNo());
				dbg.AppendFormat("state: {0} ", parms.getState().ToString());
				if(parms.getAck().Length > 0)
					dbg.AppendFormat("ack: {0} ", parms.getAck());
				dbg.AppendFormat("on connection {0}", conn.getID());
				m_glob.DPRINTV(dbg.ToString());
				#endif

				// Write data to connection
				conn.write(parms);
				sentOk = true;

				// Save msg parameters incase we need to resend later
				sentNewTunnelMsg(conn, parms);
			}
			catch(CommsError e) {
				m_glob.log().warn("CommsError sending tunnel msg: " + e.Message);
				if(!conn.closing()) {
					// Schedule error reporting event for connection
					m_tunneler.scheduleConnectionError(conn.getID(), TunnelErrorType.GENERAL_COMMS_ERROR);
				}
			}
			catch(TunnelError e) {
				m_glob.log().error("General failure sending tunnel msg: " + e.Message);
			}

			// Error sending?
			if(!sentOk) {
				// Let caller know
				throw new TunnelError("Failed to send tunnel msg");
			}
		}


		// Process received notification msgs - Caller holds mutex
		private void processRecvNotifs(HTTPMsgParms parms)
		{
			// Get notification msgs
			ArrayList notifs = parms.getNotifs();

			// Add them to our list of received notifications
			for(int nit = 0; nit < notifs.Count; ++nit) {
				string notif = (string)notifs[nit];
				m_recvNotifMsgs.Add(notif);

				// Limit the no. of old notifications we keep
				if(m_recvNotifMsgs.Count > MAX_RECV_NOTIFS)
					m_recvNotifMsgs.RemoveAt(0);
			}
		}


		/*************/
		/* Constants */
		/*************/

		// Amount to decrease leaky bucket error count by each period
		private const uint		ERROR_COUNT_DECREASE		= 1;

		// Level for leaky bucket error count which we consider fatal
		private const uint		FATAL_ERROR_COUNT			= 20;

		// Ack timeout in seconds
		private const int			MIN_ACK_TIMEOUT			= 5;
		private const int			MAX_ACK_TIMEOUT			= 20;

		// Seconds tunnel can stay completely silent before we call it an error
		private const int			MAX_SILENCE_TIME			= 60;

		// Seconds we go with no data received before considering tunnel silent
		private const int			MAX_NO_RECV_TIME			= 60;

		// Time after sending last tunnel msg we act on unviable errors (seconds)
		private const int			UNVIABLE_ERROR_TIME		= 5;

		// Maximum length we allow the old connection list to grow to
		private const int			MAX_OLD_CONNS_LENGTH		= 100;

		// Number of ack round trip times to keep for averaging
		private const int			ACKRTT_AVG_LENGTH			= (int)TuneSettings.MAX_UNACKED_MSGS * 3;

		// Maximum no. of recvd notification msgs we keep
		private const int			MAX_RECV_NOTIFS			= 25;


		/********/
		/* Data */
		/********/

		// Remember IDs of connections we've explicitly dropped
		private ArrayList					m_oldConns;

		// Sequence numbers for msgs sent/received
		private SeqNo						m_nextSendSeqNo;
		private SeqNo						m_nextRecvSeqNo;

		// Leaky bucket error counter
		private uint						m_errorCount;

		// Flag to indicate if the last dead connections ever connected
		private bool						m_leftEverConnected;
		private bool						m_rightEverConnected;

		// In which direction does the tunnel lie?
		private Direction					m_tunnelDirection;

		// Flag to indicate whether we have scheduled a shutdown
		private bool						m_shutdownScheduled;

		// Things to determine whether we can be shutdown
		private DateTime					m_lastTunnelSendTime;
		private bool						m_sentCloseMsg;

		// Time we last sent a standard CONNECTED msg down the tunnel
		private Time						m_lastTunnelConnectedSendTime;

		// Time we last received data
		private Time						m_lastRawRecvTime;
		private Time						m_lastTunRecvTime;

		// Count of bytes sent/received
		private uint						m_rawBytesIn;
		private uint						m_rawBytesOut;
		private uint						m_httpBytesIn;
		private uint						m_httpBytesOut;

		// Lists of notification msgs for sending and been received
		private ArrayList					m_sendNotifMsgs;
		private ArrayList					m_recvNotifMsgs;

	}

}	// namespace

