/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Somewhere to store all configuration for a tunnel instance.
 *
 * $Id: TunnelConfig.cs,v 1.2 2013/02/17 10:53:20 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel {

	using System;
	using System.Text;

	public class TunnelConfig {

		// Construction
		public TunnelConfig(TunnelID tid, string tune, ProxySettings proxyCfg,
		ServerAuthSettings saCfg, IPAddr destIP)
		{
			m_tid = tid;
			m_httpVer = "1.1";
			m_destIP = destIP;
			m_useHeaders = true;
			m_proxyCfg = null;
			m_serverAuthCfg = null;
			m_tuning = new TuneSettings(tune);

			// Copy proxy/serverauth config
			if(proxyCfg != null)
				m_proxyCfg = (ProxySettings)proxyCfg.Clone();
			if(saCfg != null)
				m_serverAuthCfg = (ServerAuthSettings)saCfg.Clone();
		}

		public TunnelConfig(TunnelID tid, string tune, string destAddrStr, ServerAuthSettings saCfg)
		{
			m_tid = tid;
			m_httpVer = "1.1";
			m_destIP = new IPAddr("", 0);
			m_useHeaders = true;
			m_proxyCfg = null;
			m_serverAuthCfg = null;
			m_tuning = new TuneSettings(tune);

			// Set destination IP and port from address string
			setDestAddress(destAddrStr);

			// Copy server auth config
			if(saCfg != null) {
				m_serverAuthCfg = (ServerAuthSettings)saCfg.Clone();

				// Create new authentication nonce for this tunnel
				m_serverAuthCfg.newNonce();
			}
		}


		/*************/
		/* Accessors */
		/*************/

		// Get tunnel ID
		public TunnelID getTunnelID()
		{
			return m_tid;
		}

		// Get HTTP version string
		public string getHTTPVersion()
		{
			return m_httpVer;
		}

		// Get IP address of tunnel destination
		public IPAddr getDestIP()
		{
			return m_destIP;
		}

		// Get protocol being tunneled
		public Enums.Protocol getDestProtocol()
		{
			return m_tid.getProtocol();
		}

		// Get IP address/port of tunnel destination
		public string getDestAddress()
		{
			StringBuilder addr = new StringBuilder();
			addr.AppendFormat("{0};{1}", m_destIP.getHost(), m_destIP.getPort());
			return addr.ToString();
		}

		// Get whether we are using headers for tunneling
		public bool getUseHeaders()
		{
			return m_useHeaders;
		}

		// Get tunnels proxy config
		public ProxySettings getProxySettings()
		{
			return m_proxyCfg;
		}

		// Get tunnels serverauth config
		public ServerAuthSettings getServerAuthSettings()
		{
			return m_serverAuthCfg;
		}

		// Get tunnels tuning
		public TuneSettings getTuning()
		{
			return m_tuning;
		}


		/************/
		/* Mutators */
		/************/

		// Set HTTP version string
		public void setHTTPVersion(string ver)
		{
			m_httpVer = ver;
		}

		// Set destination IP address
		public void setDestIP(IPAddr destIP)
		{
			m_destIP = destIP;
		}

		// Set destination IP address & port
		public void setDestAddress(string addr)
		{
			string[] parms = addr.Split(';');
			m_destIP = new IPAddr(parms[0], UInt32.Parse(parms[1]));
		}

		// Set whether we are using headers for tunneling
		public void setUseHeaders(bool uh)
		{
			m_useHeaders = uh;
		}


		/*********************/
		/* Private Interface */
		/*********************/


		/********/
		/* Data */
		/********/

		// The tunnel ID
		private TunnelID					m_tid;

		// HTTP version we're using over this tunnel
		private string						m_httpVer;

		// Address of ultimate destination
		private IPAddr						m_destIP;

		// Are we using headers for tunneling?
		private bool						m_useHeaders;

		// Proxy/serverauth settings for this tunnel
		private ProxySettings			m_proxyCfg;
		private ServerAuthSettings		m_serverAuthCfg;

		// Tunnel tuning
		private TuneSettings				m_tuning;

	}

}	// namespace

