/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Uniquely identify an individual tunneled connection.
 *
 * $Id: TunnelID.cs,v 1.2 2013/02/17 10:53:48 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System.Text;
	using System.Threading;

	public class TunnelID {

		// Construction
		public TunnelID()
		{
			m_id = 0;
			m_cliProcID = 0;
			m_tunnelerID = new TunnelerID();
		}

		public TunnelID(Globals glob, ClientSideConfig cfg)
		{
			m_id = getNextID();
			m_cliProcID = (uint)glob.cfg().getPid();
			m_tunnelerID = new TunnelerID(cfg.getTunnelerID());
		}

		public TunnelID(string id)
		{
			m_id = 0;
			m_cliProcID = 0;
			m_tunnelerID = new TunnelerID(id);

			// Split composite ID
			char[] sep = new char[]{';'};
			string[] parts = id.Split(sep);

			// Got everything we need?
			if(parts.Length != 6) {
				throw new InvalidTunnelID("Invalid tunnel ID string: " + id);
			}

			// Construct individual ID values
			m_id = uint.Parse(parts[0]);
			m_cliProcID = uint.Parse(parts[1]);
		}

		public TunnelID(TunnelID that)
		{
			m_id = that.m_id;
			m_cliProcID = that.m_cliProcID;
			m_tunnelerID = new TunnelerID(that.m_tunnelerID);
		}


		/********************/
		/* Public Interface */
		/********************/


		// Create a representation of ourselves as a string
		public override string ToString()
		{
			StringBuilder val = new StringBuilder();
			val.AppendFormat("{0};{1};{2}", m_id, m_cliProcID, m_tunnelerID);
			return val.ToString();
		}

		// Operators
		public static bool operator==(TunnelID t1, TunnelID t2)
		{
			return (t1.Equals(t2));
		}
		public static bool operator!=(TunnelID t1, TunnelID t2)
		{
			return !(t1.Equals(t2));
		}

		// As recommended, also override Equals() & GetHashCode()
		public override bool Equals(object that)
		{
			bool res = false;

			if(that != null && that is TunnelID) {
				TunnelID tid = (TunnelID)that;

				// Everything must match
				res = (m_id == tid.m_id);
				res &= (m_cliProcID == tid.m_cliProcID);
				res &= (m_tunnelerID == tid.m_tunnelerID);
			}

			return res;
		}

		public override int GetHashCode()
		{
			return ToString().GetHashCode();
		}


		/*************/
		/* Accessors */
		/*************/

		public uint getID()
		{
			return m_id;
		}

		public uint getProcID()
		{
			return m_cliProcID;
		}

		public TunnelerID getTunnelerID()
		{
			return m_tunnelerID;
		}

		public Enums.Protocol getProtocol()
		{
			return m_tunnelerID.getProtocol();
		}


		/*********************/
		/* Private Interface */
		/*********************/


		/***********/
		/* Methods */
		/***********/

		// Get next unique ID
		private static uint getNextID()
		{
			lock(c_idCountMtx) {
				if(++c_idCount == 0)
					c_idCount = 1;
				return c_idCount;
			}
		}


		/********/
		/* Data */
		/********/

		// Unique tunnel count ID from creating client process
		private uint					m_id;

		// Client tunneler process ID
		private uint					m_cliProcID;

		// Tunneler ID
		private TunnelerID			m_tunnelerID;

		// A counter for creating unique IDs
		private static uint			c_idCount = 0;
		private static Mutex			c_idCountMtx = new Mutex();

	}

}	// namespace


