/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Base class for client and server tunnel controllers.
 *
 * $Id: Tunneler.cs,v 1.9 2013/04/20 09:53:09 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Collections;
	using System.Text;
	using System.Threading;

	public abstract class Tunneler : EventReceiver {

		// Construction
		public Tunneler(Globals glob, Server server) : base(glob)
		{
			m_glob.DPRINTV("Tunneler::Tunneler()");

			m_server = server;
			m_tunnels = new ArrayList();
			m_closingTunnels = new ArrayList();
			m_connections = new ArrayList();
			m_closingConnections = new ArrayList();
			m_connectionMgr = new ConnectionManager(m_glob);
			m_lastTunParmsCheckTime = Time.now();

			// Schedule initial events
			scheduleEvent(new CheckAckEvent(m_glob, new Time(ACK_CHECK_PERIOD,0), this));
			scheduleEvent(new ConnectionManagementEvent(m_glob, new Time(0,CONN_MANAGE_PERIOD), this));
			scheduleEvent(new TunnelManagementEvent(m_glob, new Time(TUN_MANAGE_PERIOD,0), this));
			scheduleEvent(new LeakTunnelErrorEvent(m_glob, new Time(LEAK_ERROR_PERIOD,0), this));

			#if DEBUG_TUN
			scheduleEvent(new AuditEvent(m_glob, new Time(AUDIT_PERIOD,0), this));
			#endif
		}

		// Destruction
		public virtual void destroyed()
		{
			m_glob.DPRINTV("Tunneler::~Tunneler()");

			// Clear up list of Tunnels
			lock(m_tunnels) {
				while(m_tunnels.Count > 0) {
					Tunnel tun = (Tunnel)m_tunnels[0];
					tun.destroyed();
					m_tunnels.RemoveAt(0);
				}
			}

			// Clear up connections
			lock(m_connections) {
				lock(m_closingConnections) {
					// Destroy connection manager
					m_connectionMgr.destroyed();
	
					// Clear up list of connections
					while(m_connections.Count > 0) {
						Connection conn = (Connection)m_connections[0];
						conn.destroyed();
						m_connections.RemoveAt(0);
					}

					// Clear up list of closing connections
					while(m_closingConnections.Count > 0) {
						Connection conn = (Connection)m_closingConnections[0];
						conn.destroyed();
						m_closingConnections.RemoveAt(0);
					}
				}
			}
		}


		/*************/
		/* Accessors */
		/*************/

		// Get owning server
		public Server getServer()
		{
			return m_server;
		}


		// Get owning server configuration
		public ClientSideConfig getClientSideConfig()
		{
			// Get basic config and cast it
			SideConfig cfg = m_server.getConfig();
			if(cfg is ClientSideConfig) {
				return (ClientSideConfig)cfg;
			}
			else {
				throw new ConfigError("Unable to retrieve client side config");
			}
		}

		public ServerSideConfig getServerSideConfig()
		{
			// Get basic config and cast it
			SideConfig cfg = m_server.getConfig();
			if(cfg is ServerSideConfig) {
				return (ServerSideConfig)cfg;
			}
			else {
				throw new ConfigError("Unable to retrieve server side config");
			}
		}


		// Get number of active managed tunnels
		public int getNoActiveTunnels()
		{
			lock(m_tunnels) {
				return m_tunnels.Count;
			}
		}

		// Get IDs of all active managed tunnels
		public void getActiveTunnelIDs(ArrayList ids)
		{
			// Start with an empty list
			ids.Clear();

			lock(m_tunnels) {
				// Iterate over all active tunnels
				for(int tit = 0; tit < m_tunnels.Count; ++tit) {
					Tunnel tun = (Tunnel)m_tunnels[tit];
					ids.Add(tun.getTid());
				}
			}
		}


		// Get connection counts for some tunnel
		public int getTunnelConnectionCount(TunnelID tid, ArrayList crit)
		{
			int res = 0;

			lock(m_tunnels) {
				// Lookup tunnel
				Tunnel tun = getTunFromPred(new PredTunID(tid));
				if(tun != null) {
					// Get connection count for specified criteria
					res = tun.noConnFromSel(crit);
				}
			}

			return res;
		}

		// Get stats for some tunnel
		public TunnelStats getTunnelStats(TunnelID tid)
		{
			TunnelStats stats = null;

			lock(m_tunnels) {
				// Lookup tunnel
				Tunnel tun = getTunFromPred(new PredTunID(tid));
				if(tun != null) {
					// Get stats
					bool isTest = tun.isTestTunnel();
					string stateStr = "Unknown";
					string tstateStr = "Unknown";

					switch(tun.getState()) {
						case(Tunnel.TunnelState.CONNECT):
						case(Tunnel.TunnelState.AUTHREQ):
							stateStr = "Connecting";
							break;
						case(Tunnel.TunnelState.CLOSING):
						case(Tunnel.TunnelState.CLOSED):
							stateStr = "Closing";
							break;
						case(Tunnel.TunnelState.RECONNECT):
							stateStr = "Reconnecting";
							break;
						case(Tunnel.TunnelState.CONNECTED):
							stateStr = "Connected";
							break;
						default:
							break;
					}

					switch(tun.getTestState()) {
						case(Tunnel.TunnelTestState.NO_TESTING):
							tstateStr = "Not Testing";
							break;
						case(Tunnel.TunnelTestState.BEGIN_TESTING):
							tstateStr = "Beginning Testing";
							break;
						case(Tunnel.TunnelTestState.CLIENT_TESTING):
							tstateStr = "Client Testing";
							break;
						case(Tunnel.TunnelTestState.SERVER_TESTING):
							tstateStr = "Server Testing";
							break;
						case(Tunnel.TunnelTestState.COMPLETED_TESTING):
							tstateStr = "Testing Complete";
							break;
						default:
							break;
					}

					TunnelConfig tunCfg = tun.getConfig();
					string destAddr = tunCfg.getDestAddress();
					uint rawBytesIn = tun.getRawBytesIn();
					uint rawBytesOut = tun.getRawBytesOut();
					uint httpBytesIn = tun.getHTTPBytesIn();
					uint httpBytesOut = tun.getHTTPBytesOut();
					User user = tun.getUserDetails();
					ArrayList notifs = tun.getRecvNotifMsgs();

					// Create stats
					stats = new TunnelStats(tid, isTest, stateStr, tstateStr, destAddr,
						rawBytesIn, rawBytesOut, httpBytesIn, httpBytesOut, user, notifs);
				}
			}

			return stats;
		}


		/************/
		/* Mutators */
		/************/

		// Initiate a shutdown of the named tunnel
		public void initiateTunnelShutdown(TunnelID tid)
		{
			try {
				// Lookup tunnel
				lock(m_tunnels) {
					Tunnel tun = getTunFromPred(new PredTunID(tid));
					if(tun != null) {
						// Schedule a shutdown for this tunnel
						m_glob.DPRINTI("Tunneler::initiateTunnelShutdown() - " + tid.ToString());
						scheduleEvent(new ShutdownTunnelEvent(m_glob, new Time(0,0), this, tid, false));
					}
					else {
						StringBuilder err = new StringBuilder();
						err.Append("Tunneler::initiateTunnelShutdown() - ");
						err.AppendFormat("Unknown tunnel {0}", tid.ToString());
						m_glob.log().warn(err.ToString());
					}
				}
			}
			catch(LookupManagementError e) {
				m_glob.log().warn("Error initiating tunnel shutdown: " + e.Message);
			}
		}

		// Initiate a shutdown of all tunnels
		public void initiateAllTunnelsShutdown()
		{
			lock(m_tunnels) {
				// Iterate over all tunnels
				for(int tit = 0; tit < m_tunnels.Count; ++tit) {
					Tunnel tun = (Tunnel)m_tunnels[tit];

					// Initiate shutdown on this tunnel
					TunnelID tid = tun.getTid();
					initiateTunnelShutdown(tid);
				}
			}
		}


		// Enable/disable testing based on current config setting
		public virtual void setServerTesting()
		{
			// Do noting in base class
		}


		// Add connection error to list of scheduled events
		public void scheduleConnectionError(uint connID, Tunnel.TunnelErrorType type)
		{
			// Cancel any existing event
			ConnectionErrorEvent evnt = cancelAndGetEvent<ConnectionErrorEvent, uint>(connID);
			if(evnt != null) {
				// Add error to this event
				evnt.addError(type);
			}
			else {
				// Create new event
				evnt = new ConnectionErrorEvent(m_glob, connID, type, this);
			}

			// Schedule new event
			scheduleEvent(evnt);
		}


		/**********/
		/* Events */
		/**********/

		// Notification of a new socket connection
		public abstract void newConnectionArrived(Sock insock);

		// Notification that a tunnel has insufficient left/right connections to be viable
		public abstract void unviableTunnel(TunnelID tid);

		// Notification of new tunneled/raw data arrival
		public abstract void rawDataArrived(uint connID, TunnelID tid, RawBuff raw);

		// Check we have received an ack for some sent message
		public abstract void checkAckReceived();


		// Notification of new HTTP message arrival
		public virtual void messageArrived(uint connID, HTTPMsg msg)
		{
			#if DEBUG_TUN
			StringBuilder nfo = new StringBuilder();
			nfo.AppendFormat("Tunneler::messageArrived() - On connection {0}", connID);
			m_glob.DPRINTV(nfo.ToString());
			#endif

			lock(m_connections) {
				lock(m_tunnels) {
					// Lookup connection
					int cit = getConnectionByID(connID);

					// Let connection know it has received a new HTTP msg
					HTTPConnection httpConn = (HTTPConnection)m_connections[cit];
					httpConn.msgRecvd();

					// Log any new notifications here incase msg never makes it to Tunnel
					logMsgNotifs(msg);
				}
			}
		}


		// Notification of some new default tunnel parameters
		public virtual void newTunnelParms(TunnelParms parms, TunnelerID tid)
		{
			// Nothing to do in base class
		}


		// Notification of a new UDP 'connection'
		public virtual void newUDPConnectionArrived(Sock insock)
		{
			m_glob.DPRINTI("Tunneler::newUDPConnectionArrived()");

			// Error to receive this in a non-overriden class
			throw new CommsError("UDP tunneling not supported in this tunnel type");
		}


		// Notification that a tunnel needs to be cleared down
		public virtual void shutdownTunnel(TunnelID tid, bool otherEndInitiated)
		{
			#if DEBUG_TUN
			StringBuilder dbg = new StringBuilder();
			dbg.AppendFormat("Tunneler::shutdownTunnel() - {0}", tid.ToString());
			dbg.AppendFormat(" OEI {0}", otherEndInitiated);
			m_glob.DPRINTI(dbg.ToString());
			#endif

			try {
				// Lookup tunnel
				lock(m_tunnels) {
					Tunnel tun = getTunFromPred(new PredTunID(tid));
					if(tun != null) {
						// Check for test tunnel
						checkScheduleServerTest(tun);

						// How was this tunnel shutdown?
						if(otherEndInitiated) {
							// Other end not expecting anything else from us, so just destroy tunnel
							destroyTunnel(tun);
						}
						else {
							// Try to shutdown tunnel cleanly
							tunnelClosing(tun);
						}
					}
					else {
						StringBuilder err = new StringBuilder();
						err.AppendFormat("Tunneler::shutdownTunnel() - Unknown tunnel {0}", tid.ToString());
						m_glob.log().warn(err.ToString());
					}
				}
			}
			catch(LookupManagementError e) {
				m_glob.log().warn("Error shutting down tunnel: " + e.Message);
			}
		}


		// Notification that a tunnel has been shutdown (destroyed)
		public virtual void tunnelDestroyed(TunnelID tid)
		{
			#if DEBUG_TUN
			StringBuilder dbg = new StringBuilder();
			dbg.AppendFormat("Tunneler::tunnelDestroyed() - {0}", tid.ToString());
			m_glob.DPRINTV(dbg.ToString());
			#endif
		}


		// Notification of a poll timeout on a connection
		public virtual void pollTimeout(TunnelID tid)
		{
			// Not interested in base class
		}


		// Check UDP datagram timeout
		public virtual void checkDatagramTimeout(TunnelID tid)
		{
			m_glob.DPRINTI("Tunneler::checkDatagramTimeout() - " + tid.ToString());

			try {
				// Lookup tunnel
				lock(m_tunnels) {
					Tunnel tun = getTunFromPred(new PredTunID(tid));
					if(tun != null) {
						// UDP silence threshold exceeded?
						if(tun.rawRecvTimeout() && tun.tunRecvTimeout()) {
							m_glob.DPRINTV("UDP silence threshold exceeded for tunnel " +
								tid.ToString());

							// Try to shutdown tunnel cleanly
							tunnelClosing(tun);
						}
						else {
							// Schedule next UDP silent check event
							scheduleEvent(new UDPSilentEvent(m_glob, this, tid));
						}
					}
					else {
						StringBuilder err = new StringBuilder();
						err.AppendFormat("Tunneler::checkDatagramTimeout() - Unknown tunnel {0}",
							tid.ToString());
						m_glob.log().warn(err.ToString());
					}
				}
			}
			catch(LookupManagementError e) {
				m_glob.log().warn("Error processing UDP datagram timeout: " + e.Message);
			}
		}


		// Notification that a connection has changed state
		public virtual void connectionStateChange(uint connID,
			Connection.ConnectionState state)
		{
			#if DEBUG_TUN
			StringBuilder nfo = new StringBuilder();
			nfo.AppendFormat("Tunneler::connectionStateChange () - connection {0}", connID);
			nfo.AppendFormat(" state: {0}", state.ToString());
			m_glob.DPRINTI(nfo.ToString());
			#endif

			try {
				// Lookup connection
				lock(m_connections) {
					lock(m_tunnels) {
						int cit = getConnectionByID(connID);
						Connection conn = (Connection)m_connections[cit];

						// Do whatever the state change requires
						if(state == Connection.ConnectionState.CONNECTED) {
							conn.tellTunnelsAboutConnect();
						}
					}
				}
			}
			catch(LookupManagementError e) {
				m_glob.log().warn("Error handling connection state change: " + e.Message);
			}
		}


		// Notification that a tunnel has changed state
		public virtual void tunnelStateChanged(TunnelID tid, Tunnel.TunnelState state)
		{
			// Not interested in base class
		}


		// Notification of a socket connect failure
		public virtual void connectResult(ConnectResultEvent result)
		{
			m_glob.DPRINTV("Tunneler::connectResult()");

			try {
				// Lookup connection
				lock(m_connections) {
					lock(m_tunnels) {
						int cit = getConnectionByID(result.getConnectionID());
						Connection conn = (Connection)m_connections[cit];

						// Let connection manager clear up after connect attempt
						m_connectionMgr.delProxyConnect(conn);

						// Connect failed?
						ConnectResultEvent.ConnectResultType resType = result.getType();
						if(resType != ConnectResultEvent.ConnectResultType.SUCCESS) {
							// Record failed connect attempt
							conn.incConnectFailedCount();

							// Was this a test connection?
							Tunnel tun = getTunFromPred(new PredTunConnection(conn));
							if(tun != null && tun.isTestTunnel()) {
								// Keep trying
								m_glob.DPRINTV("Test connection failed");
								uint failedCount = conn.getConnectFailedCount();
								uint reconnDelay = MIN_TEST_RECONNECT_DELAY * failedCount;
								if(reconnDelay > MAX_TEST_RECONNECT_DELAY)
									reconnDelay = MAX_TEST_RECONNECT_DELAY;
								m_connectionMgr.addProxyConnect(conn, new Time((int)reconnDelay,0),
									result.getAddr(), result.getProtocol(), result.getSSL());
							}
							else {
								// Shall we try again?
								bool retry = true;
								uint failedCount = conn.getConnectFailedCount();
								uint maxAttempts = 0, reconnDelay = 0;
								getReconnectParms(ref maxAttempts, ref reconnDelay, failedCount);
								if(resType == ConnectResultEvent.ConnectResultType.FIREWALLED ||
								failedCount >= maxAttempts) {
									retry = false;
								}

								if(retry) {
									// Schedule reconnect attempt
									m_connectionMgr.addProxyConnect(conn, new Time((int)reconnDelay,0),
										result.getAddr(), result.getProtocol(), result.getSSL());
								}
								else {
									// Tell tunnel connect attempt failed
									if(tun != null)
										tun.connectAttemptFailed(conn);

									// Clear up connection
									destroyConnectionByIndex(cit);
								}
							}
						}
					}
				}
			}
			catch(LookupManagementError e) {
				m_glob.log().warn("Error handling connect result: " + e.Message);
			}
		}


		// Notification of error(s) on a socket connection
		public virtual void connectionErrors(uint connID, ArrayList errors)
		{
			#if DEBUG_TUN
			StringBuilder nfo = new StringBuilder();
			nfo.AppendFormat("Tunneler::connectionError() - Connection id {0}", connID);
			m_glob.DPRINTI(nfo.ToString());
			#endif

			try {
				// Lookup connection (also lock tunnels to prevent race with tunnel shutdown)
				lock(m_connections) {
					lock(m_tunnels) {
						int index = getConnectionByID(connID);
						Connection conn = (Connection)m_connections[index];

						// Record errors
						for(int eit = 0; eit < errors.Count; ++eit) {
							Tunnel.TunnelErrorType error = (Tunnel.TunnelErrorType)errors[eit];

							#if DEBUG_TUN
							StringBuilder nfo2 = new StringBuilder();
							nfo2.AppendFormat("Connection error type {0}", error.ToString());
							m_glob.DPRINTI(nfo2.ToString());
							#endif

							conn.tellTunnelsAboutError(error);
						}

						// Move connection to the closing list
						connectionClosing(index);
					}
				}
			}
			catch(LookupManagementError e) {
				m_glob.log().warn("Error handling connection error: " + e.Message);
			}
		}


		// Notification that a connection needs to be cleared down
		public virtual void shutdownConnection(uint connID)
		{
			#if DEBUG_TUN
			StringBuilder nfo = new StringBuilder();
			nfo.AppendFormat("Tunneler::shutdownConnection() - connection {0}", connID);
			m_glob.DPRINTI(nfo.ToString());
			#endif

			try {
				// Move connection to the closing list
				int index = getConnectionByID(connID);
				connectionClosing(index);
			}
			catch(LookupManagementError e) {
				m_glob.log().warn("Error shutting down connection: " + e.Message);
			}
		}


		// Notification we should force a send of a message down a tunnel/connection
		public virtual void sendTunnelMsg(TunnelID tid, uint connID, bool retry)
		{
			m_glob.DPRINTV("Tunneler::sendTunnelMsg() - " + tid.ToString());

			try {
				// Lookup connection
				lock(m_connections) {
					int index = getConnectionByID(connID);
					Connection conn = (Connection)m_connections[index];

					// Lookup tunnel
					lock(m_tunnels) {
						Tunnel tun = getTunFromPred(new PredTunID(tid));
						if(tun != null) {
							// Force send
							tun.sendTunnelMsg(retry, conn);
						}
						else {
							StringBuilder wrn = new StringBuilder();
							wrn.Append("Failed to force send of tunnel msg ");
							wrn.AppendFormat("down unknown tunnel {0}", tid.ToString());
							m_glob.log().warn(wrn.ToString());
						}
					}
				}
			}
			catch(LookupManagementError e) {
				StringBuilder wrn = new StringBuilder();
				wrn.Append("Error forcing send of tunnel msg down connection ");
				wrn.AppendFormat("{0}: {1}", connID,  e.Message);
				m_glob.log().warn(wrn.ToString());
			}
		}


		// Notification we should try to flush a connection
		public virtual void flushConnection(uint connID, bool retry)
		{
			#if DEBUG_TUN
			StringBuilder nfo = new StringBuilder();
			nfo.AppendFormat("Tunneler::flushConnection() - Connection id {0}", connID);
			m_glob.DPRINTI(nfo.ToString());
			#endif

			try {
				// Lookup connection
				lock(m_connections) {
					int index = getConnectionByID(connID);
					Connection conn = (Connection)m_connections[index];

					try {
						// Flush connection
						conn.flush(retry);
					}
					catch(CommsError e) {
						m_glob.log().warn("Error flushing connection: " + e.Message);
					}
				}
			}
			catch(LookupManagementError e) {
				m_glob.log().warn("Error flushing connection: " + e.Message);
			}
		}


		// Notification that a tunnel should decrease its leaky error count
		public virtual void decreaseTunnelsErrorCount()
		{
			lock(m_tunnels) {
				// Iterate over all tunnels
				for(int tit = 0; tit < m_tunnels.Count; ++tit) {
					// Decrease error count
					Tunnel tun = (Tunnel)m_tunnels[tit];
					tun.decreaseErrorCount();
				}
			}

			// Schedule next error leak event
			scheduleEvent(new LeakTunnelErrorEvent(m_glob, new Time(LEAK_ERROR_PERIOD,0),
				this));
		}


		// Notification that authentication has succeeded for a tunnel
		public virtual void authenticationSuccess(TunnelID tid)
		{
			// Nothing to do in base class
		}


		// Notification that we should perform connection manangement
		public virtual void performConnectionManagement()
		{
			lock(m_closingConnections) {
				#if DEBUG_TUN
				if(m_closingConnections.Count > 0) {
					StringBuilder dbg = new StringBuilder();
					dbg.Append("Tunneler::performConnectionManagement() - Managing ");
					dbg.AppendFormat("{0} connections", m_closingConnections.Count);
					m_glob.DPRINTV(dbg.ToString());
				}
				#endif

				// Look for any connections that can be shutdown
				PredConnShutdown pred = new PredConnShutdown();
				bool keepLooking = false;
				do {
					keepLooking = false;
					for(int cit = 0; cit < m_closingConnections.Count; ++cit) {
						Connection conn = (Connection)m_closingConnections[cit];
						if(pred.match(conn)) {
							#if DEBUG_TUN
							StringBuilder dbg = new StringBuilder();
							dbg.AppendFormat("Shutting down connection {0}", conn.getID());
							m_glob.DPRINTV(dbg.ToString());
							#endif

							// Let connection manager know
							m_connectionMgr.delConnection(conn);

							// Remove & destroy connection
							m_closingConnections.RemoveAt(cit);
							conn.destroyed();
							keepLooking = true;
							break;
						}
					}
				} while(keepLooking);

				// Schedule next event
				scheduleEvent(new ConnectionManagementEvent(m_glob, new Time(0,CONN_MANAGE_PERIOD), this));
			}
		}


		// Notification that we should perform tunnel manangement
		public virtual void performTunnelManagement()
		{
			lock(m_tunnels) {
				lock(m_closingTunnels) {
					#if DEBUG_TUN
					if(m_closingTunnels.Count > 0) {
						StringBuilder dbg = new StringBuilder();
						dbg.Append("Tunneler::performTunnelManagement() - Managing ");
						dbg.AppendFormat("{0} tunnels", m_closingTunnels.Count);
						m_glob.DPRINTV(dbg.ToString());
					}
					#endif

					// Find all tunnels from closing list that can be shutdown
					ArrayList sdTunnels = new ArrayList();
					for(int ctit = 0; ctit < m_closingTunnels.Count; ++ctit) {
						// Can this tunnel be shutdown now?
						Tunnel tun = (Tunnel)m_closingTunnels[ctit];
						if(tun.canBeShutdown())
							sdTunnels.Add(tun);
					}

					// Destroy all tunnels that can be shutdown
					for(int ctit = 0; ctit < sdTunnels.Count; ++ctit) {
						try {
							Tunnel tun = (Tunnel)sdTunnels[ctit];
							destroyTunnel(tun);
						}
						catch(LookupManagementError e) {
							m_glob.log().warn("Error destroying tunnel: " + e.Message);
						}
					}

					// Schedule next event
					scheduleEvent(new TunnelManagementEvent(m_glob,
						new Time(TUN_MANAGE_PERIOD,0), this));
				}
			}
		}


		// Perform some audit
		public virtual void audit()
		{
			m_glob.DPRINTV("Tunneler::audit()");

			// Currently active connections
			lock(m_connections) {
				string str = "Active connection count: " + m_connections.Count;
				m_glob.DPRINTV(str);
			}

			// Closing connections
			lock(m_closingConnections) {
				string str = "Closing connection count: " + m_closingConnections.Count;
				m_glob.DPRINTV(str);
			}

			// Currently active tunnels
			lock(m_tunnels) {
				string str = "Active tunnels count: " + m_tunnels.Count;
				m_glob.DPRINTV(str);
			}

			// Closing tunnels
			lock(m_closingTunnels) {
				string str = "Closing tunnels count: " + m_closingTunnels.Count;
				m_glob.DPRINTV(str);
			}

			// ConnectionManager audit
			lock(m_connections) {
				lock(m_closingConnections) {
					m_connectionMgr.audit();
				}
			}

			// Schedule next audit
			scheduleEvent(new AuditEvent(m_glob, new Time(AUDIT_PERIOD,0), this));
		}


		/***********************/
		/* Protected Interface */
		/***********************/


		/*************/
		/* Constants */
		/*************/

		// Time between test reconnection attempts
		protected const int			MIN_TEST_RECONNECT_DELAY	= 2;
		protected const int			MAX_TEST_RECONNECT_DELAY	= 10;

		// Time between acknowledgement checks
		protected const int			ACK_CHECK_PERIOD				= 1;

		// Time between management events
		protected const int			CONN_MANAGE_PERIOD			= 500;
		protected const int			TUN_MANAGE_PERIOD				= 2;

		// Period in seconds between leaky bucket error count decreases
		protected const int			LEAK_ERROR_PERIOD				= 2;

		// Time between audits
		protected const int			AUDIT_PERIOD					= 60 * 10;

		// Control when we clear out old tunnel parms
		#if DEBUG
		protected const int			OLD_TUN_PARMS_AGE				= 60 * 5;
		#else
		protected const int			OLD_TUN_PARMS_AGE				= 60 * 60;
		#endif
		protected const int			CHECK_TUN_PARMS_PERIOD		= OLD_TUN_PARMS_AGE / 10;


		/**************/
		/* Predicates */
		/**************/

		// Match tunnel from an ID
		protected class PredTunID : Predicate<Tunnel> {
			public PredTunID(TunnelID tid) { m_tid = tid; }
			public bool match(Tunnel tun) {
				return (tun.getTid() == m_tid);
			}
			private TunnelID m_tid;
		}

		// Match tunnel from a connection
		protected class PredTunConnection : Predicate<Tunnel> {
			public PredTunConnection(Connection conn) { m_conn = conn; }
			public bool match(Tunnel tun) {
				return (tun.ownsConnection(m_conn));
			}
			private Connection m_conn;
		}

		// Match test tunnels
		protected class PredTestTun : Predicate<Tunnel> {
			public PredTestTun() {}
			public bool match(Tunnel tun) {
				return (tun.isTestTunnel());
			}
		}

		// Match test tunnel from a tunneler ID
		protected class PredTestTunID : Predicate<Tunnel> {
			public PredTestTunID(TunnelerID tid) { m_tid = tid; }
			public bool match(Tunnel tun) {
				return (tun.isTestTunnel() && tun.getTid().getTunnelerID() == m_tid);
			}
			private TunnelerID m_tid;
		}

		// Match connections that can be shutdown
		protected class PredConnShutdown : Predicate<Connection> {
			public PredConnShutdown() {}
			public bool match(Connection conn) {
				bool res = false;
				try { res = conn.canBeShutdown(); }
				catch(CommsError) {
					// If we can't even check this connection, go ahead & kill it
					res = true;
				}
				return res;
			}
		}


		/***********/
		/* Methods */
		/***********/

		// Get reconnection attempt parameters
		protected abstract void getReconnectParms(ref uint maxConnectAttempts,
		ref uint reconnectDelay, uint failedAttempts);


		// Check whether we should relaunch a server test - Caller holds mutex
		protected virtual void checkScheduleServerTest(Tunnel tun)
		{
			// Nothing to do in base class
		}


		// Add a new connection to managed list
		protected void addConnection(Connection conn)
		{
			lock(m_connections) {
				// Add to list
				m_connections.Add(conn);

				// Let connection manager know
				m_connectionMgr.addConnection(conn);
			}
		}


		// Destroy a managed connection by index - Caller holds mutex
		protected void destroyConnectionByIndex(int index)
		{
			// Get connection
			Connection conn = (Connection)m_connections[index];

			#if DEBUG_TUN
			StringBuilder dbg = new StringBuilder();
			dbg.AppendFormat("Tunneler::destroyConnection() - connection id {0}", conn.getID());
			m_glob.DPRINTV(dbg.ToString());
			#endif

			// Let connection manager know
			m_connectionMgr.delConnection(conn);

			// Remove from list
			m_connections.RemoveAt(index);

			// Cancel any events that might still be pending for this connection
			cancelEvents<ConnectionErrorEvent, uint>(conn.getID());
			cancelEvents<SendMsgEvent, uint>(conn.getID());
			cancelEvents<FlushConnectionEvent, uint>(conn.getID());

			// Tell connection it is no longer required
			conn.destroyed();
		}


		// Make a managed connection a candidate for shutting down
		protected void connectionClosing(int index)
		{
			// Get connection
			Connection conn = (Connection)m_connections[index];

			#if DEBUG_TUN
			StringBuilder dbg = new StringBuilder();
			dbg.Append("Tunneler::connectionClosing() - Moving connection id ");
			dbg.AppendFormat("{0} to closing list", conn.getID());
			m_glob.DPRINTV(dbg.ToString());
			#endif

			// Remove from list
			m_connections.RemoveAt(index);

			// Cancel any events that might still be pending for this connection
			cancelEvents<ConnectionErrorEvent, uint>(conn.getID());
			cancelEvents<SendMsgEvent, uint>(conn.getID());
			cancelEvents<FlushConnectionEvent, uint>(conn.getID());

			// Let connection know it is now inactive
			conn.setInactive();

			// Place connection on closing list
			lock(m_closingConnections) {
				m_closingConnections.Add(conn);
			}
		}


		// Lookup managed connection index by ID - Caller holds mutex
		protected int getConnectionByID(uint id)
		{
			// Find connection in managed list
			for(int i = 0; i < m_connections.Count; ++i) {
				Connection conn = (Connection)m_connections[i];
				if(conn.getID() == id)
					return i;
			}

			// Not found
			StringBuilder err = new StringBuilder();
			err.AppendFormat("Unknown connection id {0}", id);
			throw new LookupManagementError(err.ToString());
		}


		// Destroy a managed tunnel - Caller holds tunnels mutex
		protected void destroyTunnel(Tunnel tun)
		{
			// Get tunnel ID
			TunnelID tid = tun.getTid();
			m_glob.DPRINTV("Destroying tunnel " + tid.ToString());

			// Find tunnel in managed list
			int index = m_tunnels.IndexOf(tun);
			if(index >= 0) {
				// Remove from list
				m_tunnels.RemoveAt(index);
			}
			else {
				// Not found
				StringBuilder err = new StringBuilder();
				err.AppendFormat("Tunneler::destroyTunnel() - Unknown tunnel {0}",
					tid.ToString());
				throw new LookupManagementError(err.ToString());
			}

			// Cancel any events that might still be pending for this tunnel
			cancelEvents<PollTimeoutEvent, TunnelID>(tid);
			cancelEvents<TunnelStateEvent, TunnelID>(tid);
			cancelEvents<ShutdownTunnelEvent, TunnelID>(tid);
			cancelEvents<TunnelUnviableEvent, TunnelID>(tid);
			cancelEvents<UDPSilentEvent, TunnelID>(tid);
			cancelEvents<AuthSuccessEvent, TunnelID>(tid);

			// Remove tunnel ID from closing list if it's there
			lock(m_closingTunnels) {
				index = m_closingTunnels.IndexOf(tun);
				if(index >= 0) {
					m_closingTunnels.RemoveAt(index);
				}
			}

			// Now destroy tunnel itself
			tun.destroyed();

			// Schedule notification
			scheduleEvent(new TunnelDestroyedEvent(m_glob, this, tid));
		}


		// Tag a tunnel as closing - Caller holds tunnels mutex
		protected void tunnelClosing(Tunnel tun)
		{
			m_glob.DPRINTV("Tunneler::tunnelClosing() - " + tun.getTidStr());

			// Find tunnel in managed list
			int tit = m_tunnels.IndexOf(tun);
			if(tit >= 0) {
				// Check we don't already know this tunnel is closing
				lock(m_closingTunnels) {
					int ctit = m_closingTunnels.IndexOf(tun);
					if(ctit == -1) {
						// Copy tunnel to closing list
						m_closingTunnels.Add(tun);

						// Tell tunnel it is closing
						tun.closing();
					}
				}
			}
			else {
				// Not found
				StringBuilder err = new StringBuilder();
				err.AppendFormat("Tunneler::tunnelClosing() - Unknown tunnel {0}",
					tun.getTidStr());
				throw new LookupManagementError(err.ToString());
			}
		}


		// Locate a tunnel matching some predicate
		protected Tunnel getTunFromPred<P>(P pred) where P : Predicate<Tunnel>
		{
			lock(m_tunnels) {
				for(int i = 0; i < m_tunnels.Count; ++i) {
					Tunnel tun = (Tunnel)m_tunnels[i];
					if(pred.match(tun))
						return tun;
				}
			}

			// Not found
			return null;
		}

		// Locate all tunnels matching some predicate
		protected ArrayList getAllTunsFromPred<P>(P pred) where P : Predicate<Tunnel>
		{
			ArrayList res = new ArrayList();

			lock(m_tunnels) {
				for(int i = 0; i < m_tunnels.Count; ++i) {
					Tunnel tun = (Tunnel)m_tunnels[i];
					if(pred.match(tun))
						res.Add(tun);
				}
			}

			return res;
		}


		/********/
		/* Data */
		/********/

		// Our owning server
		protected Server							m_server;

		// All the currently active tunnels
		protected ArrayList						m_tunnels;

		// Tunnels we are trying to shutdown
		protected ArrayList						m_closingTunnels;

		// All active connections
		protected ArrayList						m_connections;

		// Connections we are trying to shutdown
		protected ArrayList						m_closingConnections;

		// A connection manager
		protected ConnectionManager			m_connectionMgr;

		// Time of last old tunnel parms check
		protected Time								m_lastTunParmsCheckTime;


		/*********************/
		/* Private Interface */
		/*********************/


		/***********/
		/* Methods */
		/***********/

		// Log notifications from other end
		private void logMsgNotifs(HTTPMsg msg)
		{
			// Somewhere to build log msgs
			StringBuilder ntferr = new StringBuilder();

			// Iterate over newly arrived notifications
			HTTPMsgParms parms = msg.getParms();
			ArrayList notifs = parms.getNotifs();
			for(int nit = 0; nit < notifs.Count; ++nit) {
				string notif = (string)notifs[nit];

				// Full notif or start of partial notif?
				if(notif.EndsWith(":")) {
					ntferr.AppendFormat("{0} ", notif);
				}
				else {
					ntferr.Append(notif);
					m_glob.log().notif(ntferr.ToString());
					ntferr = new StringBuilder();
				}
			}
		}

	}

}	// namespace

