/* (C) 2007 Richard Mills / The Linux Academy Ltd.
 *
 * Uniquely identify an individual tunneler.
 *
 * $Id: TunnelerID.cs,v 1.2 2013/02/17 10:53:48 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System.Text;
	using System.Threading;

	public class TunnelerID {

		// Construction
		public TunnelerID()
		{
			m_cliName = "";
			m_cliIP = "";
			m_cliPort = 0;
			m_protocol = Enums.Protocol.TCP;
		}

		public TunnelerID(Globals glob, SideConfig cfg)
		{
			m_cliName = glob.cfg().getUniqueHostName();
			m_cliIP = cfg.getListenIP().getHost();
			m_cliPort = (uint)cfg.getListenIP().getPort();
			m_protocol = cfg.getProtocol();

			// Are we for an ephemeral tunneler?
			if(m_cliIP == "-") {
				// Ensure uniquness by replacing port with unique ID
				m_cliPort = getNextID();
			}
		}

		public TunnelerID(string id)
		{
			m_cliName = "";
			m_cliIP = "";
			m_cliPort = 0;
			m_protocol = Enums.Protocol.TCP;

			// Split composite ID
			char[] sep = new char[]{';'};
			string[] parts = id.Split(sep);

			// Got everything we need?
			if(parts.Length != 6) {
				throw new InvalidTunnelerID("Invalid tunneler ID string: " + id);
			}

			// Construct individual ID values
			m_cliName = parts[2];
			m_cliIP = parts[3];
			m_cliPort = uint.Parse(parts[4]);
			m_protocol = (Enums.Protocol)(uint.Parse(parts[5]));
		}

		public TunnelerID(TunnelerID that)
		{
			m_cliName = that.m_cliName;
			m_cliIP = that.m_cliIP;
			m_cliPort = that.m_cliPort;
			m_protocol = that.m_protocol;
		}


		/********************/
		/* Public Interface */
		/********************/


		// Create a representation of ourselves as a string
		public override string ToString()
		{
			StringBuilder val = new StringBuilder();
			val.AppendFormat("{0};{1};{2};{3}", m_cliName, m_cliIP, m_cliPort, (uint)m_protocol);
			return val.ToString();
		}

		// Operators
		public static bool operator==(TunnelerID t1, TunnelerID t2)
		{
			return (t1.Equals(t2));
		}
		public static bool operator!=(TunnelerID t1, TunnelerID t2)
		{
			return !(t1.Equals(t2));
		}

		// As recommended, also override Equals() & GetHashCode()
		public override bool Equals(object that)
		{
			bool res = false;

			if(that != null && that is TunnelerID) {
				TunnelerID tid = (TunnelerID)that;

				// Everything must match
				res = (m_cliName == tid.m_cliName);
				res &= (m_cliIP == tid.m_cliIP);
				res &= (m_cliPort == tid.m_cliPort);
				res &= (m_protocol == tid.m_protocol);
			}

			return res;
		}

		public override int GetHashCode()
		{
			return ToString().GetHashCode();
		}


		/*************/
		/* Accessors */
		/*************/

		public string getClientName()
		{
			return m_cliName;
		}

		public Enums.Protocol getProtocol()
		{
			return m_protocol;
		}


		/*********************/
		/* Private Interface */
		/*********************/


		/***********/
		/* Methods */
		/***********/

		// Get next unique ID
		private static uint getNextID()
		{
			lock(c_idCountMtx) {
				if(++c_idCount == 0)
					c_idCount = 65535;
				return c_idCount;
			}
		}


		/********/
		/* Data */
		/********/

		// Client host name
		private string					m_cliName;

		// Client IP and port
		private string					m_cliIP;
		private uint					m_cliPort;

		// Protocol being tunneled
		private Enums.Protocol		m_protocol;

		// A counter for creating unique IDs
		private static uint			c_idCount = 65535;
		private static Mutex			c_idCountMtx = new Mutex();

	}

}	// namespace


