/* (C) 2007 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate a UDP socket.
 *
 * $Id: UDPSock.cs,v 1.5 2013/02/17 10:53:48 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{

	using System;
	using System.Net;
	using System.Net.Sockets;

	public class UDPSock : Sock {

		// Construction
		public UDPSock(Globals glob, bool blocking) : base(glob, blocking)
		{
			m_sockAddr = null;
			m_ownSocket = true;
		}

		public UDPSock(Globals glob, Socket fd, string cliAddr)
		: base(glob, fd, cliAddr)
		{
			m_sockAddr = new IPEndPoint(IPAddress.Any, 0);
			m_ownSocket = false;
		}


		// We are no longer required
		public override void destroyed()
		{
			m_glob.DPRINTV("UDPSock::~UDPSock()");

			lock(this) {
				// Should we close?
				if(m_ownSocket) {
					try {
						// Close socket descriptor
						close();
					}
					catch(SockError e) {
						m_glob.log().info(e.Message);
					}
				}
			}

			// Base class
			base.destroyed();
		}


		/********************/
		/* Public Interface */
		/********************/

		// Wait for a new incoming TCP connection
		public override void newConnection(Tunneler tunneler)
		{
			lock(this) {
				// Is this really a new UDP connection, or simply a continuation?
				if(tunneler.getNoActiveTunnels() == 0) {
					// Accept connection
					Sock connectedSock = accept();

					try {
						// Tell tunneler
						if(tunneler != null)
							tunneler.newUDPConnectionArrived(connectedSock);
						else
							throw new SockError("No tunneler to handle new UDP connection", 0);
					}
					catch(TunnelError e) {
						m_glob.log().info(e.Message);
						connectedSock.destroyed();
					}
				}
			}
		}


		// Set socket listening
		public override void listen()
		{
			// Do nothing for UDP
		}


		// Wait for a new incoming connection
		public override Sock accept()
		{
			// Return connected UDP socket
			return new UDPSock(m_glob, m_fd, "Unknown");
		}


		// Establish connection out
		public override void connect(IPAddr ipaddr)
		{
			// Create underlying socket
			createSockFD(ipaddr, SocketType.Dgram, ProtocolType.Udp);

			// Get address details
			string addr = ipaddr.getHost();
			uint port = ipaddr.getPort();

			lock(this) {
				// Lookup destination address
				IPAddress destaddr = null;
				if(!Utils.getIPByName(addr, ref destaddr)) {
					throw new ConnectError("Failed to resolve UDP server address " + addr);
				}

				// Setup socket address structure
				m_sockAddr = new IPEndPoint(destaddr, (int)port);
			}
		}


		// Read from socket
		public override int read(byte[] buff)
		{
			int noread = 0;

			lock(this) {
				// Try the read
				try {
					noread = m_fd.ReceiveFrom(buff, ref m_sockAddr);
				}
				catch(SocketException e) {
					throw new SockError("Read error from UDP socket: " + e.Message,
						e.SocketErrorCode);
				}
				catch(Exception e) {
					throw new SockError("Read error from UDP socket: " + e.Message, 0);
				}

				// Get connected client address
				IPEndPoint connectedEndPoint = (IPEndPoint)m_sockAddr;
				m_connectedClientAddr = connectedEndPoint.Address.ToString();
			}

			return noread;
		}


		// Write to socket
		public override int write(RawBuff buff)
		{
			int nowritten = 0;

			lock(this) {
				// Try the write
				try {
					nowritten = m_fd.SendTo(buff.getData(), m_sockAddr);
				}
				catch(SocketException e) {
					throw new SockError("Write error to UDP socket: " + e.Message,
						e.SocketErrorCode);
				}
				catch(Exception e) {
					throw new SockError("Write error to UDP socket: " + e.Message, 0);
				}
			}

			return nowritten;
		}


		// Get the result of a non-blocking connect attempt
		public override void getConnectResult(ref bool resultAvailable, ref bool connected)
		{
			// Do nothing for UDP
			resultAvailable = connected = true;
		}



		// Bind socket to an address
		public override void bind(IPAddr ipaddr)
		{
			// Create underlying socket
			createSockFD(ipaddr, SocketType.Dgram, ProtocolType.Udp);

			// Do the bind
			base.bind(ipaddr);
		}


		/*********************/
		/* Private Interface */
		/*********************/


		/********/
		/* Data */
		/********/

		// Socket address
		private EndPoint			m_sockAddr;

		// Do we own the socket descriptor?
		private bool				m_ownSocket;

	}

}	// namespace

