/* (C) 2008 Richard Mills / The Linux Academy Ltd.
 *
 * A server to accept incoming UDP socket tunnel connections.
 *
 * $Id: UDPSockTunServer.cs,v 1.5 2013/02/17 10:53:48 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Net.Sockets;
	using System.Text;
	using System.Threading;

	public class UDPSockTunServer : SockTunServer {

		// Construction
		public UDPSockTunServer(Globals glob, SockTunCallback userCallback)
		: base(glob, userCallback)
		{
			m_glob.DPRINTV("UDPSockTunServer::UDPSockTunServer()");

			// Create blocking UDP socket
			m_prot = Enums.Protocol.UDP;
			m_sock = new UDPSock(m_glob, true);

			// Bind to ephemeral port on the loopback interface
			m_sock.bind(m_listenIP);

			#if DEBUG_TUN
			StringBuilder dbg = new StringBuilder();
			dbg.AppendFormat("UDPSockTunServer bound to {0}", m_listenIP);
			m_glob.DPRINTV(dbg.ToString());
			#endif

			// Launch server to process UDP connections
			m_thread = new Thread(new ThreadStart(startServer));
			m_thread.Start();
		}

		// We are no longer required
		public override void destroyed()
		{
			m_glob.DPRINTV("UDPSockTunServer::~UDPSockTunServer()");

			// Take mutexes
			Monitor.Enter(m_serverMtx);
			Monitor.Enter(m_shutDownMtx);

			// Signal we are no longer running
			m_shutDown = true;

			// Wakeup server thread & wait for it to finish
			if(m_thread != null) {
				if(m_thread.IsAlive) {
					Monitor.Pulse(m_serverMtx);
					Monitor.Exit(m_serverMtx);
					Monitor.Wait(m_shutDownMtx);
					Monitor.Exit(m_shutDownMtx);
					m_thread.Join();
				}
				else {
					Monitor.Exit(m_serverMtx);
					Monitor.Exit(m_shutDownMtx);
				}
				m_thread = null;
			}
			else {
				Monitor.Exit(m_serverMtx);
				Monitor.Exit(m_shutDownMtx);
			}

			// Base class
			base.destroyed();
		}


		/*********************/
		/* Private interface */
		/*********************/


		/***********/
		/* Methods */
		/***********/

		// Wait for arriving connections - Thread entry point
		private void startServer()
		{
			m_glob.DPRINTI("UDPSockTunServer::startServer() - Waiting for connection");

			// Take mutex
			Monitor.Enter(m_serverMtx);
			bool accepted = false;

			while(!m_shutDown && !accepted) {
				try {
					// Poll socket until we think it's worth trying to accept()
					bool pollError = false, pollHup = false;
					if(m_sock.pollRead(0, ref pollError, ref pollHup)) {
						// Accept pending connection
						Sock connectedSock = m_sock.accept();

						// Pass connected socket back to user
						callbackToUser(connectedSock);
						accepted = true;
					}
					else {
						// Poll timeout, wait a while before retrying
						TimeSpan pollTime = new TimeSpan(0,0,0,0,POLL_TIMEOUT);
						Monitor.Wait(m_serverMtx, pollTime);
					}
				}
				catch(SockError) {
					// General error, just carry on
				}
			}

			// Signal we have finished
			Monitor.Exit(m_serverMtx);
			Monitor.Enter(m_shutDownMtx);
			Monitor.Pulse(m_shutDownMtx);
			Monitor.Exit(m_shutDownMtx);

			m_glob.DPRINTI("UDPSockTunServer closing down");
		}

	}

}	// namespace


