/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Encapsulate a user of a tunnel.
 *
 * $Id: User.cs,v 1.1 2008/07/25 15:22:38 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Collections;
	using System.Text;

	public class User : ICloneable {

		// Construction
		public User(Globals glob)
		{
			m_glob = glob;
			m_alwaysAuthenticated = true;
			m_name = "";
			m_passwd = "";
			m_quotas = new Quotas();
			m_sentData = new ArrayList();
			m_recvData = new ArrayList();
			m_activeTunnels = new ArrayList();
			m_authenticatedTunnels = new ArrayList();
			m_keepCached = false;
		}

		public User(Globals glob, bool auth, string name, string passwd, Quotas quotas,
		bool keepCached)
		{
			m_glob = glob;
			m_alwaysAuthenticated = auth;
			m_name = name;
			m_passwd = passwd;
			m_quotas = quotas;
			m_sentData = new ArrayList();
			m_recvData = new ArrayList();
			m_activeTunnels = new ArrayList();
			m_authenticatedTunnels = new ArrayList();
			m_keepCached = keepCached;
		}

		// Clone
		public object Clone()
		{
			User clone = (User)this.MemberwiseClone();
			clone.m_quotas = (Quotas)m_quotas.Clone();
			clone.m_sentData = (ArrayList)m_sentData.Clone();
			clone.m_recvData = (ArrayList)m_recvData.Clone();
			m_activeTunnels = (ArrayList)m_activeTunnels.Clone();
			m_authenticatedTunnels = (ArrayList)m_authenticatedTunnels.Clone();
			return clone;
		}


		// Create a representation of ourselves as a string
		public override string ToString()
		{
			string name = getName();
			bool authenticated = getAuthenticated(null);
			uint bandwidth = getQuotas().getBandwidth();

			StringBuilder str = new StringBuilder();
			str.AppendFormat("Name: {0}, ", name);
			str.AppendFormat("Authenticated: {0}, ", authenticated);
			str.Append("Bandwidth: ");
			if(bandwidth == 0)
				str.Append("Unlimited");
			else
				str.AppendFormat("{0}", bandwidth);

			return str.ToString();
		}


		// Equality/inequality
		public static bool operator==(User u1, User u2)
		{
			return (u1.Equals(u2));
		}
		public static bool operator!=(User u1, User u2)
		{
			return !(u1.Equals(u2));
		}

		// As recommended, also override Equals() & GetHashCode()
		public override bool Equals(object that)
		{
			bool res = false;

			if(that != null && that is User) {
				User user = (User)that;

				// Everything must match
				res = (m_alwaysAuthenticated == user.m_alwaysAuthenticated);
				res &= (m_name == user.m_name);
				res &= (m_passwd == user.m_passwd);
				res &= (m_quotas == user.m_quotas);
				res &= (m_keepCached == user.m_keepCached);
			}

			return res;
		}

		public override int GetHashCode()
		{
			return ToString().GetHashCode();
		}


		/********************/
		/* Public Interface */
		/********************/


		/*************/
		/* Accessors */
		/*************/

		// Is this user authenticated?
		public bool getAuthenticated(Tunnel tun)
		{
			bool auth = false;

			// Allways authenticated?
			if(m_alwaysAuthenticated) {
				auth = true;
			}
			else {
				// For a specific tunnel?
				if(tun != null) {
					// Find tunnel index
					for(int tit = 0; tit < m_activeTunnels.Count; ++tit) {
						if((Tunnel)m_activeTunnels[tit] == tun) {
							// Cross reference to authenticated tunnels
							auth = (bool)m_authenticatedTunnels[tit];
						}
					}
				}
				else {
					// Non-specific tunnel, so are we authenticated for all tunnels?
					auth = true;
					for(int tit = 0; tit < m_activeTunnels.Count; ++tit) {
						auth &= (bool)m_authenticatedTunnels[tit];
					}
				}
			}

			return auth;
		}

		// Get users login name
		public string getName()
		{
			return m_name;
		}

		// Get users passwd
		public string getPasswd()
		{
			return m_passwd;
		}

		// Get users quotas
		public Quotas getQuotas()
		{
			return m_quotas;
		}

		// Get number of active tunnels for this user
		public uint getNoActiveTunnels()
		{
			return (uint)m_activeTunnels.Count;
		}

		// Get whether this user is using the specified tunnel
		public bool usingTunnel(Tunnel tun)
		{
			int tit = m_activeTunnels.IndexOf(tun);
			return (tit >= 0);
		}

		// Get whether this user should remain cached
		public bool keepCached()
		{
			return (m_keepCached || getNoActiveTunnels() > 0);
		}

		// Get current average upload bandwidth
		public uint getAvgUpBandwidth()
		{
			// Get current average bandwidth usage
			uint avgBandwidth = 0;
			uint totBandwidth = 0;
			uint avgMillis = 0;
			getAvgBandwidth(m_sentData, ref avgBandwidth, ref totBandwidth, ref avgMillis);
			return avgBandwidth;
		}

		// Get current average download bandwidth
		public uint getAvgDownBandwidth()
		{
			// Get current average bandwidth usage
			uint avgBandwidth = 0;
			uint totBandwidth = 0;
			uint avgMillis = 0;
			getAvgBandwidth(m_recvData, ref avgBandwidth, ref totBandwidth, ref avgMillis);
			return avgBandwidth;
		}


		/************/
		/* Mutators */
		/************/

		// Set whether this user is authenticated
		public void setAuthenticated(bool auth, Tunnel tun)
		{
			// For a specific tunnel?
			if(tun != null) {
				// If we're not auth'd for this tunnel we're not always auth'd
				if(!auth)
					m_alwaysAuthenticated = false;

				// Find tunnel index
				for(int tit = 0; tit < m_activeTunnels.Count; ++tit) {
					if((Tunnel)m_activeTunnels[tit] == tun) {
						// Cross reference to authenticated tunnels
						m_authenticatedTunnels[tit] = auth;
					}
				}
			}
			else {
				// Non-specific tunnel, so set always auth'd
				m_alwaysAuthenticated = auth;
			}
		}

		// Set users quotas
		public void setQuotas(Quotas quotas)
		{
			m_quotas = quotas;
		}


		// Update all our pertinant details
		public void setAll(string name, string passwd, bool auth, Quotas quotas, Tunnel newTun)
		{
			m_name = name;
			m_passwd = passwd;
			m_quotas = quotas;
			addTunnel(newTun, auth);
		}


		// Add a new tunnel to list of those carried
		public void addTunnel(Tunnel tun, bool auth)
		{
			// Check we don't already have this tunnel
			int tit = m_activeTunnels.IndexOf(tun);
			if(tit == -1) {
				m_glob.DPRINTV("User::addTunnel() - User " + m_name + " owns tunnel " +
					tun.getTidStr());
				m_activeTunnels.Add(tun);
				m_authenticatedTunnels.Add(auth);
			}
			else {
				m_glob.DPRINTV("User::addTunnel() - User " + m_name +
					" already knows about tunnel " + tun.getTidStr());
			}

			#if DEBUG_TUN
			StringBuilder nfo = new StringBuilder();
			nfo.AppendFormat("User::addTunnel() - User {0} now owns {1} active tunnels",
				m_name, getNoActiveTunnels());
			m_glob.DPRINTI(nfo.ToString());
			#endif
		}

		// Remove a tunnel from list of those carried
		public void removeTunnel(Tunnel tun)
		{
			// Check we know about this tunnel
			int tindex = m_activeTunnels.IndexOf(tun);
			if(tindex >= 0) {
				// Remove tunnel details
				m_activeTunnels.RemoveAt(tindex);
				m_authenticatedTunnels.RemoveAt(tindex);

				m_glob.DPRINTV("User::removeTunnel() - User " + m_name +
					" no longer owns tunnel " + tun.getTidStr());
			}
			else {
				m_glob.DPRINTV("User::removeTunnel() - Unknown tunnel " + tun.getTidStr());
			}

			#if DEBUG_TUN
			StringBuilder nfo = new StringBuilder();
			nfo.AppendFormat("User::removeTunnel() - User {0} now owns {1} active tunnels",
				m_name, getNoActiveTunnels());
			m_glob.DPRINTI(nfo.ToString());
			#endif
		}


		/*************/
		/* Utilities */
		/*************/

		// Get the maximum no. of bytes our quotas will let us send
		public uint getBytesCanSend()
		{
			return getBytesBelowQuota(m_sentData);
		}

		// We have just sent some no. of bytes
		public void sentBytes(uint bytes)
		{
			if(bytes > 0) {
				// Record new bytes sent
				trackBandwidth(bytes, m_sentData);
			}
		}


		// Do our quotas allow receiving some no. of bytes?
		public bool canRecvBytes(uint bytes)
		{
			bool res = true;

			if(bytes > 0) {
				// Record new bytes received
				trackBandwidth(bytes, m_recvData);

				// Does average received bandwidth exceed quota?
				res = bandwidthWithinQuota(m_recvData);
			}

			return res;
		}


		// Tunnel creation according to our quotas
		public bool canCreateNewTunnel()
		{
			bool res = true;

			// Get tunnel count limit (if set) for this user
			uint maxTunnels = m_quotas.getNoTunnels();
			if(maxTunnels > 0) {
				// User below limit?
				res = (getNoActiveTunnels() < maxTunnels);
			}

			return res;
		}


		// Force a shutdown for all our tunnels
		public void closeTunnels()
		{
			m_glob.DPRINTV("User::closeTunnels() - User " + m_name);

			for(int tit = 0; tit < m_activeTunnels.Count; ++tit) {
				Tunnel tun = (Tunnel)m_activeTunnels[tit];
				TunnelID tid = tun.getConfig().getTunnelID();
				m_glob.DPRINTV("Scheduling close for tunnel " + tid.ToString());

				// Schedule a tunnel shutdown
				Tunneler tunneler = tun.getTunneler();
				tunneler.scheduleEvent(new ShutdownTunnelEvent(m_glob, new Time(0,0),
					tunneler, tid, false));
			}
		}


		/*********************/
		/* Private Interface */
		/*********************/


		/*********/
		/* Types */
		/*********/

		// Track amount of data transfered
		private class DataTrack {
			// Construction/destruction
			public DataTrack(Time tm, uint bytes)
			{
				m_tm = tm;
				m_bytes = bytes;
			}

			// Data
			public Time		m_tm;
			public uint		m_bytes;
		}


		/*************/
		/* Constants */
		/*************/

		// How long do we track sent/recvd data?
		private const uint			DATA_TRACK_PERIOD		= 10;

		// What percentage above quotas do we tolerate?
		private const uint			QUOTA_TOLERENCE		= 10;

		// Maximum bandwidth usage we will allow
		private const uint			MAX_BANDWIDTH			= 65536;


		/***********/
		/* Methods */
		/***********/

		// Update bandwidth usage tracking
		private void trackBandwidth(uint bytes, ArrayList track)
		{
			// Drop any old data points
			trimBandwidthTrack(track);

			// Append new track point
			DataTrack tp = new DataTrack(Time.now(), bytes);
			track.Add(tp);
		}

		// Discard any tracked bandwidth outside our range
		private void trimBandwidthTrack(ArrayList track)
		{
			if(track.Count > 0) {
				// Get cutoff time
				Time cutoff = Time.now();
				cutoff.subInterval((int)DATA_TRACK_PERIOD, 0);

				// Find any tracked items outside cutoff
				bool keepLooking = true;
				while(keepLooking && track.Count > 0) {
					DataTrack tit = (DataTrack)track[0];
					if(tit.m_tm.getTimeVal() < cutoff.getTimeVal()) {
						track.RemoveAt(0);
					}
					else {
						keepLooking = false;
					}
				}
			}
		}


		// Get average bandwidth used
		private void getAvgBandwidth(ArrayList track, ref uint avg, ref uint tot,
		ref uint millis)
		{
			avg = tot = millis = 0;

			// Trim tracked data
			trimBandwidthTrack(track);

			// Calculate average
			if(track.Count > 0) {
				// Get time over which to calculate average
				Time now = Time.now();
				Time start = ((DataTrack)track[0]).m_tm;
				millis = (uint)(now.getTotalMillis() - start.getTotalMillis());

				// Do we have long enough sample time for estimate?
				if(millis > 0) {
					// Get total bandwidth
					for(int bit = 0; bit < track.Count; ++bit) {
						tot += ((DataTrack)track[bit]).m_bytes;
					}

					// Average bandwidth in bytes/s
					double avgd = ((double)tot * 1000) / (double)millis;
					avg = (uint)avgd;
				}
			}
		}


		// Is the users bandwidth usage within their quota?
		private bool bandwidthWithinQuota(ArrayList track)
		{
			bool res = true;

			// Convert quota bandwidth to bytes/sec
			uint quotaBandwidth = m_quotas.getBandwidth() * 1024;
			if(quotaBandwidth > 0) {
				// Get current average bandwidth usage
				uint avgBandwidth = 0;
				uint totBandwidth = 0;
				uint avgMillis = 0;
				getAvgBandwidth(track, ref avgBandwidth, ref totBandwidth, ref avgMillis);

				// Is usage within tolerated range?
				uint maxBandwidth = quotaBandwidth;
				maxBandwidth += (quotaBandwidth * QUOTA_TOLERENCE) / 100;
				res = (avgBandwidth <= maxBandwidth);

				#if DEBUG_TUN
				StringBuilder dbg = new StringBuilder();
				dbg.AppendFormat("bandwidthWithinQuota: Max = {0}", maxBandwidth);
				dbg.AppendFormat(", Avg = {0}", avgBandwidth);
				m_glob.DPRINTV(dbg.ToString());
				#endif
			}

			return res;
		}


		// Get how far we are currently below bandwidth quota
		private uint getBytesBelowQuota(ArrayList track)
		{
			uint res = MAX_BANDWIDTH;

			// Get quota gandwidth in bytes/sec
			uint quotaBandwidth = m_quotas.getBandwidth() * 1024;

			if(quotaBandwidth > 0) {
				res = quotaBandwidth;

				// Get current average rate
				uint avgBandwidth = 0;
				uint totBandwidth = 0;
				uint avgMillis = 0;
				getAvgBandwidth(track, ref avgBandwidth, ref totBandwidth, ref avgMillis);
				if(avgMillis > 0) {
					// Currently below quota?
					if(avgBandwidth < quotaBandwidth) {
						// How much can we send to take us up to quota?
						res = ((quotaBandwidth * avgMillis) / 1000) - totBandwidth;

						// Still apply absolute maximum
						if(res > MAX_BANDWIDTH)
							res = MAX_BANDWIDTH;
					}
					else {
						// Currently at or over quota, so throttle back
						res = 0;
					}
				}

				#if DEBUG_TUN
				StringBuilder dbg = new StringBuilder();
				dbg.AppendFormat("getBytesBelowQuota: Max = {0}", quotaBandwidth);
				dbg.AppendFormat(", Allowed = {0}", res);
				m_glob.DPRINTV(dbg.ToString());
				#endif
			}

			return res;
		}


		/********/
		/* Data */
		/********/

		// Globals
		private Globals				m_glob;

		// Are we always authenticated?
		private bool					m_alwaysAuthenticated;

		// Name/passwd
		private string					m_name;
		private string					m_passwd;

		// Quotas
		private Quotas					m_quotas;

		// Track bandwidth usage
		private ArrayList				m_sentData;
		private ArrayList				m_recvData;

		// Our current active tunnels
		private ArrayList				m_activeTunnels;
		private ArrayList				m_authenticatedTunnels;

		// Should this user always be kept cached?
		private bool					m_keepCached;

	}

}	// namespace

