/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Place to hold useful utility methods.
 *
 * $Id: Utils.cs,v 1.4 2013/03/24 09:55:58 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Collections;
	using System.IO;
	using System.Net;
	using System.Net.Sockets;
	using System.Security.Cryptography;
	using System.Text;

	public class Utils {

		// Convert to hex string
		public static string ToHex(byte[] data)
		{
			StringBuilder dmp = new StringBuilder();
			dmp.Append("0x");
			foreach(byte b in data) {
				dmp.Append(b.ToString("x2").ToLower());
			}

			return dmp.ToString();
		}


		// Extract the value of a parm="value" pair
		public static bool getParmValue(ArrayList args, string parm, ref string res)
		{
			bool ok = false;
			if(args != null) {
				for(int i = 0; i < args.Count; ++i) {
					string argParm = (string)args[i];
					if(String.Compare(argParm, 0, parm, 0, parm.Length) == 0) {
						// Split into id=value
						string[] idvalue = argParm.Split('=');
						if(idvalue.Length == 2) {
							res = idvalue[1];
							ok = true;

							// Strip any quotes if found
							char[] trim = { '"' };
							res = res.TrimStart(trim);
							res = res.TrimEnd(trim);

							break;
						}
					}
				}
			}

			return ok;
		}


		// UUEncode some of our data
		public static string UUEncode(string src)
		{
			int length = src.Length;
			StringBuilder uue = new StringBuilder((int)(length * 2));

			for(int i = 0; i < length; i += 3) {
				if(i < length - 2) {
					// Got 3 bytes to encode
					byte a = (byte)src[i];
					byte b = (byte)src[i + 1];
					byte c = (byte)src[i + 2];
					uue.Append((char)(((a >> 2) & 0x3F) + 35));
					uue.Append((char)((((a << 4) | ((b >> 4) & 0x0F)) & 0x3F) + 35));
					uue.Append((char)((((b << 2) | ((c >> 6) & 0x03)) & 0x3F) + 35));
					uue.Append((char)((c & 0x3F) + 35));
				}
				else if(i < length - 1) {
					// Only got 2 bytes to encode
					byte a = (byte)src[i];
					byte b = (byte)src[i + 1];
					uue.Append((char)(((a >> 2) & 0x3F) + 35));
					uue.Append((char)((((a << 4) | ((b >> 4) & 0x0F)) & 0x3F) + 35));
					uue.Append((char)(((b << 2) & 0x3F) + 35));
					uue.Append('~');
				}
				else {
					// Only got 1 byte to encode
					byte a = (byte)src[i];
					uue.Append((char)(((a >> 2) & 0x3F) + 35));
					uue.Append((char)(((a << 4) & 0x3F) + 35));
					uue.Append("~~");
				}
			}

			return uue.ToString();
		}

		// UUDecode some string
		public static RawBuff UUDecode(string src)
		{
			int uuestart = 0;
			int uueend = src.Length;
			byte[] uud = new byte[src.Length];
			uint uudlength = 0;

			for(int uue = uuestart; uue < uueend; uue += 4) {
				// Allways need first 2 chars
				int a = (int)src[uue] - 35;
				int b = (int)src[uue + 1] - 35;
				uud[uudlength++] = (byte)((a << 2) | (b >> 4));

				if(src[uue + 2] != '~') {
					// Got at least another byte of encoded data
					int c = (int)src[uue + 2] - 35;
					uud[uudlength++] = (byte)(((b & 0x0F) << 4) | ((c & 0x3C) >> 2));

					if(src[uue + 3] != '~') {
						// Got the full 3 bytes of encoded data
						int d = (int)src[uue + 3] - 35;
						uud[uudlength++] = (byte)(((c & 0x03) << 6) | d);
					}
				}
			}

			// Construct RawBuff
			return new RawBuff(uud, uudlength);
		}


		// Base64 encode some string
		public static string Base64Encode(string src)
		{
			int length = src.Length;
			StringBuilder uue = new StringBuilder((int)(length * 2));

			for(int i = 0; i < length; i += 3) {
				if(i < length - 2) {
					// Got 3 bytes to encode
					byte a = (byte)src[i];
					byte b = (byte)src[i + 1];
					byte c = (byte)src[i + 2];
					uue.Append(c_base64Table[((a >> 2) & 0x3F)]);
					uue.Append(c_base64Table[(((a << 4) | ((b >> 4) & 0x0F)) & 0x3F)]);
					uue.Append(c_base64Table[(((b << 2) | ((c >> 6) & 0x03)) & 0x3F)]);
					uue.Append(c_base64Table[(c & 0x3F)]);
				}
				else if(i < length - 1) {
					// Only got 2 bytes to encode
					byte a = (byte)src[i];
					byte b = (byte)src[i + 1];
					uue.Append(c_base64Table[((a >> 2) & 0x3F)]);
					uue.Append(c_base64Table[(((a << 4) | ((b >> 4) & 0x0F)) & 0x3F)]);
					uue.Append(c_base64Table[((b << 2) & 0x3F)]);
					uue.Append('=');
				}
				else {
					// Only got 1 byte to encode
					byte a = (byte)src[i];
					uue.Append(c_base64Table[((a >> 2) & 0x3F)]);
					uue.Append(c_base64Table[((a << 4) & 0x3F)]);
					uue.Append("==");
				}
			}

			// Construct string
			return uue.ToString();
		}

		// Base64 decode some string
		public static RawBuff Base64Decode(string src)
		{
			int start = 0;
			int end = src.Length;
			byte[] decoded = new byte[src.Length];
			uint dlength = 0;

			for(int enc = start; enc < end; enc += 4) {
				// Allways need first 2 chars
				byte a = base64DecodeChar(src[enc]);
				byte b = base64DecodeChar(src[enc + 1]);
				decoded[dlength++] = (byte)((a << 2) | (b >> 4));

				if(src[enc + 2] != '=') {
					// Got at least another byte of encoded data
					byte c = base64DecodeChar(src[enc + 2]);
					decoded[dlength++] = (byte)(((b & 0x0F) << 4) | ((c & 0x3C) >> 2));

					if(src[enc + 3] != '=') {
						// Got the full 3 bytes of encoded data
						byte d = base64DecodeChar(src[enc + 3]);
						decoded[dlength++] = (byte)(((c & 0x03) << 6) | d);
					}
				}
			}

			// Construct RawBuff
			return new RawBuff(decoded, dlength);
		}


		// Modified Base64 encode some string
		public static string ModBase64Encode(string src)
		{
			int length = src.Length;
			StringBuilder uue = new StringBuilder((int)(length * 2));

			for(int i = 0; i < length; i += 3) {
				if(i < length - 2) {
					// Got 3 bytes to encode
					byte a = (byte)src[i];
					byte b = (byte)src[i + 1];
					byte c = (byte)src[i + 2];
					uue.Append(c_modBase64Table[((a >> 2) & 0x3F)]);
					uue.Append(c_modBase64Table[(((a << 4) | ((b >> 4) & 0x0F)) & 0x3F)]);
					uue.Append(c_modBase64Table[(((b << 2) | ((c >> 6) & 0x03)) & 0x3F)]);
					uue.Append(c_modBase64Table[(c & 0x3F)]);
				}
				else if(i < length - 1) {
					// Only got 2 bytes to encode
					byte a = (byte)src[i];
					byte b = (byte)src[i + 1];
					uue.Append(c_modBase64Table[((a >> 2) & 0x3F)]);
					uue.Append(c_modBase64Table[(((a << 4) | ((b >> 4) & 0x0F)) & 0x3F)]);
					uue.Append(c_modBase64Table[((b << 2) & 0x3F)]);
					uue.Append('*');
				}
				else {
					// Only got 1 byte to encode
					byte a = (byte)src[i];
					uue.Append(c_modBase64Table[((a >> 2) & 0x3F)]);
					uue.Append(c_modBase64Table[((a << 4) & 0x3F)]);
					uue.Append("**");
				}
			}

			// Construct string
			return uue.ToString();
		}

		// Modified Base64 decode some string
		public static RawBuff ModBase64Decode(string src)
		{
			int start = 0;
			int end = src.Length;
			byte[] decoded = new byte[src.Length];
			uint dlength = 0;

			for(int enc = start; enc < end; enc += 4) {
				// Allways need first 2 chars
				byte a = modBase64DecodeChar(src[enc]);
				byte b = modBase64DecodeChar(src[enc + 1]);
				decoded[dlength++] = (byte)((a << 2) | (b >> 4));

				if(src[enc + 2] != '*') {
					// Got at least another byte of encoded data
					byte c = modBase64DecodeChar(src[enc + 2]);
					decoded[dlength++] = (byte)(((b & 0x0F) << 4) | ((c & 0x3C) >> 2));

					if(src[enc + 3] != '*') {
						// Got the full 3 bytes of encoded data
						byte d = modBase64DecodeChar(src[enc + 3]);
						decoded[dlength++] = (byte)(((c & 0x03) << 6) | d);
					}
				}
			}

			// Construct RawBuff
			return new RawBuff(decoded, dlength);
		}


		// URL encode some string
		public static string URLEncode(string src)
		{
			URLCode coder = new URLCode();
			return coder.encode(src);
		}

		// URL decode some string
		public static string URLDecode(string src)
		{
			URLCode coder = new URLCode();
			return coder.decode(src);
		}


		// Resolve an address to a string IP address
		public static string resolveIP(string addr)
		{
			string ip = "";

			try {
				// Get address from address info
				AddrInfo info = getAddrInfo(addr);
				ip = info.getAddrStr();
			}
			catch(AddrLookupError) {}

			return ip;
		}


		// Translate a socket address to its hostname or numeric IP address
		public static string getNameInfo(EndPoint sockAddr)
		{
			return getIPInfo(sockAddr);
		}

		// Translate a socket address to its numeric IP address
		public static string getIPInfo(EndPoint sockAddr)
		{
			IPEndPoint ipEndPoint = (IPEndPoint)sockAddr;
			return ipEndPoint.Address.ToString();
		}


		// Resolve an address to a host entry
		public static int getHostByName(string addr, ref IPHostEntry host)
		{
			int res = 0;

			try {
				// Lookup address
				host = Dns.GetHostEntry(addr);

				// Did we get it?
				if(host.AddressList.Length == 0) {
					res = -1;
				}
			}
			catch(Exception) {
				res = -1;
			}

			return res;
		}


		// Resolve a name to an IP address
		public static bool getIPByName(string addr, ref IPAddress ip)
		{
			bool ok = false;

			try {
				// Lookup IP
				IPAddress[] ips = Dns.GetHostAddresses(addr);

				// Did we get one?
				if(ips.Length > 0) {
					ip = ips[0];
					ok = true;
				}
			}
			catch(Exception) {}

			return ok;
		}


		// Get address info
		public static AddrInfo getAddrInfo(string addr)
		{
			bool ok = false;

			// Address info to populate & return
			AddrInfo addressinfo = new AddrInfo(addr);

			try {
				// Try to get address info
				IPAddress address = null;
				if(getIPByName(addr, ref address)) {
					// Set family
					addressinfo.setFamily(address.AddressFamily);

					// Resolve address as a string
					addressinfo.setAddrStr(address.ToString());

					// Got everything
					ok = true;
				}
			}
			catch(Exception) {} 

			// All ok?
			if(!ok) {
				throw new AddrLookupError("Failed to get address info for host " + addr);
			}

			// Return address info
			return addressinfo;
		}


		// Get peer address from socket descriptor
		public static string getPeerName(Socket sockfd)
		{
			string peerName = "Unknown";

			try {
				// Get socket address
				IPEndPoint endPoint = (IPEndPoint)sockfd.RemoteEndPoint;
				peerName = endPoint.Address.ToString();
			}
			catch(Exception) {}

			return peerName;
		}

		// Get local port socket is using
		public static int getSockPort(Socket sockfd)
		{
			int port = 0;

			try {
				// Get socket address
				IPEndPoint endPoint = (IPEndPoint)sockfd.LocalEndPoint;
				port = endPoint.Port;
			}
			catch(Exception) {}

			return port;
		}


		// DES encrypt/decrypt
		public static void DESCrypt(byte[] input, byte[] key, byte[] output, bool encrypt)
		{
			// Setup DES key
			DESCryptoServiceProvider desAlg = new DESCryptoServiceProvider();
			desAlg.Key = key;
			desAlg.Mode = CipherMode.ECB;
			desAlg.Padding = PaddingMode.None;

			// Encrypt/decrypt
			if(encrypt) {
				MemoryStream ms = new MemoryStream();
				CryptoStream desStream = new CryptoStream(ms, desAlg.CreateEncryptor(), CryptoStreamMode.Write);
				BinaryWriter bw = new BinaryWriter(desStream);
				bw.Write(input);
				bw.Close();
				desStream.Close();
				Array.Copy(ms.ToArray(), output, 8);
				ms.Close();
			}
			else {
				MemoryStream ms = new MemoryStream(input);
				CryptoStream desStream = new CryptoStream(ms, desAlg.CreateDecryptor(), CryptoStreamMode.Read);
				BinaryReader br = new BinaryReader(desStream);
				br.Read(output, 0, 8);
				br.Close();
				desStream.Close();
				ms.Close();
			}
		}

		// Create an 8 byte DES key from 7 byte input
		public static void CreateDESKey(byte[] input, byte[] key)
		{
			key[0] = input[0];
			key[1] = (byte)((input[0] << 7 | (input[1] & 0xff) >> 1));
			key[2] = (byte)((input[1] << 6 | (input[2] & 0xff) >> 2));
			key[3] = (byte)((input[2] << 5 | (input[3] & 0xff) >> 3));
			key[4] = (byte)((input[3] << 4 | (input[4] & 0xff) >> 4));
			key[5] = (byte)((input[4] << 3 | (input[5] & 0xff) >> 5));
			key[6] = (byte)((input[5] << 2 | (input[6] & 0xff) >> 6));
			key[7] = (byte)((input[6] << 1));
			oddParity(key, 8);
		}


		/*********************/
		/* Private Interface */
		/*********************/


		/***********/
		/* Methods */
		/***********/

		// Decode a single base64 byte
		private static byte base64DecodeChar(char enc)
		{
			// Find char in encoding array
			int pos = Array.IndexOf(c_base64Table, enc);
			return (byte)pos;
		}

		// Decode a single modified base64 byte
		private static byte modBase64DecodeChar(char enc)
		{
			// Find char in encoding array
			int pos = Array.IndexOf(c_modBase64Table, enc);
			return (byte)pos;
		}


		// Apply odd parity to a byte array
		private static void oddParity(byte[] buff, uint len)
		{
			for(uint i = 0; i < len; i++) {
				byte b = buff[i];
				bool needsParity = (((b >> 7) ^ (b >> 6) ^ (b >> 5) ^ (b >> 4) ^
					(b >> 3) ^ (b >> 2) ^ (b >> 1)) & 0x01) == 0;

				if(needsParity)
					buff[i] |= 0x01;
				else
					buff[i] &= 0xfe;
			}
		}


		/********/
		/* Data */
		/********/

		// Base64 encoding lookup table
		private static char[]			c_base64Table = new char[64] {
			'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q',
			'R','S','T','U','V','W','X','Y','Z','a','b','c','d','e','f','g','h',
			'i','j','k','l','m','n','o','p','q','r','s','t','u','v','w','x','y',
			'z','0','1','2','3','4','5','6','7','8','9','+','/'
		};

		// Modified Base64 encoding lookup table
		private static char[]			c_modBase64Table = new char[64] {
			'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q',
			'R','S','T','U','V','W','X','Y','Z','a','b','c','d','e','f','g','h',
			'i','j','k','l','m','n','o','p','q','r','s','t','u','v','w','x','y',
			'z','0','1','2','3','4','5','6','7','8','9','-','.'
		};

	}

}	// namespace

