/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Base class for data writers.
 *
 * $Id: Writer.cs,v 1.1 2008/07/25 15:22:39 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace HTTPTunnel
{
	using System;
	using System.Net.Sockets;
	using System.Text;

	public abstract class Writer {

		// Construction
		public Writer(Globals glob, Connection conn)
		{
			m_glob = glob;
			m_conn = conn;
			m_buff = new RawBuff();
		}


		/********************/
		/* Public Interface */
		/********************/


		// Process and write some data to our socket
		public virtual void write(MsgParms parms)
		{
			// Lock data buffer while we process more data into it
			lock(this) {
				// Process data into whatever form is required
				processData(parms, m_buff);

				// Flush all buffered data
				flush(parms.getSendRetry());
			}
		}

		// Write some out-of-band data
		public void writeOutOfBand(RawBuff outOfBand)
		{
			// Lock data buffer while we process more data into it
			lock(this) {
				// Append out of band data to write buffer
				m_buff.append(outOfBand);

				// Send the data
				rawWrite();
			}
		}

		// Try to flush all buffered data
		public bool flush(bool retry)
		{
			bool flushed = false;

			lock(this) {
				// See if we can write any queued data
				rawWrite();

				// Did we manage to flush all data?
				flushed = (m_buff.getLength() == 0);
				if(!flushed && retry) {
					// Retry flushing
					m_glob.DPRINTV("Flush failed, scheduling retry");
					Tunneler tunneler = m_conn.getTunneler();
					uint connID = m_conn.getID();

					// Check we don't already have a flush pending
					if(!tunneler.isScheduled<FlushConnectionEvent, uint>(connID)) {
						// Schedule flush
						tunneler.scheduleEvent(new FlushConnectionEvent(m_glob,
							new Time(0,FLUSH_TIME), connID, true, tunneler));
					}
				}
			}

			return flushed;
		}


		/*************/
		/* Accessors */
		/*************/

		// Do we have unprocessed data remaining?
		public virtual bool dataRemaining()
		{
			return (m_buff.getLength() > 0);
		}


		/***********************/
		/* Protected Interface */
		/***********************/


		/*************/
		/* Constants */
		/*************/

		// Time (in milliseconds) between flush attempts
		private const int				FLUSH_TIME			= 200;


		/***********/
		/* Methods */
		/***********/

		// Attempt to process some data to send out
		protected abstract void processData(MsgParms parms, RawBuff dest);


		/********/
		/* Data */
		/********/

		// Global data
		protected Globals				m_glob;

		// Our controlling connection
		protected Connection			m_conn;


		/*********************/
		/* Private Interface */
		/*********************/


		/***********/
		/* Methods */
		/***********/

		// Write raw data to our socket
		private void rawWrite()
		{
			// Any data to send?
			uint size = m_buff.getLength();
			if(size > 0) {
				#if DEBUG_TUN
				StringBuilder nfo = new StringBuilder();
				nfo.AppendFormat("Writer::rawWrite() - Buff contains {0} bytes", size);
				m_glob.DPRINTV(nfo.ToString());
				#endif

				// Somewhere to record progress
				StringBuilder err = new StringBuilder();
				bool ok = true;
				uint totalWritten = 0;
				Sock sock = m_conn.getSocket();

				// Try to write as much as we can
				while(totalWritten < size) {
					try {
						// Check write will not block
						bool pollError = false;
						if(sock.pollWrite(0, ref pollError)) {
							// Check for error condition on socket
							if(pollError) {
								// Error on socket
								ok = false;
								err.Append("Poll returned error for write on socket for connection ");
								err.AppendFormat("{0}", m_conn.getID());
								break;
							}

							try {
								// Write should not block
								uint nowritten = (uint)sock.write(m_buff);
								totalWritten += nowritten;

								// Erase written data from buffer
								m_buff.erase(nowritten);
							}
							catch(SockError e) {
								if(e.m_errno == SocketError.WouldBlock) {
									// Write would block
									StringBuilder nfo2 = new StringBuilder();
									nfo2.AppendFormat("Socket write says it would block for connection {0}",
										m_conn.getID());
									m_glob.DPRINTV(nfo2.ToString());
									break;
								}
								else {
									// Rethrow
									throw;
								}
							}
						}
						else {
							// Poll timeout
							StringBuilder nfo2 = new StringBuilder();
							nfo2.AppendFormat("Poll says write would block for connection {0}",
								m_conn.getID());
							m_glob.DPRINTV(nfo2.ToString());
							break;
						}
					}
					catch(SockError e) {
						ok = false;
						err.AppendFormat("SockError writing for connection {0}: {1} ({2})",
							m_conn.getID(), e.Message, e.m_errno.ToString());
						break;
					}
					catch(CommsError e) {
						ok = false;
						err.AppendFormat("CommsError writing for connection {0}: {2}",
							m_conn.getID(), e.Message);
						break;
					}
				}

				// Any problems?
				if(!ok) {
					throw new WriteError(err.ToString());
				}
			}
		}


		/********/
		/* Data */
		/********/

		// A buffer for storing processed data prior to writing
		private RawBuff				m_buff;

	}

}	// namespace

