/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Wrap a socket connection in HTTP and tunnel it to some destination.
 *
 * $Id: testcli.cs,v 1.4 2009/04/03 11:30:21 rich Exp $
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace TestClient
{
	using System;
	using System.IO;
	using System.Net.Sockets;
	using System.Text;
	using System.Threading;
	using HTTPTunnel;

	public class TestClient {

		// Read and write between stdin and HTTP socket
		private static void ReadWrite(Sock tunsock)
		{
			// Save socket
			m_sock = tunsock;

			// Launch reading threads
			Thread stdinThread = new Thread(new ThreadStart(processStdin));
			Thread socketThread = new Thread(new ThreadStart(processSocket));
			stdinThread.Start();
			socketThread.Start();

			// Loop until told to quit
			while(m_keepRunning) {
				// Wait
				Thread.Sleep(100);
			}

			// Kill reading threads
			stdinThread.Abort();
			socketThread.Abort();
		}

		// Thread to read from the console
		private static void processStdin()
		{
			while(m_keepRunning) {
				// Read next line from the console
				string line = Console.ReadLine();
				if(line.Length == 0) {
					// A blank line is the signal to quit
					Console.WriteLine("Got quit from STDIN - shutting down");
					m_keepRunning = false;
				}
				else {
					// Send down socket to server
					try {
						RawBuff msgData = new RawBuff(line);
						m_sock.write(msgData);
					}
					catch(SocketException) {
						m_keepRunning = false;
					}
				}
			}
		}

		// Thread to read from the socket
		private static void processSocket()
		{
			while(m_keepRunning) {
				try {
					// Poll socket until something arrives
					bool err = false, hup = false;
					bool ready = m_sock.pollRead(0, ref err, ref hup);
					if(ready) {
						// Read from socket
						byte[] buff = new byte[1024];
						int noread = m_sock.read(buff);

						// Get msg string
						StringBuilder msgStr = new StringBuilder();
						for(int i = 0; i < noread; ++i)
							msgStr.Append((char)buff[i]);

						// Write to console
						Console.WriteLine("Read {0} from server", msgStr.ToString());
					}
					else if(err || hup) {
						Console.WriteLine("Error reading from socket - shutting down");
						m_keepRunning = false;
					}
					else {
						// Sleep
						Thread.Sleep(10);
					}
				}
				catch(ConnectionClosed) {
					Console.WriteLine("Got EOF from server - shutting down");
					m_keepRunning = false;
				}
				catch(SockError) {
					Console.WriteLine("Exception reading from socket - shutting down");
					m_keepRunning = false;
				}
			}
		}


		// Main() entry point
		[STAThread]
		public static void Main(string[] args)
		{
			// Are we going to use TCP or UDP?
			Enums.Protocol protocol = Enums.Protocol.TCP;
			if(args.Length == 1) {
				if(args[0] == "-u")
					protocol = Enums.Protocol.UDP;
			}

			// Initialise
			try {
				// Configure global settings
				HTTPTunneler tunneler = HTTPTunneler.Get();
				GlobalConfigBlob globalBlob = new GlobalConfigBlob(Globals.QUIET);
				tunneler.ConfigureGlobal(globalBlob);

				// Start the tunneling library
				tunneler.Start();
				Console.WriteLine("Tunneling library started");

				// Configure client tunnel socket
				SockTunClientConfigBlob cliBlob = new SockTunClientConfigBlob(protocol,
					false, "localhost", 8080);
				ClientSockTunHandle cliHandle = tunneler.LaunchClientSocket(cliBlob);
				TunnelerID cliID = cliHandle.getTunID();
				Console.WriteLine("Client tunneler ID: {0}", cliID);

				// Do some simple I/O
				ReadWrite(cliHandle.getSock());

				// If our tunneler isn't already shutting down, shut it down now
				tunneler.ShutdownTunneler(cliID);

				// Wait for tunneler to tidy up and become inactive
				while(tunneler.IsTunnelerActive(cliID))
					Thread.Sleep(new TimeSpan(0,0,1));

				// Remove configured tunnels
				tunneler.RemoveTunneler(cliID);

				// Stop the tunneling library
				tunneler.Stop();
				Console.WriteLine("Tunneling library stopped");
			}
			catch(TunnelError e) {
				Console.WriteLine("Error running tunneler: {0}", e.Message);
			}

			// Tidy up and exit
			try {
				HTTPTunneler.Release();
			}
			catch(TunnelError e) {
				Console.WriteLine("Error releasing tunneler: {0}", e.Message);
			}
		}


		/********/
		/* Data */
		/********/

		// Flag to indicate when to stop
		private static bool			m_keepRunning = true;

		// Socket to communicate over
		private static Sock			m_sock;

	}

}	// namespace

