/* (C) 2005 Richard Mills / The Linux Academy Ltd.
 *
 * Wrap a socket connection in HTTP and tunnel it to some destination.
 *
 * $Id: testsvr.cs,v 1.7 2013/02/17 10:53:49 rich Exp $
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

namespace TestServer
{
	using System;
	using System.Collections;
	using System.IO;
	using System.Net.Sockets;
	using System.Text;
	using System.Threading;
	using HTTPTunnel;

	public class TestServer {

		// Read and write between stdin and HTTP socket
		private static void ReadWrite()
		{
			// Launch reading threads
			Thread stdinThread = new Thread(new ThreadStart(processStdin));
			Thread socketsThread = new Thread(new ThreadStart(processSockets));
			stdinThread.Start();
			socketsThread.Start();

			// Loop until told to quit
			while(m_keepRunning) {
				Thread.Sleep(100);
			}

			// Kill reading threads
			stdinThread.Abort();
			socketsThread.Abort();

			// Close sockets for any remaining clients
			lock(m_clientSocks) {
				for(int i = 0; i < m_clientSocks.Count; ++i) {
					Sock sock = (Sock)m_clientSocks[i];
					sock.getFD().Close();
				}
				m_clientSocks.Clear();
			}
		}

		// Thread to read from the console
		private static void processStdin()
		{
			while(m_keepRunning) {
				// Read next line from the console
				string line = Console.ReadLine();
				if(line.Length == 0) {
					// A blank line is the signal to quit
					Console.WriteLine("Got quit from STDIN - shutting down");
					m_keepRunning = false;
				}
				else {
					// Send to all connected clients
					lock(m_clientSocks) {
						RawBuff msgData = new RawBuff(line);
						try {
							for(int i = 0; i < m_clientSocks.Count; ++i) {
								Sock sock = (Sock)m_clientSocks[i];
								sock.write(msgData);
							}
						}
						catch(SocketException) {
							m_keepRunning = false;
						}
					}
				}
			}
		}

		// Thread to read from the connected sockets
		private static void processSockets()
		{
			while(m_keepRunning) {
				bool read = false;

				lock(m_clientSocks) {
					// Test each connected client
					for(int i = 0; i < m_clientSocks.Count; ++i) {
						Sock sock = (Sock)m_clientSocks[i];

						try {
							// Poll socket until something arrives
							bool err = false, hup = false;
							bool ready = sock.pollRead(0, ref err, ref hup);
							if(ready) {
								// Read from socket
								read = true;
								byte[] buff = new byte[1024];
								int noread = sock.read(buff);

								// Get msg string
								StringBuilder msgStr = new StringBuilder();
								for(int j = 0; j < noread; ++j)
									msgStr.Append((char)buff[j]);

								// Write to console
								Console.WriteLine("Read {0} from client", msgStr.ToString());
							}
							else if(err || hup) {
								Console.WriteLine("Error reading from connected client - shutting down");
								m_keepRunning = false;
							}
						}
						catch(ConnectionClosed) {
							Console.WriteLine("Got EOF from connected client");

							// Tidy up closed client connection
							sock.getFD().Close();
							m_clientSocks.RemoveAt(i);
							break;
						}
						catch(SockError) {
							Console.WriteLine("Exception reading from connected client - shutting down");
							m_keepRunning = false;
						}
					}
				}

				// Sleep
				if(!read)
					Thread.Sleep(10);
			}
		}


		// Callback to notify us of a new connected client
		private static void NewClientConnection(Sock sock)
		{
			Console.WriteLine("Got new connection from client");

			// Add client socket to list of connected clients
			lock(m_clientSocks) {
				m_clientSocks.Add(sock);
			}
		}


		// Main() entry point
		[STAThread]
		public static void Main(string[] args)
		{
			// Initialise
			try {
				// Configure global settings
				HTTPTunneler tunneler = HTTPTunneler.Get();
				GlobalConfigBlob globalBlob = new GlobalConfigBlob(Globals.QUIET);
				tunneler.ConfigureGlobal(globalBlob);

				// Start the tunneling library
				tunneler.Start();
				Console.WriteLine("Tunneling library started");

				// Configure client tunnel socket
				SockTunServerConfigBlob svrBlob = new SockTunServerConfigBlob("localhost", 8080,
					AddrInfo.Family.IPV4, false, new SockTunServer.SockTunCallback(NewClientConnection));
				ServerSockTunHandle svrHandle = tunneler.LaunchServerSocket(svrBlob);
				TunnelerID svrID = svrHandle.getTunID();
				Console.WriteLine("Server tunneler ID: {0}", svrID);

				// Do some simple I/O
				ReadWrite();

				// If our tunneler isn't already shutting down, shut it down now
				tunneler.ShutdownTunneler(svrID);

				// Wait for tunneler to tidy up and become inactive
				Console.WriteLine("Waiting for tunneler to become inactive");
				while(tunneler.IsTunnelerActive(svrID))
					Thread.Sleep(new TimeSpan(0,0,1));

				// Remove configured tunnels
				tunneler.RemoveTunneler(svrID);

				// Stop the tunneling library
				tunneler.Stop();
				Console.WriteLine("Tunneling library stopped");
			}
			catch(TunnelError e) {
				Console.WriteLine("Error running tunneler: {0}", e.Message);
			}

			// Tidy up and exit
			try {
				HTTPTunneler.Release();
			}
			catch(TunnelError e) {
				Console.WriteLine("Error releasing tunneler: {0}", e.Message);
			}
		}


		/********/
		/* Data */
		/********/

		// Flag to indicate when to stop
		private static bool			m_keepRunning = true;

		// List of connected client streams
		private static ArrayList	m_clientSocks = new ArrayList();

	}

}	// namespace

